@@ -13,15 +13,17 @@ namespace Luau
 
 struct GlobalTypes
 {
-    explicit GlobalTypes(NotNull<BuiltinTypes> builtinTypes);
+    explicit GlobalTypes(NotNull<BuiltinTypes> builtinTypes, SolverMode mode);
 
     NotNull<BuiltinTypes> builtinTypes; // Global types are based on builtin types
 
     TypeArena globalTypes;
     SourceModule globalNames; // names for symbols entered into globalScope
 
-    ScopePtr globalScope;     // shared by all modules
+    ScopePtr globalScope;             // shared by all modules
     ScopePtr globalTypeFunctionScope; // shared by all modules
+
+    SolverMode mode = SolverMode::Old;
 };
 
 } // namespace Luau
