@@ -20,6 +20,11 @@ struct TypeArena
     // Owning module, if any
     Module* owningModule = nullptr;
 
+    bool collectSingletonStats = false;
+    size_t boolSingletonsMinted = 0;
+    size_t strSingletonsMinted = 0;
+    DenseHashSet<std::optional<std::string>> uniqueStrSingletonsMinted{std::nullopt};
+
     void clear();
 
     template<typename T>
@@ -28,6 +33,12 @@ struct TypeArena
         if constexpr (std::is_same_v<T, UnionType>)
             LUAU_ASSERT(tv.options.size() >= 2);
 
+        if constexpr (std::is_same_v<T, SingletonType>)
+        {
+            if (collectSingletonStats)
+                recordSingletonStats(NotNull{&tv});
+        }
+
         return addTV(Type(std::move(tv)));
     }
 
@@ -37,10 +48,6 @@ struct TypeArena
     TypeId freshType(NotNull<BuiltinTypes> builtins, Scope* scope);
     TypeId freshType(NotNull<BuiltinTypes> builtins, Scope* scope, TypeLevel level);
 
-    TypeId freshType_DEPRECATED(TypeLevel level);
-    TypeId freshType_DEPRECATED(Scope* scope);
-    TypeId freshType_DEPRECATED(Scope* scope, TypeLevel level);
-
     TypePackId freshTypePack(Scope* scope, Polarity polarity = Polarity::Unknown);
 
     TypePackId addTypePack(std::initializer_list<TypeId> types);
@@ -58,6 +65,8 @@ struct TypeArena
     TypeId addTypeFunction(const TypeFunction& function, std::vector<TypeId> typeArguments, std::vector<TypePackId> packArguments = {});
     TypePackId addTypePackFunction(const TypePackFunction& function, std::initializer_list<TypeId> types);
     TypePackId addTypePackFunction(const TypePackFunction& function, std::vector<TypeId> typeArguments, std::vector<TypePackId> packArguments = {});
+
+    void recordSingletonStats(NotNull<SingletonType> singleton);
 };
 
 void freeze(TypeArena& arena);
