@@ -1,20 +1,12 @@
 // This file is part of the Luau programming language and is licensed under MIT License; see LICENSE.txt for details
 #pragma once
 
-#include "Luau/Type.h"
 #include "Luau/TypeFunction.h"
-#include "Luau/TypeFunctionRuntime.h"
 
 namespace Luau
 {
 
-using Kind = Variant<TypeId, TypePackId>;
-
-template<typename T>
-const T* get(const Kind& kind)
-{
-    return get_if<T>(&kind);
-}
+struct TypeFunctionContext;
 
 using TypeFunctionKind = Variant<TypeFunctionTypeId, TypeFunctionTypePackId>;
 
@@ -39,6 +31,9 @@ struct TypeFunctionRuntimeBuilderState
 };
 
 TypeFunctionTypeId serialize(TypeId ty, TypeFunctionRuntimeBuilderState* state);
+TypeFunctionTypePackId serialize(TypePackId tp, TypeFunctionRuntimeBuilderState* state);
+
 TypeId deserialize(TypeFunctionTypeId ty, TypeFunctionRuntimeBuilderState* state);
+TypePackId deserialize(TypeFunctionTypePackId tp, TypeFunctionRuntimeBuilderState* state);
 
 } // namespace Luau
