@@ -2,6 +2,7 @@
 #pragma once
 
 #include "Luau/Common.h"
+#include "Luau/DenseHash.h"
 #include "Luau/NotNull.h"
 #include "Luau/Polarity.h"
 #include "Luau/TypeFwd.h"
@@ -46,6 +47,8 @@ struct GenericTypePack
     explicit GenericTypePack(Scope* scope, Polarity polarity = Polarity::Unknown);
     GenericTypePack(TypeLevel level, const Name& name);
     GenericTypePack(Scope* scope, const Name& name);
+    GenericTypePack(Scope* scope, Name name, Polarity polarity);
+    explicit GenericTypePack(Polarity polarity);
 
     int index;
     TypeLevel level;
@@ -115,7 +118,7 @@ struct TypePackVar
 
     TypePackVar& operator=(const TypePackVar& rhs);
 
-    // Re-assignes the content of the pack, but doesn't change the owning arena and can't make pack persistent.
+    // Re-assigns the content of the pack, but doesn't change the owning arena and can't make pack persistent.
     void reassign(const TypePackVar& rhs)
     {
         ty = rhs.ty;
@@ -185,9 +188,7 @@ TypePackIterator begin(TypePackId tp);
 TypePackIterator begin(TypePackId tp, const TxnLog* log);
 TypePackIterator end(TypePackId tp);
 
-using SeenSet = std::set<std::pair<const void*, const void*>>;
-
-bool areEqual(SeenSet& seen, const TypePackVar& lhs, const TypePackVar& rhs);
+TypePackId getTail(TypePackId tp);
 
 TypePackId follow(TypePackId tp);
 TypePackId follow(TypePackId t, const void* context, TypePackId (*mapper)(const void*, TypePackId));
@@ -229,7 +230,7 @@ bool isEmpty(TypePackId tp);
 std::pair<std::vector<TypeId>, std::optional<TypePackId>> flatten(TypePackId tp);
 std::pair<std::vector<TypeId>, std::optional<TypePackId>> flatten(TypePackId tp, const TxnLog& log);
 
-/// Returs true if the type pack arose from a function that is declared to be variadic.
+/// Returns true if the type pack arose from a function that is declared to be variadic.
 /// Returns *false* for function argument packs that are inferred to be safe to oversaturate!
 bool isVariadic(TypePackId tp);
 bool isVariadic(TypePackId tp, const TxnLog& log);
@@ -253,4 +254,16 @@ LUAU_NOINLINE T* emplaceTypePack(TypePackVar* ty, Args&&... args)
 template<>
 LUAU_NOINLINE Unifiable::Bound<TypePackId>* emplaceTypePack<BoundTypePack>(TypePackVar* ty, TypePackId& tyArg);
 
+/*
+ * Takes a slice of a TypePack, starting at sliceIndex, and up to and including the tail. toBeSliced should be already decomposed into head and tail.
+ */
+TypePackId sliceTypePack(
+    size_t sliceIndex,
+    TypePackId toBeSliced,
+    const std::vector<TypeId>& head,
+    std::optional<TypePackId> tail,
+    NotNull<BuiltinTypes> builtinTypes,
+    NotNull<TypeArena> arena
+);
+
 } // namespace Luau
