@@ -12,6 +12,7 @@
 #include <algorithm>
 
 LUAU_FASTFLAG(LuauSolverV2)
+LUAU_FASTFLAGVARIABLE(LuauQueryLocalFunctionBinding)
 
 namespace Luau
 {
@@ -31,7 +32,8 @@ struct AutocompleteNodeFinder : public AstVisitor
 
     bool visit(AstExpr* expr) override
     {
-        if (expr->location.begin <= pos && pos <= expr->location.end)
+        // If the expression size is 0 (begin == end), we don't want to include it in the ancestry
+        if (expr->location.begin <= pos && pos <= expr->location.end && expr->location.begin != expr->location.end)
         {
             ancestry.push_back(expr);
             return true;
@@ -347,34 +349,73 @@ static std::optional<AstStatLocal*> findBindingLocalStatement(const SourceModule
 
 std::optional<Binding> findBindingAtPosition(const Module& module, const SourceModule& source, Position pos)
 {
-    AstExpr* expr = findExprAtPosition(source, pos);
-    if (!expr)
-        return std::nullopt;
-
-    Symbol name;
-    if (auto g = expr->as<AstExprGlobal>())
-        name = g->name;
-    else if (auto l = expr->as<AstExprLocal>())
-        name = l->local;
-    else
-        return std::nullopt;
-
-    ScopePtr currentScope = findScopeAtPosition(module, pos);
-
-    while (currentScope)
+    if (FFlag::LuauQueryLocalFunctionBinding)
     {
-        auto iter = currentScope->bindings.find(name);
-        if (iter != currentScope->bindings.end() && iter->second.location.begin <= pos)
-        {
-            // Ignore this binding if we're inside its definition. e.g. local abc = abc -- Will take the definition of abc from outer scope
-            std::optional<AstStatLocal*> bindingStatement = findBindingLocalStatement(source, iter->second);
-            if (!bindingStatement || !(*bindingStatement)->location.contains(pos))
-                return iter->second;
-        }
-        currentScope = currentScope->parent;
-    }
+        ExprOrLocal exprOrLocal = findExprOrLocalAtPosition(source, pos);
 
-    return std::nullopt;
+        Symbol name;
+        if (auto expr = exprOrLocal.getExpr())
+        {
+            if (auto g = expr->as<AstExprGlobal>())
+                name = g->name;
+            else if (auto l = expr->as<AstExprLocal>())
+                name = l->local;
+            else
+                return std::nullopt;
+        }
+        else if (auto local = exprOrLocal.getLocal())
+            name = local;
+        else
+            return std::nullopt;
+
+        ScopePtr currentScope = findScopeAtPosition(module, pos);
+
+        while (currentScope)
+        {
+            auto iter = currentScope->bindings.find(name);
+            if (iter != currentScope->bindings.end() && iter->second.location.begin <= pos)
+            {
+                // Ignore this binding if we're inside its definition. e.g. local abc = abc -- Will take the definition of abc from outer scope
+                std::optional<AstStatLocal*> bindingStatement = findBindingLocalStatement(source, iter->second);
+                if (!bindingStatement || !(*bindingStatement)->location.contains(pos))
+                    return iter->second;
+            }
+            currentScope = currentScope->parent;
+        }
+
+        return std::nullopt;
+    }
+    else
+    {
+        AstExpr* expr = findExprAtPosition(source, pos);
+        if (!expr)
+            return std::nullopt;
+
+        Symbol name;
+        if (auto g = expr->as<AstExprGlobal>())
+            name = g->name;
+        else if (auto l = expr->as<AstExprLocal>())
+            name = l->local;
+        else
+            return std::nullopt;
+
+        ScopePtr currentScope = findScopeAtPosition(module, pos);
+
+        while (currentScope)
+        {
+            auto iter = currentScope->bindings.find(name);
+            if (iter != currentScope->bindings.end() && iter->second.location.begin <= pos)
+            {
+                // Ignore this binding if we're inside its definition. e.g. local abc = abc -- Will take the definition of abc from outer scope
+                std::optional<AstStatLocal*> bindingStatement = findBindingLocalStatement(source, iter->second);
+                if (!bindingStatement || !(*bindingStatement)->location.contains(pos))
+                    return iter->second;
+            }
+            currentScope = currentScope->parent;
+        }
+
+        return std::nullopt;
+    }
 }
 
 namespace
@@ -446,7 +487,7 @@ struct FindExprOrLocal : public AstVisitor
         return true;
     }
 
-    virtual bool visit(AstExprFunction* fn) override
+    bool visit(AstExprFunction* fn) override
     {
         for (size_t i = 0; i < fn->args.size; ++i)
         {
@@ -455,13 +496,13 @@ struct FindExprOrLocal : public AstVisitor
         return visit((class AstExpr*)fn);
     }
 
-    virtual bool visit(AstStatFor* forStat) override
+    bool visit(AstStatFor* forStat) override
     {
         visitLocal(forStat->var);
         return true;
     }
 
-    virtual bool visit(AstStatForIn* forIn) override
+    bool visit(AstStatForIn* forIn) override
     {
         for (AstLocal* var : forIn->vars)
         {
@@ -483,7 +524,7 @@ static std::optional<DocumentationSymbol> checkOverloadedDocumentationSymbol(
     const Module& module,
     const TypeId ty,
     const AstExpr* parentExpr,
-    const std::optional<DocumentationSymbol> documentationSymbol
+    std::optional<DocumentationSymbol> documentationSymbol
 )
 {
     if (!documentationSymbol)
@@ -524,7 +565,18 @@ static std::optional<DocumentationSymbol> getMetatableDocumentation(
     if (indexIt == mtable->props.end())
         return std::nullopt;
 
-    TypeId followed = follow(indexIt->second.type());
+    TypeId followed;
+    if (FFlag::LuauSolverV2)
+    {
+        if (indexIt->second.readTy)
+            followed = follow(*indexIt->second.readTy);
+        else if (indexIt->second.writeTy)
+            followed = follow(*indexIt->second.writeTy);
+        else
+            return std::nullopt;
+    }
+    else
+        followed = follow(indexIt->second.type_DEPRECATED());
     const TableType* ttv = get<TableType>(followed);
     if (!ttv)
         return std::nullopt;
@@ -539,7 +591,7 @@ static std::optional<DocumentationSymbol> getMetatableDocumentation(
             return checkOverloadedDocumentationSymbol(module, *ty, parentExpr, propIt->second.documentationSymbol);
     }
     else
-        return checkOverloadedDocumentationSymbol(module, propIt->second.type(), parentExpr, propIt->second.documentationSymbol);
+        return checkOverloadedDocumentationSymbol(module, propIt->second.type_DEPRECATED(), parentExpr, propIt->second.documentationSymbol);
 
     return std::nullopt;
 }
@@ -551,8 +603,11 @@ std::optional<DocumentationSymbol> getDocumentationSymbolAtPosition(const Source
     AstExpr* targetExpr = ancestry.size() >= 1 ? ancestry[ancestry.size() - 1]->asExpr() : nullptr;
     AstExpr* parentExpr = ancestry.size() >= 2 ? ancestry[ancestry.size() - 2]->asExpr() : nullptr;
 
-    if (std::optional<Binding> binding = findBindingAtPosition(module, source, position))
-        return checkOverloadedDocumentationSymbol(module, binding->typeId, parentExpr, binding->documentationSymbol);
+    if (!FFlag::LuauQueryLocalFunctionBinding)
+    {
+        if (std::optional<Binding> binding = findBindingAtPosition(module, source, position))
+            return checkOverloadedDocumentationSymbol(module, binding->typeId, parentExpr, binding->documentationSymbol);
+    }
 
     if (targetExpr)
     {
@@ -571,14 +626,16 @@ std::optional<DocumentationSymbol> getDocumentationSymbolAtPosition(const Source
                                 return checkOverloadedDocumentationSymbol(module, *ty, parentExpr, propIt->second.documentationSymbol);
                         }
                         else
-                            return checkOverloadedDocumentationSymbol(module, propIt->second.type(), parentExpr, propIt->second.documentationSymbol);
+                            return checkOverloadedDocumentationSymbol(
+                                module, propIt->second.type_DEPRECATED(), parentExpr, propIt->second.documentationSymbol
+                            );
                     }
                 }
-                else if (const ClassType* ctv = get<ClassType>(parentTy))
+                else if (const ExternType* etv = get<ExternType>(parentTy))
                 {
-                    while (ctv)
+                    while (etv)
                     {
-                        if (auto propIt = ctv->props.find(indexName->index.value); propIt != ctv->props.end())
+                        if (auto propIt = etv->props.find(indexName->index.value); propIt != etv->props.end())
                         {
                             if (FFlag::LuauSolverV2)
                             {
@@ -587,10 +644,10 @@ std::optional<DocumentationSymbol> getDocumentationSymbolAtPosition(const Source
                             }
                             else
                                 return checkOverloadedDocumentationSymbol(
-                                    module, propIt->second.type(), parentExpr, propIt->second.documentationSymbol
+                                    module, propIt->second.type_DEPRECATED(), parentExpr, propIt->second.documentationSymbol
                                 );
                         }
-                        ctv = ctv->parent ? Luau::get<Luau::ClassType>(*ctv->parent) : nullptr;
+                        etv = etv->parent ? Luau::get<Luau::ExternType>(*etv->parent) : nullptr;
                     }
                 }
                 else if (const PrimitiveType* ptv = get<PrimitiveType>(parentTy); ptv && ptv->metatable)
@@ -636,6 +693,12 @@ std::optional<DocumentationSymbol> getDocumentationSymbolAtPosition(const Source
         }
     }
 
+    if (FFlag::LuauQueryLocalFunctionBinding)
+    {
+        if (std::optional<Binding> binding = findBindingAtPosition(module, source, position))
+            return checkOverloadedDocumentationSymbol(module, binding->typeId, parentExpr, binding->documentationSymbol);
+    }
+
     if (std::optional<TypeId> ty = findTypeAtPosition(module, source, position))
     {
         if ((*ty)->documentationSymbol)
