@@ -13,9 +13,9 @@
 
 LUAU_FASTFLAG(DebugLuauFreezeArena)
 LUAU_FASTFLAG(LuauSolverV2)
-LUAU_FASTFLAGVARIABLE(LuauPreprocessTypestatedArgument)
-LUAU_FASTFLAGVARIABLE(LuauDfgScopeStackTrueReset)
-LUAU_FASTFLAGVARIABLE(LuauDfgScopeStackNotNull)
+LUAU_FASTFLAG(LuauExplicitTypeInstantiationSyntax)
+LUAU_FASTFLAG(LuauExplicitTypeInstantiationSupport)
+
 namespace Luau
 {
 
@@ -37,18 +37,11 @@ struct PushScope
 
     ~PushScope()
     {
-        if (FFlag::LuauDfgScopeStackTrueReset)
-        {
-            // If somehow this stack has _shrunk_ to be smaller than we expect,
-            // something very strange has happened.
-            LUAU_ASSERT(stack.size() > previousSize);
-            while (stack.size() > previousSize)
-                stack.pop_back();
-        }
-        else
-        {
+        // If somehow this stack has _shrunk_ to be smaller than we expect,
+        // something very strange has happened.
+        LUAU_ASSERT(stack.size() > previousSize);
+        while (stack.size() > previousSize)
             stack.pop_back();
-        }
     }
 };
 
@@ -112,6 +105,14 @@ const RefinementKey* DataFlowGraph::getRefinementKey(const AstExpr* expr) const
     return nullptr;
 }
 
+std::optional<Symbol> DataFlowGraph::getSymbolFromDef(const Def* def) const
+{
+    if (auto ref = defToSymbol.find(def))
+        return *ref;
+
+    return std::nullopt;
+}
+
 std::optional<DefId> DfgScope::lookup(Symbol symbol) const
 {
     for (const DfgScope* current = this; current; current = current->parent)
@@ -152,32 +153,6 @@ void DfgScope::inherit(const DfgScope* childScope)
     }
 }
 
-bool DfgScope::canUpdateDefinition(Symbol symbol) const
-{
-    for (const DfgScope* current = this; current; current = current->parent)
-    {
-        if (current->bindings.find(symbol))
-            return true;
-        else if (current->scopeType == DfgScope::Loop)
-            return false;
-    }
-
-    return true;
-}
-
-bool DfgScope::canUpdateDefinition(DefId def, const std::string& key) const
-{
-    for (const DfgScope* current = this; current; current = current->parent)
-    {
-        if (auto props = current->props.find(def))
-            return true;
-        else if (current->scopeType == DfgScope::Loop)
-            return false;
-    }
-
-    return true;
-}
-
 DataFlowGraphBuilder::DataFlowGraphBuilder(NotNull<DefArena> defArena, NotNull<RefinementKeyArena> keyArena)
     : graph{defArena, keyArena}
     , defArena{defArena}
@@ -197,14 +172,7 @@ DataFlowGraph DataFlowGraphBuilder::build(
     DataFlowGraphBuilder builder(defArena, keyArena);
     builder.handle = handle;
 
-    DfgScope* moduleScope;
-    // We're not explicitly calling makeChildScope here because that function relies on currentScope
-    // which guarantees that the scope being returned is NotNull
-    // This means that while the scope stack is empty, we'll have to manually initialize the global scope
-    if (FFlag::LuauDfgScopeStackNotNull)
-        moduleScope = builder.scopes.emplace_back(new DfgScope{nullptr, DfgScope::ScopeType::Linear}).get();
-    else
-        moduleScope = builder.makeChildScope();
+    DfgScope* moduleScope = builder.scopes.emplace_back(new DfgScope{nullptr, DfgScope::ScopeType::Linear}).get();
     PushScope ps{builder.scopeStack, moduleScope};
     builder.visitBlockWithoutChildScope(block);
     builder.resolveCaptures();
@@ -218,6 +186,11 @@ DataFlowGraph DataFlowGraphBuilder::build(
     return std::move(builder.graph);
 }
 
+DataFlowGraph DataFlowGraphBuilder::empty(NotNull<DefArena> defArena, NotNull<RefinementKeyArena> keyArena)
+{
+    return DataFlowGraph{defArena, keyArena};
+}
+
 void DataFlowGraphBuilder::resolveCaptures()
 {
     for (const auto& [_, capture] : captures)
@@ -242,19 +215,9 @@ NotNull<DfgScope> DataFlowGraphBuilder::currentScope()
     return NotNull{scopeStack.back()};
 }
 
-DfgScope* DataFlowGraphBuilder::currentScope_DEPRECATED()
-{
-    if (scopeStack.empty())
-        return nullptr; // nullptr is the root DFG scope.
-    return scopeStack.back();
-}
-
 DfgScope* DataFlowGraphBuilder::makeChildScope(DfgScope::ScopeType scopeType)
 {
-    if (FFlag::LuauDfgScopeStackNotNull)
-        return scopes.emplace_back(new DfgScope{currentScope(), scopeType}).get();
-    else
-        return scopes.emplace_back(new DfgScope{currentScope_DEPRECATED(), scopeType}).get();
+    return scopes.emplace_back(new DfgScope{currentScope(), scopeType}).get();
 }
 
 void DataFlowGraphBuilder::join(DfgScope* p, DfgScope* a, DfgScope* b)
@@ -331,7 +294,7 @@ void DataFlowGraphBuilder::joinProps(DfgScope* result, const DfgScope& a, const
 
 DefId DataFlowGraphBuilder::lookup(Symbol symbol, Location location)
 {
-    DfgScope* scope = FFlag::LuauDfgScopeStackNotNull ? currentScope() : currentScope_DEPRECATED();
+    DfgScope* scope = currentScope();
 
     // true if any of the considered scopes are a loop.
     bool outsideLoopScope = false;
@@ -364,7 +327,7 @@ DefId DataFlowGraphBuilder::lookup(Symbol symbol, Location location)
 
 DefId DataFlowGraphBuilder::lookup(DefId def, const std::string& key, Location location)
 {
-    DfgScope* scope = FFlag::LuauDfgScopeStackNotNull ? currentScope() : currentScope_DEPRECATED();
+    DfgScope* scope = currentScope();
     for (DfgScope* current = scope; current; current = current->parent)
     {
         if (auto props = current->props.find(def))
@@ -410,10 +373,7 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatBlock* b)
         cf = visitBlockWithoutChildScope(b);
     }
 
-    if (FFlag::LuauDfgScopeStackNotNull)
-        currentScope()->inherit(child);
-    else
-        currentScope_DEPRECATED()->inherit(child);
+    currentScope()->inherit(child);
     return cf;
 }
 
@@ -470,7 +430,7 @@ ControlFlow DataFlowGraphBuilder::visit(AstStat* s)
         return visit(d);
     else if (auto d = s->as<AstStatDeclareFunction>())
         return visit(d);
-    else if (auto d = s->as<AstStatDeclareClass>())
+    else if (auto d = s->as<AstStatDeclareExternType>())
         return visit(d);
     else if (auto error = s->as<AstStatError>())
         return visit(error);
@@ -498,11 +458,13 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatIf* i)
         elsecf = visit(i->elsebody);
     }
 
-    DfgScope* scope = FFlag::LuauDfgScopeStackNotNull ? currentScope() : currentScope_DEPRECATED();
+    DfgScope* scope = currentScope();
+    // If the control flow from the `if` or `else` block is non-linear,
+    // then we should assume that the _other_ branch is the one taken.
     if (thencf != ControlFlow::None && elsecf == ControlFlow::None)
-        join(scope, scope, elseScope);
+        scope->inherit(elseScope);
     else if (thencf == ControlFlow::None && elsecf != ControlFlow::None)
-        join(scope, thenScope, scope);
+        scope->inherit(thenScope);
     else if ((thencf | elsecf) == ControlFlow::None)
         join(scope, thenScope, elseScope);
 
@@ -516,40 +478,59 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatIf* i)
 
 ControlFlow DataFlowGraphBuilder::visit(AstStatWhile* w)
 {
-    // TODO(controlflow): entry point has a back edge from exit point
+    // FIXME: This is unsound, as it does not consider the _second_ loop
+    // iteration. Consider something like:
+    //
+    //   local function f(_: number) end
+    //   local x = 42
+    //   while math.random () > 0.5 do
+    //       f(x)
+    //       x = ""
+    //   end
+    //
+    // While the first iteration is fine, the second iteration would
+    // allow a string to flow into a position that expects.
     DfgScope* whileScope = makeChildScope(DfgScope::Loop);
 
+    ControlFlow cf;
     {
         PushScope ps{scopeStack, whileScope};
         visitExpr(w->condition);
-        visit(w->body);
+        cf = visit(w->body);
     }
 
-    if (FFlag::LuauDfgScopeStackNotNull)
-        currentScope()->inherit(whileScope);
-    else
-        currentScope_DEPRECATED()->inherit(whileScope);
+    auto scope = currentScope();
+    // If the inner loop unconditionally returns or throws we shouldn't
+    // consume any type state from the loop body.
+    if (!matches(cf, ControlFlow::Returns | ControlFlow::Throws))
+        join(scope, scope, whileScope);
 
     return ControlFlow::None;
 }
 
 ControlFlow DataFlowGraphBuilder::visit(AstStatRepeat* r)
 {
-    // TODO(controlflow): entry point has a back edge from exit point
+    // See comment in visit(AstStatWhile*): this is unsound as it
+    // does not consider the _second_ loop iteration.
     DfgScope* repeatScope = makeChildScope(DfgScope::Loop);
 
+    ControlFlow cf;
+
     {
         PushScope ps{scopeStack, repeatScope};
-        visitBlockWithoutChildScope(r->body);
+        cf = visitBlockWithoutChildScope(r->body);
         visitExpr(r->condition);
     }
 
-    if (FFlag::LuauDfgScopeStackNotNull)
-        currentScope()->inherit(repeatScope);
-    else
-        currentScope_DEPRECATED()->inherit(repeatScope);
+    // Ultimately: the options for a repeat-until loop are more
+    // straightforward.
+    currentScope()->inherit(repeatScope);
 
-    return ControlFlow::None;
+    // `repeat` loops will unconditionally fire: if the internal control
+    // flow is unconditionally a break or continue, then we have linear
+    // control flow, but if it's throws or returns, then we need to
+    // return _that_ to the parent.
+    return matches(cf, ControlFlow::Breaks | ControlFlow::Continues) ? ControlFlow::None : cf;
 }
 
 ControlFlow DataFlowGraphBuilder::visit(AstStatBreak* b)
@@ -600,16 +581,13 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatLocal* l)
         if (i < l->values.size)
         {
             AstExpr* e = l->values.data[i];
-            if (const AstExprTable* tbl = e->as<AstExprTable>())
+            if (e->is<AstExprTable>())
             {
                 def = defs[i];
             }
         }
         graph.localDefs[local] = def;
-        if (FFlag::LuauDfgScopeStackNotNull)
-            currentScope()->bindings[local] = def;
-        else
-            currentScope_DEPRECATED()->bindings[local] = def;
+        currentScope()->bindings[local] = def;
         captures[local].allVersions.push_back(def);
     }
 
@@ -618,6 +596,8 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatLocal* l)
 
 ControlFlow DataFlowGraphBuilder::visit(AstStatFor* f)
 {
+    // See comment in visit(AstStatWhile*): this is unsound as it
+    // does not consider the _second_ loop iteration.
     DfgScope* forScope = makeChildScope(DfgScope::Loop);
 
     visitExpr(f->from);
@@ -625,6 +605,7 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatFor* f)
     if (f->step)
         visitExpr(f->step);
 
+    ControlFlow cf;
     {
         PushScope ps{scopeStack, forScope};
 
@@ -633,20 +614,17 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatFor* f)
 
         DefId def = defArena->freshCell(f->var, f->var->location);
         graph.localDefs[f->var] = def;
-        if (FFlag::LuauDfgScopeStackNotNull)
-            currentScope()->bindings[f->var] = def;
-        else
-            currentScope_DEPRECATED()->bindings[f->var] = def;
+        currentScope()->bindings[f->var] = def;
         captures[f->var].allVersions.push_back(def);
 
-        // TODO(controlflow): entry point has a back edge from exit point
-        visit(f->body);
+        cf = visit(f->body);
     }
 
-    if (FFlag::LuauDfgScopeStackNotNull)
-        currentScope()->inherit(forScope);
-    else
-        currentScope_DEPRECATED()->inherit(forScope);
+    auto scope = currentScope();
+    // If the inner loop unconditionally returns or throws we shouldn't
+    // consume any type state from the loop body.
+    if (!matches(cf, ControlFlow::Returns | ControlFlow::Throws))
+        join(scope, scope, forScope);
 
     return ControlFlow::None;
 }
@@ -655,6 +633,7 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatForIn* f)
 {
     DfgScope* forScope = makeChildScope(DfgScope::Loop);
 
+    ControlFlow cf;
     {
         PushScope ps{scopeStack, forScope};
 
@@ -665,10 +644,7 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatForIn* f)
 
             DefId def = defArena->freshCell(local, local->location);
             graph.localDefs[local] = def;
-            if (FFlag::LuauDfgScopeStackNotNull)
-                currentScope()->bindings[local] = def;
-            else
-                currentScope_DEPRECATED()->bindings[local] = def;
+            currentScope()->bindings[local] = def;
             captures[local].allVersions.push_back(def);
         }
 
@@ -677,12 +653,14 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatForIn* f)
         for (AstExpr* e : f->values)
             visitExpr(e);
 
-        visit(f->body);
+        cf = visit(f->body);
     }
-    if (FFlag::LuauDfgScopeStackNotNull)
-        currentScope()->inherit(forScope);
-    else
-        currentScope_DEPRECATED()->inherit(forScope);
+
+    auto scope = currentScope();
+    // If the inner loop unconditionally returns or throws we shouldn't
+    // consume any type state from the loop body.
+    if (!matches(cf, ControlFlow::Returns | ControlFlow::Throws))
+        join(scope, scope, forScope);
 
     return ControlFlow::None;
 }
@@ -745,10 +723,7 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatLocalFunction* l)
 {
     DefId def = defArena->freshCell(l->name, l->location);
     graph.localDefs[l->name] = def;
-    if (FFlag::LuauDfgScopeStackNotNull)
-        currentScope()->bindings[l->name] = def;
-    else
-        currentScope_DEPRECATED()->bindings[l->name] = def;
+    currentScope()->bindings[l->name] = def;
     captures[l->name].allVersions.push_back(def);
     visitExpr(l->func);
 
@@ -781,10 +756,7 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatDeclareGlobal* d)
 {
     DefId def = defArena->freshCell(d->name, d->nameLocation);
     graph.declaredDefs[d] = def;
-    if (FFlag::LuauDfgScopeStackNotNull)
-        currentScope()->bindings[d->name] = def;
-    else
-        currentScope_DEPRECATED()->bindings[d->name] = def;
+    currentScope()->bindings[d->name] = def;
     captures[d->name].allVersions.push_back(def);
 
     visitType(d->type);
@@ -796,10 +768,7 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatDeclareFunction* d)
 {
     DefId def = defArena->freshCell(d->name, d->nameLocation);
     graph.declaredDefs[d] = def;
-    if (FFlag::LuauDfgScopeStackNotNull)
-        currentScope()->bindings[d->name] = def;
-    else
-        currentScope_DEPRECATED()->bindings[d->name] = def;
+    currentScope()->bindings[d->name] = def;
     captures[d->name].allVersions.push_back(def);
 
     DfgScope* unreachable = makeChildScope();
@@ -808,12 +777,12 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatDeclareFunction* d)
     visitGenerics(d->generics);
     visitGenericPacks(d->genericPacks);
     visitTypeList(d->params);
-    visitTypeList(d->retTypes);
+    visitTypePack(d->retTypes);
 
     return ControlFlow::None;
 }
 
-ControlFlow DataFlowGraphBuilder::visit(AstStatDeclareClass* d)
+ControlFlow DataFlowGraphBuilder::visit(AstStatDeclareExternType* d)
 {
     // This declaration does not "introduce" any bindings in value namespace,
     // so there's no symbolic value to begin with. We'll traverse the properties
@@ -821,7 +790,7 @@ ControlFlow DataFlowGraphBuilder::visit(AstStatDeclareClass* d)
     DfgScope* unreachable = makeChildScope();
     PushScope ps{scopeStack, unreachable};
 
-    for (AstDeclaredClassProp prop : d->props)
+    for (AstDeclaredExternTypeProperty prop : d->props)
         visitType(prop.ty);
 
     return ControlFlow::None;
@@ -887,6 +856,11 @@ DataFlowResult DataFlowGraphBuilder::visitExpr(AstExpr* e)
             return visitExpr(i);
         else if (auto i = e->as<AstExprInterpString>())
             return visitExpr(i);
+        else if (auto i = e->as<AstExprInstantiate>())
+        {
+            LUAU_ASSERT(FFlag::LuauExplicitTypeInstantiationSyntax);
+            return visitExpr(i);
+        }
         else if (auto error = e->as<AstExprError>())
             return visitExpr(error);
         else
@@ -909,12 +883,14 @@ DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprLocal* l)
 {
     DefId def = lookup(l->local, l->local->location);
     const RefinementKey* key = keyArena->leaf(def);
+    graph.defToSymbol[def] = l->local;
     return {def, key};
 }
 
 DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprGlobal* g)
 {
     DefId def = lookup(g->name, g->location);
+    graph.defToSymbol[def] = g->name;
     return {def, keyArena->leaf(def)};
 }
 
@@ -922,11 +898,8 @@ DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprCall* c)
 {
     visitExpr(c->func);
 
-    if (FFlag::LuauPreprocessTypestatedArgument)
-    {
-        for (AstExpr* arg : c->args)
-            visitExpr(arg);
-    }
+    for (AstExpr* arg : c->args)
+        visitExpr(arg);
 
     if (shouldTypestateForFirstArgument(*c) && c->args.size > 1 && isLValue(*c->args.begin()))
     {
@@ -958,12 +931,6 @@ DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprCall* c)
         visitLValue(firstArg, def);
     }
 
-    if (!FFlag::LuauPreprocessTypestatedArgument)
-    {
-        for (AstExpr* arg : c->args)
-            visitExpr(arg);
-    }
-
     // We treat function calls as "subscripted" as they could potentially
     // return a subscripted value, consider:
     //
@@ -1033,7 +1000,7 @@ DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprFunction* f)
         visitTypePack(f->varargAnnotation);
 
     if (f->returnAnnotation)
-        visitTypeList(*f->returnAnnotation);
+        visitTypePack(f->returnAnnotation);
 
     // TODO: function body can be re-entrant, as in mutations that occurs at the end of the function can also be
     // visible to the beginning of the function, so statically speaking, the body of the function has an exit point
@@ -1051,10 +1018,7 @@ DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprFunction* f)
 DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprTable* t)
 {
     DefId tableCell = defArena->freshCell(Symbol{}, t->location);
-    if (FFlag::LuauDfgScopeStackNotNull)
-        currentScope()->props[tableCell] = {};
-    else
-        currentScope_DEPRECATED()->props[tableCell] = {};
+    currentScope()->props[tableCell] = {};
     for (AstExprTable::Item item : t->items)
     {
         DataFlowResult result = visitExpr(item.value);
@@ -1063,10 +1027,7 @@ DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprTable* t)
             visitExpr(item.key);
             if (auto string = item.key->as<AstExprConstantString>())
             {
-                if (FFlag::LuauDfgScopeStackNotNull)
-                    currentScope()->props[tableCell][string->value.data] = result.def;
-                else
-                    currentScope_DEPRECATED()->props[tableCell][string->value.data] = result.def;
+                currentScope()->props[tableCell][string->value.data] = result.def;
             }
         }
     }
@@ -1083,10 +1044,13 @@ DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprUnary* u)
 
 DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprBinary* b)
 {
-    visitExpr(b->left);
-    visitExpr(b->right);
-
-    return {defArena->freshCell(Symbol{}, b->location), nullptr};
+    auto left = visitExpr(b->left);
+    auto right = visitExpr(b->right);
+    // I think there's some subtlety here. There are probably cases where
+    // X or Y / X and Y can _never_ "be subscripted."
+    auto subscripted = (b->op == AstExprBinary::And || b->op == AstExprBinary::Or) &&
+                       (containsSubscriptedDefinition(left.def) || containsSubscriptedDefinition(right.def));
+    return {defArena->freshCell(Symbol{}, b->location, subscripted), nullptr};
 }
 
 DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprTypeAssertion* t)
@@ -1114,6 +1078,28 @@ DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprInterpString* i)
     return {defArena->freshCell(Symbol{}, i->location), nullptr};
 }
 
+DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprInstantiate* i)
+{
+    if (FFlag::LuauExplicitTypeInstantiationSupport)
+    {
+        for (const AstTypeOrPack& typeOrPack : i->typeArguments)
+        {
+            if (typeOrPack.type)
+            {
+                visitType(typeOrPack.type);
+            }
+            else
+            {
+                LUAU_ASSERT(typeOrPack.typePack);
+                visitTypePack(typeOrPack.typePack);
+            }
+        }
+    }
+
+    return visitExpr(i->expr);
+}
+
+
 DataFlowResult DataFlowGraphBuilder::visitExpr(AstExprError* error)
 {
     DfgScope* unreachable = makeChildScope();
@@ -1148,10 +1134,10 @@ void DataFlowGraphBuilder::visitLValue(AstExpr* e, DefId incomingDef)
 
 DefId DataFlowGraphBuilder::visitLValue(AstExprLocal* l, DefId incomingDef)
 {
-    DfgScope* scope = FFlag::LuauDfgScopeStackNotNull ? currentScope() : currentScope_DEPRECATED();
+    DfgScope* scope = currentScope();
 
     // In order to avoid alias tracking, we need to clip the reference to the parent def.
-    if (scope->canUpdateDefinition(l->local))
+    if (!l->upvalue)
     {
         DefId updated = defArena->freshCell(l->local, l->location, containsSubscriptedDefinition(incomingDef));
         scope->bindings[l->local] = updated;
@@ -1164,33 +1150,22 @@ DefId DataFlowGraphBuilder::visitLValue(AstExprLocal* l, DefId incomingDef)
 
 DefId DataFlowGraphBuilder::visitLValue(AstExprGlobal* g, DefId incomingDef)
 {
-    DfgScope* scope = FFlag::LuauDfgScopeStackNotNull ? currentScope() : currentScope_DEPRECATED();
+    DfgScope* scope = currentScope();
 
-    // In order to avoid alias tracking, we need to clip the reference to the parent def.
-    if (scope->canUpdateDefinition(g->name))
-    {
-        DefId updated = defArena->freshCell(g->name, g->location, containsSubscriptedDefinition(incomingDef));
-        scope->bindings[g->name] = updated;
-        captures[g->name].allVersions.push_back(updated);
-        return updated;
-    }
-    else
-        return visitExpr(static_cast<AstExpr*>(g)).def;
+    DefId updated = defArena->freshCell(g->name, g->location, containsSubscriptedDefinition(incomingDef));
+    scope->bindings[g->name] = updated;
+    captures[g->name].allVersions.push_back(updated);
+    return updated;
 }
 
 DefId DataFlowGraphBuilder::visitLValue(AstExprIndexName* i, DefId incomingDef)
 {
     DefId parentDef = visitExpr(i->expr).def;
 
-    DfgScope* scope = FFlag::LuauDfgScopeStackNotNull ? currentScope() : currentScope_DEPRECATED();
-    if (scope->canUpdateDefinition(parentDef, i->index.value))
-    {
-        DefId updated = defArena->freshCell(i->index, i->location, containsSubscriptedDefinition(incomingDef));
-        scope->props[parentDef][i->index.value] = updated;
-        return updated;
-    }
-    else
-        return visitExpr(static_cast<AstExpr*>(i)).def;
+    DfgScope* scope = currentScope();
+    DefId updated = defArena->freshCell(i->index, i->location, containsSubscriptedDefinition(incomingDef));
+    scope->props[parentDef][i->index.value] = updated;
+    return updated;
 }
 
 DefId DataFlowGraphBuilder::visitLValue(AstExprIndexExpr* i, DefId incomingDef)
@@ -1198,17 +1173,12 @@ DefId DataFlowGraphBuilder::visitLValue(AstExprIndexExpr* i, DefId incomingDef)
     DefId parentDef = visitExpr(i->expr).def;
     visitExpr(i->index);
 
-    DfgScope* scope = FFlag::LuauDfgScopeStackNotNull ? currentScope() : currentScope_DEPRECATED();
+    DfgScope* scope = currentScope();
     if (auto string = i->index->as<AstExprConstantString>())
     {
-        if (scope->canUpdateDefinition(parentDef, string->value.data))
-        {
-            DefId updated = defArena->freshCell(Symbol{}, i->location, containsSubscriptedDefinition(incomingDef));
-            scope->props[parentDef][string->value.data] = updated;
-            return updated;
-        }
-        else
-            return visitExpr(static_cast<AstExpr*>(i)).def;
+        DefId updated = defArena->freshCell(Symbol{}, i->location, containsSubscriptedDefinition(incomingDef));
+        scope->props[parentDef][string->value.data] = updated;
+        return updated;
     }
     else
         return defArena->freshCell(Symbol{}, i->location, /*subscripted=*/true);
@@ -1229,7 +1199,7 @@ void DataFlowGraphBuilder::visitType(AstType* t)
         return visitType(f);
     else if (auto tyof = t->as<AstTypeTypeof>())
         return visitType(tyof);
-    else if (auto o = t->as<AstTypeOptional>())
+    else if (t->is<AstTypeOptional>())
         return;
     else if (auto u = t->as<AstTypeUnion>())
         return visitType(u);
@@ -1237,9 +1207,9 @@ void DataFlowGraphBuilder::visitType(AstType* t)
         return visitType(i);
     else if (auto e = t->as<AstTypeError>())
         return visitType(e);
-    else if (auto s = t->as<AstTypeSingletonBool>())
+    else if (t->is<AstTypeSingletonBool>())
         return; // ok
-    else if (auto s = t->as<AstTypeSingletonString>())
+    else if (t->is<AstTypeSingletonString>())
         return; // ok
     else if (auto g = t->as<AstTypeGroup>())
         return visitType(g->type);
@@ -1275,7 +1245,7 @@ void DataFlowGraphBuilder::visitType(AstTypeFunction* f)
     visitGenerics(f->generics);
     visitGenericPacks(f->genericPacks);
     visitTypeList(f->argTypes);
-    visitTypeList(f->returnTypes);
+    visitTypePack(f->returnTypes);
 }
 
 void DataFlowGraphBuilder::visitType(AstTypeTypeof* t)
@@ -1307,7 +1277,7 @@ void DataFlowGraphBuilder::visitTypePack(AstTypePack* p)
         return visitTypePack(e);
     else if (auto v = p->as<AstTypePackVariadic>())
         return visitTypePack(v);
-    else if (auto g = p->as<AstTypePackGeneric>())
+    else if (p->is<AstTypePackGeneric>())
         return; // ok
     else
         handle->ice("Unknown AstTypePack in DataFlowGraphBuilder::visitTypePack");
