@@ -15,7 +15,6 @@
 #include <algorithm>
 
 LUAU_FASTFLAG(LuauSolverV2);
-LUAU_FASTFLAG(LuauRetainDefinitionAliasLocations)
 
 namespace Luau
 {
@@ -92,10 +91,35 @@ bool isWithinComment(const ParseResult& result, Position pos)
     return isWithinComment(result.commentLocations, pos);
 }
 
+bool isWithinHotComment(const std::vector<HotComment>& hotComments, Position pos)
+{
+    for (const HotComment& hotComment : hotComments)
+    {
+        if (hotComment.location.containsClosed(pos))
+            return true;
+    }
+
+    return false;
+}
+
+bool isWithinHotComment(const SourceModule& sourceModule, Position pos)
+{
+    return isWithinHotComment(sourceModule.hotcomments, pos);
+}
+
+bool isWithinHotComment(const ParseResult& result, Position pos)
+{
+    return isWithinHotComment(result.hotcomments, pos);
+}
+
 struct ClonePublicInterface : Substitution
 {
     NotNull<BuiltinTypes> builtinTypes;
     NotNull<Module> module;
+    // NOTE: This can be made non-optional after
+    // LuauUseWorkspacePropToChooseSolver is clipped.
+    std::optional<SolverMode> solverMode{std::nullopt};
+    bool internalTypeEscaped = false;
 
     ClonePublicInterface(const TxnLog* log, NotNull<BuiltinTypes> builtinTypes, Module* module)
         : Substitution(log, &module->interfaceTypes)
@@ -105,6 +129,20 @@ struct ClonePublicInterface : Substitution
         LUAU_ASSERT(module);
     }
 
+    ClonePublicInterface(const TxnLog* log, NotNull<BuiltinTypes> builtinTypes, Module* module, SolverMode solverMode)
+        : Substitution(log, &module->interfaceTypes)
+        , builtinTypes(builtinTypes)
+        , module(module)
+        , solverMode(solverMode)
+    {
+        LUAU_ASSERT(module);
+    }
+
+    bool isNewSolver() const
+    {
+        return FFlag::LuauSolverV2 || solverMode == SolverMode::New;
+    }
+
     bool isDirty(TypeId ty) override
     {
         if (ty->owningArena == &module->internalTypes)
@@ -158,21 +196,19 @@ struct ClonePublicInterface : Substitution
         else if (TableType* ttv = getMutable<TableType>(result))
         {
             ttv->level = TypeLevel{0, 0};
-            if (FFlag::LuauSolverV2)
+            if (isNewSolver())
+            {
                 ttv->scope = nullptr;
+                ttv->state = TableState::Sealed;
+            }
         }
 
-        if (FFlag::LuauSolverV2)
+        if (isNewSolver())
         {
-            if (auto freety = getMutable<FreeType>(result))
+            if (is<FreeType, BlockedType, PendingExpansionType>(ty))
             {
-                module->errors.emplace_back(
-                    freety->scope->location,
-                    module->name,
-                    InternalError{"Free type is escaping its module; please report this bug at "
-                                  "https://github.com/luau-lang/luau/issues"}
-                );
-                result = builtinTypes->errorRecoveryType();
+                internalTypeEscaped = true;
+                result = builtinTypes->errorType;
             }
             else if (auto genericty = getMutable<GenericType>(result))
             {
@@ -185,20 +221,16 @@ struct ClonePublicInterface : Substitution
 
     TypePackId clean(TypePackId tp) override
     {
-        if (FFlag::LuauSolverV2)
+        if (isNewSolver())
         {
-            auto clonedTp = clone(tp);
-            if (auto ftp = getMutable<FreeTypePack>(clonedTp))
+            if (is<FreeTypePack, BlockedTypePack>(tp))
             {
-                module->errors.emplace_back(
-                    ftp->scope->location,
-                    module->name,
-                    InternalError{"Free type pack is escaping its module; please report this bug at "
-                                  "https://github.com/luau-lang/luau/issues"}
-                );
-                clonedTp = builtinTypes->errorRecoveryTypePack();
+                internalTypeEscaped = true;
+                return builtinTypes->errorTypePack;
             }
-            else if (auto gtp = getMutable<GenericTypePack>(clonedTp))
+
+            auto clonedTp = clone(tp);
+            if (auto gtp = getMutable<GenericTypePack>(clonedTp))
                 gtp->scope = nullptr;
             return clonedTp;
         }
@@ -217,8 +249,9 @@ struct ClonePublicInterface : Substitution
         }
         else
         {
-            module->errors.push_back(TypeError{module->scopes[0].first, UnificationTooComplex{}});
-            return builtinTypes->errorRecoveryType();
+
+            module->errors.emplace_back(module->scopes[0].first, UnificationTooComplex{});
+            return builtinTypes->errorType;
         }
     }
 
@@ -231,8 +264,8 @@ struct ClonePublicInterface : Substitution
         }
         else
         {
-            module->errors.push_back(TypeError{module->scopes[0].first, UnificationTooComplex{}});
-            return builtinTypes->errorRecoveryTypePack();
+            module->errors.emplace_back(module->scopes[0].first, UnificationTooComplex{});
+            return builtinTypes->errorTypePack;
         }
     }
 
@@ -265,10 +298,7 @@ struct ClonePublicInterface : Substitution
 
         TypeId type = cloneType(tf.type);
 
-        if (FFlag::LuauRetainDefinitionAliasLocations)
-            return TypeFun{typeParams, typePackParams, type, tf.definitionLocation};
-        else
-            return TypeFun{typeParams, typePackParams, type};
+        return TypeFun{std::move(typeParams), std::move(typePackParams), type, tf.definitionLocation};
     }
 };
 
@@ -278,17 +308,17 @@ Module::~Module()
     unfreeze(internalTypes);
 }
 
-void Module::clonePublicInterface(NotNull<BuiltinTypes> builtinTypes, InternalErrorReporter& ice)
+void Module::clonePublicInterface(NotNull<BuiltinTypes> builtinTypes, InternalErrorReporter& ice, SolverMode mode)
 {
     CloneState cloneState{builtinTypes};
 
     ScopePtr moduleScope = getModuleScope();
 
     TypePackId returnType = moduleScope->returnType;
-    std::optional<TypePackId> varargPack = FFlag::LuauSolverV2 ? std::nullopt : moduleScope->varargPack;
+    std::optional<TypePackId> varargPack = mode == SolverMode::New ? std::nullopt : moduleScope->varargPack;
 
     TxnLog log;
-    ClonePublicInterface clonePublicInterface{&log, builtinTypes, this};
+    ClonePublicInterface clonePublicInterface{&log, builtinTypes, this, mode};
 
     returnType = clonePublicInterface.cloneTypePack(returnType);
 
@@ -309,6 +339,21 @@ void Module::clonePublicInterface(NotNull<BuiltinTypes> builtinTypes, InternalEr
         ty = clonePublicInterface.cloneType(ty);
     }
 
+    for (auto& tf : typeFunctionAliases)
+    {
+        *tf = clonePublicInterface.cloneTypeFun(*tf);
+    }
+
+    if (clonePublicInterface.internalTypeEscaped)
+    {
+        errors.emplace_back(
+            Location{}, // Not amazing but the best we can do.
+            name,
+            InternalError{"An internal type is escaping this module; please report this bug at "
+                          "https://github.com/luau-lang/luau/issues"}
+        );
+    }
+
     // Copy external stuff over to Module itself
     this->returnType = moduleScope->returnType;
     this->exportedTypeBindings = moduleScope->exportedTypeBindings;
