@@ -16,11 +16,12 @@
 #include "Luau/TypeOrPack.h"
 
 #include <algorithm>
-#include <stdexcept>
 #include <string>
 
+LUAU_FASTFLAGVARIABLE(LuauEnableDenseTableAlias)
+LUAU_FASTFLAGVARIABLE(LuauToStringDecomposition)
+
 LUAU_FASTFLAG(LuauSolverV2)
-LUAU_FASTFLAGVARIABLE(LuauSyntheticErrors)
 
 /*
  * Enables increasing levels of verbosity for Luau type names when stringifying.
@@ -41,6 +42,8 @@ LUAU_FASTFLAGVARIABLE(LuauSyntheticErrors)
 LUAU_FASTINTVARIABLE(DebugLuauVerboseTypeNames, 0)
 LUAU_FASTFLAGVARIABLE(DebugLuauToStringNoLexicalSort)
 
+LUAU_FASTFLAGVARIABLE(LuauToStringIgnoresSyntheticName)
+
 namespace Luau
 {
 
@@ -49,7 +52,11 @@ namespace
 
 struct FindCyclicTypes final : TypeVisitor
 {
-    FindCyclicTypes() = default;
+    FindCyclicTypes()
+        : TypeVisitor("FindCyclicTypes", /* skipBoundTypes */ true)
+    {
+    }
+
     FindCyclicTypes(const FindCyclicTypes&) = delete;
     FindCyclicTypes& operator=(const FindCyclicTypes&) = delete;
 
@@ -83,21 +90,10 @@ struct FindCyclicTypes final : TypeVisitor
     {
         if (!visited.insert(ty))
             return false;
-
-        if (FFlag::LuauSolverV2)
-        {
-            // TODO: Replace these if statements with assert()s when we
-            // delete FFlag::LuauSolverV2.
-            //
-            // When the old solver is used, these pointers are always
-            // unused. When the new solver is used, they are never null.
-
-            if (ft.lowerBound)
-                traverse(ft.lowerBound);
-            if (ft.upperBound)
-                traverse(ft.upperBound);
-        }
-
+        LUAU_ASSERT(ft.lowerBound);
+        LUAU_ASSERT(ft.upperBound);
+        traverse(ft.lowerBound);
+        traverse(ft.upperBound);
         return false;
     }
 
@@ -120,7 +116,7 @@ struct FindCyclicTypes final : TypeVisitor
         return true;
     }
 
-    bool visit(TypeId ty, const ClassType&) override
+    bool visit(TypeId ty, const ExternType&) override
     {
         return false;
     }
@@ -146,7 +142,7 @@ void findCyclicTypes(std::set<TypeId>& cycles, std::set<TypePackId>& cycleTPs, T
 
 static std::pair<bool, std::optional<Luau::Name>> canUseTypeNameInScope(ScopePtr scope, const std::string& name)
 {
-    for (ScopePtr curr = scope; curr; curr = curr->parent)
+    for (ScopePtr curr = std::move(scope); curr; curr = curr->parent)
     {
         for (const auto& [importName, nameTable] : curr->importedTypeBindings)
         {
@@ -161,6 +157,12 @@ static std::pair<bool, std::optional<Luau::Name>> canUseTypeNameInScope(ScopePtr
     return {false, std::nullopt};
 }
 
+struct ElementResult
+{
+    std::string str;
+    std::vector<ToStringSpan> spans;
+};
+
 struct StringifierState
 {
     ToStringOptions& opts;
@@ -175,12 +177,16 @@ struct StringifierState
     size_t indentation = 0;
 
     bool exhaustive;
+    bool ignoreSyntheticName = false;
 
     StringifierState(ToStringOptions& opts, ToStringResult& result)
         : opts(opts)
         , result(result)
         , exhaustive(opts.exhaustive)
     {
+        if (FFlag::LuauToStringIgnoresSyntheticName)
+            ignoreSyntheticName = opts.ignoreSyntheticName;
+
         for (const auto& [_, v] : opts.nameMap.types)
             usedNames.insert(v);
         for (const auto& [_, v] : opts.nameMap.typePacks)
@@ -301,6 +307,18 @@ struct StringifierState
         emit(std::to_string(i).c_str());
     }
 
+    void emitAndRecordSpan(const std::string& s, TypeId ty)
+    {
+        LUAU_ASSERT(FFlag::LuauToStringDecomposition);
+
+        size_t startPos = result.name.length();
+        emit(s);
+        size_t endPos = result.name.length();
+
+        if (endPos > startPos)
+            result.typeSpans.emplace_back(ToStringSpan{startPos, endPos, ty});
+    }
+
     void emit(Polarity p)
     {
         switch (p)
@@ -407,7 +425,7 @@ struct TypeStringifier
         if (prop.isShared())
         {
             emitKey(name);
-            stringify(prop.type());
+            stringify(*prop.readTy);
             return;
         }
 
@@ -434,11 +452,7 @@ struct TypeStringifier
 
     void stringify(const std::string& name, const Property& prop)
     {
-        if (FFlag::LuauSolverV2)
-            return _newStringify(name, prop);
-
-        emitKey(name);
-        stringify(prop.type());
+        return _newStringify(name, prop);
     }
 
     void stringify(TypePackId tp);
@@ -477,6 +491,8 @@ struct TypeStringifier
 
             bool wrap = !singleTp && get<TypePack>(follow(tp));
 
+            wrap &= !isEmpty(tp);
+
             if (wrap)
                 state.emit("(");
 
@@ -494,59 +510,40 @@ struct TypeStringifier
     {
         state.result.invalid = true;
 
-        // TODO: ftv.lowerBound and ftv.upperBound should always be non-nil when
-        // the new solver is used. This can be replaced with an assert.
-        if (FFlag::LuauSolverV2 && ftv.lowerBound && ftv.upperBound)
+        // Free types are guaranteed to have upper and lower bounds now.
+        LUAU_ASSERT(ftv.lowerBound);
+        LUAU_ASSERT(ftv.upperBound);
+        const TypeId lowerBound = follow(ftv.lowerBound);
+        const TypeId upperBound = follow(ftv.upperBound);
+        if (get<NeverType>(lowerBound) && get<UnknownType>(upperBound))
         {
-            const TypeId lowerBound = follow(ftv.lowerBound);
-            const TypeId upperBound = follow(ftv.upperBound);
-            if (get<NeverType>(lowerBound) && get<UnknownType>(upperBound))
-            {
-                state.emit("'");
-                state.emit(state.getName(ty));
-                if (FInt::DebugLuauVerboseTypeNames >= 1)
-                    state.emit(ftv.polarity);
-            }
-            else
-            {
-                state.emit("(");
-                if (!get<NeverType>(lowerBound))
-                {
-                    stringify(lowerBound);
-                    state.emit(" <: ");
-                }
-                state.emit("'");
-                state.emit(state.getName(ty));
-
-                if (FInt::DebugLuauVerboseTypeNames >= 1)
-                    state.emit(ftv.polarity);
-
-                if (!get<UnknownType>(upperBound))
-                {
-                    state.emit(" <: ");
-                    stringify(upperBound);
-                }
-                state.emit(")");
-            }
-            return;
+            state.emit("'");
+            state.emit(state.getName(ty));
+            if (FInt::DebugLuauVerboseTypeNames >= 1)
+                state.emit(ftv.polarity);
         }
-
-        if (FInt::DebugLuauVerboseTypeNames >= 1)
-            state.emit("free-");
-
-        state.emit(state.getName(ty));
-
-        if (FFlag::LuauSolverV2 && FInt::DebugLuauVerboseTypeNames >= 1)
-            state.emit(ftv.polarity);
-
-        if (FInt::DebugLuauVerboseTypeNames >= 2)
+        else
         {
-            state.emit("-");
-            if (FFlag::LuauSolverV2)
-                state.emitLevel(ftv.scope);
-            else
-                state.emit(ftv.level);
+            state.emit("(");
+            if (!get<NeverType>(lowerBound))
+            {
+                stringify(lowerBound);
+                state.emit(" <: ");
+            }
+            state.emit("'");
+            state.emit(state.getName(ty));
+
+            if (FInt::DebugLuauVerboseTypeNames >= 1)
+                state.emit(ftv.polarity);
+
+            if (!get<UnknownType>(upperBound))
+            {
+                state.emit(" <: ");
+                stringify(upperBound);
+            }
+            state.emit(")");
         }
+        return;
     }
 
     void operator()(TypeId, const BoundType& btv)
@@ -574,10 +571,7 @@ struct TypeStringifier
         if (FInt::DebugLuauVerboseTypeNames >= 2)
         {
             state.emit("-");
-            if (FFlag::LuauSolverV2)
-                state.emitLevel(gtv.scope);
-            else
-                state.emit(gtv.level);
+            state.emitLevel(gtv.scope);
         }
     }
 
@@ -677,22 +671,23 @@ struct TypeStringifier
             state.emit(">");
         }
 
-        if (FFlag::LuauSolverV2)
-        {
-            if (ftv.isCheckedFunction)
-                state.emit("@checked ");
-        }
+        if (ftv.isCheckedFunction)
+            state.emit("@checked ");
 
         state.emit("(");
 
-        if (state.opts.functionTypeArguments)
+        if (isEmpty(ftv.argTypes))
+        {
+            // if we've got an empty argument pack, we're done.
+        }
+        else if (state.opts.functionTypeArguments)
             stringify(ftv.argTypes, ftv.argNames);
         else
             stringify(ftv.argTypes);
 
         state.emit(") -> ");
 
-        bool plural = true;
+        bool plural = !isEmpty(ftv.retTypes);
 
         auto retBegin = begin(ftv.retTypes);
         auto retEnd = end(ftv.retTypes);
@@ -714,12 +709,18 @@ struct TypeStringifier
         state.unsee(&ftv);
     }
 
-    void operator()(TypeId, const TableType& ttv)
+    void operator()(TypeId ty, const TableType& ttv)
     {
         if (ttv.boundTo)
             return stringify(*ttv.boundTo);
 
-        if (!state.exhaustive)
+        bool showName = !state.exhaustive;
+        if (FFlag::LuauEnableDenseTableAlias)
+        {
+            // if hide table alias expansions are enabled and there is a name found for the table, use it
+            showName = !state.exhaustive || state.opts.hideTableAliasExpansions;
+        }
+        if (showName)
         {
             if (ttv.name)
             {
@@ -738,14 +739,40 @@ struct TypeStringifier
                     }
                 }
 
-                state.emit(*ttv.name);
+                if (FFlag::LuauToStringDecomposition)
+                    state.emitAndRecordSpan(*ttv.name, ty);
+                else
+                    state.emit(*ttv.name);
                 stringify(ttv.instantiatedTypeParams, ttv.instantiatedTypePackParams);
                 return;
             }
+        }
+
+        if (FFlag::LuauToStringIgnoresSyntheticName)
+        {
+            if (!state.exhaustive && !state.ignoreSyntheticName)
+            {
+                if (ttv.syntheticName)
+                {
+                    state.result.invalid = true;
+                    if (FFlag::LuauToStringDecomposition)
+                        state.emitAndRecordSpan(*ttv.syntheticName, ty);
+                    else
+                        state.emit(*ttv.syntheticName);
+                    stringify(ttv.instantiatedTypeParams, ttv.instantiatedTypePackParams);
+                    return;
+                }
+            }
+        }
+        else if (!state.exhaustive)
+        {
             if (ttv.syntheticName)
             {
                 state.result.invalid = true;
-                state.emit(*ttv.syntheticName);
+                if (FFlag::LuauToStringDecomposition)
+                    state.emitAndRecordSpan(*ttv.syntheticName, ty);
+                else
+                    state.emit(*ttv.syntheticName);
                 stringify(ttv.instantiatedTypeParams, ttv.instantiatedTypePackParams);
                 return;
             }
@@ -760,33 +787,16 @@ struct TypeStringifier
 
         std::string openbrace = "@@@";
         std::string closedbrace = "@@@?!";
-        switch (state.opts.hideTableKind ? (FFlag::LuauSolverV2 ? TableState::Sealed : TableState::Unsealed) : ttv.state)
+        switch (state.opts.hideTableKind ? TableState::Sealed : ttv.state)
         {
         case TableState::Sealed:
-            if (FFlag::LuauSolverV2)
-            {
-                openbrace = "{";
-                closedbrace = "}";
-            }
-            else
-            {
-                state.result.invalid = true;
-                openbrace = "{|";
-                closedbrace = "|}";
-            }
+            openbrace = "{";
+            closedbrace = "}";
             break;
         case TableState::Unsealed:
-            if (FFlag::LuauSolverV2)
-            {
-                state.result.invalid = true;
-                openbrace = "{|";
-                closedbrace = "|}";
-            }
-            else
-            {
-                openbrace = "{";
-                closedbrace = "}";
-            }
+            state.result.invalid = true;
+            openbrace = "{|";
+            closedbrace = "|}";
             break;
         case TableState::Free:
             state.result.invalid = true;
@@ -863,12 +873,15 @@ struct TypeStringifier
         state.unsee(&ttv);
     }
 
-    void operator()(TypeId, const MetatableType& mtv)
+    void operator()(TypeId ty, const MetatableType& mtv)
     {
         state.result.invalid = true;
         if (!state.exhaustive && mtv.syntheticName)
         {
-            state.emit(*mtv.syntheticName);
+            if (FFlag::LuauToStringDecomposition)
+                state.emitAndRecordSpan(*mtv.syntheticName, ty);
+            else
+                state.emit(*mtv.syntheticName);
             return;
         }
 
@@ -880,9 +893,12 @@ struct TypeStringifier
         state.emit(" }");
     }
 
-    void operator()(TypeId, const ClassType& ctv)
+    void operator()(TypeId ty, const ExternType& etv)
     {
-        state.emit(ctv.name);
+        if (FFlag::LuauToStringDecomposition)
+            state.emitAndRecordSpan(etv.name, ty);
+        else
+            state.emit(etv.name);
     }
 
     void operator()(TypeId, const AnyType&)
@@ -909,71 +925,184 @@ struct TypeStringifier
         bool optional = false;
         bool hasNonNilDisjunct = false;
 
-        std::vector<std::string> results = {};
-        for (auto el : &uv)
+        if (FFlag::LuauToStringDecomposition)
         {
-            el = follow(el);
+            std::vector<ElementResult> results = {};
+            size_t resultsLength = 0;
+            bool lengthLimitHit = false;
 
-            if (state.opts.useQuestionMarks && isNil(el))
+            for (auto el : &uv)
             {
-                optional = true;
-                continue;
-            }
-            else
-            {
-                hasNonNilDisjunct = true;
+                el = follow(el);
+
+                if (state.opts.useQuestionMarks && isNil(el))
+                {
+                    optional = true;
+                    continue;
+                }
+                else
+                {
+                    hasNonNilDisjunct = true;
+                }
+
+                std::string saved = std::move(state.result.name);
+                size_t savedSpansSize = state.result.typeSpans.size();
+
+                bool needParens = !state.cycleNames.contains(el) && (get<IntersectionType>(el) != nullptr || get<FunctionType>(el) != nullptr);
+
+                if (needParens)
+                    state.emit("(");
+
+                stringify(el);
+
+                if (needParens)
+                    state.emit(")");
+
+                ElementResult elem;
+                elem.str = std::move(state.result.name);
+
+                for (size_t i = savedSpansSize; i < state.result.typeSpans.size(); ++i)
+                    elem.spans.push_back(state.result.typeSpans[i]);
+                state.result.typeSpans.resize(savedSpansSize);
+
+                resultsLength += elem.str.length();
+                results.push_back(std::move(elem));
+
+                state.result.name = std::move(saved);
+
+                lengthLimitHit = state.opts.maxTypeLength > 0 && resultsLength > state.opts.maxTypeLength;
+
+                if (lengthLimitHit)
+                    break;
             }
 
-            std::string saved = std::move(state.result.name);
+            state.unsee(&uv);
 
-            bool needParens = !state.cycleNames.contains(el) && (get<IntersectionType>(el) || get<FunctionType>(el));
+            if (!lengthLimitHit && !FFlag::DebugLuauToStringNoLexicalSort)
+                std::sort(
+                    results.begin(),
+                    results.end(),
+                    [](const ElementResult& a, const ElementResult& b)
+                    {
+                        return a.str < b.str;
+                    }
+                );
 
-            if (needParens)
+            if (optional && results.size() > 1)
                 state.emit("(");
 
-            stringify(el);
-
-            if (needParens)
-                state.emit(")");
-
-            results.push_back(std::move(state.result.name));
-            state.result.name = std::move(saved);
-        }
-
-        state.unsee(&uv);
-
-        if (!FFlag::DebugLuauToStringNoLexicalSort)
-            std::sort(results.begin(), results.end());
-
-        if (optional && results.size() > 1)
-            state.emit("(");
-
-        bool first = true;
-        bool shouldPlaceOnNewlines = results.size() > state.opts.compositeTypesSingleLineLimit;
-        for (std::string& ss : results)
-        {
-            if (!first)
+            bool first = true;
+            bool shouldPlaceOnNewlines = results.size() > state.opts.compositeTypesSingleLineLimit;
+            for (ElementResult& elem : results)
             {
-                if (shouldPlaceOnNewlines)
-                    state.newline();
-                else
-                    state.emit(" ");
-                state.emit("| ");
+                if (!first)
+                {
+                    if (shouldPlaceOnNewlines)
+                        state.newline();
+                    else
+                        state.emit(" ");
+                    state.emit("| ");
+                }
+
+                size_t basePos = state.result.name.length();
+                state.emit(elem.str);
+                for (const auto& [start, end, ty] : elem.spans)
+                    state.result.typeSpans.emplace_back(ToStringSpan{basePos + start, basePos + end, ty});
+
+                first = false;
+            }
+
+            if (optional)
+            {
+                const char* s = "?";
+                if (results.size() > 1)
+                    s = ")?";
+
+                if (!hasNonNilDisjunct)
+                    s = "nil";
+
+                state.emit(s);
             }
-            state.emit(ss);
-            first = false;
         }
-
-        if (optional)
+        else
         {
-            const char* s = "?";
-            if (results.size() > 1)
-                s = ")?";
+            std::vector<std::string> results = {};
+            size_t resultsLength = 0;
+            bool lengthLimitHit = false;
 
-            if (!hasNonNilDisjunct)
-                s = "nil";
+            for (auto el : &uv)
+            {
+                el = follow(el);
 
-            state.emit(s);
+                if (state.opts.useQuestionMarks && isNil(el))
+                {
+                    optional = true;
+                    continue;
+                }
+                else
+                {
+                    hasNonNilDisjunct = true;
+                }
+
+                std::string saved = std::move(state.result.name);
+
+                bool needParens = !state.cycleNames.contains(el) &&
+                                  (get<IntersectionType>(el) || get<FunctionType>(el)); // NOLINT(readability-implicit-bool-conversion)
+
+                if (needParens)
+                    state.emit("(");
+
+                stringify(el);
+
+                if (needParens)
+                    state.emit(")");
+
+                resultsLength += state.result.name.length();
+                results.push_back(std::move(state.result.name));
+
+                state.result.name = std::move(saved);
+
+                lengthLimitHit = state.opts.maxTypeLength > 0 && resultsLength > state.opts.maxTypeLength;
+
+                if (lengthLimitHit)
+                    break;
+            }
+
+            state.unsee(&uv);
+
+            if (!lengthLimitHit && !FFlag::DebugLuauToStringNoLexicalSort)
+                std::sort(results.begin(), results.end());
+
+            if (optional && results.size() > 1)
+                state.emit("(");
+
+            bool first = true;
+            bool shouldPlaceOnNewlines = results.size() > state.opts.compositeTypesSingleLineLimit;
+            for (std::string& ss : results)
+            {
+                if (!first)
+                {
+                    if (shouldPlaceOnNewlines)
+                        state.newline();
+                    else
+                        state.emit(" ");
+                    state.emit("| ");
+                }
+                state.emit(ss);
+                first = false;
+            }
+
+            if (optional)
+            {
+                const char* s = "?";
+                if (results.size() > 1)
+                    s = ")?";
+
+                if (!hasNonNilDisjunct)
+                    s = "nil";
+
+                state.emit(s);
+            }
         }
     }
 
@@ -986,46 +1115,134 @@ struct TypeStringifier
             return;
         }
 
-        std::vector<std::string> results = {};
-        for (auto el : uv.parts)
+        if (FFlag::LuauToStringDecomposition)
         {
-            el = follow(el);
+            std::vector<ElementResult> results = {};
+            size_t resultsLength = 0;
+            bool lengthLimitHit = false;
 
-            std::string saved = std::move(state.result.name);
-
-            bool needParens = !state.cycleNames.contains(el) && (get<UnionType>(el) || get<FunctionType>(el));
-
-            if (needParens)
-                state.emit("(");
-
-            stringify(el);
-
-            if (needParens)
-                state.emit(")");
-
-            results.push_back(std::move(state.result.name));
-            state.result.name = std::move(saved);
-        }
-
-        state.unsee(&uv);
-
-        if (!FFlag::DebugLuauToStringNoLexicalSort)
-            std::sort(results.begin(), results.end());
-
-        bool first = true;
-        bool shouldPlaceOnNewlines = results.size() > state.opts.compositeTypesSingleLineLimit || isOverloadedFunction(ty);
-        for (std::string& ss : results)
-        {
-            if (!first)
+            for (auto el : uv.parts)
             {
-                if (shouldPlaceOnNewlines)
-                    state.newline();
-                else
-                    state.emit(" ");
-                state.emit("& ");
+                el = follow(el);
+
+                std::string saved = std::move(state.result.name);
+                size_t savedSpansSize = state.result.typeSpans.size();
+
+                bool needParens = !state.cycleNames.contains(el) && (get<UnionType>(el) != nullptr || get<FunctionType>(el) != nullptr);
+
+                if (needParens)
+                    state.emit("(");
+
+                stringify(el);
+
+                if (needParens)
+                    state.emit(")");
+
+                ElementResult elem;
+                elem.str = std::move(state.result.name);
+
+                for (size_t i = savedSpansSize; i < state.result.typeSpans.size(); ++i)
+                    elem.spans.push_back(state.result.typeSpans[i]);
+                state.result.typeSpans.resize(savedSpansSize);
+
+                resultsLength += elem.str.length();
+                results.push_back(std::move(elem));
+
+                state.result.name = std::move(saved);
+
+                lengthLimitHit = state.opts.maxTypeLength > 0 && resultsLength > state.opts.maxTypeLength;
+
+                if (lengthLimitHit)
+                    break;
+            }
+
+            state.unsee(&uv);
+
+            if (!lengthLimitHit && !FFlag::DebugLuauToStringNoLexicalSort)
+                std::sort(
+                    results.begin(),
+                    results.end(),
+                    [](const ElementResult& a, const ElementResult& b)
+                    {
+                        return a.str < b.str;
+                    }
+                );
+
+            bool first = true;
+            bool shouldPlaceOnNewlines = results.size() > state.opts.compositeTypesSingleLineLimit || isOverloadedFunction(ty);
+            for (ElementResult& elem : results)
+            {
+                if (!first)
+                {
+                    if (shouldPlaceOnNewlines)
+                        state.newline();
+                    else
+                        state.emit(" ");
+                    state.emit("& ");
+                }
+
+                size_t basePos = state.result.name.length();
+                state.emit(elem.str);
+                for (const auto& [start, end, spanTy] : elem.spans)
+                    state.result.typeSpans.emplace_back(ToStringSpan{basePos + start, basePos + end, spanTy});
+
+                first = false;
+            }
+        }
+        else
+        {
+            std::vector<std::string> results = {};
+            size_t resultsLength = 0;
+            bool lengthLimitHit = false;
+
+            for (auto el : uv.parts)
+            {
+                el = follow(el);
+
+                std::string saved = std::move(state.result.name);
+
+                bool needParens =
+                    !state.cycleNames.contains(el) && (get<UnionType>(el) || get<FunctionType>(el)); // NOLINT(readability-implicit-bool-conversion)
+
+                if (needParens)
+                    state.emit("(");
+
+                stringify(el);
+
+                if (needParens)
+                    state.emit(")");
+
+                resultsLength += state.result.name.length();
+                results.push_back(std::move(state.result.name));
+
+                state.result.name = std::move(saved);
+
+                lengthLimitHit = state.opts.maxTypeLength > 0 && resultsLength > state.opts.maxTypeLength;
+
+                if (lengthLimitHit)
+                    break;
+            }
+
+            state.unsee(&uv);
+
+            if (!lengthLimitHit && !FFlag::DebugLuauToStringNoLexicalSort)
+                std::sort(results.begin(), results.end());
+
+            bool first = true;
+            bool shouldPlaceOnNewlines = results.size() > state.opts.compositeTypesSingleLineLimit || isOverloadedFunction(ty);
+            for (std::string& ss : results)
+            {
+                if (!first)
+                {
+                    if (shouldPlaceOnNewlines)
+                        state.newline();
+                    else
+                        state.emit(" ");
+                    state.emit("& ");
+                }
+                state.emit(ss);
+                first = false;
             }
-            state.emit(ss);
-            first = false;
         }
     }
 
@@ -1033,7 +1250,7 @@ struct TypeStringifier
     {
         state.result.error = true;
 
-        if (FFlag::LuauSyntheticErrors && tv.synthetic)
+        if (tv.synthetic)
         {
             state.emit("*error-type<");
             stringify(*tv.synthetic);
@@ -1177,6 +1394,13 @@ struct TypePackStringifier
             return;
         }
 
+        if (tp.head.empty() && (!tp.tail || isEmpty(*tp.tail)))
+        {
+            state.emit("()");
+            state.unsee(&tp);
+            return;
+        }
+
         bool first = true;
 
         for (const auto& typeId : tp.head)
@@ -1219,7 +1443,7 @@ struct TypePackStringifier
     {
         state.result.error = true;
 
-        if (FFlag::LuauSyntheticErrors && error.synthetic)
+        if (error.synthetic)
         {
             state.emit("*");
             stringify(*error.synthetic);
@@ -1261,10 +1485,7 @@ struct TypePackStringifier
         if (FInt::DebugLuauVerboseTypeNames >= 2)
         {
             state.emit("-");
-            if (FFlag::LuauSolverV2)
-                state.emitLevel(pack.scope);
-            else
-                state.emit(pack.level);
+            state.emitLevel(pack.scope);
         }
 
         state.emit("...");
@@ -1283,10 +1504,7 @@ struct TypePackStringifier
         if (FInt::DebugLuauVerboseTypeNames >= 2)
         {
             state.emit("-");
-            if (FFlag::LuauSolverV2)
-                state.emitLevel(pack.scope);
-            else
-                state.emit(pack.level);
+            state.emitLevel(pack.scope);
         }
 
         state.emit("...");
@@ -1389,6 +1607,80 @@ static void assignCycleNames(
     }
 }
 
+enum class IgnoreSyntheticName
+{
+    Yes,
+    No
+};
+
+static void tableTypeToStringDetailed(
+    TypeId ty,
+    const TableType* ttv,
+    const IgnoreSyntheticName ignoreSyntheticName,
+    ToStringResult& result,
+    const std::shared_ptr<Scope>& scope,
+    const std::string_view nameToUse,
+    TypeStringifier& tvs
+)
+{
+    LUAU_ASSERT(FFlag::LuauToStringIgnoresSyntheticName);
+
+    if (ignoreSyntheticName == IgnoreSyntheticName::No && ttv->syntheticName)
+        result.invalid = true;
+
+    // If scope is provided, add module name and check visibility
+    if (ttv->name && scope)
+    {
+        auto [success, moduleName] = canUseTypeNameInScope(scope, *ttv->name);
+
+        if (!success)
+            result.invalid = true;
+
+        if (moduleName)
+            result.name = format("%s.", moduleName->c_str());
+    }
+
+    size_t startPos = result.name.length();
+    result.name += nameToUse;
+    size_t endPos = result.name.length();
+
+    if (endPos > startPos)
+        result.typeSpans.emplace_back(ToStringSpan{startPos, endPos, ty});
+
+    tvs.stringify(ttv->instantiatedTypeParams, ttv->instantiatedTypePackParams);
+}
+
+static void tableTypeToStringDetailed_DEPRECATED(
+    const TableType* ttv,
+    const IgnoreSyntheticName ignoreSyntheticName,
+    ToStringResult& result,
+    const std::shared_ptr<Scope>& scope,
+    const std::string_view nameToUse,
+    TypeStringifier& tvs
+)
+{
+    LUAU_ASSERT(FFlag::LuauToStringIgnoresSyntheticName);
+
+    if (ignoreSyntheticName == IgnoreSyntheticName::No && ttv->syntheticName)
+        result.invalid = true;
+
+    // If scope is provided, add module name and check visibility
+    if (ttv->name && scope)
+    {
+        auto [success, moduleName] = canUseTypeNameInScope(scope, *ttv->name);
+
+        if (!success)
+            result.invalid = true;
+
+        if (moduleName)
+            result.name = format("%s.", moduleName->c_str());
+    }
+
+    result.name += nameToUse;
+
+    tvs.stringify(ttv->instantiatedTypeParams, ttv->instantiatedTypePackParams);
+}
+
 ToStringResult toStringDetailed(TypeId ty, ToStringOptions& opts)
 {
     /*
@@ -1413,12 +1705,46 @@ ToStringResult toStringDetailed(TypeId ty, ToStringOptions& opts)
 
     if (!opts.exhaustive)
     {
-        if (auto ttv = get<TableType>(ty); ttv && (ttv->name || ttv->syntheticName))
+        if (FFlag::LuauToStringIgnoresSyntheticName)
+        {
+            if (state.ignoreSyntheticName)
+            {
+                if (auto ttv = get<TableType>(ty); ttv && ttv->name)
+                {
+                    if (FFlag::LuauToStringDecomposition)
+                        tableTypeToStringDetailed(ty, ttv, IgnoreSyntheticName::Yes, result, opts.scope, *ttv->name, tvs);
+                    else
+                        tableTypeToStringDetailed_DEPRECATED(ttv, IgnoreSyntheticName::Yes, result, opts.scope, *ttv->name, tvs);
+
+                    return result;
+                }
+            }
+            else if (auto ttv = get<TableType>(ty); ttv && (ttv->name || ttv->syntheticName))
+            {
+                if (FFlag::LuauToStringDecomposition)
+                    tableTypeToStringDetailed(
+                        ty, ttv, IgnoreSyntheticName::No, result, opts.scope, ttv->name ? *ttv->name : *ttv->syntheticName, tvs
+                    );
+                else
+                    tableTypeToStringDetailed_DEPRECATED(
+                        ttv, IgnoreSyntheticName::No, result, opts.scope, ttv->name ? *ttv->name : *ttv->syntheticName, tvs
+                    );
+
+                return result;
+            }
+            else if (auto mtv = get<MetatableType>(ty); mtv && mtv->syntheticName)
+            {
+                result.invalid = true;
+                result.name = *mtv->syntheticName;
+                return result;
+            }
+        }
+        else if (auto ttv = get<TableType>(ty); ttv && (ttv->name || ttv->syntheticName))
         {
             if (ttv->syntheticName)
                 result.invalid = true;
 
-            // If scope if provided, add module name and check visibility
+            // If scope is provided, add module name and check visibility
             if (ttv->name && opts.scope)
             {
                 auto [success, moduleName] = canUseTypeNameInScope(opts.scope, *ttv->name);
@@ -1430,7 +1756,10 @@ ToStringResult toStringDetailed(TypeId ty, ToStringOptions& opts)
                     result.name = format("%s.", moduleName->c_str());
             }
 
-            result.name += ttv->name ? *ttv->name : *ttv->syntheticName;
+            if (FFlag::LuauToStringDecomposition)
+                state.emitAndRecordSpan(ttv->name ? *ttv->name : *ttv->syntheticName, ty);
+            else
+                result.name += ttv->name ? *ttv->name : *ttv->syntheticName;
 
             tvs.stringify(ttv->instantiatedTypeParams, ttv->instantiatedTypePackParams);
 
@@ -1439,7 +1768,10 @@ ToStringResult toStringDetailed(TypeId ty, ToStringOptions& opts)
         else if (auto mtv = get<MetatableType>(ty); mtv && mtv->syntheticName)
         {
             result.invalid = true;
-            result.name = *mtv->syntheticName;
+            if (FFlag::LuauToStringDecomposition)
+                state.emitAndRecordSpan(*mtv->syntheticName, ty);
+            else
+                result.name = *mtv->syntheticName;
             return result;
         }
     }
@@ -1726,7 +2058,7 @@ std::string toStringNamedFunction(const std::string& funcName, const FunctionTyp
 
     size_t retSize = size(ftv.retTypes);
     bool hasTail = !finite(ftv.retTypes);
-    bool wrap = get<TypePack>(follow(ftv.retTypes)) && (hasTail ? retSize != 0 : retSize != 1);
+    bool wrap = get<TypePack>(follow(ftv.retTypes)) && (hasTail ? retSize != 0 : retSize > 1);
 
     if (wrap)
         state.emit("(");
@@ -1736,6 +2068,7 @@ std::string toStringNamedFunction(const std::string& funcName, const FunctionTyp
     if (wrap)
         state.emit(")");
 
+
     return result.name;
 }
 
@@ -1785,6 +2118,44 @@ std::string dump(const std::optional<TypePackId>& ty)
     return "nullopt";
 }
 
+std::string dump(const std::vector<TypeId>& types)
+{
+    return toStringVector(types, dumpOptions());
+}
+
+std::string dump(const std::vector<TypePackId>& typePacks)
+{
+    return toStringVector(typePacks, dumpOptions());
+}
+
+std::string dump(DenseHashMap<TypeId, TypeId>& types)
+{
+    std::string s = "{";
+    ToStringOptions& opts = dumpOptions();
+    for (const auto& [key, value] : types)
+    {
+        if (s.length() > 1)
+            s += ", ";
+        s += toString(key, opts) + " : " + toString(value, opts);
+    }
+    s += "}";
+    return s;
+}
+
+std::string dump(DenseHashMap<TypePackId, TypePackId>& types)
+{
+    std::string s = "{";
+    ToStringOptions& opts = dumpOptions();
+    for (const auto& [key, value] : types)
+    {
+        if (s.length() > 1)
+            s += ", ";
+        s += toString(key, opts) + " : " + toString(value, opts);
+    }
+    s += "}";
+    return s;
+}
+
 std::string dump(const ScopePtr& scope, const char* name)
 {
     auto binding = scope->linearSearchForBinding(name);
@@ -1821,6 +2192,18 @@ std::string toStringVector(const std::vector<TypeId>& types, ToStringOptions& op
     return s;
 }
 
+std::string toStringVector(const std::vector<TypePackId>& typePacks, ToStringOptions& opts)
+{
+    std::string s;
+    for (TypePackId typePack : typePacks)
+    {
+        if (!s.empty())
+            s += ", ";
+        s += toString(typePack, opts);
+    }
+    return s;
+}
+
 std::string toString(const Constraint& constraint, ToStringOptions& opts)
 {
     auto go = [&opts](auto&& c) -> std::string
@@ -1884,14 +2267,17 @@ std::string toString(const Constraint& constraint, ToStringOptions& opts)
         }
         else if constexpr (std::is_same_v<T, HasPropConstraint>)
         {
-            return tos(c.resultType) + " ~ hasProp " + tos(c.subjectType) + ", \"" + c.prop + "\" ctx=" + std::to_string(int(c.context));
+            std::string s = tos(c.resultType) + " ~ hasProp " + tos(c.subjectType) + ", \"" + c.prop + "\" ctx=" + std::to_string(int(c.context));
+            if (c.inConditional)
+                s += " (inConditional)";
+            return s;
         }
         else if constexpr (std::is_same_v<T, HasIndexerConstraint>)
         {
             return tos(c.resultType) + " ~ hasIndexer " + tos(c.subjectType) + " " + tos(c.indexType);
         }
         else if constexpr (std::is_same_v<T, AssignPropConstraint>)
-            return "assignProp " + tos(c.lhsType) + " " + c.propName + " " + tos(c.rhsType);
+            return tos(c.propType) + " ~ assignProp " + tos(c.lhsType) + " " + c.propName + " " + tos(c.rhsType);
         else if constexpr (std::is_same_v<T, AssignIndexConstraint>)
             return "assignIndex " + tos(c.lhsType) + " " + tos(c.indexType) + " " + tos(c.rhsType);
         else if constexpr (std::is_same_v<T, UnpackConstraint>)
@@ -1904,8 +2290,15 @@ std::string toString(const Constraint& constraint, ToStringOptions& opts)
         }
         else if constexpr (std::is_same_v<T, EqualityConstraint>)
             return "equality: " + tos(c.resultType) + " ~ " + tos(c.assignmentType);
-        else if constexpr (std::is_same_v<T, TableCheckConstraint>)
-            return "table_check " + tos(c.expectedType) + " :> " + tos(c.exprType);
+        else if constexpr (std::is_same_v<T, SimplifyConstraint>)
+            return "simplify " + tos(c.ty);
+        else if constexpr (std::is_same_v<T, PushFunctionTypeConstraint>)
+            return "push_function_type " + tos(c.expectedFunctionType) + " => " + tos(c.functionType);
+        else if constexpr (std::is_same_v<T, TypeInstantiationConstraint>)
+            return "explicitly_specified_constraints " + tos(c.functionType) + " (typeArguments = " + dump(c.typeArguments) +
+                   "), (typePackArguments = " + dump(c.typePackArguments) + ")";
+        else if constexpr (std::is_same_v<T, PushTypeConstraint>)
+            return "push_type " + tos(c.expectedType) + " => " + tos(c.targetType);
         else
             static_assert(always_false_v<T>, "Non-exhaustive constraint switch");
     };
@@ -1928,6 +2321,29 @@ std::string dump(const Constraint& c)
     return s;
 }
 
+// Converts the given number index into a human-readable string for that index to be used in errors.
+// e.g. the index `0` becomes `1st`, `1` becomes `2nd`, `11` becomes `12th`, etc.
+std::string toHumanReadableIndex(size_t number)
+{
+    size_t humanIndex = number + 1;
+    size_t finalDigit = humanIndex % 10;
+
+    if (humanIndex > 10 && humanIndex < 20)
+        return std::to_string(humanIndex) + "th";
+
+    switch (finalDigit)
+    {
+    case 1:
+        return std::to_string(humanIndex) + "st";
+    case 2:
+        return std::to_string(humanIndex) + "nd";
+    case 3:
+        return std::to_string(humanIndex) + "rd";
+    default:
+        return std::to_string(humanIndex) + "th";
+    }
+}
+
 std::optional<std::string> getFunctionNameAsString(const AstExpr& expr)
 {
     const AstExpr* curr = &expr;
