@@ -2,8 +2,9 @@
 #include "Luau/Type.h"
 
 #include "Luau/BuiltinDefinitions.h"
+#include "Luau/BuiltinTypeFunctions.h"
 #include "Luau/Common.h"
-#include "Luau/ConstraintSolver.h"
+#include "Luau/StructuralTypeEquality.h"
 #include "Luau/DenseHash.h"
 #include "Luau/Error.h"
 #include "Luau/RecursionCounter.h"
@@ -16,6 +17,7 @@
 #include "Luau/VisitType.h"
 
 #include <algorithm>
+#include <memory>
 #include <optional>
 #include <stdexcept>
 #include <unordered_map>
@@ -23,11 +25,12 @@
 
 LUAU_FASTFLAG(DebugLuauFreezeArena)
 
+LUAU_FASTFLAG(LuauSolverV2)
+
 LUAU_FASTINTVARIABLE(LuauTypeMaximumStringifierLength, 500)
 LUAU_FASTINTVARIABLE(LuauTableTypeMaximumStringifierLength, 0)
 LUAU_FASTINT(LuauTypeInferRecursionLimit)
 LUAU_FASTFLAG(LuauInstantiateInSubtyping)
-LUAU_FASTFLAGVARIABLE(LuauFreeTypesMustHaveBounds)
 
 namespace Luau
 {
@@ -282,8 +285,8 @@ std::optional<TypeId> getMetatable(TypeId type, NotNull<BuiltinTypes> builtinTyp
 
     if (const MetatableType* mtType = get<MetatableType>(type))
         return mtType->metatable;
-    else if (const ClassType* classType = get<ClassType>(type))
-        return classType->metatable;
+    else if (const ExternType* externType = get<ExternType>(type))
+        return externType->metatable;
     else if (isString(type))
     {
         auto ptv = get<PrimitiveType>(builtinTypes->stringType);
@@ -346,10 +349,10 @@ std::optional<ModuleName> getDefinitionModuleName(TypeId type)
         if (ftv->definition)
             return ftv->definition->definitionModuleName;
     }
-    else if (auto ctv = get<ClassType>(type))
+    else if (auto etv = get<ExternType>(type))
     {
-        if (!ctv->definitionModuleName.empty())
-            return ctv->definitionModuleName;
+        if (!etv->definitionModuleName.empty())
+            return etv->definitionModuleName;
     }
 
     return std::nullopt;
@@ -440,7 +443,7 @@ bool maybeSingleton(TypeId ty)
 
 bool hasLength(TypeId ty, DenseHashSet<TypeId>& seen, int* recursionCount)
 {
-    RecursionLimiter _rl(recursionCount, FInt::LuauTypeInferRecursionLimit);
+    RecursionLimiter _rl("Type::hasLength", recursionCount, FInt::LuauTypeInferRecursionLimit);
 
     ty = follow(ty);
 
@@ -506,31 +509,6 @@ FreeType::FreeType(Scope* scope, TypeLevel level, TypeId lowerBound, TypeId uppe
 {
 }
 
-// Old constructors
-FreeType::FreeType(TypeLevel level)
-    : index(Unifiable::freshIndex())
-    , level(level)
-    , scope(nullptr)
-{
-    LUAU_ASSERT(!FFlag::LuauFreeTypesMustHaveBounds);
-}
-
-FreeType::FreeType(Scope* scope)
-    : index(Unifiable::freshIndex())
-    , level{}
-    , scope(scope)
-{
-    LUAU_ASSERT(!FFlag::LuauFreeTypesMustHaveBounds);
-}
-
-FreeType::FreeType(Scope* scope, TypeLevel level)
-    : index(Unifiable::freshIndex())
-    , level(level)
-    , scope(scope)
-{
-    LUAU_ASSERT(!FFlag::LuauFreeTypesMustHaveBounds);
-}
-
 GenericType::GenericType()
     : index(Unifiable::freshIndex())
     , name("g" + std::to_string(index))
@@ -575,6 +553,16 @@ GenericType::GenericType(Scope* scope, const Name& name)
 {
 }
 
+GenericType::GenericType(Scope* scope, Name name, Polarity polarity)
+    : index(Unifiable::freshIndex())
+    , scope(scope)
+    , name(std::move(name))
+    , explicitName(true)
+    , polarity(polarity)
+{
+}
+
+
 BlockedType::BlockedType()
     : index(Unifiable::freshIndex())
 {
@@ -608,8 +596,8 @@ PendingExpansionType::PendingExpansionType(
 )
     : prefix(prefix)
     , name(name)
-    , typeArguments(typeArguments)
-    , packArguments(packArguments)
+    , typeArguments(std::move(typeArguments))
+    , packArguments(std::move(packArguments))
     , index(++nextIndex)
 {
 }
@@ -642,8 +630,8 @@ FunctionType::FunctionType(
     bool hasSelf
 )
     : definition(std::move(defn))
-    , generics(generics)
-    , genericPacks(genericPacks)
+    , generics(std::move(generics))
+    , genericPacks(std::move(genericPacks))
     , argTypes(argTypes)
     , retTypes(retTypes)
     , hasSelf(hasSelf)
@@ -660,8 +648,8 @@ FunctionType::FunctionType(
     bool hasSelf
 )
     : definition(std::move(defn))
-    , generics(generics)
-    , genericPacks(genericPacks)
+    , generics(std::move(generics))
+    , genericPacks(std::move(genericPacks))
     , level(level)
     , argTypes(argTypes)
     , retTypes(retTypes)
@@ -731,7 +719,7 @@ Property Property::create(std::optional<TypeId> read, std::optional<TypeId> writ
     }
 }
 
-TypeId Property::type() const
+TypeId Property::type_DEPRECATED() const
 {
     LUAU_ASSERT(readTy);
     return *readTy;
@@ -740,8 +728,7 @@ TypeId Property::type() const
 void Property::setType(TypeId ty)
 {
     readTy = ty;
-    if (FFlag::LuauSolverV2)
-        writeTy = ty;
+    writeTy = ty;
 }
 
 void Property::makeShared()
@@ -794,160 +781,6 @@ TableType::TableType(const Props& props, const std::optional<TableIndexer>& inde
 {
 }
 
-// Test Types for equivalence
-// More complex than we'd like because Types can self-reference.
-
-bool areSeen(SeenSet& seen, const void* lhs, const void* rhs)
-{
-    if (lhs == rhs)
-        return true;
-
-    auto p = std::make_pair(const_cast<void*>(lhs), const_cast<void*>(rhs));
-    if (seen.find(p) != seen.end())
-        return true;
-
-    seen.insert(p);
-    return false;
-}
-
-bool areEqual(SeenSet& seen, const FunctionType& lhs, const FunctionType& rhs)
-{
-    if (areSeen(seen, &lhs, &rhs))
-        return true;
-
-    // TODO: check generics CLI-39915
-
-    if (!areEqual(seen, *lhs.argTypes, *rhs.argTypes))
-        return false;
-
-    if (!areEqual(seen, *lhs.retTypes, *rhs.retTypes))
-        return false;
-
-    return true;
-}
-
-bool areEqual(SeenSet& seen, const TableType& lhs, const TableType& rhs)
-{
-    if (areSeen(seen, &lhs, &rhs))
-        return true;
-
-    if (lhs.state != rhs.state)
-        return false;
-
-    if (lhs.props.size() != rhs.props.size())
-        return false;
-
-    if (bool(lhs.indexer) != bool(rhs.indexer))
-        return false;
-
-    if (lhs.indexer && rhs.indexer)
-    {
-        if (!areEqual(seen, *lhs.indexer->indexType, *rhs.indexer->indexType))
-            return false;
-
-        if (!areEqual(seen, *lhs.indexer->indexResultType, *rhs.indexer->indexResultType))
-            return false;
-    }
-
-    auto l = lhs.props.begin();
-    auto r = rhs.props.begin();
-
-    while (l != lhs.props.end())
-    {
-        if (l->first != r->first)
-            return false;
-
-        if (!areEqual(seen, *l->second.type(), *r->second.type()))
-            return false;
-        ++l;
-        ++r;
-    }
-
-    return true;
-}
-
-static bool areEqual(SeenSet& seen, const MetatableType& lhs, const MetatableType& rhs)
-{
-    if (areSeen(seen, &lhs, &rhs))
-        return true;
-
-    return areEqual(seen, *lhs.table, *rhs.table) && areEqual(seen, *lhs.metatable, *rhs.metatable);
-}
-
-bool areEqual(SeenSet& seen, const Type& lhs, const Type& rhs)
-{
-    if (auto bound = get_if<BoundType>(&lhs.ty))
-        return areEqual(seen, *bound->boundTo, rhs);
-
-    if (auto bound = get_if<BoundType>(&rhs.ty))
-        return areEqual(seen, lhs, *bound->boundTo);
-
-    if (lhs.ty.index() != rhs.ty.index())
-        return false;
-
-    {
-        const FreeType* lf = get_if<FreeType>(&lhs.ty);
-        const FreeType* rf = get_if<FreeType>(&rhs.ty);
-        if (lf && rf)
-            return lf->index == rf->index;
-    }
-
-    {
-        const GenericType* lg = get_if<GenericType>(&lhs.ty);
-        const GenericType* rg = get_if<GenericType>(&rhs.ty);
-        if (lg && rg)
-            return lg->index == rg->index;
-    }
-
-    {
-        const PrimitiveType* lp = get_if<PrimitiveType>(&lhs.ty);
-        const PrimitiveType* rp = get_if<PrimitiveType>(&rhs.ty);
-        if (lp && rp)
-            return lp->type == rp->type;
-    }
-
-    {
-        const GenericType* lg = get_if<GenericType>(&lhs.ty);
-        const GenericType* rg = get_if<GenericType>(&rhs.ty);
-        if (lg && rg)
-            return lg->index == rg->index;
-    }
-
-    {
-        const ErrorType* le = get_if<ErrorType>(&lhs.ty);
-        const ErrorType* re = get_if<ErrorType>(&rhs.ty);
-        if (le && re)
-            return le->index == re->index;
-    }
-
-    {
-        const FunctionType* lf = get_if<FunctionType>(&lhs.ty);
-        const FunctionType* rf = get_if<FunctionType>(&rhs.ty);
-        if (lf && rf)
-            return areEqual(seen, *lf, *rf);
-    }
-
-    {
-        const TableType* lt = get_if<TableType>(&lhs.ty);
-        const TableType* rt = get_if<TableType>(&rhs.ty);
-        if (lt && rt)
-            return areEqual(seen, *lt, *rt);
-    }
-
-    {
-        const MetatableType* lmt = get_if<MetatableType>(&lhs.ty);
-        const MetatableType* rmt = get_if<MetatableType>(&rhs.ty);
-
-        if (lmt && rmt)
-            return areEqual(seen, *lmt, *rmt);
-    }
-
-    if (get_if<AnyType>(&lhs.ty) && get_if<AnyType>(&rhs.ty))
-        return true;
-
-    return false;
-}
-
 Type* asMutable(TypeId ty)
 {
     return const_cast<Type*>(ty);
@@ -1002,11 +835,12 @@ TypeId makeFunction(
     std::initializer_list<TypeId> retTypes
 );
 
-TypeId makeStringMetatable(NotNull<BuiltinTypes> builtinTypes); // BuiltinDefinitions.cpp
+TypeId makeStringMetatable(NotNull<BuiltinTypes> builtinTypes, SolverMode mode); // BuiltinDefinitions.cpp
 
 BuiltinTypes::BuiltinTypes()
     : arena(new TypeArena)
     , debugFreezeArena(FFlag::DebugLuauFreezeArena)
+    , typeFunctions(std::make_unique<BuiltinTypeFunctions>())
     , nilType(arena->addType(Type{PrimitiveType{PrimitiveType::NilType}, /*persistent*/ true}))
     , numberType(arena->addType(Type{PrimitiveType{PrimitiveType::Number}, /*persistent*/ true}))
     , stringType(arena->addType(Type{PrimitiveType{PrimitiveType::String}, /*persistent*/ true}))
@@ -1014,7 +848,7 @@ BuiltinTypes::BuiltinTypes()
     , threadType(arena->addType(Type{PrimitiveType{PrimitiveType::Thread}, /*persistent*/ true}))
     , bufferType(arena->addType(Type{PrimitiveType{PrimitiveType::Buffer}, /*persistent*/ true}))
     , functionType(arena->addType(Type{PrimitiveType{PrimitiveType::Function}, /*persistent*/ true}))
-    , classType(arena->addType(Type{ClassType{"class", {}, std::nullopt, std::nullopt, {}, {}, {}, {}}, /*persistent*/ true}))
+    , externType(arena->addType(Type{ExternType{"userdata", {}, std::nullopt, std::nullopt, {}, {}, {}, {}}, /*persistent*/ true}))
     , tableType(arena->addType(Type{PrimitiveType{PrimitiveType::Table}, /*persistent*/ true}))
     , emptyTableType(arena->addType(Type{TableType{TableState::Sealed, TypeLevel{}, nullptr}, /*persistent*/ true}))
     , trueType(arena->addType(Type{SingletonType{BooleanSingleton{true}}, /*persistent*/ true}))
@@ -1026,6 +860,7 @@ BuiltinTypes::BuiltinTypes()
     , noRefineType(arena->addType(Type{NoRefineType{}, /*persistent*/ true}))
     , falsyType(arena->addType(Type{UnionType{{falseType, nilType}}, /*persistent*/ true}))
     , truthyType(arena->addType(Type{NegationType{falsyType}, /*persistent*/ true}))
+    , notNilType(arena->addType(Type{NegationType{nilType}, /*persistent*/ true}))
     , optionalNumberType(arena->addType(Type{UnionType{{numberType, nilType}}, /*persistent*/ true}))
     , optionalStringType(arena->addType(Type{UnionType{{stringType, nilType}}, /*persistent*/ true}))
     , emptyTypePack(arena->addTypePack(TypePackVar{TypePack{{}}, /*persistent*/ true}))
@@ -1050,16 +885,6 @@ BuiltinTypes::~BuiltinTypes()
     FFlag::DebugLuauFreezeArena.value = prevFlag;
 }
 
-TypeId BuiltinTypes::errorRecoveryType() const
-{
-    return errorType;
-}
-
-TypePackId BuiltinTypes::errorRecoveryTypePack() const
-{
-    return errorTypePack;
-}
-
 TypeId BuiltinTypes::errorRecoveryType(TypeId guess) const
 {
     return guess;
@@ -1096,7 +921,18 @@ void persist(TypeId ty)
             LUAU_ASSERT(ttv->state != TableState::Free && ttv->state != TableState::Unsealed);
 
             for (const auto& [_name, prop] : ttv->props)
-                queue.push_back(prop.type());
+            {
+                if (FFlag::LuauSolverV2)
+                {
+                    if (prop.readTy)
+                        queue.push_back(*prop.readTy);
+                    if (prop.writeTy)
+                        queue.push_back(*prop.writeTy);
+                }
+                else
+                    queue.push_back(prop.type_DEPRECATED());
+            }
+
 
             if (ttv->indexer)
             {
@@ -1104,10 +940,20 @@ void persist(TypeId ty)
                 queue.push_back(ttv->indexer->indexResultType);
             }
         }
-        else if (auto ctv = get<ClassType>(t))
+        else if (auto etv = get<ExternType>(t))
         {
-            for (const auto& [_name, prop] : ctv->props)
-                queue.push_back(prop.type());
+            for (const auto& [_name, prop] : etv->props)
+            {
+                if (FFlag::LuauSolverV2)
+                {
+                    if (prop.readTy)
+                        queue.push_back(*prop.readTy);
+                    if (prop.writeTy)
+                        queue.push_back(*prop.writeTy);
+                }
+                else
+                    queue.push_back(prop.type_DEPRECATED());
+            }
         }
         else if (auto utv = get<UnionType>(t))
         {
@@ -1206,7 +1052,7 @@ std::optional<TypeLevel> getLevel(TypePackId tp)
         return std::nullopt;
 }
 
-const Property* lookupClassProp(const ClassType* cls, const Name& name)
+const Property* lookupExternTypeProp(const ExternType* cls, const Name& name)
 {
     while (cls)
     {
@@ -1215,7 +1061,7 @@ const Property* lookupClassProp(const ClassType* cls, const Name& name)
             return &it->second;
 
         if (cls->parent)
-            cls = get<ClassType>(*cls->parent);
+            cls = get<ExternType>(*cls->parent);
         else
             return nullptr;
 
@@ -1225,7 +1071,7 @@ const Property* lookupClassProp(const ClassType* cls, const Name& name)
     return nullptr;
 }
 
-bool isSubclass(const ClassType* cls, const ClassType* parent)
+bool isSubclass(const ExternType* cls, const ExternType* parent)
 {
     while (cls)
     {
@@ -1234,7 +1080,7 @@ bool isSubclass(const ClassType* cls, const ClassType* parent)
         else if (!cls->parent)
             return false;
 
-        cls = get<ClassType>(*cls->parent);
+        cls = get<ExternType>(*cls->parent);
         LUAU_ASSERT(cls);
     }
 
@@ -1303,8 +1149,8 @@ static Tags* getTags(TypeId ty)
         return &ftv->tags;
     else if (auto ttv = getMutable<TableType>(ty))
         return &ttv->tags;
-    else if (auto ctv = getMutable<ClassType>(ty))
-        return &ctv->tags;
+    else if (auto etv = getMutable<ExternType>(ty))
+        return &etv->tags;
 
     return nullptr;
 }
@@ -1334,19 +1180,19 @@ bool hasTag(TypeId ty, const std::string& tagName)
 {
     ty = follow(ty);
 
-    // We special case classes because getTags only returns a pointer to one vector of tags.
-    // But classes has multiple vector of tags, represented throughout the hierarchy.
-    if (auto ctv = get<ClassType>(ty))
+    // We special case extern types because getTags only returns a pointer to one vector of tags.
+    // But extern types has multiple vector of tags, represented throughout the hierarchy.
+    if (auto etv = get<ExternType>(ty))
     {
-        while (ctv)
+        while (etv)
         {
-            if (hasTag(ctv->tags, tagName))
+            if (hasTag(etv->tags, tagName))
                 return true;
-            else if (!ctv->parent)
+            else if (!etv->parent)
                 return false;
 
-            ctv = get<ClassType>(*ctv->parent);
-            LUAU_ASSERT(ctv);
+            etv = get<ExternType>(*etv->parent);
+            LUAU_ASSERT(etv);
         }
     }
     else if (auto tags = getTags(ty))
