@@ -3,7 +3,6 @@
 #include "Luau/TypeArena.h"
 
 LUAU_FASTFLAGVARIABLE(DebugLuauFreezeArena);
-LUAU_FASTFLAG(LuauFreeTypesMustHaveBounds)
 
 namespace Luau
 {
@@ -50,33 +49,6 @@ TypeId TypeArena::freshType(NotNull<BuiltinTypes> builtins, Scope* scope, TypeLe
     return allocated;
 }
 
-TypeId TypeArena::freshType_DEPRECATED(TypeLevel level)
-{
-    TypeId allocated = types.allocate(FreeType{level});
-
-    asMutable(allocated)->owningArena = this;
-
-    return allocated;
-}
-
-TypeId TypeArena::freshType_DEPRECATED(Scope* scope)
-{
-    TypeId allocated = types.allocate(FreeType{scope});
-
-    asMutable(allocated)->owningArena = this;
-
-    return allocated;
-}
-
-TypeId TypeArena::freshType_DEPRECATED(Scope* scope, TypeLevel level)
-{
-    TypeId allocated = types.allocate(FreeType{scope, level});
-
-    asMutable(allocated)->owningArena = this;
-
-    return allocated;
-}
-
 TypePackId TypeArena::freshTypePack(Scope* scope, Polarity polarity)
 {
     TypePackId allocated = typePacks.allocate(FreeTypePack{scope, polarity});
@@ -142,6 +114,27 @@ TypePackId TypeArena::addTypePackFunction(const TypePackFunction& function, std:
     return addTypePack(TypeFunctionInstanceTypePack{NotNull{&function}, std::move(typeArguments), std::move(packArguments)});
 }
 
+void TypeArena::recordSingletonStats(const NotNull<SingletonType> singleton)
+{
+    auto record = [this](auto&& s)
+    {
+        using T = std::decay_t<decltype(s)>;
+
+        if constexpr (std::is_same_v<T, BooleanSingleton>)
+            boolSingletonsMinted += 1;
+        else if constexpr (std::is_same_v<T, StringSingleton>)
+        {
+            strSingletonsMinted += 1;
+            if (!s.value.empty())
+                uniqueStrSingletonsMinted.insert(s.value);
+        }
+        else
+            static_assert(always_false_v<T>, "Missing allocation count support for this kind of singleton");
+    };
+
+    visit(record, singleton->variant);
+}
+
 void freeze(TypeArena& arena)
 {
     if (!FFlag::DebugLuauFreezeArena)
