@@ -2,9 +2,17 @@
 
 #include "Luau/TypeFunctionRuntime.h"
 
+#include "Luau/Allocator.h"
+#include "Luau/Common.h"
+#include "Luau/Lexer.h"
+#include "Luau/BytecodeBuilder.h"
+#include "Luau/ParseResult.h"
+#include "Luau/Compiler.h"
 #include "Luau/DenseHash.h"
 #include "Luau/StringUtils.h"
+#include "Luau/Type.h"
 #include "Luau/TypeFunction.h"
+#include "Luau/TypeFunctionRuntimeBuilder.h"
 
 #include "lua.h"
 #include "lualib.h"
@@ -14,11 +22,141 @@
 #include <vector>
 
 LUAU_DYNAMIC_FASTINT(LuauTypeFunctionSerdeIterationLimit)
-LUAU_FASTFLAGVARIABLE(LuauTypeFunReadWriteParents)
+LUAU_FASTFLAG(LuauTypeCheckerUdtfRenameClassToExtern)
+
+LUAU_FASTFLAGVARIABLE(LuauUnionofIntersectionofFlattens)
+LUAU_FASTFLAGVARIABLE(LuauTypeFunctionSupportsFrozen)
 
 namespace Luau
 {
 
+LuauTempThreadPopper::LuauTempThreadPopper(lua_State* L)
+    : L(L)
+{
+}
+
+LuauTempThreadPopper::~LuauTempThreadPopper()
+{
+    lua_pop(L, 1);
+}
+
+static void dummyStateClose(lua_State*) {}
+
+TypeFunctionRuntime::TypeFunctionRuntime(NotNull<InternalErrorReporter> ice, NotNull<TypeCheckLimits> limits)
+    : ice(ice)
+    , limits(limits)
+    , state(nullptr, dummyStateClose)
+{
+}
+
+TypeFunctionRuntime::~TypeFunctionRuntime() {}
+
+std::optional<std::string> TypeFunctionRuntime::registerFunction(AstStatTypeFunction* function)
+{
+    // If evaluation is disabled, we do not generate additional error messages
+    if (!allowEvaluation)
+        return std::nullopt;
+
+    // Do not evaluate type functions with parse errors inside
+    if (function->hasErrors)
+        return std::nullopt;
+
+    prepareState();
+
+    lua_State* global = state.get();
+
+    // Fetch to check if function is already registered
+    lua_pushlightuserdata(global, function);
+    lua_gettable(global, LUA_REGISTRYINDEX);
+
+    if (!lua_isnil(global, -1))
+    {
+        lua_pop(global, 1);
+        return std::nullopt;
+    }
+
+    lua_pop(global, 1);
+
+    AstName name = function->name;
+
+    // Construct ParseResult containing the type function
+    Allocator allocator;
+    AstNameTable names(allocator);
+
+    AstExpr* exprFunction = function->body;
+    AstArray<AstExpr*> exprReturns{&exprFunction, 1};
+    AstStatReturn stmtReturn{Location{}, exprReturns};
+    AstStat* stmtArray[] = {&stmtReturn};
+    AstArray<AstStat*> stmts{stmtArray, 1};
+    AstStatBlock exec{Location{}, stmts};
+    ParseResult parseResult{&exec, 1, {}, {}, {}, CstNodeMap{nullptr}};
+
+    BytecodeBuilder builder;
+    try
+    {
+        compileOrThrow(builder, parseResult, names);
+    }
+    catch (CompileError& e)
+    {
+        return format("'%s' type function failed to compile with error message: %s", name.value, e.what());
+    }
+
+    std::string bytecode = builder.getBytecode();
+
+    // Separate sandboxed thread for individual execution and private globals
+    lua_State* L = lua_newthread(global);
+    LuauTempThreadPopper popper(global);
+
+    // Create individual environment for the type function
+    luaL_sandboxthread(L);
+
+    // Do not allow global writes to that environment
+    lua_pushvalue(L, LUA_GLOBALSINDEX);
+    lua_setreadonly(L, -1, true);
+    lua_pop(L, 1);
+
+    // Load bytecode into Luau state
+    if (auto error = checkResultForError(L, name.value, luau_load(L, name.value, bytecode.data(), bytecode.size(), 0)))
+        return error;
+
+    // Execute the global function which should return our user-defined type function
+    if (auto error = checkResultForError(L, name.value, lua_resume(L, nullptr, 0)))
+        return error;
+
+    if (!lua_isfunction(L, -1))
+    {
+        lua_pop(L, 1);
+        return format("Could not find '%s' type function in the global scope", name.value);
+    }
+
+    // Store resulting function in the registry
+    lua_pushlightuserdata(global, function);
+    lua_xmove(L, global, 1);
+    lua_settable(global, LUA_REGISTRYINDEX);
+
+    return std::nullopt;
+}
+
+void TypeFunctionRuntime::prepareState()
+{
+    if (state)
+        return;
+
+    state = StateRef(lua_newstate(typeFunctionAlloc, nullptr), lua_close);
+    lua_State* L = state.get();
+
+    lua_setthreaddata(L, this);
+
+    setTypeFunctionEnvironment(L);
+
+    registerTypeUserData(L);
+
+    registerTypesLibrary(L);
+
+    luaL_sandbox(L);
+    luaL_sandboxthread(L);
+}
+
 constexpr int kTypeUserdataTag = 42;
 
 void* typeFunctionAlloc(void* ud, void* ptr, size_t osize, size_t nsize)
@@ -63,7 +201,7 @@ std::optional<std::string> checkResultForError(lua_State* L, const char* typeFun
     }
 }
 
-static TypeFunctionRuntime* getTypeFunctionRuntime(lua_State* L)
+TypeFunctionRuntime* getTypeFunctionRuntime(lua_State* L)
 {
     return static_cast<TypeFunctionRuntime*>(lua_getthreaddata(lua_mainthread(L)));
 }
@@ -80,12 +218,23 @@ TypeFunctionTypePackVar* allocateTypeFunctionTypePack(lua_State* L, TypeFunction
     return ctx->typePackArena.allocate(std::move(type));
 }
 
+void pushType(lua_State* L, TypeFunctionTypeId type)
+{
+    TypeFunctionTypeId* ptr = static_cast<TypeFunctionTypeId*>(lua_newuserdatatagged(L, sizeof(TypeFunctionTypeId), kTypeUserdataTag));
+    *ptr = type;
+
+    // set the new userdata's metatable to type metatable
+    luaL_getmetatable(L, "type");
+    lua_setmetatable(L, -2);
+}
+
 // Pushes a new type userdata onto the stack
-void allocTypeUserData(lua_State* L, TypeFunctionTypeVariant type)
+void allocTypeUserData(lua_State* L, TypeFunctionTypeVariant type, bool frozen)
 {
     // allocate a new type userdata
     TypeFunctionTypeId* ptr = static_cast<TypeFunctionTypeId*>(lua_newuserdatatagged(L, sizeof(TypeFunctionTypeId), kTypeUserdataTag));
     *ptr = allocateTypeFunctionType(L, std::move(type));
+    const_cast<TypeFunctionType*>(*ptr)->frozen = frozen;
 
     // set the new userdata's metatable to type metatable
     luaL_getmetatable(L, "type");
@@ -142,7 +291,7 @@ static std::string getTag(lua_State* L, TypeFunctionTypeId ty)
         return "never";
     else if (get<TypeFunctionAnyType>(ty))
         return "any";
-    else if (auto s = get<TypeFunctionSingletonType>(ty))
+    else if (get<TypeFunctionSingletonType>(ty))
         return "singleton";
     else if (get<TypeFunctionNegationType>(ty))
         return "negation";
@@ -154,8 +303,13 @@ static std::string getTag(lua_State* L, TypeFunctionTypeId ty)
         return "table";
     else if (get<TypeFunctionFunctionType>(ty))
         return "function";
-    else if (get<TypeFunctionClassType>(ty))
-        return "class";
+    else if (get<TypeFunctionExternType>(ty))
+    {
+        if (FFlag::LuauTypeCheckerUdtfRenameClassToExtern)
+            return "extern";
+        else
+            return "class";
+    }
     else if (get<TypeFunctionGenericType>(ty))
         return "generic";
 
@@ -265,7 +419,7 @@ static int createSingleton(lua_State* L)
 }
 
 // Luau: `types.generic(name: string, ispack: boolean?) -> type
-// Create a generic type with the specified type. If an optinal boolean is set to true, result is a generic pack
+// Create a generic type with the specified type. If an optional boolean is set to true, result is a generic pack
 static int createGeneric(lua_State* L)
 {
     const char* name = luaL_checkstring(L, 1);
@@ -315,22 +469,78 @@ static int getSingletonValue(lua_State* L)
     luaL_error(L, "type.value: can't call `value` method on `%s` type", getTag(L, self).c_str());
 }
 
+// Luau: `types.optional(ty: type) -> type`
+// Returns the type instance representing an optional version of `ty`.
+// If `ty` is a union, this adds `nil` to the components of the union.
+// Otherwise, makes a union of the two things.
+static int createOptional(lua_State* L)
+{
+    int argumentCount = lua_gettop(L);
+    if (argumentCount != 1)
+        luaL_error(L, "types.optional: expected 1 argument, but got %d", argumentCount);
+
+    TypeFunctionTypeId argument = getTypeUserData(L, 1);
+
+    std::vector<TypeFunctionTypeId> components;
+
+    if (auto unionTy = get<TypeFunctionUnionType>(argument))
+    {
+        components.reserve(unionTy->components.size() + 1);
+
+        components.insert(components.begin(), unionTy->components.begin(), unionTy->components.end());
+    }
+    else
+        components.emplace_back(argument);
+
+    components.emplace_back(allocateTypeFunctionType(L, TypeFunctionPrimitiveType(TypeFunctionPrimitiveType::NilType)));
+
+    allocTypeUserData(L, TypeFunctionUnionType{std::move(components)});
+
+    return 1;
+}
+
 // Luau: `types.unionof(...: type) -> type`
 // Returns the type instance representing union
 static int createUnion(lua_State* L)
 {
     // get the number of arguments for union
     int argSize = lua_gettop(L);
-    if (argSize < 2)
+    if (!FFlag::LuauUnionofIntersectionofFlattens && argSize < 2)
         luaL_error(L, "types.unionof: expected at least 2 types to union, but got %d", argSize);
 
     std::vector<TypeFunctionTypeId> components;
     components.reserve(argSize);
 
     for (int i = 1; i <= argSize; i++)
-        components.push_back(getTypeUserData(L, i));
+    {
+        if (FFlag::LuauUnionofIntersectionofFlattens)
+        {
+            TypeFunctionTypeId component = getTypeUserData(L, i);
 
-    allocTypeUserData(L, TypeFunctionUnionType{components});
+            if (auto unionComponent = get<TypeFunctionUnionType>(component))
+                components.insert(components.end(), unionComponent->components.begin(), unionComponent->components.end());
+            else if (get<TypeFunctionNeverType>(component))
+                continue;
+            else
+                components.push_back(component);
+        }
+        else
+        {
+            components.push_back(getTypeUserData(L, i));
+        }
+    }
+
+    if (FFlag::LuauUnionofIntersectionofFlattens)
+    {
+        if (components.size() == 0)
+            allocTypeUserData(L, TypeFunctionNeverType{});
+        else if (components.size() == 1)
+            pushType(L, components[0]);
+        else
+            allocTypeUserData(L, TypeFunctionUnionType{std::move(components)});
+    }
+    else
+        allocTypeUserData(L, TypeFunctionUnionType{std::move(components)});
 
     return 1;
 }
@@ -341,16 +551,42 @@ static int createIntersection(lua_State* L)
 {
     // get the number of arguments for intersection
     int argSize = lua_gettop(L);
-    if (argSize < 2)
+    if (!FFlag::LuauUnionofIntersectionofFlattens && argSize < 2)
         luaL_error(L, "types.intersectionof: expected at least 2 types to intersection, but got %d", argSize);
 
     std::vector<TypeFunctionTypeId> components;
     components.reserve(argSize);
 
     for (int i = 1; i <= argSize; i++)
-        components.push_back(getTypeUserData(L, i));
+    {
+        if (FFlag::LuauUnionofIntersectionofFlattens)
+        {
+            TypeFunctionTypeId component = getTypeUserData(L, i);
 
-    allocTypeUserData(L, TypeFunctionIntersectionType{components});
+            if (auto intersectionComponent = get<TypeFunctionIntersectionType>(component))
+                components.insert(components.end(), intersectionComponent->components.begin(), intersectionComponent->components.end());
+            else if (get<TypeFunctionUnknownType>(component))
+                continue;
+            else
+                components.push_back(component);
+        }
+        else
+        {
+            components.push_back(getTypeUserData(L, i));
+        }
+    }
+
+    if (FFlag::LuauUnionofIntersectionofFlattens)
+    {
+        if (components.size() == 0)
+            allocTypeUserData(L, TypeFunctionUnknownType{});
+        else if (components.size() == 1)
+            pushType(L, components[0]);
+        else
+            allocTypeUserData(L, TypeFunctionIntersectionType{std::move(components)});
+    }
+    else
+        allocTypeUserData(L, TypeFunctionIntersectionType{std::move(components)});
 
     return 1;
 }
@@ -510,7 +746,7 @@ static int createTable(lua_State* L)
     if (metatable && !get<TypeFunctionTableType>(*metatable))
         luaL_error(L, "types.newtable: expected to be given a table type as a metatable, but got %s instead", getTag(L, *metatable).c_str());
 
-    allocTypeUserData(L, TypeFunctionTableType{props, indexer, metatable});
+    allocTypeUserData(L, TypeFunctionTableType{std::move(props), indexer, metatable});
     return 1;
 }
 
@@ -527,6 +763,9 @@ static int setTableProp(lua_State* L)
     if (!tftt)
         luaL_error(L, "type.setproperty: expected self to be a table, but got %s instead", getTag(L, self).c_str());
 
+    if (FFlag::LuauTypeFunctionSupportsFrozen && self->frozen)
+        luaL_error(L, "type.setproperty: cannot be called to mutate a frozen type, use `types.copy` to make a copy");
+
     TypeFunctionTypeId key = getTypeUserData(L, 2);
     auto tfst = get<TypeFunctionSingletonType>(key);
     if (!tfst)
@@ -561,6 +800,9 @@ static int setReadTableProp(lua_State* L)
     if (!tftt)
         luaL_error(L, "type.setreadproperty: expected self to be a table, but got %s instead", getTag(L, self).c_str());
 
+    if (FFlag::LuauTypeFunctionSupportsFrozen && self->frozen)
+        luaL_error(L, "type.setreadproperty: cannot be called to mutate a frozen type, use `types.copy` to make a copy");
+
     TypeFunctionTypeId key = getTypeUserData(L, 2);
     auto tfst = get<TypeFunctionSingletonType>(key);
     if (!tfst)
@@ -606,6 +848,9 @@ static int setWriteTableProp(lua_State* L)
     if (!tftt)
         luaL_error(L, "type.setwriteproperty: expected self to be a table, but got %s instead", getTag(L, self).c_str());
 
+    if (FFlag::LuauTypeFunctionSupportsFrozen && self->frozen)
+        luaL_error(L, "type.setwriteproperty: cannot be called to mutate a frozen type, use `types.copy` to make a copy");
+
     TypeFunctionTypeId key = getTypeUserData(L, 2);
     auto tfst = get<TypeFunctionSingletonType>(key);
     if (!tfst)
@@ -727,10 +972,13 @@ static int setTableIndexer(lua_State* L)
     if (!tftt)
         luaL_error(L, "type.setindexer: expected self to be either a table, but got %s instead", getTag(L, self).c_str());
 
+    if (FFlag::LuauTypeFunctionSupportsFrozen && self->frozen)
+        luaL_error(L, "type.setindexer: cannot be called to mutate a frozen type, use `types.copy` to make a copy");
+
     TypeFunctionTypeId key = getTypeUserData(L, 2);
     TypeFunctionTypeId value = getTypeUserData(L, 3);
 
-    if (auto tfnt = get<TypeFunctionNeverType>(key))
+    if (get<TypeFunctionNeverType>(key))
     {
         tftt->indexer = std::nullopt;
         return 0;
@@ -768,6 +1016,9 @@ static int setTableMetatable(lua_State* L)
     if (!tftt)
         luaL_error(L, "type.setmetatable: expected self to be a table, but got %s instead", getTag(L, self).c_str());
 
+    if (FFlag::LuauTypeFunctionSupportsFrozen && self->frozen)
+        luaL_error(L, "type.setmetatable: cannot be called to mutate a frozen type, use `types.copy` to make a copy");
+
     TypeFunctionTypeId arg = getTypeUserData(L, 2);
     if (!get<TypeFunctionTableType>(arg))
         luaL_error(L, "type.setmetatable: expected the argument to be a table, but got %s instead", getTag(L, self).c_str());
@@ -872,12 +1123,12 @@ static TypeFunctionTypePackId getTypePack(lua_State* L, int headIdx, int tailIdx
     if (head.size() == 0 && tail.has_value())
         result = *tail;
     else
-        result = allocateTypeFunctionTypePack(L, TypeFunctionTypePack{head, tail});
+        result = allocateTypeFunctionTypePack(L, TypeFunctionTypePack{std::move(head), tail});
 
     return result;
 }
 
-static void pushTypePack(lua_State* L, TypeFunctionTypePackId tp)
+void pushTypePack(lua_State* L, TypeFunctionTypePackId tp)
 {
     if (auto tftp = get<TypeFunctionTypePack>(tp))
     {
@@ -997,6 +1248,9 @@ static int setFunctionParameters(lua_State* L)
     if (!tfft)
         luaL_error(L, "type.setparameters: expected self to be a function, but got %s instead", getTag(L, self).c_str());
 
+    if (FFlag::LuauTypeFunctionSupportsFrozen && self->frozen)
+        luaL_error(L, "type.setparameters: cannot be called to mutate a frozen type, use `types.copy` to make a copy");
+
     tfft->argTypes = getTypePack(L, 2, 3);
 
     return 0;
@@ -1033,6 +1287,9 @@ static int setFunctionReturns(lua_State* L)
     if (!tfft)
         luaL_error(L, "type.setreturns: expected self to be a function, but got %s instead", getTag(L, self).c_str());
 
+    if (FFlag::LuauTypeFunctionSupportsFrozen && self->frozen)
+        luaL_error(L, "type.setreturns: cannot be called to mutate a frozen type, use `types.copy` to make a copy");
+
     tfft->retTypes = getTypePack(L, 2, 3);
 
     return 0;
@@ -1064,6 +1321,9 @@ static int setFunctionGenerics(lua_State* L)
     if (!tfft)
         luaL_error(L, "type.setgenerics: expected self to be a function, but got %s instead", getTag(L, self).c_str());
 
+    if (FFlag::LuauTypeFunctionSupportsFrozen && self->frozen)
+        luaL_error(L, "type.setgenerics: cannot be called to mutate a frozen type, use `types.copy` to make a copy");
+
     int argumentCount = lua_gettop(L);
     if (argumentCount > 3)
         luaL_error(L, "type.setgenerics: expected 3 arguments, but got %d", argumentCount);
@@ -1105,28 +1365,6 @@ static int getFunctionGenerics(lua_State* L)
     return 1;
 }
 
-// Luau: `self:parent() -> type`
-// Returns the parent of a class type
-static int getClassParent_DEPRECATED(lua_State* L)
-{
-    int argumentCount = lua_gettop(L);
-    if (argumentCount != 1)
-        luaL_error(L, "type.parent: expected 1 arguments, but got %d", argumentCount);
-
-    TypeFunctionTypeId self = getTypeUserData(L, 1);
-    auto tfct = get<TypeFunctionClassType>(self);
-    if (!tfct)
-        luaL_error(L, "type.parent: expected self to be a class, but got %s instead", getTag(L, self).c_str());
-
-    // If the parent does not exist, we should return nil
-    if (!tfct->parent_DEPRECATED)
-        lua_pushnil(L);
-    else
-        allocTypeUserData(L, (*tfct->parent_DEPRECATED)->type);
-
-    return 1;
-}
-
 // Luau: `self:readparent() -> type`
 // Returns the read type of the class' parent
 static int getReadParent(lua_State* L)
@@ -1136,7 +1374,7 @@ static int getReadParent(lua_State* L)
         luaL_error(L, "type.parent: expected 1 arguments, but got %d", argumentCount);
 
     TypeFunctionTypeId self = getTypeUserData(L, 1);
-    auto tfct = get<TypeFunctionClassType>(self);
+    auto tfct = get<TypeFunctionExternType>(self);
     if (!tfct)
         luaL_error(L, "type.parent: expected self to be a class, but got %s instead", getTag(L, self).c_str());
 
@@ -1158,7 +1396,7 @@ static int getWriteParent(lua_State* L)
         luaL_error(L, "type.parent: expected 1 arguments, but got %d", argumentCount);
 
     TypeFunctionTypeId self = getTypeUserData(L, 1);
-    auto tfct = get<TypeFunctionClassType>(self);
+    auto tfct = get<TypeFunctionExternType>(self);
     if (!tfct)
         luaL_error(L, "type.parent: expected self to be a class, but got %s instead", getTag(L, self).c_str());
 
@@ -1242,7 +1480,7 @@ static int getProps(lua_State* L)
         return 1;
     }
 
-    if (auto tfct = get<TypeFunctionClassType>(self))
+    if (auto tfct = get<TypeFunctionExternType>(self))
     {
         lua_createtable(L, int(tfct->props.size()), 0);
         for (auto& [name, prop] : tfct->props)
@@ -1305,7 +1543,7 @@ static int getIndexer(lua_State* L)
         return 1;
     }
 
-    if (auto tfct = get<TypeFunctionClassType>(self))
+    if (auto tfct = get<TypeFunctionExternType>(self))
     {
         // if the indexer does not exist, we should return nil
         if (!tfct->indexer.has_value())
@@ -1353,7 +1591,7 @@ static int getReadIndexer(lua_State* L)
         return 1;
     }
 
-    if (auto tfct = get<TypeFunctionClassType>(self))
+    if (auto tfct = get<TypeFunctionExternType>(self))
     {
         // if the indexer does not exist, we should return nil
         if (!tfct->indexer.has_value())
@@ -1399,7 +1637,7 @@ static int getWriteIndexer(lua_State* L)
         return 1;
     }
 
-    if (auto tfct = get<TypeFunctionClassType>(self))
+    if (auto tfct = get<TypeFunctionExternType>(self))
     {
         // if the indexer does not exist, we should return nil
         if (!tfct->indexer.has_value())
@@ -1439,7 +1677,7 @@ static int getMetatable(lua_State* L)
         return 1;
     }
 
-    if (auto tfct = get<TypeFunctionClassType>(self))
+    if (auto tfct = get<TypeFunctionExternType>(self))
     {
         // if the metatable does not exist, we should return nil
         if (!tfct->metatable.has_value())
@@ -1519,6 +1757,7 @@ void registerTypesLibrary(lua_State* L)
         {"negationof", createNegation},
         {"unionof", createUnion},
         {"intersectionof", createIntersection},
+        {"optional", createOptional},
         {"newtable", createTable},
         {"newfunction", createFunction},
         {"copy", deepCopy},
@@ -1593,8 +1832,9 @@ void registerTypeUserData(lua_State* L)
         // Union and Intersection type methods
         {"components", getComponents},
 
-        // Class type methods
-        {FFlag::LuauTypeFunReadWriteParents ? "readparent" : "parent", FFlag::LuauTypeFunReadWriteParents ? getReadParent : getClassParent_DEPRECATED},
+        //  Extern type methods
+        {"readparent", getReadParent},
+        {"writeparent", getWriteParent},
 
         // Function type methods (cont.)
         {"setgenerics", setFunctionGenerics},
@@ -1604,9 +1844,6 @@ void registerTypeUserData(lua_State* L)
         {"name", getGenericName},
         {"ispack", getGenericIsPack},
 
-        // move this under Class type methods when removing FFlagLuauTypeFunReadWriteParents
-        {FFlag::LuauTypeFunReadWriteParents ? "writeparent" : nullptr, FFlag::LuauTypeFunReadWriteParents ? getWriteParent : nullptr},
-
         {nullptr, nullptr}
     };
 
@@ -1903,12 +2140,12 @@ bool areEqual(SeenSet& seen, const TypeFunctionFunctionType& lhs, const TypeFunc
     return true;
 }
 
-bool areEqual(SeenSet& seen, const TypeFunctionClassType& lhs, const TypeFunctionClassType& rhs)
+bool areEqual(SeenSet& seen, const TypeFunctionExternType& lhs, const TypeFunctionExternType& rhs)
 {
     if (seenSetContains(seen, &lhs, &rhs))
         return true;
 
-    return lhs.classTy == rhs.classTy;
+    return lhs.externTy == rhs.externTy;
 }
 
 bool areEqual(SeenSet& seen, const TypeFunctionType& lhs, const TypeFunctionType& rhs)
@@ -1976,8 +2213,8 @@ bool areEqual(SeenSet& seen, const TypeFunctionType& lhs, const TypeFunctionType
     }
 
     {
-        const TypeFunctionClassType* lf = get<TypeFunctionClassType>(&lhs);
-        const TypeFunctionClassType* rf = get<TypeFunctionClassType>(&rhs);
+        const TypeFunctionExternType* lf = get<TypeFunctionExternType>(&lhs);
+        const TypeFunctionExternType* rf = get<TypeFunctionExternType>(&rhs);
         if (lf && rf)
             return areEqual(seen, *lf, *rf);
     }
@@ -2085,6 +2322,11 @@ TypeFunctionProperty TypeFunctionProperty::rw(TypeFunctionTypeId read, TypeFunct
     return p;
 }
 
+bool TypeFunctionProperty::isShared() const
+{
+    return readTy && writeTy && readTy == writeTy;
+}
+
 bool TypeFunctionProperty::isReadOnly() const
 {
     return readTy && !writeTy;
@@ -2099,13 +2341,6 @@ bool TypeFunctionProperty::isWriteOnly() const
  * Below is a helper class for type.copy()
  * Forked version of Clone.cpp
  */
-using TypeFunctionKind = Variant<TypeFunctionTypeId, TypeFunctionTypePackId>;
-
-template<typename T>
-const T* get(const TypeFunctionKind& kind)
-{
-    return get_if<T>(&kind);
-}
 
 class TypeFunctionCloner
 {
@@ -2240,11 +2475,11 @@ private:
                 break;
             }
         }
-        else if (auto u = get<TypeFunctionUnknownType>(ty))
+        else if (get<TypeFunctionUnknownType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionUnknownType{});
-        else if (auto a = get<TypeFunctionNeverType>(ty))
+        else if (get<TypeFunctionNeverType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionNeverType{});
-        else if (auto a = get<TypeFunctionAnyType>(ty))
+        else if (get<TypeFunctionAnyType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionAnyType{});
         else if (auto s = get<TypeFunctionSingletonType>(ty))
         {
@@ -2253,20 +2488,20 @@ private:
             else if (auto ss = get<TypeFunctionStringSingleton>(s))
                 target = typeFunctionRuntime->typeArena.allocate(TypeFunctionSingletonType{TypeFunctionStringSingleton{ss->value}});
         }
-        else if (auto u = get<TypeFunctionUnionType>(ty))
+        else if (get<TypeFunctionUnionType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionUnionType{{}});
-        else if (auto i = get<TypeFunctionIntersectionType>(ty))
+        else if (get<TypeFunctionIntersectionType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionIntersectionType{{}});
-        else if (auto n = get<TypeFunctionNegationType>(ty))
+        else if (get<TypeFunctionNegationType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionNegationType{{}});
-        else if (auto t = get<TypeFunctionTableType>(ty))
+        else if (get<TypeFunctionTableType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionTableType{{}, std::nullopt, std::nullopt});
-        else if (auto f = get<TypeFunctionFunctionType>(ty))
+        else if (get<TypeFunctionFunctionType>(ty))
         {
             TypeFunctionTypePackId emptyTypePack = typeFunctionRuntime->typePackArena.allocate(TypeFunctionTypePack{});
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionFunctionType{{}, {}, emptyTypePack, emptyTypePack});
         }
-        else if (auto c = get<TypeFunctionClassType>(ty))
+        else if (get<TypeFunctionExternType>(ty))
             target = ty; // Don't copy a class since they are immutable
         else if (auto g = get<TypeFunctionGenericType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionGenericType{g->isNamed, g->isPack, g->name});
@@ -2285,9 +2520,9 @@ private:
 
         // Create a shallow serialization
         TypeFunctionTypePackId target = {};
-        if (auto tPack = get<TypeFunctionTypePack>(tp))
+        if (get<TypeFunctionTypePack>(tp))
             target = typeFunctionRuntime->typePackArena.allocate(TypeFunctionTypePack{{}});
-        else if (auto vPack = get<TypeFunctionVariadicTypePack>(tp))
+        else if (get<TypeFunctionVariadicTypePack>(tp))
             target = typeFunctionRuntime->typePackArena.allocate(TypeFunctionVariadicTypePack{});
         else if (auto gPack = get<TypeFunctionGenericTypePack>(tp))
             target = typeFunctionRuntime->typePackArena.allocate(TypeFunctionGenericTypePack{gPack->isNamed, gPack->name});
@@ -2321,7 +2556,7 @@ private:
             cloneChildren(t1, t2);
         else if (auto [f1, f2] = std::tuple{getMutable<TypeFunctionFunctionType>(ty), getMutable<TypeFunctionFunctionType>(tfti)}; f1 && f2)
             cloneChildren(f1, f2);
-        else if (auto [c1, c2] = std::tuple{getMutable<TypeFunctionClassType>(ty), getMutable<TypeFunctionClassType>(tfti)}; c1 && c2)
+        else if (auto [c1, c2] = std::tuple{getMutable<TypeFunctionExternType>(ty), getMutable<TypeFunctionExternType>(tfti)}; c1 && c2)
             cloneChildren(c1, c2);
         else if (auto [g1, g2] = std::tuple{getMutable<TypeFunctionGenericType>(ty), getMutable<TypeFunctionGenericType>(tfti)}; g1 && g2)
             cloneChildren(g1, g2);
@@ -2431,7 +2666,7 @@ private:
         f2->retTypes = shallowClone(f1->retTypes);
     }
 
-    void cloneChildren(TypeFunctionClassType* c1, TypeFunctionClassType* c2)
+    void cloneChildren(TypeFunctionExternType* c1, TypeFunctionExternType* c2)
     {
         // noop.
     }
