@@ -2,7 +2,6 @@
 
 #include "Luau/TypeFunctionRuntimeBuilder.h"
 
-#include "Luau/Ast.h"
 #include "Luau/BuiltinDefinitions.h"
 #include "Luau/Common.h"
 #include "Luau/DenseHash.h"
@@ -12,6 +11,7 @@
 #include "Luau/TypeFwd.h"
 #include "Luau/TypeFunctionRuntime.h"
 #include "Luau/TypePack.h"
+#include "Luau/TypeOrPack.h"
 #include "Luau/ToString.h"
 
 #include <optional>
@@ -19,7 +19,8 @@
 // used to control the recursion limit of any operations done by user-defined type functions
 // currently, controls serialization, deserialization, and `type.copy`
 LUAU_DYNAMIC_FASTINTVARIABLE(LuauTypeFunctionSerdeIterationLimit, 100'000);
-LUAU_FASTFLAG(LuauTypeFunReadWriteParents)
+
+LUAU_FASTFLAGVARIABLE(LuauTypeFunctionDeserializationShouldNotCrashOnGenericPacks)
 
 namespace Luau
 {
@@ -40,7 +41,7 @@ class TypeFunctionSerializer
     // queue.back() should always return two of same type in their respective sides
     // For example `auto [first, second] = queue.back()`: if first is PrimitiveType,
     // second must be TypeFunctionPrimitiveType; else there should be an error
-    std::vector<std::tuple<Kind, TypeFunctionKind>> queue;
+    std::vector<std::tuple<TypeOrPack, TypeFunctionKind>> queue;
 
     SeenTypes types;     // Mapping of TypeIds that have been shallow serialized to TypeFunctionTypeIds
     SeenTypePacks packs; // Mapping of TypePackIds that have been shallow serialized to TypeFunctionTypePackIds
@@ -120,7 +121,7 @@ private:
         return std::nullopt;
     }
 
-    std::optional<TypeFunctionKind> find(Kind kind) const
+    std::optional<TypeFunctionKind> find(TypeOrPack kind) const
     {
         if (auto ty = get<TypeId>(kind))
             return find(*ty);
@@ -173,11 +174,11 @@ private:
             }
             }
         }
-        else if (auto u = get<UnknownType>(ty))
+        else if (get<UnknownType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionUnknownType{});
-        else if (auto a = get<NeverType>(ty))
+        else if (get<NeverType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionNeverType{});
-        else if (auto a = get<AnyType>(ty))
+        else if (get<AnyType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionAnyType{});
         else if (auto s = get<SingletonType>(ty))
         {
@@ -191,28 +192,26 @@ private:
                 state->errors.push_back(error);
             }
         }
-        else if (auto u = get<UnionType>(ty))
+        else if (get<UnionType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionUnionType{{}});
-        else if (auto i = get<IntersectionType>(ty))
+        else if (get<IntersectionType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionIntersectionType{{}});
-        else if (auto n = get<NegationType>(ty))
+        else if (get<NegationType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionNegationType{{}});
-        else if (auto t = get<TableType>(ty))
+        else if (get<TableType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionTableType{{}, std::nullopt, std::nullopt});
-        else if (auto m = get<MetatableType>(ty))
+        else if (get<MetatableType>(ty))
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionTableType{{}, std::nullopt, std::nullopt});
-        else if (auto f = get<FunctionType>(ty))
+        else if (get<FunctionType>(ty))
         {
             TypeFunctionTypePackId emptyTypePack = typeFunctionRuntime->typePackArena.allocate(TypeFunctionTypePack{});
             target = typeFunctionRuntime->typeArena.allocate(TypeFunctionFunctionType{{}, {}, emptyTypePack, emptyTypePack});
         }
-        else if (auto c = get<ClassType>(ty))
+        else if (get<ExternType>(ty))
         {
             // Since there aren't any new class types being created in type functions, we will deserialize by using a direct reference to the original
             // class
-            target = typeFunctionRuntime->typeArena.allocate(
-                TypeFunctionClassType{{}, std::nullopt, std::nullopt, std::nullopt, std::nullopt, std::nullopt, ty}
-            );
+            target = typeFunctionRuntime->typeArena.allocate(TypeFunctionExternType{{}, std::nullopt, std::nullopt, std::nullopt, std::nullopt, ty});
         }
         else if (auto g = get<GenericType>(ty))
         {
@@ -221,7 +220,7 @@ private:
             if (!g->explicitName)
                 name = format("g%d", g->index);
 
-            target = typeFunctionRuntime->typeArena.allocate(TypeFunctionGenericType{g->explicitName, false, name});
+            target = typeFunctionRuntime->typeArena.allocate(TypeFunctionGenericType{g->explicitName, false, std::move(name)});
         }
         else
         {
@@ -243,9 +242,9 @@ private:
 
         // Create a shallow serialization
         TypeFunctionTypePackId target = {};
-        if (auto tPack = get<TypePack>(tp))
+        if (get<TypePack>(tp))
             target = typeFunctionRuntime->typePackArena.allocate(TypeFunctionTypePack{{}});
-        else if (auto vPack = get<VariadicTypePack>(tp))
+        else if (get<VariadicTypePack>(tp))
             target = typeFunctionRuntime->typePackArena.allocate(TypeFunctionVariadicTypePack{});
         else if (auto gPack = get<GenericTypePack>(tp))
         {
@@ -254,7 +253,7 @@ private:
             if (!gPack->explicitName)
                 name = format("g%d", gPack->index);
 
-            target = typeFunctionRuntime->typePackArena.allocate(TypeFunctionGenericTypePack{gPack->explicitName, name});
+            target = typeFunctionRuntime->typePackArena.allocate(TypeFunctionGenericTypePack{gPack->explicitName, std::move(name)});
         }
         else
         {
@@ -291,7 +290,7 @@ private:
             serializeChildren(m1, m2);
         else if (auto [f1, f2] = std::tuple{get<FunctionType>(ty), getMutable<TypeFunctionFunctionType>(tfti)}; f1 && f2)
             serializeChildren(f1, f2);
-        else if (auto [c1, c2] = std::tuple{get<ClassType>(ty), getMutable<TypeFunctionClassType>(tfti)}; c1 && c2)
+        else if (auto [c1, c2] = std::tuple{get<ExternType>(ty), getMutable<TypeFunctionExternType>(tfti)}; c1 && c2)
             serializeChildren(c1, c2);
         else if (auto [g1, g2] = std::tuple{get<GenericType>(ty), getMutable<TypeFunctionGenericType>(tfti)}; g1 && g2)
             serializeChildren(g1, g2);
@@ -317,7 +316,7 @@ private:
         }
     }
 
-    void serializeChildren(Kind kind, TypeFunctionKind tfkind)
+    void serializeChildren(TypeOrPack kind, TypeFunctionKind tfkind)
     {
         if (auto [ty, tfty] = std::tuple{get<TypeId>(kind), get<TypeFunctionTypeId>(tfkind)}; ty && tfty)
             serializeChildren(*ty, *tfty);
@@ -391,7 +390,7 @@ private:
     void serializeChildren(const MetatableType* m1, TypeFunctionTableType* m2)
     {
         // Serialize main part of the metatable immediately
-        if (auto tableTy = get<TableType>(m1->table))
+        if (auto tableTy = get<TableType>(follow(m1->table)))
             serializeChildren(tableTy, m2);
 
         m2->metatable = shallowSerialize(m1->metatable);
@@ -411,7 +410,7 @@ private:
         f2->retTypes = shallowSerialize(f1->retTypes);
     }
 
-    void serializeChildren(const ClassType* c1, TypeFunctionClassType* c2)
+    void serializeChildren(const ExternType* c1, TypeFunctionExternType* c2)
     {
         for (const auto& [k, p] : c1->props)
         {
@@ -436,16 +435,9 @@ private:
         {
             TypeFunctionTypeId parent = shallowSerialize(*c1->parent);
 
-            if (FFlag::LuauTypeFunReadWriteParents)
-            {
-                // we don't yet have read/write parents in the type inference engine.
-                c2->readParent = parent;
-                c2->writeParent = parent;
-            }
-            else
-            {
-                c2->parent_DEPRECATED = parent;
-            }
+            // we don't yet have read/write parents in the type inference engine.
+            c2->readParent = parent;
+            c2->writeParent = parent;
         }
     }
 
@@ -480,12 +472,35 @@ struct SerializedGeneric
     bool isNamed = false;
     std::string name;
     T type = nullptr;
+
+    explicit SerializedGeneric(std::string name)
+        : name(std::move(name))
+    {
+    }
+
+    SerializedGeneric(bool isNamed, std::string name, T type)
+        : isNamed(isNamed)
+        , name(std::move(name))
+        , type(std::move(type))
+    {
+    }
 };
 
 struct SerializedFunctionScope
 {
     size_t oldQueueSize = 0;
     TypeFunctionFunctionType* function = nullptr;
+
+    explicit SerializedFunctionScope(size_t oldQueueSize)
+        : oldQueueSize(oldQueueSize)
+    {
+    }
+
+    SerializedFunctionScope(size_t oldQueueSize, TypeFunctionFunctionType* function)
+        : oldQueueSize(oldQueueSize)
+        , function(function)
+    {
+    }
 };
 
 // Complete inverse of TypeFunctionSerializer
@@ -504,7 +519,7 @@ class TypeFunctionDeserializer
     // queue.back() should always return two of same type in their respective sides
     // For example `auto [first, second] = queue.back()`: if first is TypeFunctionPrimitiveType,
     // second must be PrimitiveType; else there should be an error
-    std::vector<std::tuple<TypeFunctionKind, Kind>> queue;
+    std::vector<std::tuple<TypeFunctionKind, TypeOrPack>> queue;
 
     // Generic types and packs currently in scope
     // Generics are resolved by name even if runtime generic type pointers are different
@@ -537,12 +552,12 @@ public:
 
         if (hasExceededIterationLimit() || state->errors.size() != 0)
         {
-            TypeId error = state->ctx->builtins->errorRecoveryType();
+            TypeId error = state->ctx->builtins->errorType;
             types[ty] = error;
             return error;
         }
 
-        return find(ty).value_or(state->ctx->builtins->errorRecoveryType());
+        return find(ty).value_or(state->ctx->builtins->errorType);
     }
 
     TypePackId deserialize(TypeFunctionTypePackId tp)
@@ -552,12 +567,12 @@ public:
 
         if (hasExceededIterationLimit() || state->errors.size() != 0)
         {
-            TypePackId error = state->ctx->builtins->errorRecoveryTypePack();
+            TypePackId error = state->ctx->builtins->errorTypePack;
             packs[tp] = error;
             return error;
         }
 
-        return find(tp).value_or(state->ctx->builtins->errorRecoveryTypePack());
+        return find(tp).value_or(state->ctx->builtins->errorTypePack);
     }
 
 private:
@@ -608,7 +623,7 @@ private:
         return std::nullopt;
     }
 
-    std::optional<Kind> find(TypeFunctionKind kind) const
+    std::optional<TypeOrPack> find(TypeFunctionKind kind) const
     {
         if (auto ty = get<TypeFunctionTypeId>(kind))
             return find(*ty);
@@ -669,11 +684,11 @@ private:
                 state->ctx->ice->ice("Deserializing user defined type function arguments: mysterious type is being deserialized");
             }
         }
-        else if (auto u = get<TypeFunctionUnknownType>(ty))
+        else if (get<TypeFunctionUnknownType>(ty))
             target = state->ctx->builtins->unknownType;
-        else if (auto n = get<TypeFunctionNeverType>(ty))
+        else if (get<TypeFunctionNeverType>(ty))
             target = state->ctx->builtins->neverType;
-        else if (auto a = get<TypeFunctionAnyType>(ty))
+        else if (get<TypeFunctionAnyType>(ty))
             target = state->ctx->builtins->anyType;
         else if (auto s = get<TypeFunctionSingletonType>(ty))
         {
@@ -684,11 +699,11 @@ private:
             else
                 state->ctx->ice->ice("Deserializing user defined type function arguments: mysterious type is being deserialized");
         }
-        else if (auto u = get<TypeFunctionUnionType>(ty))
+        else if (get<TypeFunctionUnionType>(ty))
             target = state->ctx->arena->addTV(Type(UnionType{{}}));
-        else if (auto i = get<TypeFunctionIntersectionType>(ty))
+        else if (get<TypeFunctionIntersectionType>(ty))
             target = state->ctx->arena->addTV(Type(IntersectionType{{}}));
-        else if (auto n = get<TypeFunctionNegationType>(ty))
+        else if (get<TypeFunctionNegationType>(ty))
             target = state->ctx->arena->addType(NegationType{state->ctx->builtins->unknownType});
         else if (auto t = get<TypeFunctionTableType>(ty); t && !t->metatable.has_value())
             target = state->ctx->arena->addType(TableType{TableType::Props{}, std::nullopt, TypeLevel{}, TableState::Sealed});
@@ -697,14 +712,14 @@ private:
             TypeId emptyTable = state->ctx->arena->addType(TableType{TableType::Props{}, std::nullopt, TypeLevel{}, TableState::Sealed});
             target = state->ctx->arena->addType(MetatableType{emptyTable, emptyTable});
         }
-        else if (auto f = get<TypeFunctionFunctionType>(ty))
+        else if (get<TypeFunctionFunctionType>(ty))
         {
             TypePackId emptyTypePack = state->ctx->arena->addTypePack(TypePack{});
             target = state->ctx->arena->addType(FunctionType{emptyTypePack, emptyTypePack, {}, false});
         }
-        else if (auto c = get<TypeFunctionClassType>(ty))
+        else if (auto c = get<TypeFunctionExternType>(ty))
         {
-            target = c->classTy;
+            target = c->externTy;
         }
         else if (auto g = get<TypeFunctionGenericType>(ty))
         {
@@ -748,11 +763,11 @@ private:
 
         // Create a shallow deserialization
         TypePackId target = {};
-        if (auto tPack = get<TypeFunctionTypePack>(tp))
+        if (get<TypeFunctionTypePack>(tp))
         {
             target = state->ctx->arena->addTypePack(TypePack{});
         }
-        else if (auto vPack = get<TypeFunctionVariadicTypePack>(tp))
+        else if (get<TypeFunctionVariadicTypePack>(tp))
         {
             target = state->ctx->arena->addTypePack(VariadicTypePack{});
         }
@@ -811,7 +826,7 @@ private:
             deserializeChildren(m2, m1);
         else if (auto [f1, f2] = std::tuple{getMutable<FunctionType>(ty), getMutable<TypeFunctionFunctionType>(tfti)}; f1 && f2)
             deserializeChildren(f2, f1);
-        else if (auto [c1, c2] = std::tuple{getMutable<ClassType>(ty), getMutable<TypeFunctionClassType>(tfti)}; c1 && c2)
+        else if (auto [c1, c2] = std::tuple{getMutable<ExternType>(ty), getMutable<TypeFunctionExternType>(tfti)}; c1 && c2)
             deserializeChildren(c2, c1);
         else if (auto [g1, g2] = std::tuple{getMutable<GenericType>(ty), getMutable<TypeFunctionGenericType>(tfti)}; g1 && g2)
             deserializeChildren(g2, g1);
@@ -832,7 +847,7 @@ private:
             state->ctx->ice->ice("Deserializing user defined type function arguments: mysterious type is being deserialized");
     }
 
-    void deserializeChildren(TypeFunctionKind tfkind, Kind kind)
+    void deserializeChildren(TypeFunctionKind tfkind, TypeOrPack kind)
     {
         if (auto [ty, tfty] = std::tuple{get<TypeId>(kind), get<TypeFunctionTypeId>(tfkind)}; ty && tfty)
             deserializeChildren(*tfty, *ty);
@@ -911,15 +926,20 @@ private:
 
     void deserializeChildren(TypeFunctionFunctionType* f2, FunctionType* f1)
     {
-        functionScopes.push_back({queue.size(), f2});
-
+        functionScopes.emplace_back(queue.size(), f2);
         std::set<std::pair<bool, std::string>> genericNames;
 
         // Introduce generic function parameters into scope
         for (auto ty : f2->generics)
         {
             auto gty = get<TypeFunctionGenericType>(ty);
-            LUAU_ASSERT(gty && !gty->isPack);
+            if (FFlag::LuauTypeFunctionDeserializationShouldNotCrashOnGenericPacks && (!gty || gty->isPack))
+            {
+                state->errors.emplace_back("Encountered unexpected generic");
+                return;
+            }
+            else
+                LUAU_ASSERT(gty && !gty->isPack);
 
             std::pair<bool, std::string> nameKey = std::make_pair(gty->isNamed, gty->name);
 
@@ -933,13 +953,19 @@ private:
             genericNames.insert(nameKey);
 
             TypeId mapping = state->ctx->arena->addTV(Type(gty->isNamed ? GenericType{state->ctx->scope.get(), gty->name} : GenericType{}));
-            genericTypes.push_back({gty->isNamed, gty->name, mapping});
+            genericTypes.emplace_back(gty->isNamed, gty->name, mapping);
         }
 
         for (auto tp : f2->genericPacks)
         {
             auto gtp = get<TypeFunctionGenericTypePack>(tp);
-            LUAU_ASSERT(gtp);
+            if (FFlag::LuauTypeFunctionDeserializationShouldNotCrashOnGenericPacks && !gtp)
+            {
+                state->errors.emplace_back("Encountered unexpected generic type pack");
+                return;
+            }
+            else
+                LUAU_ASSERT(gtp);
 
             std::pair<bool, std::string> nameKey = std::make_pair(gtp->isNamed, gtp->name);
 
@@ -954,7 +980,7 @@ private:
 
             TypePackId mapping =
                 state->ctx->arena->addTypePack(TypePackVar(gtp->isNamed ? GenericTypePack{state->ctx->scope.get(), gtp->name} : GenericTypePack{}));
-            genericPacks.push_back({gtp->isNamed, gtp->name, mapping});
+            genericPacks.emplace_back(gtp->isNamed, gtp->name, mapping);
         }
 
         f1->generics.reserve(f2->generics.size());
@@ -972,7 +998,7 @@ private:
             f1->retTypes = shallowDeserialize(f2->retTypes);
     }
 
-    void deserializeChildren(TypeFunctionClassType* c2, ClassType* c1)
+    void deserializeChildren(TypeFunctionExternType* c2, ExternType* c1)
     {
         // noop.
     }
@@ -1007,9 +1033,19 @@ TypeFunctionTypeId serialize(TypeId ty, TypeFunctionRuntimeBuilderState* state)
     return TypeFunctionSerializer(state).serialize(ty);
 }
 
+TypeFunctionTypePackId serialize(TypePackId tp, TypeFunctionRuntimeBuilderState* state)
+{
+    return TypeFunctionSerializer(state).serialize(tp);
+}
+
 TypeId deserialize(TypeFunctionTypeId ty, TypeFunctionRuntimeBuilderState* state)
 {
     return TypeFunctionDeserializer(state).deserialize(ty);
 }
 
+TypePackId deserialize(TypeFunctionTypePackId tp, TypeFunctionRuntimeBuilderState* state)
+{
+    return TypeFunctionDeserializer(state).deserialize(tp);
+}
+
 } // namespace Luau
