@@ -1,9 +1,13 @@
 // This file is part of the Luau programming language and is licensed under MIT License; see LICENSE.txt for details
 
 #include "Luau/TypePath.h"
+
+#include "Luau/Anyification.h"
 #include "Luau/Common.h"
 #include "Luau/DenseHash.h"
+#include "Luau/ToString.h"
 #include "Luau/Type.h"
+#include "Luau/TypeArena.h"
 #include "Luau/TypeFwd.h"
 #include "Luau/TypePack.h"
 #include "Luau/TypeOrPack.h"
@@ -11,10 +15,10 @@
 #include <functional>
 #include <optional>
 #include <sstream>
-#include <type_traits>
 
-LUAU_FASTFLAG(LuauSolverV2);
-LUAU_FASTFLAGVARIABLE(LuauDisableNewSolverAssertsInMixedMode)
+LUAU_FASTFLAG(LuauSolverV2)
+
+LUAU_FASTFLAG(LuauSubtypingMissingPropertiesAsNil)
 
 // Maximum number of steps to follow when traversing a path. May not always
 // equate to the number of components in a path, depending on the traversal
@@ -30,7 +34,6 @@ namespace TypePath
 Property::Property(std::string name)
     : name(std::move(name))
 {
-    LUAU_ASSERT(!FFlag::LuauSolverV2);
 }
 
 Property Property::read(std::string name)
@@ -53,11 +56,21 @@ bool Index::operator==(const Index& other) const
     return index == other.index;
 }
 
+bool PackSlice::operator==(const PackSlice& other) const
+{
+    return start_index == other.start_index;
+}
+
 bool Reduction::operator==(const Reduction& other) const
 {
     return resultType == other.resultType;
 }
 
+bool GenericPackMapping::operator==(const GenericPackMapping& other) const
+{
+    return mappedType == other.mappedType;
+}
+
 Path Path::append(const Path& suffix) const
 {
     std::vector<Component> joined(components);
@@ -130,11 +143,21 @@ size_t PathHash::operator()(const PackField& field) const
     return static_cast<size_t>(field);
 }
 
+size_t PathHash::operator()(const PackSlice& slice) const
+{
+    return slice.start_index;
+}
+
 size_t PathHash::operator()(const Reduction& reduction) const
 {
     return std::hash<TypeId>()(reduction.resultType);
 }
 
+size_t PathHash::operator()(const GenericPackMapping& mapping) const
+{
+    return std::hash<TypePackId>()(mapping.mappedType);
+}
+
 size_t PathHash::operator()(const Component& component) const
 {
     return visit(*this, component);
@@ -157,90 +180,97 @@ Path PathBuilder::build()
 
 PathBuilder& PathBuilder::readProp(std::string name)
 {
-    if (!FFlag::LuauDisableNewSolverAssertsInMixedMode)
-        LUAU_ASSERT(FFlag::LuauSolverV2);
-    components.push_back(Property{std::move(name), true});
+    components.emplace_back(Property{std::move(name), true});
     return *this;
 }
 
 PathBuilder& PathBuilder::writeProp(std::string name)
 {
-    if (!FFlag::LuauDisableNewSolverAssertsInMixedMode)
-        LUAU_ASSERT(FFlag::LuauSolverV2);
-    components.push_back(Property{std::move(name), false});
+    components.emplace_back(Property{std::move(name), false});
     return *this;
 }
 
 PathBuilder& PathBuilder::prop(std::string name)
 {
-    LUAU_ASSERT(!FFlag::LuauSolverV2);
-    components.push_back(Property{std::move(name)});
+    components.emplace_back(Property{std::move(name)});
     return *this;
 }
 
 PathBuilder& PathBuilder::index(size_t i)
 {
-    components.push_back(Index{i});
+    components.emplace_back(Index{i});
     return *this;
 }
 
 PathBuilder& PathBuilder::mt()
 {
-    components.push_back(TypeField::Metatable);
+    components.emplace_back(TypeField::Metatable);
     return *this;
 }
 
 PathBuilder& PathBuilder::lb()
 {
-    components.push_back(TypeField::LowerBound);
+    components.emplace_back(TypeField::LowerBound);
     return *this;
 }
 
 PathBuilder& PathBuilder::ub()
 {
-    components.push_back(TypeField::UpperBound);
+    components.emplace_back(TypeField::UpperBound);
     return *this;
 }
 
 PathBuilder& PathBuilder::indexKey()
 {
-    components.push_back(TypeField::IndexLookup);
+    components.emplace_back(TypeField::IndexLookup);
     return *this;
 }
 
 PathBuilder& PathBuilder::indexValue()
 {
-    components.push_back(TypeField::IndexResult);
+    components.emplace_back(TypeField::IndexResult);
     return *this;
 }
 
 PathBuilder& PathBuilder::negated()
 {
-    components.push_back(TypeField::Negated);
+    components.emplace_back(TypeField::Negated);
     return *this;
 }
 
 PathBuilder& PathBuilder::variadic()
 {
-    components.push_back(TypeField::Variadic);
+    components.emplace_back(TypeField::Variadic);
     return *this;
 }
 
 PathBuilder& PathBuilder::args()
 {
-    components.push_back(PackField::Arguments);
+    components.emplace_back(PackField::Arguments);
     return *this;
 }
 
 PathBuilder& PathBuilder::rets()
 {
-    components.push_back(PackField::Returns);
+    components.emplace_back(PackField::Returns);
     return *this;
 }
 
 PathBuilder& PathBuilder::tail()
 {
-    components.push_back(PackField::Tail);
+    components.emplace_back(PackField::Tail);
+    return *this;
+}
+
+PathBuilder& PathBuilder::packSlice(size_t start_index)
+{
+    components.emplace_back(PackSlice{start_index});
+    return *this;
+}
+
+PathBuilder& PathBuilder::mappedGenericPack(TypePackId mappedType)
+{
+    components.emplace_back(GenericPackMapping{mappedType});
     return *this;
 }
 
@@ -251,20 +281,25 @@ namespace
 
 struct TraversalState
 {
-    TraversalState(TypeId root, NotNull<BuiltinTypes> builtinTypes)
+    TraversalState(TypeId root, const NotNull<BuiltinTypes> builtinTypes, TypeArena* arena)
         : current(root)
         , builtinTypes(builtinTypes)
+        , arena(arena)
     {
     }
-    TraversalState(TypePackId root, NotNull<BuiltinTypes> builtinTypes)
+
+    TraversalState(TypePackId root, NotNull<BuiltinTypes> builtinTypes, TypeArena* arena)
         : current(root)
         , builtinTypes(builtinTypes)
+        , arena(arena)
     {
     }
 
     TypeOrPack current;
     NotNull<BuiltinTypes> builtinTypes;
+    NotNull<TypeArena> arena;
     int steps = 0;
+    bool encounteredErrorSuppression = false;
 
     void updateCurrent(TypeId ty)
     {
@@ -307,9 +342,9 @@ struct TraversalState
                 prop = &it->second;
             }
         }
-        else if (auto c = get<ClassType>(*currentType))
+        else if (auto c = get<ExternType>(*currentType))
         {
-            prop = lookupClassProp(c, property.name);
+            prop = lookupExternTypeProp(c, property.name);
         }
         // For a metatable type, the table takes priority; check that before
         // falling through to the metatable entry below.
@@ -349,7 +384,7 @@ struct TraversalState
             if (FFlag::LuauSolverV2)
                 maybeType = property.isRead ? prop->readTy : prop->writeTy;
             else
-                maybeType = prop->type();
+                maybeType = prop->type_DEPRECATED();
 
             if (maybeType)
             {
@@ -368,26 +403,52 @@ struct TraversalState
 
         if (auto currentType = get<TypeId>(current))
         {
+            bool updatedCurrent = false;
+
+            if (get<ErrorType>(*currentType))
+            {
+                encounteredErrorSuppression = true;
+                return false;
+            }
+
             if (auto u = get<UnionType>(*currentType))
             {
                 auto it = begin(u);
-                std::advance(it, index.index);
-                if (it != end(u))
+                // We want to track the index that updates the current type with `idx` while still iterating through the entire union to check for
+                // error types with `it`.
+                size_t idx = 0;
+                for (auto it = begin(u); it != end(u); ++it)
                 {
-                    updateCurrent(*it);
-                    return true;
+                    if (get<ErrorType>(*it))
+                        encounteredErrorSuppression = true;
+                    if (idx == index.index)
+                    {
+                        updateCurrent(*it);
+                        updatedCurrent = true;
+                    }
+                    ++idx;
                 }
             }
             else if (auto i = get<IntersectionType>(*currentType))
             {
                 auto it = begin(i);
-                std::advance(it, index.index);
-                if (it != end(i))
+                // We want to track the index that updates the current type with `idx` while still iterating through the entire intersection to check
+                // for error types with `it`.
+                size_t idx = 0;
+                for (auto it = begin(i); it != end(i); ++it)
                 {
-                    updateCurrent(*it);
-                    return true;
+                    if (get<ErrorType>(*it))
+                        encounteredErrorSuppression = true;
+                    if (idx == index.index)
+                    {
+                        updateCurrent(*it);
+                        updatedCurrent = true;
+                    }
+                    ++idx;
                 }
             }
+
+            return updatedCurrent;
         }
         else
         {
@@ -461,7 +522,7 @@ struct TraversalState
                     indexer = &(*mtMt->indexer);
             }
             // Note: we don't appear to walk the class hierarchy for indexers
-            else if (auto ct = get<ClassType>(current); ct && ct->indexer)
+            else if (auto ct = get<ExternType>(current); ct && ct->indexer)
                 indexer = &(*ct->indexer);
 
             if (indexer)
@@ -536,6 +597,50 @@ struct TraversalState
 
         return false;
     }
+
+    bool traverse(const TypePath::PackSlice slice)
+    {
+        if (checkInvariants())
+            return false;
+
+        const TypePackId* currentPack = get<TypePackId>(current);
+        if (!currentPack)
+            return false;
+
+        auto [flatHead, flatTail] = flatten(*currentPack);
+
+        if (flatHead.size() <= slice.start_index)
+            return false;
+
+        std::vector<TypeId> headSlice;
+        headSlice.reserve(flatHead.size() - slice.start_index);
+
+        auto headIter = begin(flatHead);
+        for (size_t i = 0; i < slice.start_index && headIter != end(flatHead); ++i)
+            ++headIter;
+
+        while (headIter != end(flatHead))
+        {
+            headSlice.push_back(*headIter);
+            ++headIter;
+        }
+
+        TypePackId packSlice = arena->addTypePack(headSlice, flatTail);
+
+        updateCurrent(packSlice);
+
+        return true;
+    }
+
+    bool traverse(const TypePath::GenericPackMapping mapping)
+    {
+        if (checkInvariants())
+            return false;
+
+        updateCurrent(mapping.mappedType);
+
+        return true;
+    }
 };
 
 } // namespace
@@ -619,12 +724,16 @@ std::string toString(const TypePath::Path& path, bool prefixDot)
             }
             result << "()";
         }
+        else if constexpr (std::is_same_v<T, TypePath::PackSlice>)
+            result << "[" << std::to_string(c.start_index) << ":]";
         else if constexpr (std::is_same_v<T, TypePath::Reduction>)
         {
             // We need to rework the TypePath system to make subtyping failures easier to understand
             // https://roblox.atlassian.net/browse/CLI-104422
             result << "~~>";
         }
+        else if constexpr (std::is_same_v<T, TypePath::GenericPackMapping>)
+            result << "~";
         else
         {
             static_assert(always_false_v<T>, "Unhandled Component variant");
@@ -641,8 +750,6 @@ std::string toString(const TypePath::Path& path, bool prefixDot)
 
 std::string toStringHuman(const TypePath::Path& path)
 {
-    LUAU_ASSERT(FFlag::LuauSolverV2);
-
     enum class State
     {
         Initial,
@@ -688,8 +795,6 @@ std::string toStringHuman(const TypePath::Path& path)
         }
         else if constexpr (std::is_same_v<T, TypePath::Index>)
         {
-            size_t humanIndex = c.index + 1;
-
             if (state == State::Initial && !last)
                 result << "in" << ' ';
             else if (state == State::PendingIs)
@@ -697,21 +802,7 @@ std::string toStringHuman(const TypePath::Path& path)
             else if (state == State::Property)
                 result << '`' << ' ' << "has" << ' ';
 
-            result << "the " << humanIndex;
-            switch (humanIndex)
-            {
-            case 1:
-                result << "st";
-                break;
-            case 2:
-                result << "nd";
-                break;
-            case 3:
-                result << "rd";
-                break;
-            default:
-                result << "th";
-            }
+            result << "the " << toHumanReadableIndex(c.index);
 
             switch (c.variant)
             {
@@ -834,6 +925,8 @@ std::string toStringHuman(const TypePath::Path& path)
                 state = State::Normal;
             }
         }
+        else if constexpr (std::is_same_v<T, TypePath::PackSlice>)
+            result << "the portion of the type pack starting at index " << c.start_index << " to the end";
         else if constexpr (std::is_same_v<T, TypePath::Reduction>)
         {
             if (state == State::Initial)
@@ -841,6 +934,8 @@ std::string toStringHuman(const TypePath::Path& path)
             result << "reduces to" << ' ';
             state = State::Normal;
         }
+        else if constexpr (std::is_same_v<T, TypePath::GenericPackMapping>)
+            result << "is a generic pack mapped to ";
         else
         {
             static_assert(always_false_v<T>, "Unhandled Component variant");
@@ -897,70 +992,219 @@ static bool traverse(TraversalState& state, const Path& path)
     return true;
 }
 
-std::optional<TypeOrPack> traverse(TypeId root, const Path& path, NotNull<BuiltinTypes> builtinTypes)
+std::optional<TypeOrPack> traverse(const TypePackId root, const Path& path, const NotNull<BuiltinTypes> builtinTypes, const NotNull<TypeArena> arena)
 {
-    TraversalState state(follow(root), builtinTypes);
+    TraversalState state(follow(root), builtinTypes, arena);
+    if (traverse(state, path))
+    {
+        if (state.encounteredErrorSuppression)
+            return builtinTypes->errorType;
+        return state.current;
+    }
+    else
+        return std::nullopt;
+}
+
+std::optional<TypeOrPack> traverse(const TypeId root, const Path& path, const NotNull<BuiltinTypes> builtinTypes, const NotNull<TypeArena> arena)
+{
+    TraversalState state(follow(root), builtinTypes, arena);
     if (traverse(state, path))
         return state.current;
     else
         return std::nullopt;
 }
 
-std::optional<TypeOrPack> traverse(TypePackId root, const Path& path, NotNull<BuiltinTypes> builtinTypes)
+std::optional<TypeId> traverseForType(const TypeId root, const Path& path, const NotNull<BuiltinTypes> builtinTypes, const NotNull<TypeArena> arena)
 {
-    TraversalState state(follow(root), builtinTypes);
-    if (traverse(state, path))
-        return state.current;
-    else
-        return std::nullopt;
-}
-
-std::optional<TypeId> traverseForType(TypeId root, const Path& path, NotNull<BuiltinTypes> builtinTypes)
-{
-    TraversalState state(follow(root), builtinTypes);
+    TraversalState state(follow(root), builtinTypes, arena);
     if (traverse(state, path))
     {
-        auto ty = get<TypeId>(state.current);
+        if (state.encounteredErrorSuppression)
+            return builtinTypes->errorType;
+
+        const TypeId* ty = get<TypeId>(state.current);
         return ty ? std::make_optional(*ty) : std::nullopt;
     }
     else
         return std::nullopt;
 }
 
-std::optional<TypeId> traverseForType(TypePackId root, const Path& path, NotNull<BuiltinTypes> builtinTypes)
+std::optional<TypeId> traverseForType(
+    const TypePackId root,
+    const Path& path,
+    const NotNull<BuiltinTypes> builtinTypes,
+    const NotNull<TypeArena> arena
+)
 {
-    TraversalState state(follow(root), builtinTypes);
+    TraversalState state(follow(root), builtinTypes, arena);
     if (traverse(state, path))
     {
-        auto ty = get<TypeId>(state.current);
+        if (state.encounteredErrorSuppression)
+            return builtinTypes->errorType;
+        const TypeId* ty = get<TypeId>(state.current);
         return ty ? std::make_optional(*ty) : std::nullopt;
     }
     else
         return std::nullopt;
 }
 
-std::optional<TypePackId> traverseForPack(TypeId root, const Path& path, NotNull<BuiltinTypes> builtinTypes)
+std::optional<TypePackId> traverseForPack(
+    const TypeId root,
+    const Path& path,
+    const NotNull<BuiltinTypes> builtinTypes,
+    const NotNull<TypeArena> arena
+)
 {
-    TraversalState state(follow(root), builtinTypes);
+    TraversalState state(follow(root), builtinTypes, arena);
     if (traverse(state, path))
     {
-        auto ty = get<TypePackId>(state.current);
+        if (state.encounteredErrorSuppression)
+            return builtinTypes->errorTypePack;
+        const TypePackId* ty = get<TypePackId>(state.current);
         return ty ? std::make_optional(*ty) : std::nullopt;
     }
     else
         return std::nullopt;
 }
 
-std::optional<TypePackId> traverseForPack(TypePackId root, const Path& path, NotNull<BuiltinTypes> builtinTypes)
+std::optional<TypePackId> traverseForPack(
+    const TypePackId root,
+    const Path& path,
+    const NotNull<BuiltinTypes> builtinTypes,
+    const NotNull<TypeArena> arena
+)
 {
-    TraversalState state(follow(root), builtinTypes);
+    TraversalState state(follow(root), builtinTypes, arena);
     if (traverse(state, path))
     {
-        auto ty = get<TypePackId>(state.current);
+        if (state.encounteredErrorSuppression)
+            return builtinTypes->errorTypePack;
+        const TypePackId* ty = get<TypePackId>(state.current);
         return ty ? std::make_optional(*ty) : std::nullopt;
     }
     else
         return std::nullopt;
 }
 
+std::optional<size_t> traverseForIndex(const Path& path)
+{
+    auto componentIter = begin(path.components);
+    size_t index = 0;
+    const auto lastComponent = end(path.components) - 1;
+
+    while (componentIter != lastComponent)
+    {
+        if (const auto packSlice = get_if<Luau::TypePath::PackSlice>(&*componentIter))
+        {
+            index += packSlice->start_index;
+        }
+        else
+        {
+            return std::nullopt;
+        }
+        ++componentIter;
+    }
+
+    if (const auto indexComponent = get_if<TypePath::Index>(&*componentIter))
+    {
+        index += indexComponent->index;
+        return index;
+    }
+    return std::nullopt;
+}
+
+TypePack flattenPackWithPath(TypePackId root, const Path& path)
+{
+    std::vector<TypeId> flattened;
+
+    std::optional<TypePackId> curr = root;
+    auto pathIter = begin(path.components);
+    const auto pathEnd = end(path.components);
+
+    while (curr)
+    {
+        TypePackIterator it(*curr);
+        // Push back curr's head
+        for (; it != end(*curr); ++it)
+            flattened.emplace_back(*it);
+
+        // Check if curr has a tail, and if the next bit of path is Tail + GenericPackMapping
+        curr = it.tail();
+        if (!curr || !get<GenericTypePack>(*curr) || pathIter == pathEnd)
+            break;
+
+        if (const TypePath::PackField* pf = get_if<TypePath::PackField>(&*pathIter); !pf || *pf != TypePath::PackField::Tail)
+            break;
+
+        ++pathIter;
+
+        const TypePath::GenericPackMapping* gpm = get_if<TypePath::GenericPackMapping>(&*pathIter);
+        if (!gpm)
+            break;
+
+        ++pathIter;
+        curr = gpm->mappedType;
+    }
+
+    return {flattened, curr};
+}
+
+TypePack traverseForFlattenedPack(const TypeId root, const Path& path, const NotNull<BuiltinTypes> builtinTypes, const NotNull<TypeArena> arena)
+{
+    // Iterate over path's components, and figure out when it turns into Tails and GenericPackMappings
+    // We want to split out the part of the path that contains the generic pack mappings we're interested in, so that we can flatten it
+    // path[splitIndex:] will contain only Tails and GenericPackMappings
+    size_t splitIndex = 0;
+    for (size_t i = path.components.size(); i > 0; --i)
+    {
+        const TypePath::Component& c = path.components[i - 1];
+
+        const TypePath::PackField* pf = get_if<TypePath::PackField>(&c);
+        const bool isNotTail = pf == nullptr || *pf != TypePath::PackField::Tail;
+        const bool isNotGPM = get_if<TypePath::GenericPackMapping>(&c) == nullptr;
+
+        if (isNotTail && isNotGPM)
+        {
+            splitIndex = i;
+            break;
+        }
+    }
+
+    // Root is a TypeId, not a TypePackId, so splitIndex should be > 0
+    LUAU_ASSERT(splitIndex > 0);
+    if (splitIndex == path.components.size() || splitIndex == 0)
+        return {{}, std::nullopt};
+
+    auto basePathIter = begin(path.components);
+    std::advance(basePathIter, splitIndex);
+    const std::optional<TypePackId> basePack = traverseForPack(root, Path(std::vector(begin(path.components), basePathIter)), builtinTypes, arena);
+
+    if (!basePack)
+    {
+        LUAU_ASSERT(!"Expected to be able to traverse to a TypePackId");
+        return {{}, std::nullopt};
+    }
+
+    return flattenPackWithPath(*basePack, Path(std::vector(basePathIter, end(path.components))));
+}
+
+bool matchesPrefix(const Path& prefix, const Path& full)
+{
+    auto prefixIter = begin(prefix.components);
+    auto fullIter = begin(full.components);
+    const auto prefixEnd = end(prefix.components);
+    const auto fullEnd = end(full.components);
+
+    while (prefixIter != prefixEnd)
+    {
+        if (fullIter == fullEnd)
+            return false;
+
+        if (*prefixIter++ != *fullIter++)
+            return false;
+    }
+
+    return true;
+}
+
 } // namespace Luau
