@@ -2,20 +2,19 @@
 #include "Luau/TypeUtils.h"
 
 #include "Luau/Common.h"
+#include "Luau/IterativeTypeVisitor.h"
 #include "Luau/Normalize.h"
 #include "Luau/Scope.h"
+#include "Luau/Simplify.h"
 #include "Luau/ToString.h"
 #include "Luau/Type.h"
 #include "Luau/TypeInfer.h"
+#include "Luau/TypePack.h"
 
 #include <algorithm>
 
-LUAU_FASTFLAG(LuauSolverV2);
-LUAU_FASTFLAG(LuauAutocompleteRefactorsForIncrementalAutocomplete);
-LUAU_FASTFLAG(LuauTrackInteriorFreeTypesOnScope);
-LUAU_FASTFLAG(LuauFreeTypesMustHaveBounds)
-LUAU_FASTFLAG(LuauNonReentrantGeneralization)
-LUAU_FASTFLAG(LuauDisableNewSolverAssertsInMixedMode)
+LUAU_FASTFLAG(LuauSolverV2)
+LUAU_FASTFLAGVARIABLE(LuauContainsAnyGenericDoesntTraverseIntoExtern)
 
 namespace Luau
 {
@@ -45,59 +44,6 @@ bool occursCheck(TypeId needle, TypeId haystack)
     return false;
 }
 
-// FIXME: Property is quite large.
-//
-// Returning it on the stack like this isn't great. We'd like to just return a
-// const Property*, but we mint a property of type any if the subject type is
-// any.
-std::optional<Property> findTableProperty(NotNull<BuiltinTypes> builtinTypes, ErrorVec& errors, TypeId ty, const std::string& name, Location location)
-{
-    if (get<AnyType>(ty))
-        return Property::rw(ty);
-
-    if (const TableType* tableType = getTableType(ty))
-    {
-        const auto& it = tableType->props.find(name);
-        if (it != tableType->props.end())
-            return it->second;
-    }
-
-    std::optional<TypeId> mtIndex = findMetatableEntry(builtinTypes, errors, ty, "__index", location);
-    int count = 0;
-    while (mtIndex)
-    {
-        TypeId index = follow(*mtIndex);
-
-        if (count >= 100)
-            return std::nullopt;
-
-        ++count;
-
-        if (const auto& itt = getTableType(index))
-        {
-            const auto& fit = itt->props.find(name);
-            if (fit != itt->props.end())
-                return fit->second.type();
-        }
-        else if (const auto& itf = get<FunctionType>(index))
-        {
-            std::optional<TypeId> r = first(follow(itf->retTypes));
-            if (!r)
-                return builtinTypes->nilType;
-            else
-                return *r;
-        }
-        else if (get<AnyType>(index))
-            return builtinTypes->anyType;
-        else
-            errors.push_back(TypeError{location, GenericError{"__index should either be a function or table. Got " + toString(index)}});
-
-        mtIndex = findMetatableEntry(builtinTypes, errors, *mtIndex, "__index", location);
-    }
-
-    return std::nullopt;
-}
-
 std::optional<TypeId> findMetatableEntry(
     NotNull<BuiltinTypes> builtinTypes,
     ErrorVec& errors,
@@ -120,13 +66,18 @@ std::optional<TypeId> findMetatableEntry(
     const TableType* mtt = getTableType(unwrapped);
     if (!mtt)
     {
-        errors.push_back(TypeError{location, GenericError{"Metatable was not a table"}});
+        errors.emplace_back(location, GenericError{"Metatable was not a table"});
         return std::nullopt;
     }
 
     auto it = mtt->props.find(entry);
     if (it != mtt->props.end())
-        return it->second.type();
+    {
+        if (it->second.readTy)
+            return it->second.readTy;
+        else
+            return it->second.writeTy;
+    }
     else
         return std::nullopt;
 }
@@ -159,18 +110,13 @@ std::optional<TypeId> findTablePropertyRespectingMeta(
         const auto& it = tableType->props.find(name);
         if (it != tableType->props.end())
         {
-            if (FFlag::LuauSolverV2)
+            switch (context)
             {
-                switch (context)
-                {
-                case ValueContext::RValue:
-                    return it->second.readTy;
-                case ValueContext::LValue:
-                    return it->second.writeTy;
-                }
+            case ValueContext::RValue:
+                return it->second.readTy;
+            case ValueContext::LValue:
+                return it->second.writeTy;
             }
-            else
-                return it->second.type();
         }
     }
 
@@ -189,7 +135,20 @@ std::optional<TypeId> findTablePropertyRespectingMeta(
         {
             const auto& fit = itt->props.find(name);
             if (fit != itt->props.end())
-                return fit->second.type();
+            {
+                if (FFlag::LuauSolverV2)
+                {
+                    switch (context)
+                    {
+                    case ValueContext::RValue:
+                        return fit->second.readTy;
+                    case ValueContext::LValue:
+                        return fit->second.writeTy;
+                    }
+                }
+                else
+                    return fit->second.type_DEPRECATED();
+            }
         }
         else if (const auto& itf = get<FunctionType>(index))
         {
@@ -202,7 +161,7 @@ std::optional<TypeId> findTablePropertyRespectingMeta(
         else if (get<AnyType>(index))
             return builtinTypes->anyType;
         else
-            errors.push_back(TypeError{location, GenericError{"__index should either be a function or table. Got " + toString(index)}});
+            errors.emplace_back(location, GenericError{"__index should either be a function or table. Got " + toString(index)});
 
         mtIndex = findMetatableEntry(builtinTypes, errors, *mtIndex, "__index", location);
     }
@@ -308,11 +267,9 @@ TypePack extendTypePack(
             TypePack newPack;
             newPack.tail = arena.freshTypePack(ftp->scope, ftp->polarity);
 
-            if (FFlag::LuauNonReentrantGeneralization)
-                trackInteriorFreeTypePack(ftp->scope, *newPack.tail);
+            trackInteriorFreeTypePack(ftp->scope, *newPack.tail);
 
-            if (FFlag::LuauSolverV2)
-                result.tail = newPack.tail;
+            result.tail = newPack.tail;
             size_t overridesIndex = 0;
             while (result.head.size() < length)
             {
@@ -323,15 +280,9 @@ TypePack extendTypePack(
                 }
                 else
                 {
-                    if (FFlag::LuauSolverV2)
-                    {
-                        FreeType ft{ftp->scope, builtinTypes->neverType, builtinTypes->unknownType, ftp->polarity};
-                        t = arena.addType(ft);
-                        if (FFlag::LuauTrackInteriorFreeTypesOnScope)
-                            trackInteriorFreeType(ftp->scope, t);
-                    }
-                    else
-                        t = FFlag::LuauFreeTypesMustHaveBounds ? arena.freshType(builtinTypes, ftp->scope) : arena.freshType_DEPRECATED(ftp->scope);
+                    FreeType ft{ftp->scope, builtinTypes->neverType, builtinTypes->unknownType, ftp->polarity};
+                    t = arena.addType(ft);
+                    trackInteriorFreeType(ftp->scope, t);
                 }
 
                 newPack.head.push_back(t);
@@ -343,10 +294,10 @@ TypePack extendTypePack(
 
             return result;
         }
-        else if (auto etp = getMutable<ErrorTypePack>(pack))
+        else if (getMutable<ErrorTypePack>(pack))
         {
             while (result.head.size() < length)
-                result.head.push_back(builtinTypes->errorRecoveryType());
+                result.head.push_back(builtinTypes->errorType);
 
             result.tail = pack;
             return result;
@@ -438,7 +389,22 @@ TypeId stripNil(NotNull<BuiltinTypes> builtinTypes, TypeArena& arena, TypeId ty)
 
 ErrorSuppression shouldSuppressErrors(NotNull<Normalizer> normalizer, TypeId ty)
 {
-    LUAU_ASSERT(FFlag::LuauSolverV2 || FFlag::LuauAutocompleteRefactorsForIncrementalAutocomplete);
+    if (auto tfit = get<TypeFunctionInstanceType>(follow(ty)))
+    {
+        for (auto ty : tfit->typeArguments)
+        {
+            std::shared_ptr<const NormalizedType> normType = normalizer->normalize(ty);
+
+            if (!normType)
+                return ErrorSuppression::NormalizationFailed;
+
+            if (normType->shouldSuppressErrors())
+                return ErrorSuppression::Suppress;
+        }
+
+        return ErrorSuppression::DoNotSuppress;
+    }
+
     std::shared_ptr<const NormalizedType> normType = normalizer->normalize(ty);
 
     if (!normType)
@@ -449,6 +415,14 @@ ErrorSuppression shouldSuppressErrors(NotNull<Normalizer> normalizer, TypeId ty)
 
 ErrorSuppression shouldSuppressErrors(NotNull<Normalizer> normalizer, TypePackId tp)
 {
+    // Flatten t where t = ...any will produce a type pack [ {}, t]
+    // which trivially fails the tail check below, which is why we need to special case here
+    if (auto tpId = get<VariadicTypePack>(follow(tp)))
+    {
+        if (get<AnyType>(follow(tpId->ty)))
+            return ErrorSuppression::Suppress;
+    }
+
     auto [tys, tail] = flatten(tp);
 
     // check the head, one type at a time
@@ -532,14 +506,6 @@ private:
     NotNull<std::vector<TypeId>> toBlock_;
 };
 
-std::vector<TypeId> findBlockedTypesIn(AstExprTable* expr, NotNull<DenseHashMap<const AstExpr*, TypeId>> astTypes)
-{
-    std::vector<TypeId> toBlock;
-    BlockedTypeInLiteralVisitor v{astTypes, NotNull{&toBlock}};
-    expr->visit(&v);
-    return toBlock;
-}
-
 std::vector<TypeId> findBlockedArgTypesIn(AstExprCall* expr, NotNull<DenseHashMap<const AstExpr*, TypeId>> astTypes)
 {
     std::vector<TypeId> toBlock;
@@ -556,10 +522,6 @@ std::vector<TypeId> findBlockedArgTypesIn(AstExprCall* expr, NotNull<DenseHashMa
 
 void trackInteriorFreeType(Scope* scope, TypeId ty)
 {
-    if (FFlag::LuauDisableNewSolverAssertsInMixedMode)
-        LUAU_ASSERT(FFlag::LuauTrackInteriorFreeTypesOnScope);
-    else
-        LUAU_ASSERT(FFlag::LuauSolverV2 && FFlag::LuauTrackInteriorFreeTypesOnScope);
     for (; scope; scope = scope->parent.get())
     {
         if (scope->interiorFreeTypes)
@@ -577,8 +539,6 @@ void trackInteriorFreeType(Scope* scope, TypeId ty)
 void trackInteriorFreeTypePack(Scope* scope, TypePackId tp)
 {
     LUAU_ASSERT(tp);
-    if (!FFlag::LuauNonReentrantGeneralization)
-        return;
 
     for (; scope; scope = scope->parent.get())
     {
@@ -594,4 +554,379 @@ void trackInteriorFreeTypePack(Scope* scope, TypePackId tp)
     LUAU_ASSERT(!"No scopes in parent chain had a present `interiorFreeTypePacks` member.");
 }
 
+bool fastIsSubtype(TypeId subTy, TypeId superTy)
+{
+    Relation r = relate(superTy, subTy);
+    return r == Relation::Coincident || r == Relation::Superset;
+}
+
+std::optional<TypeId> extractMatchingTableType(std::vector<TypeId>& tables, TypeId exprType, NotNull<BuiltinTypes> builtinTypes)
+{
+    if (tables.empty())
+        return std::nullopt;
+
+    const TableType* exprTable = get<TableType>(follow(exprType));
+    if (!exprTable)
+        return std::nullopt;
+
+    size_t tableCount = 0;
+    std::optional<TypeId> firstTable;
+
+    for (TypeId ty : tables)
+    {
+        ty = follow(ty);
+        if (auto tt = get<TableType>(ty))
+        {
+            // If the expected table has a key whose type is a string or boolean
+            // singleton and the corresponding exprType property does not match,
+            // then skip this table.
+
+            if (!firstTable)
+                firstTable = ty;
+            ++tableCount;
+
+            for (const auto& [name, expectedProp] : tt->props)
+            {
+                if (!expectedProp.readTy)
+                    continue;
+
+                const TypeId expectedType = follow(*expectedProp.readTy);
+
+                auto st = get<SingletonType>(expectedType);
+                if (!st)
+                    continue;
+
+                auto it = exprTable->props.find(name);
+                if (it == exprTable->props.end())
+                    continue;
+
+                const auto& [_name, exprProp] = *it;
+
+                if (!exprProp.readTy)
+                    continue;
+
+                const TypeId propType = follow(*exprProp.readTy);
+
+                const FreeType* ft = get<FreeType>(propType);
+
+                if (ft && get<SingletonType>(ft->lowerBound))
+                {
+                    if (fastIsSubtype(builtinTypes->booleanType, ft->upperBound) && fastIsSubtype(expectedType, builtinTypes->booleanType))
+                    {
+                        return ty;
+                    }
+
+                    if (fastIsSubtype(builtinTypes->stringType, ft->upperBound) && fastIsSubtype(expectedType, ft->lowerBound))
+                    {
+                        return ty;
+                    }
+                }
+
+                if (fastIsSubtype(propType, expectedType))
+                    return ty;
+            }
+        }
+    }
+
+    if (tableCount == 1)
+    {
+        LUAU_ASSERT(firstTable);
+        return firstTable;
+    }
+
+    return std::nullopt;
+}
+
+bool isRecord(const AstExprTable::Item& item)
+{
+    if (item.kind == AstExprTable::Item::Record)
+        return true;
+    else if (item.kind == AstExprTable::Item::General && item.key->is<AstExprConstantString>())
+        return true;
+    else
+        return false;
+}
+
+AstExpr* unwrapGroup(AstExpr* expr)
+{
+    while (auto group = expr->as<AstExprGroup>())
+        expr = group->expr;
+
+    return expr;
+}
+
+bool isOptionalType(TypeId ty, NotNull<BuiltinTypes> builtinTypes)
+{
+    ty = follow(ty);
+
+    if (ty == builtinTypes->nilType || ty == builtinTypes->anyType || ty == builtinTypes->unknownType)
+        return true;
+    else if (const PrimitiveType* pt = get<PrimitiveType>(ty))
+        return pt->type == PrimitiveType::NilType;
+    else if (const UnionType* ut = get<UnionType>(ty))
+    {
+        for (TypeId option : ut)
+        {
+            option = follow(option);
+
+            if (option == builtinTypes->nilType || option == builtinTypes->anyType || option == builtinTypes->unknownType)
+                return true;
+            else if (const PrimitiveType* pt = get<PrimitiveType>(option); pt && pt->type == PrimitiveType::NilType)
+                return true;
+        }
+
+        return false;
+    }
+
+    return false;
+}
+
+bool isApproximatelyFalsyType(TypeId ty)
+{
+    ty = follow(ty);
+    bool seenNil = false;
+    bool seenFalse = false;
+    if (auto ut = get<UnionType>(ty))
+    {
+        for (auto option : ut)
+        {
+            if (auto pt = get<PrimitiveType>(option); pt && pt->type == PrimitiveType::NilType)
+                seenNil = true;
+            else if (auto st = get<SingletonType>(option); st && st->variant == BooleanSingleton{false})
+                seenFalse = true;
+            else
+                return false;
+        }
+    }
+    return seenFalse && seenNil;
+}
+
+bool isApproximatelyTruthyType(TypeId ty)
+{
+    ty = follow(ty);
+    if (auto nt = get<NegationType>(ty))
+        return isApproximatelyFalsyType(nt->ty);
+    return false;
+}
+
+UnionBuilder::UnionBuilder(NotNull<TypeArena> arena, NotNull<BuiltinTypes> builtinTypes)
+    : arena(arena)
+    , builtinTypes(builtinTypes)
+{
+}
+
+void UnionBuilder::add(TypeId ty)
+{
+    ty = follow(ty);
+
+    if (is<NeverType>(ty) || isTop)
+        return;
+
+    if (is<UnknownType>(ty))
+    {
+        isTop = true;
+        return;
+    }
+
+    if (auto utv = get<UnionType>(ty))
+    {
+        for (auto option : utv)
+            options.insert(option);
+    }
+    else
+        options.insert(ty);
+}
+
+TypeId UnionBuilder::build()
+{
+    if (isTop)
+        return builtinTypes->unknownType;
+
+    if (options.empty())
+        return builtinTypes->neverType;
+
+    if (options.size() == 1)
+        return options.front();
+
+    return arena->addType(UnionType{options.take()});
+}
+
+size_t UnionBuilder::size() const
+{
+    return options.size();
+}
+
+void UnionBuilder::reserve(size_t size)
+{
+    options.reserve(size);
+}
+
+IntersectionBuilder::IntersectionBuilder(NotNull<TypeArena> arena, NotNull<BuiltinTypes> builtinTypes)
+    : arena(arena)
+    , builtinTypes(builtinTypes)
+{
+}
+
+void IntersectionBuilder::add(TypeId ty)
+{
+    ty = follow(ty);
+
+    if (is<NeverType>(ty))
+    {
+        isBottom = true;
+        return;
+    }
+
+    if (is<UnknownType>(ty))
+        return;
+
+    if (auto itv = get<IntersectionType>(ty))
+    {
+        for (auto part : itv)
+            parts.insert(part);
+    }
+    else
+        parts.insert(ty);
+}
+
+TypeId IntersectionBuilder::build()
+{
+    if (isBottom)
+        return builtinTypes->neverType;
+
+    if (parts.empty())
+        return builtinTypes->unknownType;
+
+    if (parts.size() == 1)
+        return parts.front();
+
+    return arena->addType(IntersectionType{parts.take()});
+}
+
+size_t IntersectionBuilder::size() const
+{
+    return parts.size();
+}
+
+void IntersectionBuilder::reserve(size_t size)
+{
+    parts.reserve(size);
+}
+
+
+TypeId addIntersection(NotNull<TypeArena> arena, NotNull<BuiltinTypes> builtinTypes, std::initializer_list<TypeId> list)
+{
+    IntersectionBuilder ib(arena, builtinTypes);
+    ib.reserve(list.size());
+    for (TypeId part : list)
+        ib.add(part);
+
+    return ib.build();
+}
+
+TypeId addUnion(NotNull<TypeArena> arena, NotNull<BuiltinTypes> builtinTypes, std::initializer_list<TypeId> list)
+{
+    UnionBuilder ub(arena, builtinTypes);
+    ub.reserve(list.size());
+    for (TypeId option : list)
+        ub.add(option);
+    return ub.build();
+}
+
+ContainsAnyGeneric::ContainsAnyGeneric()
+    : TypeOnceVisitor("ContainsAnyGeneric", /* skipBoundTypes */ true)
+{
+}
+
+bool ContainsAnyGeneric::visit(TypeId ty, const ExternType&)
+{
+    return !FFlag::LuauContainsAnyGenericDoesntTraverseIntoExtern;
+}
+
+bool ContainsAnyGeneric::visit(TypeId ty)
+{
+    found = found || is<GenericType>(ty);
+    return !found;
+}
+
+bool ContainsAnyGeneric::visit(TypePackId ty)
+{
+    found = found || is<GenericTypePack>(follow(ty));
+    return !found;
+}
+
+bool ContainsAnyGeneric::hasAnyGeneric(TypeId ty)
+{
+    ContainsAnyGeneric cg;
+    cg.traverse(ty);
+    return cg.found;
+}
+
+bool ContainsAnyGeneric::hasAnyGeneric(TypePackId tp)
+{
+    ContainsAnyGeneric cg;
+    cg.traverse(tp);
+    return cg.found;
+}
+
+struct ContainsGenerics : public IterativeTypeVisitor
+{
+    NotNull<DenseHashSet<const void*>> generics;
+
+    explicit ContainsGenerics(NotNull<DenseHashSet<const void*>> generics)
+        : IterativeTypeVisitor("ContainsGenerics", /* skipBoundTypes */ true)
+        , generics{generics}
+    {
+    }
+
+    bool found = false;
+
+    bool visit(TypeId ty) override
+    {
+        return !found;
+    }
+
+    bool visit(TypeId ty, const GenericType&) override
+    {
+        found |= generics->contains(ty);
+        return true;
+    }
+
+    bool visit(TypeId ty, const TypeFunctionInstanceType&) override
+    {
+        return !found;
+    }
+
+    bool visit(TypePackId tp, const GenericTypePack&) override
+    {
+        found |= generics->contains(tp);
+        return !found;
+    }
+};
+
+bool containsGeneric(TypeId ty, NotNull<DenseHashSet<const void*>> generics)
+{
+    ContainsGenerics cg{generics};
+    cg.run(ty);
+    return cg.found;
+}
+
+bool containsGeneric(TypePackId ty, NotNull<DenseHashSet<const void*>> generics)
+{
+    ContainsGenerics cg{generics};
+    cg.run(ty);
+    return cg.found;
+}
+
+bool isBlocked(TypeId ty)
+{
+    ty = follow(ty);
+
+    if (auto tfit = get<TypeFunctionInstanceType>(ty))
+        return tfit->state == TypeFunctionInstanceState::Unsolved;
+
+    return is<BlockedType, PendingExpansionType>(ty);
+}
+
+
 } // namespace Luau
