@@ -1,10 +1,9 @@
 // This file is part of the Luau programming language and is licensed under MIT License; see LICENSE.txt for details
 #pragma once
 
+#include "Luau/Ast.h"
 #include "Luau/Location.h"
 
-#include <string>
-
 namespace Luau
 {
 
@@ -82,6 +81,18 @@ public:
     unsigned int blockDepth;
 };
 
+// Shared between the expression and call nodes
+struct CstTypeInstantiation
+{
+    Position leftArrow1Position = {0, 0};
+    Position leftArrow2Position = {0, 0};
+
+    AstArray<Position> commaPositions = {};
+
+    Position rightArrow1Position = {0, 0};
+    Position rightArrow2Position = {0, 0};
+};
+
 class CstExprCall : public CstNode
 {
 public:
@@ -92,6 +103,7 @@ public:
     std::optional<Position> openParens;
     std::optional<Position> closeParens;
     AstArray<Position> commaPositions;
+    CstTypeInstantiation* explicitTypes = nullptr;
 };
 
 class CstExprIndexExpr : public CstNode
@@ -116,7 +128,9 @@ public:
     Position openGenericsPosition{0, 0};
     AstArray<Position> genericsCommaPositions;
     Position closeGenericsPosition{0, 0};
+    AstArray<Position> argsAnnotationColonPositions;
     AstArray<Position> argsCommaPositions;
+    Position varargAnnotationColonPosition{0, 0};
     Position returnSpecifierPosition{0, 0};
 };
 
@@ -189,12 +203,34 @@ public:
     AstArray<Position> stringPositions;
 };
 
+class CstExprExplicitTypeInstantiation : public CstNode
+{
+public:
+    LUAU_CST_RTTI(CstExprExplicitTypeInstantiation)
+
+    explicit CstExprExplicitTypeInstantiation(CstTypeInstantiation instantiation);
+
+    CstTypeInstantiation instantiation;
+};
+
 class CstStatDo : public CstNode
 {
 public:
     LUAU_CST_RTTI(CstStatDo)
 
-    explicit CstStatDo(Position endPosition);
+    explicit CstStatDo(Position statsStartPosition, Position endPosition);
+
+    Position statsStartPosition;
+    Position endPosition;
+};
+
+// Clip with FFlag::LuauCstStatBlock
+class CstStatDo_DEPRECATED : public CstNode
+{
+public:
+    LUAU_CST_RTTI(CstStatDo_DEPRECATED)
+
+    explicit CstStatDo_DEPRECATED(Position endPosition);
 
     Position endPosition;
 };
@@ -224,8 +260,9 @@ class CstStatLocal : public CstNode
 public:
     LUAU_CST_RTTI(CstStatLocal)
 
-    CstStatLocal(AstArray<Position> varsCommaPositions, AstArray<Position> valuesCommaPositions);
+    CstStatLocal(AstArray<Position> varsAnnotationColonPositions, AstArray<Position> varsCommaPositions, AstArray<Position> valuesCommaPositions);
 
+    AstArray<Position> varsAnnotationColonPositions;
     AstArray<Position> varsCommaPositions;
     AstArray<Position> valuesCommaPositions;
 };
@@ -235,8 +272,9 @@ class CstStatFor : public CstNode
 public:
     LUAU_CST_RTTI(CstStatFor)
 
-    CstStatFor(Position equalsPosition, Position endCommaPosition, std::optional<Position> stepCommaPosition);
+    CstStatFor(Position annotationColonPosition, Position equalsPosition, Position endCommaPosition, std::optional<Position> stepCommaPosition);
 
+    Position annotationColonPosition;
     Position equalsPosition;
     Position endCommaPosition;
     std::optional<Position> stepCommaPosition;
@@ -247,8 +285,9 @@ class CstStatForIn : public CstNode
 public:
     LUAU_CST_RTTI(CstStatForIn)
 
-    CstStatForIn(AstArray<Position> varsCommaPositions, AstArray<Position> valuesCommaPositions);
+    CstStatForIn(AstArray<Position> varsAnnotationColonPositions, AstArray<Position> varsCommaPositions, AstArray<Position> valuesCommaPositions);
 
+    AstArray<Position> varsAnnotationColonPositions;
     AstArray<Position> varsCommaPositions;
     AstArray<Position> valuesCommaPositions;
 };
@@ -388,6 +427,7 @@ public:
         std::optional<Position> separatorPosition;
 
         CstExprConstantString* stringInfo = nullptr; // only if Kind == StringProperty
+        Position stringPosition{0, 0};               // only if Kind == StringProperty
     };
 
     CstTypeTable(AstArray<Item> items, bool isArray);
@@ -472,8 +512,10 @@ class CstTypePackExplicit : public CstNode
 public:
     LUAU_CST_RTTI(CstTypePackExplicit)
 
-    CstTypePackExplicit(Position openParenthesesPosition, Position closeParenthesesPosition, AstArray<Position> commaPositions);
+    explicit CstTypePackExplicit();
+    explicit CstTypePackExplicit(Position openParenthesesPosition, Position closeParenthesesPosition, AstArray<Position> commaPositions);
 
+    bool hasParentheses;
     Position openParenthesesPosition;
     Position closeParenthesesPosition;
     AstArray<Position> commaPositions;
@@ -489,4 +531,4 @@ public:
     Position ellipsisPosition;
 };
 
-} // namespace Luau
\ No newline at end of file
+} // namespace Luau
