@@ -2,34 +2,32 @@
 #pragma once
 
 #include "Luau/RequireNavigator.h"
-#include "Luau/RequirerUtils.h"
+#include "Luau/VfsNavigator.h"
 
 struct FileNavigationContext : Luau::Require::NavigationContext
 {
     using NavigateResult = Luau::Require::NavigationContext::NavigateResult;
+    using ConfigStatus = Luau::Require::NavigationContext::ConfigStatus;
 
     FileNavigationContext(std::string requirerPath);
 
-    std::string getRequirerIdentifier() const override;
-
     // Navigation interface
-    NavigateResult reset(const std::string& requirerChunkname) override;
+    NavigateResult resetToRequirer() override;
     NavigateResult jumpToAlias(const std::string& path) override;
 
     NavigateResult toParent() override;
     NavigateResult toChild(const std::string& component) override;
 
-    bool isConfigPresent() const override;
+    ConfigStatus getConfigStatus() const override;
+    ConfigBehavior getConfigBehavior() const override;
+    std::optional<std::string> getAlias(const std::string& alias) const override;
     std::optional<std::string> getConfig() const override;
 
     // Custom capabilities
     bool isModulePresent() const;
     std::optional<std::string> getIdentifier() const;
 
-    std::string path;
-    std::string suffix;
-    std::string requirerPath;
-
 private:
-    NavigateResult storePathResult(PathResult result);
+    std::string requirerPath;
+    VfsNavigator vfs;
 };
