@@ -164,7 +164,7 @@ static int lua_callgrind(lua_State* L)
 }
 #endif
 
-static void* createCliRequireContext(lua_State* L)
+void* createCliRequireContext(lua_State* L)
 {
     void* ctx = lua_newuserdatadtor(
         L,
@@ -581,7 +581,7 @@ static bool runFile(const char* name, lua_State* GL, bool repl)
     // new thread needs to have the globals sandboxed
     luaL_sandboxthread(L);
 
-    std::string chunkname = "@" + std::string(name);
+    std::string chunkname = "@" + normalizePath(name);
 
     std::string bytecode = Luau::compile(*source, copts());
     int status = 0;
