@@ -7,7 +7,10 @@
 #include <stdarg.h>
 #include <stdio.h>
 
-LUAU_FASTFLAG(LuauVectorLibNativeDot);
+LUAU_FASTFLAG(LuauCodegenUpvalueLoadProp2)
+LUAU_FASTFLAG(LuauCodegenSplitFloat)
+LUAU_FASTFLAG(LuauCodegenLocationEndFix)
+LUAU_FASTFLAG(LuauCodegenUintToFloat)
 
 namespace Luau
 {
@@ -25,7 +28,7 @@ static_assert(sizeof(textForCondition) / sizeof(textForCondition[0]) == size_t(C
 
 const unsigned kMaxAlign = 32;
 
-static int getFmovImm(double value)
+static int getFmovImmFp64(double value)
 {
     uint64_t u;
     static_assert(sizeof(u) == sizeof(value), "expected double to be 64-bit");
@@ -46,6 +49,27 @@ static int getFmovImm(double value)
     return dec == int(u >> 48) ? imm : -1;
 }
 
+static int getFmovImmFp32(float value)
+{
+    uint32_t u;
+    static_assert(sizeof(u) == sizeof(value), "expected float to be 32-bit");
+    memcpy(&u, &value, sizeof(value));
+
+    // positive 0 is encodable via movi
+    if (u == 0)
+        return 256;
+
+    // early out: fmov can only encode float with 19 least significant zeros
+    if ((u & ((1ull << 19) - 1)) != 0)
+        return -1;
+
+    // f32 expansion is abcdfegh => aBbbbbbc defgh000 00000000 00000000
+    int imm = (int(u >> 24) & 0x80) | (int(u >> 19) & 0x7f);
+    int dec = ((imm & 0x80) << 5) | ((imm & 0x40) ? 0b00000111'11000000 : 0b00001000'00000000) | (imm & 0x3f);
+
+    return dec == int(u >> 19) ? imm : -1;
+}
+
 AssemblyBuilderA64::AssemblyBuilderA64(bool logText, unsigned int features)
     : logText(logText)
     , features(features)
@@ -531,6 +555,17 @@ void AssemblyBuilderA64::adr(RegisterA64 dst, uint64_t value)
     patchOffset(location, -int(location) - int((data.size() - pos) / 4), Patch::Imm19);
 }
 
+void AssemblyBuilderA64::adr(RegisterA64 dst, float value)
+{
+    size_t pos = allocateData(4, 4);
+    uint32_t location = getCodeSize();
+
+    writef32(&data[pos], value);
+    placeADR("adr", dst, 0b10000);
+
+    patchOffset(location, -int(location) - int((data.size() - pos) / 4), Patch::Imm19);
+}
+
 void AssemblyBuilderA64::adr(RegisterA64 dst, double value)
 {
     size_t pos = allocateData(8, 8);
@@ -549,19 +584,37 @@ void AssemblyBuilderA64::adr(RegisterA64 dst, Label& label)
 
 void AssemblyBuilderA64::fmov(RegisterA64 dst, RegisterA64 src)
 {
-    CODEGEN_ASSERT(dst.kind == KindA64::d && (src.kind == KindA64::d || src.kind == KindA64::x));
-
-    if (src.kind == KindA64::d)
-        placeR1("fmov", dst, src, 0b000'11110'01'1'0000'00'10000);
+    if (FFlag::LuauCodegenUpvalueLoadProp2 || FFlag::LuauCodegenSplitFloat)
+    {
+        if (dst.kind == KindA64::d && src.kind == KindA64::d)
+            placeR1("fmov", dst, src, 0b00'11110'01'1'0000'00'10000);
+        else if (dst.kind == KindA64::d && src.kind == KindA64::x)
+            placeR1("fmov", dst, src, 0b00'11110'01'1'00'111'000000);
+        else if (dst.kind == KindA64::x && src.kind == KindA64::d)
+            placeR1("fmov", dst, src, 0b00'11110'01'1'00'110'000000);
+        else if (FFlag::LuauCodegenSplitFloat && dst.kind == KindA64::s && src.kind == KindA64::s)
+            placeR1("fmov", dst, src, 0b00'11110'00'1'0000'00'10000);
+        else if (FFlag::LuauCodegenSplitFloat && dst.kind == KindA64::s && src.kind == KindA64::w)
+            placeR1("fmov", dst, src, 0b00'11110'00'1'00'111'000000);
+        else
+            CODEGEN_ASSERT(!"Unsupported fmov kind");
+    }
     else
-        placeR1("fmov", dst, src, 0b000'11110'01'1'00'111'000000);
+    {
+        CODEGEN_ASSERT(dst.kind == KindA64::d && (src.kind == KindA64::d || src.kind == KindA64::x));
+
+        if (src.kind == KindA64::d)
+            placeR1("fmov", dst, src, 0b000'11110'01'1'0000'00'10000);
+        else
+            placeR1("fmov", dst, src, 0b000'11110'01'1'00'111'000000);
+    }
 }
 
 void AssemblyBuilderA64::fmov(RegisterA64 dst, double src)
 {
     CODEGEN_ASSERT(dst.kind == KindA64::d || dst.kind == KindA64::q);
 
-    int imm = getFmovImm(src);
+    int imm = getFmovImmFp64(src);
     CODEGEN_ASSERT(imm >= 0 && imm <= 256);
 
     // fmov can't encode 0, but movi can; movi is otherwise not useful for fp immediates because it encodes repeating patterns
@@ -581,22 +634,92 @@ void AssemblyBuilderA64::fmov(RegisterA64 dst, double src)
     }
 }
 
+void AssemblyBuilderA64::fmov(RegisterA64 dst, float src)
+{
+    CODEGEN_ASSERT(dst.kind == KindA64::s || dst.kind == KindA64::q);
+
+    int imm = getFmovImmFp32(src);
+    CODEGEN_ASSERT(imm >= 0 && imm <= 256);
+
+    // fmov can't encode 0, but movi can; movi is otherwise not useful for fp immediates because it encodes repeating patterns
+    if (dst.kind == KindA64::s)
+    {
+        if (imm == 256)
+            placeFMOV("movi", dst, src, 0b001'0111100000'000'1110'01'00000);
+        else
+            placeFMOV("fmov", dst, src, 0b000'11110'00'1'00000000'100'00000 | (imm << 8));
+    }
+    else
+    {
+        if (imm == 256)
+            placeFMOV("movi.4s", dst, src, 0b010'0111100000'000'0000'01'00000);
+        else
+            placeFMOV("fmov.4s", dst, src, 0b010'0111100000'000'1111'0'1'00000 | ((imm >> 5) << 11) | (imm & 31));
+    }
+}
+
 void AssemblyBuilderA64::fabs(RegisterA64 dst, RegisterA64 src)
 {
-    CODEGEN_ASSERT(dst.kind == KindA64::d && src.kind == KindA64::d);
+    CODEGEN_ASSERT(dst.kind == src.kind);
+    CODEGEN_ASSERT(dst.kind == KindA64::d || dst.kind == KindA64::s || dst.kind == KindA64::q);
 
-    placeR1("fabs", dst, src, 0b000'11110'01'1'0000'01'10000);
+    if (dst.kind == KindA64::q)
+        placeR1("fabs", dst, src, 0b010'01110'10'1'0000'01111'10);
+    else if (dst.kind == KindA64::d)
+        placeR1("fabs", dst, src, 0b000'11110'01'1'0000'01'10000);
+    else
+        placeR1("fabs", dst, src, 0b000'11110'00'1'0000'01'10000);
 }
 
 void AssemblyBuilderA64::faddp(RegisterA64 dst, RegisterA64 src)
 {
-    LUAU_ASSERT(FFlag::LuauVectorLibNativeDot);
     CODEGEN_ASSERT(dst.kind == KindA64::d || dst.kind == KindA64::s);
     CODEGEN_ASSERT(dst.kind == src.kind);
 
     placeR1("faddp", dst, src, 0b011'11110'0'0'11000'01101'10 | ((dst.kind == KindA64::d) << 12));
 }
 
+void AssemblyBuilderA64::fmla(RegisterA64 dst, RegisterA64 src1, RegisterA64 src2)
+{
+    // There is no scalar version of FMLA instruction
+    // Vector instruction is used for both cases with proper sz bit.
+
+    //                Q U        Sz  Rm    Opcode Rn    Rd
+    uint32_t op = 0b0'0'0'011100'0'1'00000'110011'00000'00000;
+    const uint32_t QBit = 1 << 30;
+    const uint32_t SzBit = 1 << 22;
+
+    if (dst.kind == KindA64::d)
+    {
+        CODEGEN_ASSERT(src1.kind == KindA64::d && src2.kind == KindA64::d);
+
+        if (logText)
+            logAppend(" %-12sd%d,d%d,d%d\n", "fmla", dst.index, src1.index, src2.index);
+
+        place(dst.index | (src1.index << 5) | (src2.index << 16) | op | QBit | SzBit);
+    }
+    else if (dst.kind == KindA64::s)
+    {
+        CODEGEN_ASSERT(src1.kind == KindA64::s && src2.kind == KindA64::s);
+
+        if (logText)
+            logAppend(" %-12ss%d,s%d,s%d\n", "fmla", dst.index, src1.index, src2.index);
+
+        place(dst.index | (src1.index << 5) | (src2.index << 16) | op);
+    }
+    else
+    {
+        CODEGEN_ASSERT(dst.kind == KindA64::q && src1.kind == KindA64::q && src2.kind == KindA64::q);
+
+        if (logText)
+            logAppend(" %-12sv%d.4s,v%d.4s,v%d.4s\n", "fmla", dst.index, src1.index, src2.index);
+
+        place(dst.index | (src1.index << 5) | (src2.index << 16) | op | QBit);
+    }
+
+    commit();
+}
+
 void AssemblyBuilderA64::fadd(RegisterA64 dst, RegisterA64 src1, RegisterA64 src2)
 {
     if (dst.kind == KindA64::d)
@@ -687,9 +810,13 @@ void AssemblyBuilderA64::fneg(RegisterA64 dst, RegisterA64 src)
 
 void AssemblyBuilderA64::fsqrt(RegisterA64 dst, RegisterA64 src)
 {
-    CODEGEN_ASSERT(dst.kind == KindA64::d && src.kind == KindA64::d);
+    CODEGEN_ASSERT(dst.kind == src.kind);
+    CODEGEN_ASSERT(dst.kind == KindA64::d || dst.kind == KindA64::s);
 
-    placeR1("fsqrt", dst, src, 0b000'11110'01'1'0000'11'10000);
+    if (dst.kind == KindA64::d)
+        placeR1("fsqrt", dst, src, 0b000'11110'01'1'0000'11'10000);
+    else
+        placeR1("fsqrt", dst, src, 0b000'11110'00'1'0000'11'10000);
 }
 
 void AssemblyBuilderA64::fsub(RegisterA64 dst, RegisterA64 src1, RegisterA64 src2)
@@ -773,25 +900,112 @@ void AssemblyBuilderA64::dup_4s(RegisterA64 dst, RegisterA64 src, uint8_t index)
     commit();
 }
 
+void AssemblyBuilderA64::umov_4s(RegisterA64 dst, RegisterA64 src, uint8_t index)
+{
+    CODEGEN_ASSERT(dst.kind == KindA64::w);
+    CODEGEN_ASSERT(src.kind == KindA64::q);
+    CODEGEN_ASSERT(index < 4);
+
+    if (logText)
+        logAppend(" %-12sw%d,v%d.s[%d]\n", "umov", dst.index, src.index, index);
+
+    //                Q     A-SIMD SzIdx Opcode Rn    Rd
+    uint32_t op = 0b0'0'0'01110000'00100'001111'00000'00000;
+
+    place(dst.index | (src.index << 5) | op | (index << 19));
+
+    commit();
+}
+
+void AssemblyBuilderA64::fcmeq_4s(RegisterA64 dst, RegisterA64 src1, RegisterA64 src2)
+{
+    if (logText)
+        logAppend(" %-12sv%d.4s,v%d.4s,v%d.4s\n", "fcmeq", dst.index, src1.index, src2.index);
+
+    //                Q U      ESz Rm    Opcode Rn    Rd
+    uint32_t op = 0b0'1'0'01110001'00000'111001'00000'00000;
+
+    place(dst.index | (src1.index << 5) | (src2.index << 16) | op);
+
+    commit();
+}
+
+void AssemblyBuilderA64::fcmgt_4s(RegisterA64 dst, RegisterA64 src1, RegisterA64 src2)
+{
+    if (logText)
+        logAppend(" %-12sv%d.4s,v%d.4s,v%d.4s\n", "fcmgt", dst.index, src1.index, src2.index);
+
+    //                Q U      ESz Rm    Opcode Rn    Rd
+    uint32_t op = 0b0'1'1'01110101'00000'111001'00000'00000;
+
+    place(dst.index | (src1.index << 5) | (src2.index << 16) | op);
+
+    commit();
+}
+
+void AssemblyBuilderA64::bit(RegisterA64 dst, RegisterA64 src, RegisterA64 mask)
+{
+    if (logText)
+        logAppend(" %-12sv%d.16b,v%d.16b,v%d.16b\n", "bit", dst.index, src.index, mask.index);
+
+    //                Q U          Rm    Opcode Rn    Rd
+    uint32_t op = 0b0'1'1'01110101'00000'000111'00000'00000;
+
+    place(dst.index | (src.index << 5) | (mask.index << 16) | op);
+
+    commit();
+}
+
+void AssemblyBuilderA64::bif(RegisterA64 dst, RegisterA64 src, RegisterA64 mask)
+{
+    if (logText)
+        logAppend(" %-12sv%d.16b,v%d.16b,v%d.16b\n", "bif", dst.index, src.index, mask.index);
+
+    //                Q U          Rm    Opcode Rn    Rd
+    uint32_t op = 0b0'1'1'01110111'00000'000111'00000'00000;
+
+    place(dst.index | (src.index << 5) | (mask.index << 16) | op);
+
+    commit();
+}
+
 void AssemblyBuilderA64::frinta(RegisterA64 dst, RegisterA64 src)
 {
-    CODEGEN_ASSERT(dst.kind == KindA64::d && src.kind == KindA64::d);
+    CODEGEN_ASSERT(dst.kind == src.kind);
+    CODEGEN_ASSERT(dst.kind == KindA64::d || dst.kind == KindA64::s || dst.kind == KindA64::q);
 
-    placeR1("frinta", dst, src, 0b000'11110'01'1'001'100'10000);
+    if (dst.kind == KindA64::q)
+        placeR1("frinta", dst, src, 0b011'01110'00'1'0000'11000'10);
+    else if (dst.kind == KindA64::d)
+        placeR1("frinta", dst, src, 0b000'11110'01'1'001'100'10000);
+    else
+        placeR1("frinta", dst, src, 0b000'11110'00'1'001'100'10000);
 }
 
 void AssemblyBuilderA64::frintm(RegisterA64 dst, RegisterA64 src)
 {
-    CODEGEN_ASSERT(dst.kind == KindA64::d && src.kind == KindA64::d);
+    CODEGEN_ASSERT(dst.kind == src.kind);
+    CODEGEN_ASSERT(dst.kind == KindA64::d || dst.kind == KindA64::s || dst.kind == KindA64::q);
 
-    placeR1("frintm", dst, src, 0b000'11110'01'1'001'010'10000);
+    if (dst.kind == KindA64::q)
+        placeR1("frintm", dst, src, 0b010'01110'00'1'0000'11001'10);
+    else if (dst.kind == KindA64::d)
+        placeR1("frintm", dst, src, 0b000'11110'01'1'001'010'10000);
+    else
+        placeR1("frintm", dst, src, 0b000'11110'00'1'001'010'10000);
 }
 
 void AssemblyBuilderA64::frintp(RegisterA64 dst, RegisterA64 src)
 {
-    CODEGEN_ASSERT(dst.kind == KindA64::d && src.kind == KindA64::d);
+    CODEGEN_ASSERT(dst.kind == src.kind);
+    CODEGEN_ASSERT(dst.kind == KindA64::d || dst.kind == KindA64::s || dst.kind == KindA64::q);
 
-    placeR1("frintp", dst, src, 0b000'11110'01'1'001'001'10000);
+    if (dst.kind == KindA64::q)
+        placeR1("frintp", dst, src, 0b010'01110'10'1'0000'11000'10);
+    else if (dst.kind == KindA64::d)
+        placeR1("frintp", dst, src, 0b000'11110'01'1'001'001'10000);
+    else
+        placeR1("frintp", dst, src, 0b000'11110'00'1'001'001'10000);
 }
 
 void AssemblyBuilderA64::fcvt(RegisterA64 dst, RegisterA64 src)
@@ -830,10 +1044,23 @@ void AssemblyBuilderA64::scvtf(RegisterA64 dst, RegisterA64 src)
 
 void AssemblyBuilderA64::ucvtf(RegisterA64 dst, RegisterA64 src)
 {
-    CODEGEN_ASSERT(dst.kind == KindA64::d);
-    CODEGEN_ASSERT(src.kind == KindA64::w || src.kind == KindA64::x);
+    if (FFlag::LuauCodegenUintToFloat)
+    {
+        CODEGEN_ASSERT(dst.kind == KindA64::d || dst.kind == KindA64::s);
+        CODEGEN_ASSERT(src.kind == KindA64::w || src.kind == KindA64::x);
 
-    placeR1("ucvtf", dst, src, 0b000'11110'01'1'00'011'000000);
+        if (dst.kind == KindA64::d)
+            placeR1("ucvtf", dst, src, 0b000'11110'01'1'00'011'000000);
+        else
+            placeR1("ucvtf", dst, src, 0b000'11110'00'1'00'011'000000);
+    }
+    else
+    {
+        CODEGEN_ASSERT(dst.kind == KindA64::d);
+        CODEGEN_ASSERT(src.kind == KindA64::w || src.kind == KindA64::x);
+
+        placeR1("ucvtf", dst, src, 0b000'11110'01'1'00'011'000000);
+    }
 }
 
 void AssemblyBuilderA64::fjcvtzs(RegisterA64 dst, RegisterA64 src)
@@ -847,23 +1074,34 @@ void AssemblyBuilderA64::fjcvtzs(RegisterA64 dst, RegisterA64 src)
 
 void AssemblyBuilderA64::fcmp(RegisterA64 src1, RegisterA64 src2)
 {
-    CODEGEN_ASSERT(src1.kind == KindA64::d && src2.kind == KindA64::d);
+    CODEGEN_ASSERT(src1.kind == src2.kind);
+    CODEGEN_ASSERT(src1.kind == KindA64::d || src1.kind == KindA64::s);
 
-    placeFCMP("fcmp", src1, src2, 0b11110'01'1, 0b00);
+    if (src1.kind == KindA64::d)
+        placeFCMP("fcmp", src1, src2, 0b11110'01'1, 0b00);
+    else
+        placeFCMP("fcmp", src1, src2, 0b11110'00'1, 0b00);
 }
 
 void AssemblyBuilderA64::fcmpz(RegisterA64 src)
 {
-    CODEGEN_ASSERT(src.kind == KindA64::d);
+    CODEGEN_ASSERT(src.kind == KindA64::d || src.kind == KindA64::s);
 
-    placeFCMP("fcmp", src, RegisterA64{src.kind, 0}, 0b11110'01'1, 0b01);
+    if (src.kind == KindA64::d)
+        placeFCMP("fcmp", src, RegisterA64{src.kind, 0}, 0b11110'01'1, 0b01);
+    else
+        placeFCMP("fcmp", src, RegisterA64{src.kind, 0}, 0b11110'00'1, 0b01);
 }
 
 void AssemblyBuilderA64::fcsel(RegisterA64 dst, RegisterA64 src1, RegisterA64 src2, ConditionA64 cond)
 {
-    CODEGEN_ASSERT(dst.kind == KindA64::d);
+    CODEGEN_ASSERT(dst.kind == src1.kind && src1.kind == src2.kind);
+    CODEGEN_ASSERT(dst.kind == KindA64::d || dst.kind == KindA64::s);
 
-    placeCS("fcsel", dst, src1, src2, cond, 0b11110'01'1, 0b11);
+    if (src1.kind == KindA64::d)
+        placeCS("fcsel", dst, src1, src2, cond, 0b11110'01'1, 0b11);
+    else
+        placeCS("fcsel", dst, src1, src2, cond, 0b11110'00'1, 0b11);
 }
 
 void AssemblyBuilderA64::udf()
@@ -942,7 +1180,7 @@ uint32_t AssemblyBuilderA64::getCodeSize() const
 
 unsigned AssemblyBuilderA64::getInstructionCount() const
 {
-    return unsigned(getCodeSize()) / 4;
+    return FFlag::LuauCodegenLocationEndFix ? unsigned(getCodeSize()) : unsigned(getCodeSize()) / 4;
 }
 
 bool AssemblyBuilderA64::isMaskSupported(uint32_t mask)
@@ -954,9 +1192,14 @@ bool AssemblyBuilderA64::isMaskSupported(uint32_t mask)
            (mask >> rz) == (1u << (32 - lz - rz)) - 1; // sequence of 1s must be contiguous
 }
 
-bool AssemblyBuilderA64::isFmovSupported(double value)
+bool AssemblyBuilderA64::isFmovSupportedFp64(double value)
 {
-    return getFmovImm(value) >= 0;
+    return getFmovImmFp64(value) >= 0;
+}
+
+bool AssemblyBuilderA64::isFmovSupportedFp32(float value)
+{
+    return getFmovImmFp32(value) >= 0;
 }
 
 void AssemblyBuilderA64::place0(const char* name, uint32_t op)
