@@ -235,6 +235,14 @@ Udata* newUserdata(lua_State* L, size_t s, int tag)
     return u;
 }
 
+void getImport(lua_State* L, StkId res, unsigned id, unsigned pc)
+{
+    Closure* cl = clvalue(L->ci->func);
+    L->ci->savedpc = cl->l.p->code + pc;
+
+    luaV_getimport(L, cl->env, cl->l.p->k, res, id, /*propagatenil*/ false);
+}
+
 // Extracted as-is from lvmexecute.cpp with the exception of control flow (reentry) and removed interrupts/savedpc
 Closure* callFallback(lua_State* L, StkId ra, StkId argtop, int nresults)
 {
