@@ -9,6 +9,11 @@
 
 #include "lobject.h"
 
+LUAU_FASTFLAGVARIABLE(LuauCodegenGcoDse2)
+LUAU_FASTFLAG(LuauCodegenNumIntFolds2)
+LUAU_FASTFLAG(LuauCodegenBufferRangeMerge3)
+LUAU_FASTFLAGVARIABLE(LuauCodegenDsoPairTrackFix)
+
 // TODO: optimization can be improved by knowing which registers are live in at each VM exit
 
 namespace Luau
@@ -16,8 +21,6 @@ namespace Luau
 namespace CodeGen
 {
 
-constexpr uint8_t kUnknownTag = 0xff;
-
 // Luau value structure reminder:
 // [              TValue             ]
 // [     Value     ][ Extra ][  Tag  ]
@@ -40,8 +43,9 @@ struct StoreRegInfo
 
 struct RemoveDeadStoreState
 {
-    RemoveDeadStoreState(IrFunction& function)
+    RemoveDeadStoreState(IrFunction& function, std::vector<uint32_t>& remainingUses)
         : function(function)
+        , remainingUses(remainingUses)
     {
         maxReg = function.proto ? function.proto->maxstacksize : 255;
     }
@@ -68,40 +72,87 @@ struct RemoveDeadStoreState
         }
     }
 
-    void killTagAndValueStorePair(StoreRegInfo& regInfo)
+    bool tagValuePairEstablished(StoreRegInfo& regInfo)
     {
         bool tagEstablished = regInfo.tagInstIdx != ~0u || regInfo.knownTag != kUnknownTag;
 
         // When tag is 'nil', we don't need to remove the unused value store
         bool valueEstablished = regInfo.valueInstIdx != ~0u || regInfo.knownTag == LUA_TNIL;
 
-        // Partial stores can only be removed if the whole pair is established
-        if (tagEstablished && valueEstablished)
+        return tagEstablished && valueEstablished;
+    }
+
+    void killTagAndValueStorePair(StoreRegInfo& regInfo)
+    {
+        if (FFlag::LuauCodegenDsoPairTrackFix)
         {
-            if (regInfo.tagInstIdx != ~0u)
+            // Partial stores can only be removed if the whole pair is established
+            if (tagValuePairEstablished(regInfo))
             {
-                kill(function, function.instructions[regInfo.tagInstIdx]);
-                regInfo.tagInstIdx = ~0u;
-            }
+                if (regInfo.tagInstIdx != ~0u)
+                {
+                    kill(function, function.instructions[regInfo.tagInstIdx]);
+                    regInfo.tagInstIdx = ~0u;
+                }
 
-            if (regInfo.valueInstIdx != ~0u)
+                if (regInfo.valueInstIdx != ~0u)
+                {
+                    kill(function, function.instructions[regInfo.valueInstIdx]);
+                    regInfo.valueInstIdx = ~0u;
+                }
+
+                regInfo.maybeGco = false;
+            }
+        }
+        else
+        {
+            bool tagEstablished = regInfo.tagInstIdx != ~0u || regInfo.knownTag != kUnknownTag;
+
+            // When tag is 'nil', we don't need to remove the unused value store
+            bool valueEstablished = regInfo.valueInstIdx != ~0u || regInfo.knownTag == LUA_TNIL;
+
+            // Partial stores can only be removed if the whole pair is established
+            if (tagEstablished && valueEstablished)
             {
-                kill(function, function.instructions[regInfo.valueInstIdx]);
-                regInfo.valueInstIdx = ~0u;
-            }
+                if (regInfo.tagInstIdx != ~0u)
+                {
+                    kill(function, function.instructions[regInfo.tagInstIdx]);
+                    regInfo.tagInstIdx = ~0u;
+                }
 
-            regInfo.maybeGco = false;
+                if (regInfo.valueInstIdx != ~0u)
+                {
+                    kill(function, function.instructions[regInfo.valueInstIdx]);
+                    regInfo.valueInstIdx = ~0u;
+                }
+
+                regInfo.maybeGco = false;
+            }
         }
     }
 
     void killTValueStore(StoreRegInfo& regInfo)
     {
-        if (regInfo.tvalueInstIdx != ~0u)
+        if (FFlag::LuauCodegenGcoDse2)
         {
-            kill(function, function.instructions[regInfo.tvalueInstIdx]);
+            // TValue can only be killed if it is not overlayed by a partial tag/value write
+            if (regInfo.tvalueInstIdx != kInvalidInstIdx && regInfo.tagInstIdx == kInvalidInstIdx && regInfo.valueInstIdx == kInvalidInstIdx)
+            {
+                kill(function, function.instructions[regInfo.tvalueInstIdx]);
 
-            regInfo.tvalueInstIdx = ~0u;
-            regInfo.maybeGco = false;
+                regInfo.tvalueInstIdx = kInvalidInstIdx;
+                regInfo.maybeGco = false;
+            }
+        }
+        else
+        {
+            if (regInfo.tvalueInstIdx != kInvalidInstIdx)
+            {
+                kill(function, function.instructions[regInfo.tvalueInstIdx]);
+
+                regInfo.tvalueInstIdx = kInvalidInstIdx;
+                regInfo.maybeGco = false;
+            }
         }
     }
 
@@ -117,6 +168,13 @@ struct RemoveDeadStoreState
         killTagAndValueStorePair(regInfo);
         killTValueStore(regInfo);
 
+        if (FFlag::LuauCodegenGcoDse2)
+        {
+            regInfo.tagInstIdx = kInvalidInstIdx;
+            regInfo.valueInstIdx = kInvalidInstIdx;
+            regInfo.tvalueInstIdx = kInvalidInstIdx;
+        }
+
         // Opaque register definition removes the knowledge of the actual tag value
         regInfo.knownTag = kUnknownTag;
     }
@@ -270,6 +328,19 @@ struct RemoveDeadStoreState
         hasGcoToClear = false;
     }
 
+    bool hasRemainingUses(uint32_t instIdx)
+    {
+        IrInst& inst = function.instructions[instIdx];
+
+        return anyArgumentMatch(
+            inst,
+            [&](IrOp op)
+            {
+                return op.kind == IrOpKind::Inst && remainingUses[op.index] != 0;
+            }
+        );
+    }
+
     // Partial clear of information about registers that might contain a GC object
     // This is used by instructions that might perform a GC assist and GC needs all pointers to be pinned to stack
     void flushGcoRegs()
@@ -283,11 +354,32 @@ struct RemoveDeadStoreState
                 // If we happen to know the exact tag, it has to be a GCO, otherwise 'maybeGCO' should be false
                 CODEGEN_ASSERT(regInfo.knownTag == kUnknownTag || isGCO(regInfo.knownTag));
 
-                // Indirect register read by GC doesn't clear the known tag
-                regInfo.tagInstIdx = ~0u;
-                regInfo.valueInstIdx = ~0u;
-                regInfo.tvalueInstIdx = ~0u;
-                regInfo.maybeGco = false;
+                if (FFlag::LuauCodegenGcoDse2)
+                {
+                    // If the values stored are still used and might be a GCO object, we have to pin in to the stack
+                    // And we have to pin all components of the register containing GCO
+                    bool tagUsedAfter = regInfo.tagInstIdx != ~0u && hasRemainingUses(regInfo.tagInstIdx);
+                    bool valueUsedAfter = regInfo.valueInstIdx != ~0u && hasRemainingUses(regInfo.valueInstIdx);
+                    bool tvalueUsedAfter = regInfo.tvalueInstIdx != ~0u && hasRemainingUses(regInfo.tvalueInstIdx);
+
+                    if (tagUsedAfter || valueUsedAfter || tvalueUsedAfter)
+                    {
+                        regInfo.tagInstIdx = ~0u;
+                        regInfo.valueInstIdx = ~0u;
+                        regInfo.tvalueInstIdx = ~0u;
+                    }
+
+                    // Indirect register read by GC doesn't clear the known tag
+                    regInfo.maybeGco = false;
+                }
+                else
+                {
+                    // Indirect register read by GC doesn't clear the known tag
+                    regInfo.tagInstIdx = ~0u;
+                    regInfo.valueInstIdx = ~0u;
+                    regInfo.tvalueInstIdx = ~0u;
+                    regInfo.maybeGco = false;
+                }
             }
         }
 
@@ -295,12 +387,16 @@ struct RemoveDeadStoreState
     }
 
     IrFunction& function;
+    std::vector<uint32_t>& remainingUses;
 
     std::array<StoreRegInfo, 256> info;
     int maxReg = 255;
 
     // Some of the registers contain values which might be a GC object
     bool hasGcoToClear = false;
+
+    // Have there been any object allocations which might remain unused
+    bool hasAllocations = false;
 };
 
 static bool tryReplaceTagWithFullStore(
@@ -328,16 +424,16 @@ static bool tryReplaceTagWithFullStore(
 
             if (prevValueInst.cmd == IrCmd::STORE_VECTOR)
             {
-                CODEGEN_ASSERT(prevValueInst.e.kind == IrOpKind::None);
-                IrOp prevValueX = prevValueInst.b;
-                IrOp prevValueY = prevValueInst.c;
-                IrOp prevValueZ = prevValueInst.d;
-                replace(function, block, instIndex, IrInst{IrCmd::STORE_VECTOR, targetOp, prevValueX, prevValueY, prevValueZ, tagOp});
+                CODEGEN_ASSERT(OP_E(prevValueInst).kind == IrOpKind::None);
+                IrOp prevValueX = OP_B(prevValueInst);
+                IrOp prevValueY = OP_C(prevValueInst);
+                IrOp prevValueZ = OP_D(prevValueInst);
+                replace(function, block, instIndex, IrInst{IrCmd::STORE_VECTOR, {targetOp, prevValueX, prevValueY, prevValueZ, tagOp}});
             }
             else
             {
-                IrOp prevValueOp = prevValueInst.b;
-                replace(function, block, instIndex, IrInst{IrCmd::STORE_SPLIT_TVALUE, targetOp, tagOp, prevValueOp});
+                IrOp prevValueOp = OP_B(prevValueInst);
+                replace(function, block, instIndex, IrInst{IrCmd::STORE_SPLIT_TVALUE, {targetOp, tagOp, prevValueOp}});
             }
         }
 
@@ -358,13 +454,13 @@ static bool tryReplaceTagWithFullStore(
 
         if (prev.cmd == IrCmd::STORE_SPLIT_TVALUE)
         {
-            CODEGEN_ASSERT(prev.d.kind == IrOpKind::None);
+            CODEGEN_ASSERT(OP_D(prev).kind == IrOpKind::None);
 
             // If the 'nil' is stored, we keep 'STORE_TAG Rn, tnil' as it writes the 'full' TValue
             if (tag != LUA_TNIL)
             {
-                IrOp prevValueOp = prev.c;
-                replace(function, block, instIndex, IrInst{IrCmd::STORE_SPLIT_TVALUE, targetOp, tagOp, prevValueOp});
+                IrOp prevValueOp = OP_C(prev);
+                replace(function, block, instIndex, IrInst{IrCmd::STORE_SPLIT_TVALUE, {targetOp, tagOp, prevValueOp}});
             }
 
             state.killTValueStore(regInfo);
@@ -380,10 +476,10 @@ static bool tryReplaceTagWithFullStore(
             // If the 'nil' is stored, we keep 'STORE_TAG Rn, tnil' as it writes the 'full' TValue
             if (tag != LUA_TNIL)
             {
-                IrOp prevValueX = prev.b;
-                IrOp prevValueY = prev.c;
-                IrOp prevValueZ = prev.d;
-                replace(function, block, instIndex, IrInst{IrCmd::STORE_VECTOR, targetOp, prevValueX, prevValueY, prevValueZ, tagOp});
+                IrOp prevValueX = OP_B(prev);
+                IrOp prevValueY = OP_C(prev);
+                IrOp prevValueZ = OP_D(prev);
+                replace(function, block, instIndex, IrInst{IrCmd::STORE_VECTOR, {targetOp, prevValueX, prevValueY, prevValueZ, tagOp}});
             }
 
             state.killTValueStore(regInfo);
@@ -413,11 +509,11 @@ static bool tryReplaceValueWithFullStore(
     // If the tag+value pair is established, we can mark both as dead and use a single split TValue store
     if (regInfo.tagInstIdx != ~0u && regInfo.valueInstIdx != ~0u)
     {
-        IrOp prevTagOp = function.instructions[regInfo.tagInstIdx].b;
+        IrOp prevTagOp = OP_B(function.instructions[regInfo.tagInstIdx]);
         uint8_t prevTag = function.tagOp(prevTagOp);
 
         CODEGEN_ASSERT(regInfo.knownTag == prevTag);
-        replace(function, block, instIndex, IrInst{IrCmd::STORE_SPLIT_TVALUE, targetOp, prevTagOp, valueOp});
+        replace(function, block, instIndex, IrInst{IrCmd::STORE_SPLIT_TVALUE, {targetOp, prevTagOp, valueOp}});
 
         state.killTagStore(regInfo);
         state.killValueStore(regInfo);
@@ -433,12 +529,12 @@ static bool tryReplaceValueWithFullStore(
 
         if (prev.cmd == IrCmd::STORE_SPLIT_TVALUE)
         {
-            IrOp prevTagOp = prev.b;
+            IrOp prevTagOp = OP_B(prev);
             uint8_t prevTag = function.tagOp(prevTagOp);
 
             CODEGEN_ASSERT(regInfo.knownTag == prevTag);
-            CODEGEN_ASSERT(prev.d.kind == IrOpKind::None);
-            replace(function, block, instIndex, IrInst{IrCmd::STORE_SPLIT_TVALUE, targetOp, prevTagOp, valueOp});
+            CODEGEN_ASSERT(OP_D(prev).kind == IrOpKind::None);
+            replace(function, block, instIndex, IrInst{IrCmd::STORE_SPLIT_TVALUE, {targetOp, prevTagOp, valueOp}});
 
             state.killTValueStore(regInfo);
 
@@ -447,12 +543,22 @@ static bool tryReplaceValueWithFullStore(
         }
         else if (prev.cmd == IrCmd::STORE_VECTOR)
         {
-            IrOp prevTagOp = prev.e;
+            IrOp prevTagOp = OP_E(prev);
             CODEGEN_ASSERT(prevTagOp.kind != IrOpKind::None);
             uint8_t prevTag = function.tagOp(prevTagOp);
 
             CODEGEN_ASSERT(regInfo.knownTag == prevTag);
-            replace(function, block, instIndex, IrInst{IrCmd::STORE_SPLIT_TVALUE, targetOp, prevTagOp, valueOp});
+            replace(function, block, instIndex, IrInst{IrCmd::STORE_SPLIT_TVALUE, {targetOp, prevTagOp, valueOp}});
+
+            state.killTValueStore(regInfo);
+
+            regInfo.tvalueInstIdx = instIndex;
+            return true;
+        }
+        else if (FFlag::LuauCodegenDsoPairTrackFix && prev.cmd == IrCmd::STORE_TVALUE && regInfo.knownTag != kUnknownTag)
+        {
+            IrOp prevTagOp = build.constTag(regInfo.knownTag);
+            replace(function, block, instIndex, IrInst{IrCmd::STORE_SPLIT_TVALUE, {targetOp, prevTagOp, valueOp}});
 
             state.killTValueStore(regInfo);
 
@@ -476,14 +582,14 @@ static bool tryReplaceVectorValueWithFullStore(
     // If the tag+value pair is established, we can mark both as dead and use a single split TValue store
     if (regInfo.tagInstIdx != ~0u && regInfo.valueInstIdx != ~0u)
     {
-        IrOp prevTagOp = function.instructions[regInfo.tagInstIdx].b;
+        IrOp prevTagOp = OP_B(function.instructions[regInfo.tagInstIdx]);
         uint8_t prevTag = function.tagOp(prevTagOp);
 
         CODEGEN_ASSERT(regInfo.knownTag == prevTag);
 
         IrInst& storeInst = function.instructions[instIndex];
         CODEGEN_ASSERT(storeInst.cmd == IrCmd::STORE_VECTOR);
-        replace(function, storeInst.e, prevTagOp);
+        replace(function, OP_E(storeInst), prevTagOp);
 
         state.killTagStore(regInfo);
         state.killValueStore(regInfo);
@@ -499,15 +605,15 @@ static bool tryReplaceVectorValueWithFullStore(
 
         if (prev.cmd == IrCmd::STORE_SPLIT_TVALUE)
         {
-            IrOp prevTagOp = prev.b;
+            IrOp prevTagOp = OP_B(prev);
             uint8_t prevTag = function.tagOp(prevTagOp);
 
             CODEGEN_ASSERT(regInfo.knownTag == prevTag);
-            CODEGEN_ASSERT(prev.d.kind == IrOpKind::None);
+            CODEGEN_ASSERT(OP_D(prev).kind == IrOpKind::None);
 
             IrInst& storeInst = function.instructions[instIndex];
             CODEGEN_ASSERT(storeInst.cmd == IrCmd::STORE_VECTOR);
-            replace(function, storeInst.e, prevTagOp);
+            replace(function, OP_E(storeInst), prevTagOp);
 
             state.killTValueStore(regInfo);
 
@@ -516,7 +622,7 @@ static bool tryReplaceVectorValueWithFullStore(
         }
         else if (prev.cmd == IrCmd::STORE_VECTOR)
         {
-            IrOp prevTagOp = prev.e;
+            IrOp prevTagOp = OP_E(prev);
             CODEGEN_ASSERT(prevTagOp.kind != IrOpKind::None);
             uint8_t prevTag = function.tagOp(prevTagOp);
 
@@ -524,7 +630,7 @@ static bool tryReplaceVectorValueWithFullStore(
 
             IrInst& storeInst = function.instructions[instIndex];
             CODEGEN_ASSERT(storeInst.cmd == IrCmd::STORE_VECTOR);
-            replace(function, storeInst.e, prevTagOp);
+            replace(function, OP_E(storeInst), prevTagOp);
 
             state.killTValueStore(regInfo);
 
@@ -536,26 +642,50 @@ static bool tryReplaceVectorValueWithFullStore(
     return false;
 }
 
+static void updateRemainingUses(RemoveDeadStoreState& state, IrInst& inst, uint32_t index)
+{
+    state.remainingUses[index] = inst.useCount;
+
+    visitArguments(
+        inst,
+        [&](IrOp op)
+        {
+            if (op.kind == IrOpKind::Inst)
+            {
+                CODEGEN_ASSERT(state.remainingUses[op.index] != 0);
+                state.remainingUses[op.index]--;
+            }
+        }
+    );
+}
+
 static void markDeadStoresInInst(RemoveDeadStoreState& state, IrBuilder& build, IrFunction& function, IrBlock& block, IrInst& inst, uint32_t index)
 {
+    if (FFlag::LuauCodegenGcoDse2)
+        updateRemainingUses(state, inst, index);
+
     switch (inst.cmd)
     {
     case IrCmd::STORE_TAG:
-        if (inst.a.kind == IrOpKind::VmReg)
+        if (OP_A(inst).kind == IrOpKind::VmReg)
         {
-            int reg = vmRegOp(inst.a);
+            int reg = vmRegOp(OP_A(inst));
 
             if (function.cfg.captured.regs.test(reg))
                 return;
 
             StoreRegInfo& regInfo = state.info[reg];
 
-            if (tryReplaceTagWithFullStore(state, build, function, block, index, inst.a, inst.b, regInfo))
+            if (tryReplaceTagWithFullStore(state, build, function, block, index, OP_A(inst), OP_B(inst), regInfo))
                 break;
 
-            uint8_t tag = function.tagOp(inst.b);
+            uint8_t tag = function.tagOp(OP_B(inst));
 
             regInfo.tagInstIdx = index;
+
+            if (FFlag::LuauCodegenDsoPairTrackFix && state.tagValuePairEstablished(regInfo))
+                regInfo.tvalueInstIdx = kInvalidInstIdx;
+
             regInfo.maybeGco = isGCO(tag);
             regInfo.knownTag = tag;
             state.hasGcoToClear |= regInfo.maybeGco;
@@ -563,22 +693,22 @@ static void markDeadStoresInInst(RemoveDeadStoreState& state, IrBuilder& build,
         break;
     case IrCmd::STORE_EXTRA:
         // To simplify, extra field store is preserved along with all other stores made so far
-        if (inst.a.kind == IrOpKind::VmReg)
+        if (OP_A(inst).kind == IrOpKind::VmReg)
         {
-            state.useReg(vmRegOp(inst.a));
+            state.useReg(vmRegOp(OP_A(inst)));
         }
         break;
     case IrCmd::STORE_POINTER:
-        if (inst.a.kind == IrOpKind::VmReg)
+        if (OP_A(inst).kind == IrOpKind::VmReg)
         {
-            int reg = vmRegOp(inst.a);
+            int reg = vmRegOp(OP_A(inst));
 
             if (function.cfg.captured.regs.test(reg))
                 return;
 
             StoreRegInfo& regInfo = state.info[reg];
 
-            if (tryReplaceValueWithFullStore(state, build, function, block, index, inst.a, inst.b, regInfo))
+            if (tryReplaceValueWithFullStore(state, build, function, block, index, OP_A(inst), OP_B(inst), regInfo))
             {
                 regInfo.maybeGco = true;
                 state.hasGcoToClear |= true;
@@ -590,22 +720,26 @@ static void markDeadStoresInInst(RemoveDeadStoreState& state, IrBuilder& build,
                 state.killValueStore(regInfo);
 
             regInfo.valueInstIdx = index;
+
+            if (FFlag::LuauCodegenDsoPairTrackFix && state.tagValuePairEstablished(regInfo))
+                regInfo.tvalueInstIdx = kInvalidInstIdx;
+
             regInfo.maybeGco = true;
             state.hasGcoToClear = true;
         }
         break;
     case IrCmd::STORE_DOUBLE:
     case IrCmd::STORE_INT:
-        if (inst.a.kind == IrOpKind::VmReg)
+        if (OP_A(inst).kind == IrOpKind::VmReg)
         {
-            int reg = vmRegOp(inst.a);
+            int reg = vmRegOp(OP_A(inst));
 
             if (function.cfg.captured.regs.test(reg))
                 return;
 
             StoreRegInfo& regInfo = state.info[reg];
 
-            if (tryReplaceValueWithFullStore(state, build, function, block, index, inst.a, inst.b, regInfo))
+            if (tryReplaceValueWithFullStore(state, build, function, block, index, OP_A(inst), OP_B(inst), regInfo))
                 break;
 
             // Partial value store can be removed by a new one if the tag is known
@@ -613,13 +747,17 @@ static void markDeadStoresInInst(RemoveDeadStoreState& state, IrBuilder& build,
                 state.killValueStore(regInfo);
 
             regInfo.valueInstIdx = index;
+
+            if (FFlag::LuauCodegenDsoPairTrackFix && state.tagValuePairEstablished(regInfo))
+                regInfo.tvalueInstIdx = kInvalidInstIdx;
+
             regInfo.maybeGco = false;
         }
         break;
     case IrCmd::STORE_VECTOR:
-        if (inst.a.kind == IrOpKind::VmReg)
+        if (OP_A(inst).kind == IrOpKind::VmReg)
         {
-            int reg = vmRegOp(inst.a);
+            int reg = vmRegOp(OP_A(inst));
 
             if (function.cfg.captured.regs.test(reg))
                 return;
@@ -634,13 +772,17 @@ static void markDeadStoresInInst(RemoveDeadStoreState& state, IrBuilder& build,
                 state.killValueStore(regInfo);
 
             regInfo.valueInstIdx = index;
+
+            if (FFlag::LuauCodegenDsoPairTrackFix && state.tagValuePairEstablished(regInfo))
+                regInfo.tvalueInstIdx = kInvalidInstIdx;
+
             regInfo.maybeGco = false;
         }
         break;
     case IrCmd::STORE_TVALUE:
-        if (inst.a.kind == IrOpKind::VmReg)
+        if (OP_A(inst).kind == IrOpKind::VmReg)
         {
-            int reg = vmRegOp(inst.a);
+            int reg = vmRegOp(OP_A(inst));
 
             if (function.cfg.captured.regs.test(reg))
                 return;
@@ -650,31 +792,46 @@ static void markDeadStoresInInst(RemoveDeadStoreState& state, IrBuilder& build,
             state.killTagAndValueStorePair(regInfo);
             state.killTValueStore(regInfo);
 
-            regInfo.tvalueInstIdx = index;
-            regInfo.maybeGco = true;
-
-            // We do not use tag inference from the source instruction here as it doesn't provide useful opportunities for dead store removal
-            regInfo.knownTag = kUnknownTag;
-
-            // If the argument is a vector, it's not a GC object
-            // Note that for known boolean/number/GCO, we already optimize into STORE_SPLIT_TVALUE form
-            // TODO (CLI-101027): similar code is used in constant propagation optimization and should be shared in utilities
-            if (IrInst* arg = function.asInstOp(inst.b))
+            if (FFlag::LuauCodegenGcoDse2)
             {
-                if (arg->cmd == IrCmd::TAG_VECTOR)
-                    regInfo.maybeGco = false;
+                regInfo.tagInstIdx = kInvalidInstIdx;
+                regInfo.valueInstIdx = kInvalidInstIdx;
+            }
 
-                if (arg->cmd == IrCmd::LOAD_TVALUE && arg->c.kind != IrOpKind::None)
-                    regInfo.maybeGco = isGCO(function.tagOp(arg->c));
+            regInfo.tvalueInstIdx = index;
+
+            if (FFlag::LuauCodegenDsoPairTrackFix)
+            {
+                regInfo.knownTag = tryGetOperandTag(function, OP_B(inst)).value_or(kUnknownTag);
+                regInfo.maybeGco = regInfo.knownTag == kUnknownTag || isGCO(regInfo.knownTag);
+            }
+            else
+            {
+                regInfo.maybeGco = true;
+
+                // We do not use tag inference from the source instruction here as it doesn't provide useful opportunities for dead store removal
+                regInfo.knownTag = kUnknownTag;
+
+                // If the argument is a vector, it's not a GC object
+                // Note that for known boolean/number/GCO, we already optimize into STORE_SPLIT_TVALUE form
+                // TODO (CLI-101027): similar code is used in constant propagation optimization and should be shared in utilities
+                if (IrInst* arg = function.asInstOp(OP_B(inst)))
+                {
+                    if (arg->cmd == IrCmd::TAG_VECTOR)
+                        regInfo.maybeGco = false;
+
+                    if (arg->cmd == IrCmd::LOAD_TVALUE && OP_C(arg).kind != IrOpKind::None)
+                        regInfo.maybeGco = isGCO(function.tagOp(OP_C(arg)));
+                }
             }
 
             state.hasGcoToClear |= regInfo.maybeGco;
         }
         break;
     case IrCmd::STORE_SPLIT_TVALUE:
-        if (inst.a.kind == IrOpKind::VmReg)
+        if (OP_A(inst).kind == IrOpKind::VmReg)
         {
-            int reg = vmRegOp(inst.a);
+            int reg = vmRegOp(OP_A(inst));
 
             if (function.cfg.captured.regs.test(reg))
                 return;
@@ -684,65 +841,77 @@ static void markDeadStoresInInst(RemoveDeadStoreState& state, IrBuilder& build,
             state.killTagAndValueStorePair(regInfo);
             state.killTValueStore(regInfo);
 
+            if (FFlag::LuauCodegenGcoDse2)
+            {
+                regInfo.tagInstIdx = kInvalidInstIdx;
+                regInfo.valueInstIdx = kInvalidInstIdx;
+            }
+
             regInfo.tvalueInstIdx = index;
-            regInfo.maybeGco = isGCO(function.tagOp(inst.b));
-            regInfo.knownTag = function.tagOp(inst.b);
+            regInfo.maybeGco = isGCO(function.tagOp(OP_B(inst)));
+            regInfo.knownTag = function.tagOp(OP_B(inst));
             state.hasGcoToClear |= regInfo.maybeGco;
         }
         break;
 
         // Guard checks can jump to a block which might be using some or all the values we stored
     case IrCmd::CHECK_TAG:
-        state.checkLiveIns(inst.c);
+        state.checkLiveIns(OP_C(inst));
 
         // Tag guard establishes the tag value of the register in the current block
-        if (IrInst* load = function.asInstOp(inst.a); load && load->cmd == IrCmd::LOAD_TAG && load->a.kind == IrOpKind::VmReg)
+        if (IrInst* load = function.asInstOp(OP_A(inst)); load && load->cmd == IrCmd::LOAD_TAG && OP_A(load).kind == IrOpKind::VmReg)
         {
-            int reg = vmRegOp(load->a);
+            int reg = vmRegOp(OP_A(load));
 
             StoreRegInfo& regInfo = state.info[reg];
 
-            regInfo.knownTag = function.tagOp(inst.b);
+            regInfo.knownTag = function.tagOp(OP_B(inst));
         }
         break;
     case IrCmd::TRY_NUM_TO_INDEX:
-        state.checkLiveIns(inst.b);
+        state.checkLiveIns(OP_B(inst));
         break;
     case IrCmd::TRY_CALL_FASTGETTM:
-        state.checkLiveIns(inst.c);
+        state.checkLiveIns(OP_C(inst));
         break;
     case IrCmd::CHECK_FASTCALL_RES:
-        state.checkLiveIns(inst.b);
+        state.checkLiveIns(OP_B(inst));
         break;
     case IrCmd::CHECK_TRUTHY:
-        state.checkLiveIns(inst.c);
+        state.checkLiveIns(OP_C(inst));
         break;
     case IrCmd::CHECK_READONLY:
-        state.checkLiveIns(inst.b);
+        state.checkLiveIns(OP_B(inst));
         break;
     case IrCmd::CHECK_NO_METATABLE:
-        state.checkLiveIns(inst.b);
+        state.checkLiveIns(OP_B(inst));
         break;
     case IrCmd::CHECK_SAFE_ENV:
-        state.checkLiveIns(inst.a);
+        state.checkLiveIns(OP_A(inst));
         break;
     case IrCmd::CHECK_ARRAY_SIZE:
-        state.checkLiveIns(inst.c);
+        state.checkLiveIns(OP_C(inst));
         break;
     case IrCmd::CHECK_SLOT_MATCH:
-        state.checkLiveIns(inst.c);
+        state.checkLiveIns(OP_C(inst));
         break;
     case IrCmd::CHECK_NODE_NO_NEXT:
-        state.checkLiveIns(inst.b);
+        state.checkLiveIns(OP_B(inst));
         break;
     case IrCmd::CHECK_NODE_VALUE:
-        state.checkLiveIns(inst.b);
+        state.checkLiveIns(OP_B(inst));
         break;
     case IrCmd::CHECK_BUFFER_LEN:
-        state.checkLiveIns(inst.d);
+        if (FFlag::LuauCodegenBufferRangeMerge3 && FFlag::LuauCodegenNumIntFolds2)
+            state.checkLiveIns(OP_F(inst));
+        else
+            state.checkLiveIns(OP_D(inst));
         break;
     case IrCmd::CHECK_USERDATA_TAG:
-        state.checkLiveIns(inst.c);
+        state.checkLiveIns(OP_C(inst));
+        break;
+    case IrCmd::CHECK_CMP_INT:
+        state.checkLiveIns(OP_D(inst));
         break;
 
     case IrCmd::JUMP:
@@ -766,7 +935,7 @@ static void markDeadStoresInInst(RemoveDeadStoreState& state, IrBuilder& build,
     case IrCmd::DO_LEN:
     case IrCmd::GET_TABLE:
     case IrCmd::SET_TABLE:
-    case IrCmd::GET_IMPORT:
+    case IrCmd::GET_CACHED_IMPORT:
     case IrCmd::CONCAT:
     case IrCmd::INTERRUPT:
     case IrCmd::CHECK_GC:
@@ -785,6 +954,11 @@ static void markDeadStoresInInst(RemoveDeadStoreState& state, IrBuilder& build,
         visitVmRegDefsUses(state, function, inst);
         break;
 
+    case IrCmd::NEW_USERDATA:
+        if (FFlag::LuauCodegenGcoDse2)
+            state.hasAllocations = true;
+        break;
+
     default:
         // Guards have to be covered explicitly
         CODEGEN_ASSERT(!isNonTerminatingJump(inst.cmd));
@@ -807,11 +981,24 @@ static void markDeadStoresInBlock(IrBuilder& build, IrBlock& block, RemoveDeadSt
     }
 }
 
-static void markDeadStoresInBlockChain(IrBuilder& build, std::vector<uint8_t>& visited, IrBlock* block)
+static void markDeadStoresInBlockChain(
+    IrBuilder& build,
+    std::vector<uint8_t>& visited,
+    std::vector<uint32_t>& remainingUses,
+    std::vector<uint32_t>& blockIdxChain,
+    IrBlock* block
+)
 {
     IrFunction& function = build.function;
 
-    RemoveDeadStoreState state{function};
+    RemoveDeadStoreState state{function, remainingUses};
+
+    if (FFlag::LuauCodegenGcoDse2)
+    {
+        // We will be visiting this chain a few times to clean unreferenced temporaries
+        // Clear the storage we reuse
+        blockIdxChain.clear();
+    }
 
     while (block)
     {
@@ -819,6 +1006,9 @@ static void markDeadStoresInBlockChain(IrBuilder& build, std::vector<uint8_t>& v
         CODEGEN_ASSERT(!visited[blockIdx]);
         visited[blockIdx] = true;
 
+        if (FFlag::LuauCodegenGcoDse2)
+            blockIdxChain.push_back(blockIdx);
+
         markDeadStoresInBlock(build, *block, state);
 
         IrInst& termInst = function.instructions[block->finish];
@@ -827,9 +1017,9 @@ static void markDeadStoresInBlockChain(IrBuilder& build, std::vector<uint8_t>& v
 
         // Unconditional jump into a block with a single user (current block) allows us to continue optimization
         // with the information we have gathered so far (unless we have already visited that block earlier)
-        if (termInst.cmd == IrCmd::JUMP && termInst.a.kind == IrOpKind::Block)
+        if (termInst.cmd == IrCmd::JUMP && OP_A(termInst).kind == IrOpKind::Block)
         {
-            IrBlock& target = function.blockOp(termInst.a);
+            IrBlock& target = function.blockOp(OP_A(termInst));
             uint32_t targetIdx = function.getBlockIndex(target);
 
             if (target.useCount == 1 && !visited[targetIdx] && target.kind != IrBlockKind::Fallback)
@@ -838,6 +1028,74 @@ static void markDeadStoresInBlockChain(IrBuilder& build, std::vector<uint8_t>& v
 
         block = nextBlock;
     }
+
+    // If there are allocating instructions, check if they have 'read' uses after DSE
+    if (FFlag::LuauCodegenGcoDse2 && state.hasAllocations)
+    {
+        bool foundUnused = false;
+
+        // Remove uses in instructions writing to the allocations
+        for (uint32_t blockIdx : blockIdxChain)
+        {
+            IrBlock& block = function.blocks[blockIdx];
+
+            for (uint32_t index = block.start; index <= block.finish; index++)
+            {
+                IrInst& inst = function.instructions[index];
+
+                state.remainingUses[index] = inst.useCount;
+
+                switch (inst.cmd)
+                {
+                case IrCmd::BUFFER_WRITEI8:
+                case IrCmd::BUFFER_WRITEI16:
+                case IrCmd::BUFFER_WRITEI32:
+                case IrCmd::BUFFER_WRITEF32:
+                case IrCmd::BUFFER_WRITEF64:
+                    state.remainingUses[OP_A(inst).index]--;
+
+                    if (state.remainingUses[OP_A(inst).index] == 0)
+                        foundUnused = true;
+                    break;
+                default:
+                    break;
+                }
+            }
+        }
+
+        // Remove those write instructions if they were the only users of the allocation
+        if (foundUnused)
+        {
+            for (uint32_t blockIdx : blockIdxChain)
+            {
+                IrBlock& block = function.blocks[blockIdx];
+
+                for (uint32_t index = block.start; index <= block.finish; index++)
+                {
+                    IrInst& inst = function.instructions[index];
+
+                    switch (inst.cmd)
+                    {
+                    case IrCmd::BUFFER_WRITEI8:
+                    case IrCmd::BUFFER_WRITEI16:
+                    case IrCmd::BUFFER_WRITEI32:
+                    case IrCmd::BUFFER_WRITEF32:
+                    case IrCmd::BUFFER_WRITEF64:
+                        if (state.remainingUses[OP_A(inst).index] == 0)
+                        {
+                            IrInst& pointer = function.instOp(OP_A(inst));
+
+                            if (pointer.cmd == IrCmd::NEW_USERDATA)
+                                kill(function, inst);
+                        }
+                        break;
+                    default:
+                        break;
+                    }
+                }
+            }
+        }
+    }
 }
 
 void markDeadStoresInBlockChains(IrBuilder& build)
@@ -845,6 +1103,8 @@ void markDeadStoresInBlockChains(IrBuilder& build)
     IrFunction& function = build.function;
 
     std::vector<uint8_t> visited(function.blocks.size(), false);
+    std::vector<uint32_t> remainingUses(function.instructions.size(), 0u);
+    std::vector<uint32_t> blockIdxChain;
 
     for (IrBlock& block : function.blocks)
     {
@@ -854,7 +1114,7 @@ void markDeadStoresInBlockChains(IrBuilder& build)
         if (visited[function.getBlockIndex(block)])
             continue;
 
-        markDeadStoresInBlockChain(build, visited, &block);
+        markDeadStoresInBlockChain(build, visited, remainingUses, blockIdxChain, &block);
     }
 }
 
