@@ -5,10 +5,13 @@
 
 namespace Luau
 {
+
+class AstNameTable;
+
 namespace Compile
 {
 
-Constant foldBuiltin(int bfid, const Constant* args, size_t count);
+Constant foldBuiltin(AstNameTable& stringTable, int bfid, const Constant* args, size_t count);
 Constant foldBuiltinMath(AstName index);
 
 } // namespace Compile
