@@ -7,6 +7,9 @@
 
 #include <array>
 
+LUAU_FASTFLAGVARIABLE(LuauCompileMathIsNanInfFinite)
+LUAU_FASTFLAGVARIABLE(LuauCompileFastcallsSurvivePolyfills)
+
 namespace Luau
 {
 namespace Compile
@@ -22,6 +25,32 @@ Builtin getBuiltin(AstExpr* node, const DenseHashMap<AstName, Global>& globals,
     }
     else if (AstExprIndexName* expr = node->as<AstExprIndexName>())
     {
+        if (FFlag::LuauCompileFastcallsSurvivePolyfills)
+        {
+            if (AstExprLocal* object = expr->expr->as<AstExprLocal>())
+            {
+                const Variable* v = variables.find(object->local);
+
+                // Local that is initialized and not modified might hold the built-in library itself
+                if (v && !v->written && v->init)
+                {
+                    AstExprGlobal* object = nullptr;
+
+                    // Look for patterns like 'local m = math' and 'local math = math or replacement'
+                    // Fastcall is used in safe env where libraries like 'math' are built-in
+                    // This means that if we are still in safe env, 'math' was truthy and local was initialized to the library value
+                    // If safe env is false, 'math' could be a polyfill or something else and fastcall takes the fallback, preserving the behavior
+                    if (AstExprGlobal* global = v->init->as<AstExprGlobal>())
+                        object = global;
+                    else if (AstExprBinary* cond = v->init->as<AstExprBinary>(); cond && cond->op == AstExprBinary::Or)
+                        object = cond->left->as<AstExprGlobal>();
+
+                    if (object)
+                        return getGlobalState(globals, object->name) == Global::Default ? Builtin{object->name, expr->index} : Builtin();
+                }
+            }
+        }
+
         if (AstExprGlobal* object = expr->expr->as<AstExprGlobal>())
         {
             return getGlobalState(globals, object->name) == Global::Default ? Builtin{object->name, expr->index} : Builtin();
@@ -139,6 +168,16 @@ static int getBuiltinFunctionId(const Builtin& builtin, const CompileOptions& op
             return LBF_MATH_ROUND;
         if (builtin.method == "lerp")
             return LBF_MATH_LERP;
+
+        if (FFlag::LuauCompileMathIsNanInfFinite)
+        {
+            if (builtin.method == "isnan")
+                return LBF_MATH_ISNAN;
+            if (builtin.method == "isinf")
+                return LBF_MATH_ISINF;
+            if (builtin.method == "isfinite")
+                return LBF_MATH_ISFINITE;
+        }
     }
 
     if (builtin.object == "bit32")
@@ -251,6 +290,8 @@ static int getBuiltinFunctionId(const Builtin& builtin, const CompileOptions& op
             return LBF_VECTOR_MIN;
         if (builtin.method == "max")
             return LBF_VECTOR_MAX;
+        if (builtin.method == "lerp")
+            return LBF_VECTOR_LERP;
     }
 
     if (options.vectorCtor)
@@ -552,9 +593,17 @@ BuiltinInfo getBuiltinInfo(int bfid)
     case LBF_VECTOR_MIN:
     case LBF_VECTOR_MAX:
         return {-1, 1}; // variadic
+    case LBF_VECTOR_LERP:
+        return {3, 1, BuiltinInfo::Flag_NoneSafe};
 
     case LBF_MATH_LERP:
         return {3, 1, BuiltinInfo::Flag_NoneSafe};
+    case LBF_MATH_ISNAN:
+        return {1, 1, BuiltinInfo::Flag_NoneSafe};
+    case LBF_MATH_ISINF:
+        return {1, 1, BuiltinInfo::Flag_NoneSafe};
+    case LBF_MATH_ISFINITE:
+        return {1, 1, BuiltinInfo::Flag_NoneSafe};
     }
 
     LUAU_UNREACHABLE();
