@@ -2,10 +2,16 @@
 #include "ConstantFolding.h"
 
 #include "BuiltinFolding.h"
+#include "Luau/Bytecode.h"
+#include "Luau/Lexer.h"
 
 #include <vector>
 #include <math.h>
 
+LUAU_FASTFLAG(LuauExplicitTypeInstantiationSyntax)
+LUAU_FASTFLAGVARIABLE(LuauCompileFoldVectorComp)
+LUAU_FASTFLAG(LuauCompileInlinedBuiltins)
+
 namespace Luau
 {
 namespace Compile
@@ -80,7 +86,7 @@ static void foldUnary(Constant& result, AstExprUnary::Op op, const Constant& arg
     }
 }
 
-static void foldBinary(Constant& result, AstExprBinary::Op op, const Constant& la, const Constant& ra)
+static void foldBinary(Constant& result, AstExprBinary::Op op, const Constant& la, const Constant& ra, AstNameTable& stringTable)
 {
     switch (op)
     {
@@ -283,6 +289,24 @@ static void foldBinary(Constant& result, AstExprBinary::Op op, const Constant& l
         break;
 
     case AstExprBinary::Concat:
+        if (la.type == Constant::Type_String && ra.type == Constant::Type_String)
+        {
+            result.type = Constant::Type_String;
+            result.stringLength = la.stringLength + ra.stringLength;
+            if (la.stringLength == 0)
+                result.valueString = ra.valueString;
+            else if (ra.stringLength == 0)
+                result.valueString = la.valueString;
+            else
+            {
+                std::string tmp;
+                tmp.reserve(result.stringLength + 1);
+                tmp.append(la.valueString, la.stringLength);
+                tmp.append(ra.valueString, ra.stringLength);
+                AstName name = stringTable.getOrAdd(tmp.c_str(), result.stringLength);
+                result.valueString = name.value;
+            }
+        }
         break;
 
     case AstExprBinary::CompareNe:
@@ -352,6 +376,48 @@ static void foldBinary(Constant& result, AstExprBinary::Op op, const Constant& l
     }
 }
 
+static void foldInterpString(Constant& result, AstExprInterpString* expr, DenseHashMap<AstExpr*, Constant>& constants, AstNameTable& stringTable)
+{
+    LUAU_ASSERT(expr->strings.size == expr->expressions.size + 1);
+    size_t resultLength = 0;
+    for (size_t index = 0; index < expr->strings.size; ++index)
+    {
+        resultLength += expr->strings.data[index].size;
+        if (index < expr->expressions.size)
+        {
+            const Constant* c = constants.find(expr->expressions.data[index]);
+            LUAU_ASSERT(c != nullptr && c->type == Constant::Type::Type_String);
+            resultLength += c->stringLength;
+        }
+    }
+    result.type = Constant::Type_String;
+    result.stringLength = resultLength;
+
+    if (resultLength == 0)
+    {
+        result.valueString = "";
+        return;
+    }
+
+    std::string tmp;
+    tmp.reserve(resultLength);
+
+    for (size_t index = 0; index < expr->strings.size; ++index)
+    {
+        AstArray<char> string = expr->strings.data[index];
+        tmp.append(string.data, string.size);
+        if (index < expr->expressions.size)
+        {
+            const Constant* c = constants.find(expr->expressions.data[index]);
+            tmp.append(c->valueString, c->stringLength);
+        }
+    }
+    result.type = Constant::Type_String;
+    result.stringLength = resultLength;
+    AstName name = stringTable.getOrAdd(tmp.c_str(), resultLength);
+    result.valueString = name.value;
+}
+
 struct ConstantVisitor : AstVisitor
 {
     DenseHashMap<AstExpr*, Constant>& constants;
@@ -361,6 +427,7 @@ struct ConstantVisitor : AstVisitor
     const DenseHashMap<AstExprCall*, int>* builtins;
     bool foldLibraryK = false;
     LibraryMemberConstantCallback libraryMemberConstantCb;
+    AstNameTable& stringTable;
 
     bool wasEmpty = false;
 
@@ -372,7 +439,8 @@ struct ConstantVisitor : AstVisitor
         DenseHashMap<AstLocal*, Constant>& locals,
         const DenseHashMap<AstExprCall*, int>* builtins,
         bool foldLibraryK,
-        LibraryMemberConstantCallback libraryMemberConstantCb
+        LibraryMemberConstantCallback libraryMemberConstantCb,
+        AstNameTable& stringTable
     )
         : constants(constants)
         , variables(variables)
@@ -380,6 +448,7 @@ struct ConstantVisitor : AstVisitor
         , builtins(builtins)
         , foldLibraryK(foldLibraryK)
         , libraryMemberConstantCb(libraryMemberConstantCb)
+        , stringTable(stringTable)
     {
         // since we do a single pass over the tree, if the initial state was empty we don't need to clear out old entries
         wasEmpty = constants.empty() && locals.empty();
@@ -433,52 +502,132 @@ struct ConstantVisitor : AstVisitor
         {
             analyze(expr->func);
 
-            if (const int* bfid = builtins ? builtins->find(expr) : nullptr)
+            if (FFlag::LuauCompileInlinedBuiltins)
             {
-                // since recursive calls to analyze() may reuse the vector we need to be careful and preserve existing contents
-                size_t offset = builtinArgs.size();
-                bool canFold = true;
+                const int* bfid = builtins ? builtins->find(expr) : nullptr;
 
-                builtinArgs.reserve(offset + expr->args.size);
-
-                for (size_t i = 0; i < expr->args.size; ++i)
+                if (bfid && *bfid != LBF_NONE)
                 {
-                    Constant ac = analyze(expr->args.data[i]);
+                    // since recursive calls to analyze() may reuse the vector we need to be careful and preserve existing contents
+                    size_t offset = builtinArgs.size();
+                    bool canFold = true;
 
-                    if (ac.type == Constant::Type_Unknown)
-                        canFold = false;
-                    else
-                        builtinArgs.push_back(ac);
+                    builtinArgs.reserve(offset + expr->args.size);
+
+                    for (size_t i = 0; i < expr->args.size; ++i)
+                    {
+                        Constant ac = analyze(expr->args.data[i]);
+
+                        if (ac.type == Constant::Type_Unknown)
+                            canFold = false;
+                        else
+                            builtinArgs.push_back(ac);
+                    }
+
+                    if (canFold)
+                    {
+                        LUAU_ASSERT(builtinArgs.size() == offset + expr->args.size);
+                        result = foldBuiltin(stringTable, *bfid, builtinArgs.data() + offset, expr->args.size);
+                    }
+
+                    builtinArgs.resize(offset);
                 }
-
-                if (canFold)
+                else
                 {
-                    LUAU_ASSERT(builtinArgs.size() == offset + expr->args.size);
-                    result = foldBuiltin(*bfid, builtinArgs.data() + offset, expr->args.size);
+                    for (size_t i = 0; i < expr->args.size; ++i)
+                        analyze(expr->args.data[i]);
                 }
-
-                builtinArgs.resize(offset);
             }
             else
             {
-                for (size_t i = 0; i < expr->args.size; ++i)
-                    analyze(expr->args.data[i]);
+                if (const int* bfid = builtins ? builtins->find(expr) : nullptr)
+                {
+                    // since recursive calls to analyze() may reuse the vector we need to be careful and preserve existing contents
+                    size_t offset = builtinArgs.size();
+                    bool canFold = true;
+
+                    builtinArgs.reserve(offset + expr->args.size);
+
+                    for (size_t i = 0; i < expr->args.size; ++i)
+                    {
+                        Constant ac = analyze(expr->args.data[i]);
+
+                        if (ac.type == Constant::Type_Unknown)
+                            canFold = false;
+                        else
+                            builtinArgs.push_back(ac);
+                    }
+
+                    if (canFold)
+                    {
+                        LUAU_ASSERT(builtinArgs.size() == offset + expr->args.size);
+                        result = foldBuiltin(stringTable, *bfid, builtinArgs.data() + offset, expr->args.size);
+                    }
+
+                    builtinArgs.resize(offset);
+                }
+                else
+                {
+                    for (size_t i = 0; i < expr->args.size; ++i)
+                        analyze(expr->args.data[i]);
+                }
             }
         }
         else if (AstExprIndexName* expr = node->as<AstExprIndexName>())
         {
-            analyze(expr->expr);
-
-            if (foldLibraryK)
+            if (FFlag::LuauCompileFoldVectorComp)
             {
-                if (AstExprGlobal* eg = expr->expr->as<AstExprGlobal>())
-                {
-                    if (eg->name == "math")
-                        result = foldBuiltinMath(expr->index);
+                Constant value = analyze(expr->expr);
 
-                    // if we have a custom handler and the constant hasn't been resolved
-                    if (libraryMemberConstantCb && result.type == Constant::Type_Unknown)
-                        libraryMemberConstantCb(eg->name.value, expr->index.value, reinterpret_cast<Luau::CompileConstant*>(&result));
+                if (value.type == Constant::Type_Vector)
+                {
+                    if (expr->index == "x" || expr->index == "X")
+                    {
+                        result.type = Constant::Type_Number;
+                        result.valueNumber = value.valueVector[0];
+                    }
+                    else if (expr->index == "y" || expr->index == "Y")
+                    {
+                        result.type = Constant::Type_Number;
+                        result.valueNumber = value.valueVector[1];
+                    }
+                    else if (expr->index == "z" || expr->index == "Z")
+                    {
+                        result.type = Constant::Type_Number;
+                        result.valueNumber = value.valueVector[2];
+                    }
+
+                    // Do not handle 'w' component because it isn't known if the runtime will be configured in 3-wide or 4-wide mode
+                    // In 3-wide, access to 'w' will call unspecified metamethod or fail
+                }
+                else if (foldLibraryK)
+                {
+                    if (AstExprGlobal* eg = expr->expr->as<AstExprGlobal>())
+                    {
+                        if (eg->name == "math")
+                            result = foldBuiltinMath(expr->index);
+
+                        // if we have a custom handler and the constant hasn't been resolved
+                        if (libraryMemberConstantCb && result.type == Constant::Type_Unknown)
+                            libraryMemberConstantCb(eg->name.value, expr->index.value, reinterpret_cast<Luau::CompileConstant*>(&result));
+                    }
+                }
+            }
+            else
+            {
+                analyze(expr->expr);
+
+                if (foldLibraryK)
+                {
+                    if (AstExprGlobal* eg = expr->expr->as<AstExprGlobal>())
+                    {
+                        if (eg->name == "math")
+                            result = foldBuiltinMath(expr->index);
+
+                        // if we have a custom handler and the constant hasn't been resolved
+                        if (libraryMemberConstantCb && result.type == Constant::Type_Unknown)
+                            libraryMemberConstantCb(eg->name.value, expr->index.value, reinterpret_cast<Luau::CompileConstant*>(&result));
+                    }
                 }
             }
         }
@@ -518,7 +667,7 @@ struct ConstantVisitor : AstVisitor
 
             // note: ra doesn't need to be constant to fold and/or
             if (la.type != Constant::Type_Unknown)
-                foldBinary(result, expr->op, la, ra);
+                foldBinary(result, expr->op, la, ra, stringTable);
         }
         else if (AstExprTypeAssertion* expr = node->as<AstExprTypeAssertion>())
         {
@@ -537,8 +686,18 @@ struct ConstantVisitor : AstVisitor
         }
         else if (AstExprInterpString* expr = node->as<AstExprInterpString>())
         {
+            bool onlyConstantSubExpr = true;
             for (AstExpr* expression : expr->expressions)
-                analyze(expression);
+                if (analyze(expression).type != Constant::Type_String)
+                    onlyConstantSubExpr = false;
+
+            if (onlyConstantSubExpr)
+                foldInterpString(result, expr, constants, stringTable);
+        }
+        else if (AstExprInstantiate* expr = node->as<AstExprInstantiate>())
+        {
+            LUAU_ASSERT(FFlag::LuauExplicitTypeInstantiationSyntax);
+            result = analyze(expr->expr);
         }
         else
         {
@@ -628,10 +787,11 @@ void foldConstants(
     const DenseHashMap<AstExprCall*, int>* builtins,
     bool foldLibraryK,
     LibraryMemberConstantCallback libraryMemberConstantCb,
-    AstNode* root
+    AstNode* root,
+    AstNameTable& stringTable
 )
 {
-    ConstantVisitor visitor{constants, variables, locals, builtins, foldLibraryK, libraryMemberConstantCb};
+    ConstantVisitor visitor{constants, variables, locals, builtins, foldLibraryK, libraryMemberConstantCb, stringTable};
     root->visit(&visitor);
 }
 
