@@ -54,7 +54,7 @@ const char* luau_ident = "$Luau: Copyright (C) 2019-2024 Roblox Corporation $\n"
 
 #define api_update_top(L, p) \
     { \
-        api_check(L, p >= L->base && p < L->ci->top); \
+        api_check(L, p >= L->base && p <= L->ci->top); \
         L->top = p; \
     }
 
@@ -139,7 +139,29 @@ int lua_checkstack(lua_State* L, int size)
         res = 0; // stack overflow
     else if (size > 0)
     {
-        luaD_checkstack(L, size);
+        if (stacklimitreached(L, size))
+        {
+            struct CallContext
+            {
+                int size;
+
+                static void run(lua_State* L, void* ud)
+                {
+                    CallContext* ctx = (CallContext*)ud;
+
+                    luaD_growstack(L, ctx->size);
+                }
+            } ctx = {size};
+
+            // there could be no memory to extend the stack
+            if (luaD_rawrunprotected(L, &CallContext::run, &ctx) != LUA_OK)
+                return 0;
+        }
+        else
+        {
+            condhardstacktests(luaD_reallocstack(L, L->stacksize - EXTRA_STACK, 0));
+        }
+
         expandstacklimit(L, L->top + size);
     }
     return res;
@@ -773,6 +795,16 @@ int lua_rawgeti(lua_State* L, int idx, int n)
     return ttype(L->top - 1);
 }
 
+int lua_rawgetptagged(lua_State* L, int idx, void* p, int tag)
+{
+    luaC_threadbarrier(L);
+    StkId t = index2addr(L, idx);
+    api_check(L, ttistable(t));
+    setobj2s(L, L->top, luaH_getp(hvalue(t), p, tag));
+    api_incr_top(L);
+    return ttype(L->top - 1);
+}
+
 void lua_createtable(lua_State* L, int narray, int nrec)
 {
     luaC_checkGC(L);
@@ -912,6 +944,18 @@ void lua_rawseti(lua_State* L, int idx, int n)
     L->top--;
 }
 
+void lua_rawsetptagged(lua_State* L, int idx, void* p, int tag)
+{
+    api_checknelems(L, 1);
+    StkId o = index2addr(L, idx);
+    api_check(L, ttistable(o));
+    if (hvalue(o)->readonly)
+        luaG_readonlyerror(L);
+    setobj2t(L, luaH_setp(L, hvalue(o), p, tag), L->top - 1);
+    luaC_barriert(L, hvalue(o), L->top - 1);
+    L->top--;
+}
+
 int lua_setmetatable(lua_State* L, int objindex)
 {
     api_checknelems(L, 1);
@@ -1006,8 +1050,9 @@ void lua_call(lua_State* L, int nargs, int nresults)
 /*
 ** Execute a protected call.
 */
+// data to `f_call'
 struct CallS
-{ // data to `f_call'
+{
     StkId func;
     int nresults;
 };
@@ -1040,6 +1085,39 @@ int lua_pcall(lua_State* L, int nargs, int nresults, int errfunc)
     return status;
 }
 
+/*
+** Execute a protected C call.
+*/
+// data to `f_Ccall'
+struct CCallS
+{
+    lua_CFunction func;
+    void* ud;
+};
+
+static void f_Ccall(lua_State* L, void* ud)
+{
+    struct CCallS* c = cast_to(struct CCallS*, ud);
+
+    if (!lua_checkstack(L, 2))
+        luaG_runerror(L, "stack limit");
+
+    lua_pushcclosurek(L, c->func, nullptr, 0, nullptr);
+    lua_pushlightuserdata(L, c->ud);
+    luaD_call(L, L->top - 2, 0);
+}
+
+int lua_cpcall(lua_State* L, lua_CFunction func, void* ud)
+{
+    api_check(L, L->status == 0);
+
+    struct CCallS c;
+    c.func = func;
+    c.ud = ud;
+
+    return luaD_pcall(L, f_Ccall, &c, savestack(L, L->top), 0);
+}
+
 int lua_status(lua_State* L)
 {
     return L->status;
@@ -1445,8 +1523,16 @@ void lua_unref(lua_State* L, int ref)
 
     global_State* g = L->global;
     LuaTable* reg = hvalue(registry(L));
-    TValue* slot = luaH_setnum(L, reg, ref);
-    setnvalue(slot, g->registryfree); // NB: no barrier needed because value isn't collectable
+
+    const TValue* slot = luaH_getnum(reg, ref);
+    api_check(L, slot != luaO_nilobject);
+
+    // similar to how 'luaH_setnum' makes non-nil slot value mutable
+    TValue* mutableSlot = (TValue*)slot;
+
+    // NB: no barrier needed because value isn't collectable
+    setnvalue(mutableSlot, g->registryfree);
+
     g->registryfree = ref;
 }
 
@@ -1479,16 +1565,6 @@ void lua_setuserdatametatable(lua_State* L, int tag)
     L->top--;
 }
 
-void lua_setuserdatametatable_DEPRECATED(lua_State* L, int tag, int idx)
-{
-    api_check(L, unsigned(tag) < LUA_UTAG_LIMIT);
-    api_check(L, !L->global->udatamt[tag]); // reassignment not supported
-    StkId o = index2addr(L, idx);
-    api_check(L, ttistable(o));
-    L->global->udatamt[tag] = hvalue(o);
-    L->top--;
-}
-
 void lua_getuserdatametatable(lua_State* L, int tag)
 {
     api_check(L, unsigned(tag) < LUA_UTAG_LIMIT);
