@@ -1701,6 +1701,26 @@ static int luauF_vectormax(lua_State* L, StkId res, TValue* arg0, int nresults,
     return -1;
 }
 
+static int luauF_vectorlerp(lua_State* L, StkId res, TValue* arg0, int nresults, StkId args, int nparams)
+{
+    if (nparams >= 3 && nresults <= 1 && ttisvector(arg0) && ttisvector(args) && ttisnumber(args + 1))
+    {
+        const float* a = vvalue(arg0);
+        const float* b = vvalue(args);
+        const float t = static_cast<float>(nvalue(args + 1));
+
+#if LUA_VECTOR_SIZE == 4
+        setvvalue(res, luai_lerpf(a[0], b[0], t), luai_lerpf(a[1], b[1], t), luai_lerpf(a[2], b[2], t), luai_lerpf(a[3], b[3], t));
+#else
+        setvvalue(res, luai_lerpf(a[0], b[0], t), luai_lerpf(a[1], b[1], t), luai_lerpf(a[2], b[2], t), 0.0f);
+#endif
+
+        return 1;
+    }
+
+    return -1;
+}
+
 static int luauF_lerp(lua_State* L, StkId res, TValue* arg0, int nresults, StkId args, int nparams)
 {
     if (nparams >= 3 && nresults <= 1 && ttisnumber(arg0) && ttisnumber(args) && ttisnumber(args + 1))
@@ -1718,6 +1738,45 @@ static int luauF_lerp(lua_State* L, StkId res, TValue* arg0, int nresults, StkId
     return -1;
 }
 
+static int luauF_isnan(lua_State* L, StkId res, TValue* arg0, int nresults, StkId args, int nparams)
+{
+    if (nparams >= 1 && nresults <= 1 && ttisnumber(arg0))
+    {
+        double x = nvalue(arg0);
+
+        setbvalue(res, isnan(x));
+        return 1;
+    }
+
+    return -1;
+}
+
+static int luauF_isinf(lua_State* L, StkId res, TValue* arg0, int nresults, StkId args, int nparams)
+{
+    if (nparams >= 1 && nresults <= 1 && ttisnumber(arg0))
+    {
+        double x = nvalue(arg0);
+
+        setbvalue(res, isinf(x));
+        return 1;
+    }
+
+    return -1;
+}
+
+static int luauF_isfinite(lua_State* L, StkId res, TValue* arg0, int nresults, StkId args, int nparams)
+{
+    if (nparams >= 1 && nresults <= 1 && ttisnumber(arg0))
+    {
+        double x = nvalue(arg0);
+
+        setbvalue(res, isfinite(x));
+        return 1;
+    }
+
+    return -1;
+}
+
 static int luauF_missing(lua_State* L, StkId res, TValue* arg0, int nresults, StkId args, int nparams)
 {
     return -1;
@@ -1780,7 +1839,7 @@ static bool luau_hassse41()
     __cpuid(1, cpuinfo[0], cpuinfo[1], cpuinfo[2], cpuinfo[3]);
 #endif
 
-    // We requre SSE4.1 support for ROUNDSD
+    // We require SSE4.1 support for ROUNDSD
     // https://en.wikipedia.org/wiki/CPUID#EAX=1:_Processor_Info_and_Feature_Bits
     return (cpuinfo[2] & (1 << 19)) != 0;
 }
@@ -1915,6 +1974,12 @@ const luau_FastFunction luauF_table[256] = {
 
     luauF_lerp,
 
+    luauF_vectorlerp,
+
+    luauF_isnan,
+    luauF_isinf,
+    luauF_isfinite,
+
 // When adding builtins, add them above this line; what follows is 64 "dummy" entries with luauF_missing fallback.
 // This is important so that older versions of the runtime that don't support newer builtins automatically fall back via luauF_missing.
 // Given the builtin addition velocity this should always provide a larger compatibility window than bytecode versions suggest.
