@@ -366,7 +366,7 @@ static TValue* arrayornewkey(lua_State* L, LuaTable* t, const TValue* key)
         int k;
         double n = nvalue(key);
         luai_num2int(k, n);
-        if (luai_numeq(cast_num(k), n) && cast_to(unsigned int, k - 1) < cast_to(unsigned int, t->sizearray))
+        if (luai_numeq(cast_num(k), n) && unsigned(k) - 1 < unsigned(t->sizearray))
             return &t->array[k - 1];
     }
 
@@ -604,7 +604,7 @@ static TValue* newkey(lua_State* L, LuaTable* t, const TValue* key)
 const TValue* luaH_getnum(LuaTable* t, int key)
 {
     // (1 <= key && key <= t->sizearray)
-    if (cast_to(unsigned int, key - 1) < cast_to(unsigned int, t->sizearray))
+    if (unsigned(key) - 1 < unsigned(t->sizearray))
         return &t->array[key - 1];
     else if (t->node != dummynode)
     {
@@ -641,6 +641,24 @@ const TValue* luaH_getstr(LuaTable* t, TString* key)
     return luaO_nilobject;
 }
 
+/*
+** search function for lightuserdata
+*/
+const TValue* luaH_getp(LuaTable* t, void* key, int tag)
+{
+    LuaNode* n = hashpointer(t, key);
+    for (;;)
+    { // check whether `key' is somewhere in the chain
+        const TKey* nk = gkey(n);
+        if (ttislightuserdata(nk) && pvalue(nk) == key && lightuserdatatag(nk) == tag)
+            return gval(n); // that's it
+        if (gnext(n) == 0)
+            break;
+        n += gnext(n);
+    }
+    return luaO_nilobject;
+}
+
 /*
 ** main search function
 */
@@ -701,7 +719,7 @@ TValue* luaH_newkey(lua_State* L, LuaTable* t, const TValue* key)
 TValue* luaH_setnum(lua_State* L, LuaTable* t, int key)
 {
     // (1 <= key && key <= t->sizearray)
-    if (cast_to(unsigned int, key - 1) < cast_to(unsigned int, t->sizearray))
+    if (unsigned(key) - 1 < unsigned(t->sizearray))
         return &t->array[key - 1];
     // hash fallback
     const TValue* p = luaH_getnum(t, key);
@@ -729,6 +747,19 @@ TValue* luaH_setstr(lua_State* L, LuaTable* t, TString* key)
     }
 }
 
+TValue* luaH_setp(lua_State* L, LuaTable* t, void* key, int tag)
+{
+    const TValue* p = luaH_getp(t, key, tag);
+    if (p != luaO_nilobject)
+        return cast_to(TValue*, p);
+    else
+    {
+        TValue k;
+        setpvalue(&k, key, tag);
+        return newkey(L, t, &k);
+    }
+}
+
 static int updateaboundary(LuaTable* t, int boundary)
 {
     if (boundary < t->sizearray && ttisnil(&t->array[boundary - 1]))
