@@ -10,7 +10,8 @@
 // special tag value is used for newproxy-created user data (all other user data objects are host-exposed)
 #define UTAG_PROXY (LUA_UTAG_LIMIT + 1)
 
-#define sizeudata(len) (offsetof(Udata, data) + len)
+// userdata larger than 16 bytes will be extended to guarantee 16 byte alignment of subsequent blocks
+#define sizeudata(len) (offsetof(Udata, data) + (len > 16 ? ((len + 15) & ~15) : len))
 
 LUAI_FUNC Udata* luaU_newudata(lua_State* L, size_t s, int tag);
 LUAI_FUNC void luaU_freeudata(lua_State* L, Udata* u, struct lua_Page* page);
