@@ -9,10 +9,10 @@
 #include "Luau/ModuleResolver.h"
 #include "Luau/NotNull.h"
 #include "Luau/Parser.h"
+#include "Luau/PrettyPrinter.h"
 #include "Luau/Type.h"
 #include "Luau/TypeAttach.h"
 #include "Luau/TypeInfer.h"
-#include "Luau/Transpiler.h"
 
 #include "doctest.h"
 
@@ -29,6 +29,7 @@ LUAU_FASTFLAG(LuauSolverV2);
 LUAU_FASTFLAG(DebugLuauLogSolverToJsonFile)
 
 LUAU_FASTFLAGVARIABLE(DebugLuauForceAllNewSolverTests);
+LUAU_FASTINT(LuauStackGuardThreshold)
 
 extern std::optional<unsigned> randomSeed; // tests/main.cpp
 
@@ -131,7 +132,7 @@ std::vector<std::unique_ptr<RequireNode>> TestRequireNode::getChildren() const
 
 std::vector<RequireAlias> TestRequireNode::getAvailableAliases() const
 {
-    return {{"defaultalias"}};
+    return {RequireAlias("defaultalias")};
 }
 
 std::unique_ptr<RequireNode> TestRequireSuggester::getNode(const ModuleName& name) const
@@ -174,7 +175,7 @@ std::optional<SourceCode> TestFileResolver::readSource(const ModuleName& name)
     return SourceCode{it->second, sourceType};
 }
 
-std::optional<ModuleInfo> TestFileResolver::resolveModule(const ModuleInfo* context, AstExpr* expr)
+std::optional<ModuleInfo> TestFileResolver::resolveModule(const ModuleInfo* context, AstExpr* expr, const TypeCheckLimits& limits)
 {
     if (AstExprGlobal* g = expr->as<AstExprGlobal>())
     {
@@ -247,7 +248,7 @@ std::optional<std::string> TestFileResolver::getEnvironmentForModule(const Modul
     return std::nullopt;
 }
 
-const Config& TestConfigResolver::getConfig(const ModuleName& name) const
+const Config& TestConfigResolver::getConfig(const ModuleName& name, const TypeCheckLimits& limits) const
 {
     auto it = configFiles.find(name);
     if (it != configFiles.end())
@@ -257,37 +258,8 @@ const Config& TestConfigResolver::getConfig(const ModuleName& name) const
 }
 
 Fixture::Fixture(bool prepareAutocomplete)
-    : frontend(
-          &fileResolver,
-          &configResolver,
-          {/* retainFullTypeGraphs= */ true, /* forAutocomplete */ false, /* runLintChecks */ false, /* randomConstraintResolutionSeed */ randomSeed}
-      )
-    , builtinTypes(frontend.builtinTypes)
+    : forAutocomplete(prepareAutocomplete)
 {
-    configResolver.defaultConfig.mode = Mode::Strict;
-    configResolver.defaultConfig.enabledLint.warningMask = ~0ull;
-    configResolver.defaultConfig.parseOptions.captureComments = true;
-
-    Luau::freeze(frontend.globals.globalTypes);
-    Luau::freeze(frontend.globalsForAutocomplete.globalTypes);
-
-    Luau::setPrintLine([](auto s) {});
-
-    if (FFlag::DebugLuauLogSolverToJsonFile)
-    {
-        frontend.writeJsonLog = [&](const Luau::ModuleName& moduleName, std::string log)
-        {
-            std::string path = moduleName + ".log.json";
-            size_t pos = moduleName.find_last_of('/');
-            if (pos != std::string::npos)
-                path = moduleName.substr(pos + 1);
-
-            std::ofstream os(path);
-
-            os << log << std::endl;
-            MESSAGE("Wrote JSON log to ", path);
-        };
-    }
 }
 
 Fixture::~Fixture()
@@ -314,31 +286,33 @@ AstStatBlock* Fixture::parse(const std::string& source, const ParseOptions& pars
             if (FFlag::LuauSolverV2)
             {
                 Mode mode = sourceModule->mode ? *sourceModule->mode : Mode::Strict;
+                Frontend::Stats stats;
                 ModulePtr module = Luau::check(
                     *sourceModule,
                     mode,
                     {},
-                    builtinTypes,
+                    getBuiltins(),
                     NotNull{&ice},
                     NotNull{&moduleResolver},
                     NotNull{&fileResolver},
-                    frontend.globals.globalScope,
-                    frontend.globals.globalTypeFunctionScope,
+                    getFrontend().globals.globalScope,
+                    getFrontend().globals.globalTypeFunctionScope,
                     /*prepareModuleScope*/ nullptr,
-                    frontend.options,
+                    getFrontend().options,
                     {},
                     false,
+                    stats,
                     {}
                 );
 
-                Luau::lint(sourceModule->root, *sourceModule->names, frontend.globals.globalScope, module.get(), sourceModule->hotcomments, {});
+                Luau::lint(sourceModule->root, *sourceModule->names, getFrontend().globals.globalScope, module.get(), sourceModule->hotcomments, {});
             }
             else
             {
-                TypeChecker typeChecker(frontend.globals.globalScope, &moduleResolver, builtinTypes, &frontend.iceHandler);
+                TypeChecker typeChecker(getFrontend().globals.globalScope, &moduleResolver, getBuiltins(), &getFrontend().iceHandler);
                 ModulePtr module = typeChecker.check(*sourceModule, sourceModule->mode.value_or(Luau::Mode::Nonstrict), std::nullopt);
 
-                Luau::lint(sourceModule->root, *sourceModule->names, frontend.globals.globalScope, module.get(), sourceModule->hotcomments, {});
+                Luau::lint(sourceModule->root, *sourceModule->names, getFrontend().globals.globalScope, module.get(), sourceModule->hotcomments, {});
             }
         }
 
@@ -350,12 +324,15 @@ AstStatBlock* Fixture::parse(const std::string& source, const ParseOptions& pars
 
 CheckResult Fixture::check(Mode mode, const std::string& source, std::optional<FrontendOptions> options)
 {
+    // Force the frontend here
+    getFrontend();
     ModuleName mm = fromString(mainModuleName);
     configResolver.defaultConfig.mode = mode;
     fileResolver.source[mm] = std::move(source);
-    frontend.markDirty(mm);
+    getFrontend().markDirty(mm);
+    getFrontend().clearStats();
 
-    CheckResult result = frontend.check(mm, options);
+    CheckResult result = getFrontend().check(mm, options);
 
     return result;
 }
@@ -370,18 +347,18 @@ LintResult Fixture::lint(const std::string& source, const std::optional<LintOpti
     ModuleName mm = fromString(mainModuleName);
     configResolver.defaultConfig.mode = Mode::Strict;
     fileResolver.source[mm] = std::move(source);
-    frontend.markDirty(mm);
+    getFrontend().markDirty(mm);
 
     return lintModule(mm, lintOptions);
 }
 
 LintResult Fixture::lintModule(const ModuleName& moduleName, const std::optional<LintOptions>& lintOptions)
 {
-    FrontendOptions options = frontend.options;
+    FrontendOptions options = getFrontend().options;
     options.runLintChecks = true;
     options.enabledLintWarnings = lintOptions;
 
-    CheckResult result = frontend.check(moduleName, options);
+    CheckResult result = getFrontend().check(moduleName, options);
 
     return result.lintResult;
 }
@@ -450,14 +427,14 @@ ParseResult Fixture::matchParseErrorPrefix(const std::string& source, const std:
 ModulePtr Fixture::getMainModule(bool forAutocomplete)
 {
     if (forAutocomplete && !FFlag::LuauSolverV2)
-        return frontend.moduleResolverForAutocomplete.getModule(fromString(mainModuleName));
+        return getFrontend().moduleResolverForAutocomplete.getModule(fromString(mainModuleName));
 
-    return frontend.moduleResolver.getModule(fromString(mainModuleName));
+    return getFrontend().moduleResolver.getModule(fromString(mainModuleName));
 }
 
 SourceModule* Fixture::getMainSourceModule()
 {
-    return frontend.getSourceModule(fromString(mainModuleName));
+    return getFrontend().getSourceModule(fromString(mainModuleName));
 }
 
 std::optional<PrimitiveType::Type> Fixture::getPrimitiveType(TypeId ty)
@@ -497,7 +474,7 @@ TypeId Fixture::requireType(const std::string& name)
 
 TypeId Fixture::requireType(const ModuleName& moduleName, const std::string& name)
 {
-    ModulePtr module = frontend.moduleResolver.getModule(moduleName);
+    ModulePtr module = getFrontend().moduleResolver.getModule(moduleName);
     REQUIRE(module);
     return requireType(module, name);
 }
@@ -573,7 +550,7 @@ TypeId Fixture::requireTypeAlias(const std::string& name)
 
 TypeId Fixture::requireExportedType(const ModuleName& moduleName, const std::string& name)
 {
-    ModulePtr module = frontend.moduleResolver.getModule(moduleName);
+    ModulePtr module = getFrontend().moduleResolver.getModule(moduleName);
     REQUIRE(module);
 
     auto it = module->exportedTypeBindings.find(name);
@@ -582,16 +559,23 @@ TypeId Fixture::requireExportedType(const ModuleName& moduleName, const std::str
     return it->second.type;
 }
 
+TypeId Fixture::parseType(std::string_view src)
+{
+    return getFrontend().parseType(
+        NotNull{&allocator}, NotNull{&nameTable}, NotNull{&getFrontend().iceHandler}, TypeCheckLimits{}, NotNull{&arena}, src
+    );
+}
+
 std::string Fixture::decorateWithTypes(const std::string& code)
 {
     fileResolver.source[mainModuleName] = code;
 
-    Luau::CheckResult typeInfo = frontend.check(mainModuleName);
+    Luau::CheckResult typeInfo = getFrontend().check(mainModuleName);
 
-    SourceModule* sourceModule = frontend.getSourceModule(mainModuleName);
-    attachTypeData(*sourceModule, *frontend.moduleResolver.getModule(mainModuleName));
+    SourceModule* sourceModule = getFrontend().getSourceModule(mainModuleName);
+    attachTypeData(*sourceModule, *getFrontend().moduleResolver.getModule(mainModuleName));
 
-    return transpileWithTypes(*sourceModule->root);
+    return prettyPrintWithTypes(*sourceModule->root);
 }
 
 void Fixture::dumpErrors(std::ostream& os, const std::vector<TypeError>& errors)
@@ -621,9 +605,9 @@ void Fixture::dumpErrors(std::ostream& os, const std::vector<TypeError>& errors)
 
 void Fixture::registerTestTypes()
 {
-    addGlobalBinding(frontend.globals, "game", builtinTypes->anyType, "@luau");
-    addGlobalBinding(frontend.globals, "workspace", builtinTypes->anyType, "@luau");
-    addGlobalBinding(frontend.globals, "script", builtinTypes->anyType, "@luau");
+    addGlobalBinding(getFrontend().globals, "game", getBuiltins()->anyType, "@luau");
+    addGlobalBinding(getFrontend().globals, "workspace", getBuiltins()->anyType, "@luau");
+    addGlobalBinding(getFrontend().globals, "script", getBuiltins()->anyType, "@luau");
 }
 
 void Fixture::dumpErrors(const CheckResult& cr)
@@ -682,9 +666,9 @@ void Fixture::validateErrors(const std::vector<Luau::TypeError>& errors)
 
 LoadDefinitionFileResult Fixture::loadDefinition(const std::string& source, bool forAutocomplete)
 {
-    GlobalTypes& globals = forAutocomplete ? frontend.globalsForAutocomplete : frontend.globals;
+    GlobalTypes& globals = forAutocomplete ? getFrontend().globalsForAutocomplete : getFrontend().globals;
     unfreeze(globals.globalTypes);
-    LoadDefinitionFileResult result = frontend.loadDefinitionFile(
+    LoadDefinitionFileResult result = getFrontend().loadDefinitionFile(
         globals, globals.globalScope, source, "@test", /* captureComments */ false, /* typecheckForAutocomplete */ forAutocomplete
     );
     freeze(globals.globalTypes);
@@ -695,19 +679,96 @@ LoadDefinitionFileResult Fixture::loadDefinition(const std::string& source, bool
     return result;
 }
 
+NotNull<BuiltinTypes> Fixture::getBuiltins()
+{
+    if (!builtinTypes)
+        getFrontend();
+    return NotNull{builtinTypes};
+}
+
+const BuiltinTypeFunctions& Fixture::getBuiltinTypeFunctions()
+{
+    return *getBuiltins()->typeFunctions;
+}
+
+Frontend& Fixture::getFrontend()
+{
+    if (frontend)
+        return *frontend;
+
+    Frontend& f = frontend.emplace(
+        &fileResolver,
+        &configResolver,
+        FrontendOptions{
+            /* retainFullTypeGraphs= */ true, /* forAutocomplete */ false, /* runLintChecks */ false, /* randomConstraintResolutionSeed */ randomSeed
+        }
+    );
+
+    builtinTypes = f.builtinTypes;
+    // Fixture::Fixture begins here
+    configResolver.defaultConfig.mode = Mode::Strict;
+    configResolver.defaultConfig.enabledLint.warningMask = ~0ull;
+    configResolver.defaultConfig.parseOptions.captureComments = true;
+
+    Luau::freeze(f.globals.globalTypes);
+    Luau::freeze(f.globalsForAutocomplete.globalTypes);
+
+    Luau::setPrintLine([](auto s) {});
+
+    if (FFlag::DebugLuauLogSolverToJsonFile)
+    {
+        f.writeJsonLog = [&](const Luau::ModuleName& moduleName, std::string log)
+        {
+            std::string path = moduleName + ".log.json";
+            size_t pos = moduleName.find_last_of('/');
+            if (pos != std::string::npos)
+                path = moduleName.substr(pos + 1);
+
+            std::ofstream os(path);
+
+            os << log << std::endl;
+            MESSAGE("Wrote JSON log to ", path);
+        };
+    }
+    return *frontend;
+}
+
+void Fixture::limitStackSize(size_t size)
+{
+    // The FInt is designed to trip when the amount of available address
+    // space goes below some threshold, but for this API, the convenient thing
+    // is to specify how much the test should be allowed to use.  We need to
+    // do a tiny amount of arithmetic to convert.
+
+    uintptr_t addressSpaceSize = getStackAddressSpaceSize();
+
+    dynamicScopedInts.emplace_back(FInt::LuauStackGuardThreshold, (int)(addressSpaceSize - size));
+}
+
 BuiltinsFixture::BuiltinsFixture(bool prepareAutocomplete)
     : Fixture(prepareAutocomplete)
 {
-    Luau::unfreeze(frontend.globals.globalTypes);
-    Luau::unfreeze(frontend.globalsForAutocomplete.globalTypes);
+}
 
-    registerBuiltinGlobals(frontend, frontend.globals);
-    if (prepareAutocomplete)
-        registerBuiltinGlobals(frontend, frontend.globalsForAutocomplete, /*typeCheckForAutocomplete*/ true);
+Frontend& BuiltinsFixture::getFrontend()
+{
+    if (frontend)
+        return *frontend;
+
+    Frontend& f = Fixture::getFrontend();
+    // Do builtins fixture things now
+    Luau::unfreeze(f.globals.globalTypes);
+    Luau::unfreeze(f.globalsForAutocomplete.globalTypes);
+
+    registerBuiltinGlobals(f, f.globals);
+    if (forAutocomplete)
+        registerBuiltinGlobals(f, f.globalsForAutocomplete, /*typeCheckForAutocomplete*/ true);
     registerTestTypes();
 
-    Luau::freeze(frontend.globals.globalTypes);
-    Luau::freeze(frontend.globalsForAutocomplete.globalTypes);
+    Luau::freeze(f.globals.globalTypes);
+    Luau::freeze(f.globalsForAutocomplete.globalTypes);
+
+    return *frontend;
 }
 
 static std::vector<std::string_view> parsePathExpr(const AstExpr& pathExpr)
@@ -831,59 +892,59 @@ std::optional<TypeId> linearSearchForBinding(Scope* scope, const char* name)
     return std::nullopt;
 }
 
-void registerHiddenTypes(Frontend* frontend)
+void registerHiddenTypes(Frontend& frontend)
 {
-    GlobalTypes& globals = frontend->globals;
+    GlobalTypes& globals = frontend.globals;
 
     unfreeze(globals.globalTypes);
 
-    TypeId t = globals.globalTypes.addType(GenericType{"T"});
+    TypeId t = globals.globalTypes.addType(GenericType{"T", Polarity::Mixed});
     GenericTypeDefinition genericT{t};
 
-    TypeId u = globals.globalTypes.addType(GenericType{"U"});
+    TypeId u = globals.globalTypes.addType(GenericType{"U", Polarity::Mixed});
     GenericTypeDefinition genericU{u};
 
     ScopePtr globalScope = globals.globalScope;
     globalScope->exportedTypeBindings["Not"] = TypeFun{{genericT}, globals.globalTypes.addType(NegationType{t})};
     globalScope->exportedTypeBindings["Mt"] = TypeFun{{genericT, genericU}, globals.globalTypes.addType(MetatableType{t, u})};
-    globalScope->exportedTypeBindings["fun"] = TypeFun{{}, frontend->builtinTypes->functionType};
-    globalScope->exportedTypeBindings["cls"] = TypeFun{{}, frontend->builtinTypes->classType};
-    globalScope->exportedTypeBindings["err"] = TypeFun{{}, frontend->builtinTypes->errorType};
-    globalScope->exportedTypeBindings["tbl"] = TypeFun{{}, frontend->builtinTypes->tableType};
+    globalScope->exportedTypeBindings["fun"] = TypeFun{{}, frontend.builtinTypes->functionType};
+    globalScope->exportedTypeBindings["cls"] = TypeFun{{}, frontend.builtinTypes->externType};
+    globalScope->exportedTypeBindings["err"] = TypeFun{{}, frontend.builtinTypes->errorType};
+    globalScope->exportedTypeBindings["tbl"] = TypeFun{{}, frontend.builtinTypes->tableType};
 
     freeze(globals.globalTypes);
 }
 
-void createSomeClasses(Frontend* frontend)
+void createSomeExternTypes(Frontend& frontend)
 {
-    GlobalTypes& globals = frontend->globals;
+    GlobalTypes& globals = frontend.globals;
 
     TypeArena& arena = globals.globalTypes;
     unfreeze(arena);
 
     ScopePtr moduleScope = globals.globalScope;
 
-    TypeId parentType = arena.addType(ClassType{"Parent", {}, frontend->builtinTypes->classType, std::nullopt, {}, nullptr, "Test", {}});
+    TypeId parentType = arena.addType(ExternType{"Parent", {}, frontend.builtinTypes->externType, std::nullopt, {}, nullptr, "Test", {}});
 
-    ClassType* parentClass = getMutable<ClassType>(parentType);
-    parentClass->props["method"] = {makeFunction(arena, parentType, {}, {})};
+    ExternType* parentExternType = getMutable<ExternType>(parentType);
+    parentExternType->props["method"] = {makeFunction(arena, parentType, {}, {})};
 
-    parentClass->props["virtual_method"] = {makeFunction(arena, parentType, {}, {})};
+    parentExternType->props["virtual_method"] = {makeFunction(arena, parentType, {}, {})};
 
     addGlobalBinding(globals, "Parent", {parentType});
     moduleScope->exportedTypeBindings["Parent"] = TypeFun{{}, parentType};
 
-    TypeId childType = arena.addType(ClassType{"Child", {}, parentType, std::nullopt, {}, nullptr, "Test", {}});
+    TypeId childType = arena.addType(ExternType{"Child", {}, parentType, std::nullopt, {}, nullptr, "Test", {}});
 
     addGlobalBinding(globals, "Child", {childType});
     moduleScope->exportedTypeBindings["Child"] = TypeFun{{}, childType};
 
-    TypeId anotherChildType = arena.addType(ClassType{"AnotherChild", {}, parentType, std::nullopt, {}, nullptr, "Test", {}});
+    TypeId anotherChildType = arena.addType(ExternType{"AnotherChild", {}, parentType, std::nullopt, {}, nullptr, "Test", {}});
 
     addGlobalBinding(globals, "AnotherChild", {anotherChildType});
     moduleScope->exportedTypeBindings["AnotherChild"] = TypeFun{{}, anotherChildType};
 
-    TypeId unrelatedType = arena.addType(ClassType{"Unrelated", {}, frontend->builtinTypes->classType, std::nullopt, {}, nullptr, "Test", {}});
+    TypeId unrelatedType = arena.addType(ExternType{"Unrelated", {}, frontend.builtinTypes->externType, std::nullopt, {}, nullptr, "Test", {}});
 
     addGlobalBinding(globals, "Unrelated", {unrelatedType});
     moduleScope->exportedTypeBindings["Unrelated"] = TypeFun{{}, unrelatedType};
