@@ -7,9 +7,9 @@
 
 #include "doctest.h"
 
-LUAU_FASTFLAG(LuauSolverV2);
-LUAU_FASTFLAG(LintRedundantNativeAttribute);
-LUAU_FASTFLAG(LuauDeprecatedAttribute);
+LUAU_FASTFLAG(LuauSolverV2)
+
+LUAU_FASTFLAG(LuauExplicitTypeInstantiationSyntax)
 
 using namespace Luau;
 
@@ -44,13 +44,13 @@ TEST_CASE_FIXTURE(Fixture, "UnknownGlobal")
     LintResult result = lint("--!nocheck\nreturn foo");
 
     REQUIRE(1 == result.warnings.size());
-    CHECK_EQ(result.warnings[0].text, "Unknown global 'foo'");
+    CHECK_EQ(result.warnings[0].text, "Unknown global 'foo'; consider assigning to it first");
 }
 
 TEST_CASE_FIXTURE(Fixture, "DeprecatedGlobal")
 {
     // Normally this would be defined externally, so hack it in for testing
-    addGlobalBinding(frontend.globals, "Wait", Binding{builtinTypes->anyType, {}, true, "wait", "@test/global/Wait"});
+    addGlobalBinding(getFrontend().globals, "Wait", Binding{getBuiltins()->anyType, {}, true, "wait", "@test/global/Wait"});
 
     LintResult result = lint("Wait(5)");
 
@@ -62,7 +62,7 @@ TEST_CASE_FIXTURE(Fixture, "DeprecatedGlobalNoReplacement")
 {
     // Normally this would be defined externally, so hack it in for testing
     const char* deprecationReplacementString = "";
-    addGlobalBinding(frontend.globals, "Version", Binding{builtinTypes->anyType, {}, true, deprecationReplacementString});
+    addGlobalBinding(getFrontend().globals, "Version", Binding{getBuiltins()->anyType, {}, true, deprecationReplacementString});
 
     LintResult result = lint("Version()");
 
@@ -389,7 +389,7 @@ return bar()
 TEST_CASE_FIXTURE(Fixture, "ImportUnused")
 {
     // Normally this would be defined externally, so hack it in for testing
-    addGlobalBinding(frontend.globals, "game", builtinTypes->anyType, "@test");
+    addGlobalBinding(getFrontend().globals, "game", getBuiltins()->anyType, "@test");
 
     LintResult result = lint(R"(
 local Roact = require(game.Packages.Roact)
@@ -620,16 +620,16 @@ return foo1
 
 TEST_CASE_FIXTURE(Fixture, "UnknownType")
 {
-    unfreeze(frontend.globals.globalTypes);
+    unfreeze(getFrontend().globals.globalTypes);
     TableType::Props instanceProps{
-        {"ClassName", {builtinTypes->anyType}},
+        {"ClassName", {getBuiltins()->anyType}},
     };
 
-    TableType instanceTable{instanceProps, std::nullopt, frontend.globals.globalScope->level, Luau::TableState::Sealed};
-    TypeId instanceType = frontend.globals.globalTypes.addType(instanceTable);
+    TableType instanceTable{instanceProps, std::nullopt, getFrontend().globals.globalScope->level, Luau::TableState::Sealed};
+    TypeId instanceType = getFrontend().globals.globalTypes.addType(instanceTable);
     TypeFun instanceTypeFun{{}, instanceType};
 
-    frontend.globals.globalScope->exportedTypeBindings["Part"] = instanceTypeFun;
+    getFrontend().globals.globalScope->exportedTypeBindings["Part"] = instanceTypeFun;
 
     LintResult result = lint(R"(
 local game = ...
@@ -1252,15 +1252,19 @@ _ = {
     [2] = 2,
     [1] = 3,
 }
+
+function _foo(): { first: number, second: string, first: boolean }
+end
 )");
 
-    REQUIRE(6 == result.warnings.size());
+    REQUIRE(7 == result.warnings.size());
     CHECK_EQ(result.warnings[0].text, "Table field 'first' is a duplicate; previously defined at line 3");
     CHECK_EQ(result.warnings[1].text, "Table field 'first' is a duplicate; previously defined at line 9");
     CHECK_EQ(result.warnings[2].text, "Table index 1 is a duplicate; previously defined as a list entry");
     CHECK_EQ(result.warnings[3].text, "Table index 3 is a duplicate; previously defined as a list entry");
     CHECK_EQ(result.warnings[4].text, "Table type field 'first' is a duplicate; previously defined at line 24");
     CHECK_EQ(result.warnings[5].text, "Table index 1 is a duplicate; previously defined at line 36");
+    CHECK_EQ(result.warnings[6].text, "Table type field 'first' is a duplicate; previously defined at line 41");
 }
 
 TEST_CASE_FIXTURE(Fixture, "read_write_table_props")
@@ -1299,6 +1303,19 @@ TEST_CASE_FIXTURE(Fixture, "ImportOnlyUsedInTypeAnnotation")
     CHECK_EQ(result.warnings[0].text, "Variable 'x' is never used; prefix with '_' to silence");
 }
 
+TEST_CASE_FIXTURE(Fixture, "ImportOnlyUsedInReturnType")
+{
+    LintResult result = lint(R"(
+        local Foo = require(script.Parent.Foo)
+
+        function foo(): Foo.Y
+        end
+    )");
+
+    REQUIRE(1 == result.warnings.size());
+    CHECK_EQ(result.warnings[0].text, "Function 'foo' is never used; prefix with '_' to silence");
+}
+
 TEST_CASE_FIXTURE(Fixture, "DisableUnknownGlobalWithTypeChecking")
 {
     LintResult result = lint(R"(
@@ -1323,10 +1340,10 @@ TEST_CASE_FIXTURE(Fixture, "no_spurious_warning_after_a_function_type_alias")
 
 TEST_CASE_FIXTURE(Fixture, "use_all_parent_scopes_for_globals")
 {
-    ScopePtr testScope = frontend.addEnvironment("Test");
-    unfreeze(frontend.globals.globalTypes);
-    frontend.loadDefinitionFile(
-        frontend.globals,
+    ScopePtr testScope = getFrontend().addEnvironment("Test");
+    unfreeze(getFrontend().globals.globalTypes);
+    getFrontend().loadDefinitionFile(
+        getFrontend().globals,
         testScope,
         R"(
         declare Foo: number
@@ -1334,7 +1351,7 @@ TEST_CASE_FIXTURE(Fixture, "use_all_parent_scopes_for_globals")
         "@test",
         /* captureComments */ false
     );
-    freeze(frontend.globals.globalTypes);
+    freeze(getFrontend().globals.globalTypes);
 
     fileResolver.environments["A"] = "Test";
 
@@ -1503,32 +1520,32 @@ TEST_CASE_FIXTURE(Fixture, "LintHygieneUAF")
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "DeprecatedApiTyped")
 {
-    unfreeze(frontend.globals.globalTypes);
-    TypeId instanceType = frontend.globals.globalTypes.addType(ClassType{"Instance", {}, std::nullopt, std::nullopt, {}, {}, "Test", {}});
+    unfreeze(getFrontend().globals.globalTypes);
+    TypeId instanceType = getFrontend().globals.globalTypes.addType(ExternType{"Instance", {}, std::nullopt, std::nullopt, {}, {}, "Test", {}});
     persist(instanceType);
-    frontend.globals.globalScope->exportedTypeBindings["Instance"] = TypeFun{{}, instanceType};
+    getFrontend().globals.globalScope->exportedTypeBindings["Instance"] = TypeFun{{}, instanceType};
 
-    getMutable<ClassType>(instanceType)->props = {
-        {"Name", {builtinTypes->stringType}},
-        {"DataCost", {builtinTypes->numberType, /* deprecated= */ true}},
-        {"Wait", {builtinTypes->anyType, /* deprecated= */ true}},
+    getMutable<ExternType>(instanceType)->props = {
+        {"Name", {getBuiltins()->stringType}},
+        {"DataCost", {getBuiltins()->numberType, /* deprecated= */ true}},
+        {"Wait", {getBuiltins()->anyType, /* deprecated= */ true}},
     };
 
     TypeId colorType =
-        frontend.globals.globalTypes.addType(TableType{{}, std::nullopt, frontend.globals.globalScope->level, Luau::TableState::Sealed});
+        getFrontend().globals.globalTypes.addType(TableType{{}, std::nullopt, getFrontend().globals.globalScope->level, Luau::TableState::Sealed});
 
-    getMutable<TableType>(colorType)->props = {{"toHSV", {builtinTypes->anyType, /* deprecated= */ true, "Color3:ToHSV"}}};
+    getMutable<TableType>(colorType)->props = {{"toHSV", {getBuiltins()->anyType, /* deprecated= */ true, "Color3:ToHSV"}}};
 
-    addGlobalBinding(frontend.globals, "Color3", Binding{colorType, {}});
+    addGlobalBinding(getFrontend().globals, "Color3", Binding{colorType, {}});
 
-    if (TableType* ttv = getMutable<TableType>(getGlobalBinding(frontend.globals, "table")))
+    if (TableType* ttv = getMutable<TableType>(getGlobalBinding(getFrontend().globals, "table")))
     {
         ttv->props["foreach"].deprecated = true;
         ttv->props["getn"].deprecated = true;
         ttv->props["getn"].deprecatedSuggestion = "#";
     }
 
-    freeze(frontend.globals.globalTypes);
+    freeze(getFrontend().globals.globalTypes);
 
     LintResult result = lint(R"(
 return function (i: Instance)
@@ -1553,7 +1570,7 @@ end
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "DeprecatedApiUntyped")
 {
-    if (TableType* ttv = getMutable<TableType>(getGlobalBinding(frontend.globals, "table")))
+    if (TableType* ttv = getMutable<TableType>(getGlobalBinding(getFrontend().globals, "table")))
     {
         ttv->props["foreach"].deprecated = true;
         ttv->props["getn"].deprecated = true;
@@ -1610,7 +1627,7 @@ static void checkDeprecatedWarning(const Luau::LintWarning& warning, const Luau:
 
 TEST_CASE_FIXTURE(Fixture, "DeprecatedAttribute")
 {
-    ScopedFastFlag sff[] = {{FFlag::LuauDeprecatedAttribute, true}, {FFlag::LuauSolverV2, true}};
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
 
     // @deprecated works on local functions
     {
@@ -1827,7 +1844,7 @@ Account = { balance=0 }
 function Account:deposit(v)
     self.balance = self.balance + v
 end
-    
+
 Account:deposit(200.00)
 )");
 
@@ -1848,7 +1865,7 @@ end
 function Account:deposit (v)
     self.balance = self.balance + v
 end
-    
+
 (getAccount()):deposit(200.00)
 )");
 
@@ -1857,9 +1874,196 @@ end
     }
 }
 
+TEST_CASE_FIXTURE(Fixture, "DeprecatedAttributeWithParams")
+{
+    // @deprecated works on local functions
+    {
+        LintResult result = lint(R"(
+@[deprecated{ use = "prodfun", reason = "Too old." }]
+local function testfun(x)
+    return x + 1
+end
+
+testfun(1)
+)");
+
+        REQUIRE(1 == result.warnings.size());
+        checkDeprecatedWarning(
+            result.warnings[0], Position(6, 0), Position(6, 7), "Function 'testfun' is deprecated, use 'prodfun' instead. Too old."
+        );
+    }
+
+    // @deprecated works on globals functions
+    {
+        LintResult result = lint(R"(
+@[deprecated{ use = "prodfun", reason = "Too old." }]
+function testfun(x)
+    return x + 1
+end
+
+testfun(1)
+)");
+
+        REQUIRE(1 == result.warnings.size());
+        checkDeprecatedWarning(
+            result.warnings[0], Position(6, 0), Position(6, 7), "Function 'testfun' is deprecated, use 'prodfun' instead. Too old."
+        );
+    }
+
+    // @deprecated with only 'use' works on local functions
+    {
+        LintResult result = lint(R"(
+@[deprecated{ use = "prodfun" }]
+local function testfun(x)
+    return x + 1
+end
+
+testfun(1)
+)");
+
+        REQUIRE(1 == result.warnings.size());
+        checkDeprecatedWarning(result.warnings[0], Position(6, 0), Position(6, 7), "Function 'testfun' is deprecated, use 'prodfun' instead");
+    }
+
+    // @deprecated with only 'use' works on globals functions
+    {
+        LintResult result = lint(R"(
+@[deprecated{ use = "prodfun" }]
+function testfun(x)
+    return x + 1
+end
+
+testfun(1)
+)");
+        REQUIRE(1 == result.warnings.size());
+        checkDeprecatedWarning(result.warnings[0], Position(6, 0), Position(6, 7), "Function 'testfun' is deprecated, use 'prodfun' instead");
+    }
+
+
+    // @deprecated with only 'reason' works on local functions
+    {
+        LintResult result = lint(R"(
+@[deprecated{ reason = "Too old." }]
+local function testfun(x)
+    return x + 1
+end
+
+testfun(1)
+)");
+
+        REQUIRE(1 == result.warnings.size());
+        checkDeprecatedWarning(result.warnings[0], Position(6, 0), Position(6, 7), "Function 'testfun' is deprecated. Too old.");
+    }
+
+    // @deprecated with only 'reason' works on globals functions
+    {
+        LintResult result = lint(R"(
+@[deprecated{ reason = "Too old." }]
+function testfun(x)
+    return x + 1
+end
+
+testfun(1)
+)");
+
+        REQUIRE(1 == result.warnings.size());
+        checkDeprecatedWarning(result.warnings[0], Position(6, 0), Position(6, 7), "Function 'testfun' is deprecated. Too old.");
+    }
+
+    // @deprecated works for methods with a literal class name
+    {
+        LintResult result = lint(R"(
+Account = { balance=0 }
+
+@[deprecated{use = 'credit', reason = 'It sounds cool'}]
+function Account:deposit(v)
+    self.balance = self.balance + v
+end
+
+Account:deposit(200.00)
+)");
+
+        REQUIRE(1 == result.warnings.size());
+        checkDeprecatedWarning(
+            result.warnings[0], Position(8, 0), Position(8, 15), "Member 'Account.deposit' is deprecated, use 'credit' instead. It sounds cool"
+        );
+    }
+
+    // @deprecated works for methods with a compound expression class name
+    {
+        LintResult result = lint(R"(
+Account = { balance=0 }
+
+function getAccount()
+    return Account
+end
+
+@[deprecated{use = 'credit', reason = 'It sounds cool'}]
+function Account:deposit (v)
+    self.balance = self.balance + v
+end
+
+(getAccount()):deposit(200.00)
+)");
+
+        REQUIRE(1 == result.warnings.size());
+        checkDeprecatedWarning(
+            result.warnings[0], Position(12, 0), Position(12, 22), "Member 'deposit' is deprecated, use 'credit' instead. It sounds cool"
+        );
+    }
+
+    {
+        loadDefinition(R"(
+@[deprecated{use = 'foo', reason = 'Do better.'}] declare function bar(x: number): string
+)");
+
+        LintResult result = lint(R"(
+bar(2)
+)");
+
+        REQUIRE(1 == result.warnings.size());
+        checkDeprecatedWarning(result.warnings[0], Position(1, 0), Position(1, 3), "Function 'bar' is deprecated, use 'foo' instead. Do better.");
+    }
+
+    {
+        loadDefinition(R"(
+declare Hooty : {
+    tooty : @[deprecated{use = 'foo', reason = 'bar'}] @checked (number) -> number
+}
+)");
+        LintResult result = lint(R"(
+print(Hooty:tooty(2.0))
+)");
+
+        REQUIRE(1 == result.warnings.size());
+        checkDeprecatedWarning(result.warnings[0], Position(1, 6), Position(1, 17), "Member 'Hooty.tooty' is deprecated, use 'foo' instead. bar");
+    }
+
+    {
+        loadDefinition(R"(
+declare class Foo
+   @[deprecated{use = 'foo', reason = 'baz'}]
+   function bar(self, value: number) : number
+end
+
+declare Foo: {
+   new: () -> Foo
+}
+)");
+
+        LintResult result = lint(R"(
+local foo = Foo.new()
+print(foo:bar(2.0))
+)");
+
+        REQUIRE(1 == result.warnings.size());
+        checkDeprecatedWarning(result.warnings[0], Position(2, 6), Position(2, 13), "Member 'bar' is deprecated, use 'foo' instead. baz");
+    }
+}
+
 TEST_CASE_FIXTURE(Fixture, "DeprecatedAttributeFunctionDeclaration")
 {
-    ScopedFastFlag sff[] = {{FFlag::LuauDeprecatedAttribute, true}, {FFlag::LuauSolverV2, true}};
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
 
     // @deprecated works on function type declarations
 
@@ -1877,7 +2081,7 @@ bar(2)
 
 TEST_CASE_FIXTURE(Fixture, "DeprecatedAttributeTableDeclaration")
 {
-    ScopedFastFlag sff[] = {{FFlag::LuauDeprecatedAttribute, true}, {FFlag::LuauSolverV2, true}};
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
 
     // @deprecated works on table type declarations
 
@@ -1897,7 +2101,7 @@ print(Hooty:tooty(2.0))
 
 TEST_CASE_FIXTURE(Fixture, "DeprecatedAttributeMethodDeclaration")
 {
-    ScopedFastFlag sff[] = {{FFlag::LuauDeprecatedAttribute, true}, {FFlag::LuauSolverV2, true}};
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
 
     // @deprecated works on table type declarations
 
@@ -2319,8 +2523,6 @@ local _ = a <= (b == 0)
 
 TEST_CASE_FIXTURE(Fixture, "RedundantNativeAttribute")
 {
-    ScopedFastFlag sff[] = {{FFlag::LintRedundantNativeAttribute, true}};
-
     LintResult result = lint(R"(
 --!native
 
@@ -2345,4 +2547,19 @@ f(3)(4)
     CHECK_EQ(result.warnings[1].location, Location(Position(5, 4), Position(5, 11)));
 }
 
+TEST_CASE_FIXTURE(Fixture, "type_instantiation_lints")
+{
+    ScopedFastFlag sff{FFlag::LuauExplicitTypeInstantiationSyntax, true};
+
+    LintResult result = lint(R"(
+local function a<b>(cool: b)
+    print(cool)
+end
+
+a<<"hi">>("hi")
+)");
+
+    REQUIRE(0 == result.warnings.size());
+}
+
 TEST_SUITE_END();
