@@ -21,30 +21,30 @@ struct SimplifyFixture : Fixture
 
     ToStringOptions opts;
 
-    Scope scope{builtinTypes->anyTypePack};
+    Scope scope{getBuiltins()->anyTypePack};
 
-    const TypeId anyTy = builtinTypes->anyType;
-    const TypeId unknownTy = builtinTypes->unknownType;
-    const TypeId neverTy = builtinTypes->neverType;
-    const TypeId errorTy = builtinTypes->errorType;
+    const TypeId anyTy = getBuiltins()->anyType;
+    const TypeId unknownTy = getBuiltins()->unknownType;
+    const TypeId neverTy = getBuiltins()->neverType;
+    const TypeId errorTy = getBuiltins()->errorType;
 
-    const TypeId functionTy = builtinTypes->functionType;
-    const TypeId tableTy = builtinTypes->tableType;
+    const TypeId functionTy = getBuiltins()->functionType;
+    const TypeId tableTy = getBuiltins()->tableType;
 
-    const TypeId numberTy = builtinTypes->numberType;
-    const TypeId stringTy = builtinTypes->stringType;
-    const TypeId booleanTy = builtinTypes->booleanType;
-    const TypeId nilTy = builtinTypes->nilType;
+    const TypeId numberTy = getBuiltins()->numberType;
+    const TypeId stringTy = getBuiltins()->stringType;
+    const TypeId booleanTy = getBuiltins()->booleanType;
+    const TypeId nilTy = getBuiltins()->nilType;
 
-    const TypeId classTy = builtinTypes->classType;
+    const TypeId classTy = getBuiltins()->externType;
 
-    const TypeId trueTy = builtinTypes->trueType;
-    const TypeId falseTy = builtinTypes->falseType;
+    const TypeId trueTy = getBuiltins()->trueType;
+    const TypeId falseTy = getBuiltins()->falseType;
 
-    const TypeId truthyTy = builtinTypes->truthyType;
-    const TypeId falsyTy = builtinTypes->falsyType;
+    const TypeId truthyTy = getBuiltins()->truthyType;
+    const TypeId falsyTy = getBuiltins()->falsyType;
 
-    const TypeId freeTy = freshType(arena, builtinTypes, &scope);
+    const TypeId freeTy = freshType(arena, getBuiltins(), &scope);
     const TypeId genericTy = arena->addType(GenericType{});
     const TypeId blockedTy = arena->addType(BlockedType{});
     const TypeId pendingTy = arena->addType(PendingExpansionType{{}, {}, {}, {}});
@@ -55,28 +55,29 @@ struct SimplifyFixture : Fixture
     const TypePackId emptyTypePack = arena->addTypePack({});
 
     const TypeId fn1Ty = arena->addType(FunctionType{emptyTypePack, emptyTypePack});
-    const TypeId fn2Ty = arena->addType(FunctionType{builtinTypes->anyTypePack, emptyTypePack});
+    const TypeId fn2Ty = arena->addType(FunctionType{getBuiltins()->anyTypePack, emptyTypePack});
 
     TypeId parentClassTy = nullptr;
     TypeId childClassTy = nullptr;
     TypeId anotherChildClassTy = nullptr;
     TypeId unrelatedClassTy = nullptr;
 
+    // This only affects type stringification.
     ScopedFastFlag sff{FFlag::LuauSolverV2, true};
 
     SimplifyFixture()
     {
-        createSomeClasses(&frontend);
+        createSomeExternTypes(getFrontend());
 
-        parentClassTy = frontend.globals.globalScope->linearSearchForBinding("Parent")->typeId;
-        childClassTy = frontend.globals.globalScope->linearSearchForBinding("Child")->typeId;
-        anotherChildClassTy = frontend.globals.globalScope->linearSearchForBinding("AnotherChild")->typeId;
-        unrelatedClassTy = frontend.globals.globalScope->linearSearchForBinding("Unrelated")->typeId;
+        parentClassTy = getFrontend().globals.globalScope->linearSearchForBinding("Parent")->typeId;
+        childClassTy = getFrontend().globals.globalScope->linearSearchForBinding("Child")->typeId;
+        anotherChildClassTy = getFrontend().globals.globalScope->linearSearchForBinding("AnotherChild")->typeId;
+        unrelatedClassTy = getFrontend().globals.globalScope->linearSearchForBinding("Unrelated")->typeId;
     }
 
     TypeId intersect(TypeId a, TypeId b)
     {
-        return simplifyIntersection(builtinTypes, arena, a, b).result;
+        return simplifyIntersection(getBuiltins(), arena, a, b).result;
     }
 
     std::string intersectStr(TypeId a, TypeId b)
@@ -89,11 +90,11 @@ struct SimplifyFixture : Fixture
         return bool(get<IntersectionType>(follow(a)));
     }
 
-    TypeId mkTable(std::map<Name, TypeId> propTypes)
+    TypeId mkTable(std::map<Name, Property> propTypes)
     {
         TableType::Props props;
-        for (const auto& [name, ty] : propTypes)
-            props[name] = Property{ty};
+        for (const auto& [name, prop] : propTypes)
+            props[name] = prop;
 
         return arena->addType(TableType{props, {}, TypeLevel{}, TableState::Sealed});
     }
@@ -110,7 +111,7 @@ struct SimplifyFixture : Fixture
 
     TypeId union_(TypeId a, TypeId b)
     {
-        return simplifyUnion(builtinTypes, arena, a, b).result;
+        return simplifyUnion(getBuiltins(), arena, a, b).result;
     }
 };
 
@@ -476,7 +477,7 @@ TEST_CASE_FIXTURE(SimplifyFixture, "union")
     CHECK(nilTy == intersect(t1, nilTy));
     // CHECK(nilTy == intersect(nilTy, t1)); // TODO?
 
-    CHECK(builtinTypes->stringType == intersect(builtinTypes->optionalStringType, truthyTy));
+    CHECK(getBuiltins()->stringType == intersect(getBuiltins()->optionalStringType, truthyTy));
 }
 
 TEST_CASE_FIXTURE(SimplifyFixture, "two_unions")
@@ -512,7 +513,7 @@ TEST_CASE_FIXTURE(SimplifyFixture, "top_class_type")
     CHECK(neverTy == intersect(classTy, stringTy));
 }
 
-TEST_CASE_FIXTURE(SimplifyFixture, "classes")
+TEST_CASE_FIXTURE(SimplifyFixture, "extern_types")
 {
     CHECK(childClassTy == intersect(childClassTy, parentClassTy));
     CHECK(childClassTy == intersect(parentClassTy, childClassTy));
@@ -523,7 +524,7 @@ TEST_CASE_FIXTURE(SimplifyFixture, "classes")
     CHECK(neverTy == intersect(childClassTy, unrelatedClassTy));
 }
 
-TEST_CASE_FIXTURE(SimplifyFixture, "negations_of_classes")
+TEST_CASE_FIXTURE(SimplifyFixture, "negations_of_extern_types")
 {
     TypeId notChildClassTy = mkNegation(childClassTy);
     TypeId notParentClassTy = mkNegation(parentClassTy);
@@ -611,12 +612,77 @@ TEST_CASE_FIXTURE(SimplifyFixture, "bound_intersected_by_itself_should_be_itself
 TEST_CASE_FIXTURE(SimplifyFixture, "cyclic_never_union_and_string")
 {
     // t1 where t1 = never | t1
-    TypeId leftType = arena->addType(UnionType{{builtinTypes->neverType, builtinTypes->neverType}});
+    TypeId leftType = arena->addType(UnionType{{getBuiltins()->neverType, getBuiltins()->neverType}});
     UnionType* leftUnion = getMutable<UnionType>(leftType);
     REQUIRE(leftUnion);
     leftUnion->options[0] = leftType;
 
-    CHECK(builtinTypes->stringType == union_(leftType, builtinTypes->stringType));
+    CHECK(getBuiltins()->stringType == union_(leftType, getBuiltins()->stringType));
+}
+
+TEST_CASE_FIXTURE(SimplifyFixture, "any & (error | string)")
+{
+    TypeId errStringTy = arena->addType(UnionType{{getBuiltins()->errorType, getBuiltins()->stringType}});
+
+    auto res = intersect(builtinTypes->anyType, errStringTy);
+
+    CHECK("*error-type* | string" == toString(res));
+}
+
+TEST_CASE_FIXTURE(SimplifyFixture, "(error | string) & any")
+{
+    TypeId errStringTy = arena->addType(UnionType{{getBuiltins()->errorType, getBuiltins()->stringType}});
+
+    auto res = intersect(errStringTy, builtinTypes->anyType);
+
+    CHECK("*error-type* | string" == toString(res));
+}
+
+TEST_CASE_FIXTURE(SimplifyFixture, "{ x: number, y: number } & { x: unknown }")
+{
+    TypeId leftTy = mkTable({{"x", builtinTypes->numberType}, {"y", builtinTypes->numberType}});
+    TypeId rightTy = mkTable({{"x", Property::rw(builtinTypes->unknownType)}});
+
+    CHECK(leftTy == intersect(leftTy, rightTy));
+}
+
+TEST_CASE_FIXTURE(SimplifyFixture, "{ x: number, y: number } & { read x: unknown }")
+{
+    TypeId leftTy = mkTable({{"x", builtinTypes->numberType}, {"y", builtinTypes->numberType}});
+    TypeId rightTy = mkTable({{"x", Property::readonly(builtinTypes->unknownType)}});
+
+    CHECK(leftTy == intersect(leftTy, rightTy));
+}
+
+TEST_CASE_FIXTURE(SimplifyFixture, "{ read x: Child } & { x: Parent }")
+{
+    createSomeExternTypes(getFrontend());
+
+    TypeId parentTy = getFrontend().globals.globalScope->exportedTypeBindings["Parent"].type;
+    REQUIRE(parentTy);
+
+    TypeId childTy = getFrontend().globals.globalScope->exportedTypeBindings["Child"].type;
+    REQUIRE(childTy);
+
+    TypeId leftTy = mkTable({{"x", Property::readonly(childTy)}});
+    TypeId rightTy = mkTable({{"x", parentTy}});
+
+    // TODO: This could be { read x: Child, write x: Parent }
+    CHECK("{ read x: Child } & { x: Parent }" == toString(intersect(leftTy, rightTy)));
+}
+
+TEST_CASE_FIXTURE(SimplifyFixture, "intersect_parts_empty_table_non_empty")
+{
+    TableType empty;
+    empty.state = TableState::Sealed;
+    TypeId emptyTable = arena->addType(std::move(empty));
+
+    TableType nonEmpty;
+    nonEmpty.props["p"] = arena->addType(UnionType{{getBuiltins()->numberType, getBuiltins()->stringType}});
+    nonEmpty.state = TableState::Sealed;
+    TypeId nonEmptyTable = arena->addType(std::move(nonEmpty));
+
+    CHECK("{ p: number | string }" == toString(simplifyIntersection(getBuiltins(), arena, {nonEmptyTable, emptyTable}).result));
 }
 
 TEST_SUITE_END();
