@@ -6,6 +6,7 @@
 #include "Fixture.h"
 
 #include "Luau/TypeChecker2.h"
+#include "Luau/TypePack.h"
 #include "ScopedFlags.h"
 #include "doctest.h"
 
@@ -13,8 +14,8 @@ using namespace Luau;
 
 LUAU_FASTFLAG(LuauRecursiveTypeParameterRestriction)
 LUAU_FASTFLAG(LuauSolverV2)
-LUAU_FASTFLAG(LuauAttributeSyntax)
-LUAU_FASTFLAG(LuauImproveTypePathsInErrors)
+LUAU_FASTFLAG(LuauBetterTypeMismatchErrors)
+LUAU_FASTFLAG(LuauToStringDecomposition)
 
 TEST_SUITE_BEGIN("ToString");
 
@@ -51,7 +52,16 @@ TEST_CASE_FIXTURE(Fixture, "free_types")
     CheckResult result = check("local a");
     LUAU_REQUIRE_NO_ERRORS(result);
 
-    CHECK_EQ("a", toString(requireType("a")));
+    CHECK_EQ("'a", toString(requireType("a")));
+}
+
+TEST_CASE_FIXTURE(Fixture, "free_types_stringify_the_same_regardless_of_solver")
+{
+    TypeArena a;
+    TypeId t =
+        a.addType(FreeType{getFrontend().globals.globalScope.get(), getFrontend().builtinTypes->neverType, getFrontend().builtinTypes->unknownType});
+
+    CHECK_EQ("'a", toString(t));
 }
 
 TEST_CASE_FIXTURE(Fixture, "cyclic_table")
@@ -60,10 +70,7 @@ TEST_CASE_FIXTURE(Fixture, "cyclic_table")
     TableType* tableOne = getMutable<TableType>(&cyclicTable);
     tableOne->props["self"] = {&cyclicTable};
 
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ("t1 where t1 = {| self: t1 |}", toString(&cyclicTable));
-    else
-        CHECK_EQ("t1 where t1 = { self: t1 }", toString(&cyclicTable));
+    CHECK_EQ("t1 where t1 = {| self: t1 |}", toString(&cyclicTable));
 }
 
 TEST_CASE_FIXTURE(Fixture, "named_table")
@@ -81,18 +88,12 @@ TEST_CASE_FIXTURE(Fixture, "empty_table")
         local a: {}
     )");
 
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ("{  }", toString(requireType("a")));
-    else
-        CHECK_EQ("{|  |}", toString(requireType("a")));
+    CHECK_EQ("{  }", toString(requireType("a")));
 
     // Should stay the same with useLineBreaks enabled
     ToStringOptions opts;
     opts.useLineBreaks = true;
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ("{  }", toString(requireType("a"), opts));
-    else
-        CHECK_EQ("{|  |}", toString(requireType("a"), opts));
+    CHECK_EQ("{  }", toString(requireType("a"), opts));
 }
 
 TEST_CASE_FIXTURE(Fixture, "table_respects_use_line_break")
@@ -104,24 +105,14 @@ TEST_CASE_FIXTURE(Fixture, "table_respects_use_line_break")
     ToStringOptions opts;
     opts.useLineBreaks = true;
 
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ(
-            "{\n"
-            "    anotherProp: number,\n"
-            "    prop: string,\n"
-            "    thirdProp: boolean\n"
-            "}",
-            toString(requireType("a"), opts)
-        );
-    else
-        CHECK_EQ(
-            "{|\n"
-            "    anotherProp: number,\n"
-            "    prop: string,\n"
-            "    thirdProp: boolean\n"
-            "|}",
-            toString(requireType("a"), opts)
-        );
+    CHECK_EQ(
+        "{\n"
+        "    anotherProp: number,\n"
+        "    prop: string,\n"
+        "    thirdProp: boolean\n"
+        "}",
+        toString(requireType("a"), opts)
+    );
 }
 
 TEST_CASE_FIXTURE(Fixture, "nil_or_nil_is_nil_not_question_mark")
@@ -151,10 +142,7 @@ TEST_CASE_FIXTURE(Fixture, "metatable")
     Type table{TypeVariant(TableType())};
     Type metatable{TypeVariant(TableType())};
     Type mtv{TypeVariant(MetatableType{&table, &metatable})};
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ("{ @metatable {|  |}, {|  |} }", toString(&mtv));
-    else
-        CHECK_EQ("{ @metatable {  }, {  } }", toString(&mtv));
+    CHECK_EQ("{ @metatable {|  |}, {|  |} }", toString(&mtv));
 }
 
 TEST_CASE_FIXTURE(Fixture, "named_metatable")
@@ -220,8 +208,8 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "exhaustive_toString_of_cyclic_table")
     {
         CHECK_EQ(
             "t2 where "
-            "t1 = { __index: t1, __mul: ((t2, number) -> t2) & ((t2, t2) -> t2), new: () -> t2 } ; "
-            "t2 = { @metatable t1, {| x: number, y: number, z: number |} }",
+            "t1 = {| __index: t1, __mul: ((t2, number) -> t2) & ((t2, t2) -> t2), new: () -> t2 |} ; "
+            "t2 = { @metatable t1, { x: number, y: number, z: number } }",
             a
         );
     }
@@ -230,27 +218,27 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "exhaustive_toString_of_cyclic_table")
 
 TEST_CASE_FIXTURE(Fixture, "intersection_parenthesized_only_if_needed")
 {
-    auto utv = Type{UnionType{{builtinTypes->numberType, builtinTypes->stringType}}};
-    auto itv = Type{IntersectionType{{&utv, builtinTypes->booleanType}}};
+    auto utv = Type{UnionType{{getBuiltins()->numberType, getBuiltins()->stringType}}};
+    auto itv = Type{IntersectionType{{&utv, getBuiltins()->booleanType}}};
 
     CHECK_EQ(toString(&itv), "(number | string) & boolean");
 }
 
 TEST_CASE_FIXTURE(Fixture, "union_parenthesized_only_if_needed")
 {
-    auto itv = Type{IntersectionType{{builtinTypes->numberType, builtinTypes->stringType}}};
-    auto utv = Type{UnionType{{&itv, builtinTypes->booleanType}}};
+    auto itv = Type{IntersectionType{{getBuiltins()->numberType, getBuiltins()->stringType}}};
+    auto utv = Type{UnionType{{&itv, getBuiltins()->booleanType}}};
 
     CHECK_EQ(toString(&utv), "(number & string) | boolean");
 }
 
 TEST_CASE_FIXTURE(Fixture, "functions_are_always_parenthesized_in_unions_or_intersections")
 {
-    auto stringAndNumberPack = TypePackVar{TypePack{{builtinTypes->stringType, builtinTypes->numberType}}};
-    auto numberAndStringPack = TypePackVar{TypePack{{builtinTypes->numberType, builtinTypes->stringType}}};
+    auto stringAndNumberPack = TypePackVar{TypePack{{getBuiltins()->stringType, getBuiltins()->numberType}}};
+    auto numberAndStringPack = TypePackVar{TypePack{{getBuiltins()->numberType, getBuiltins()->stringType}}};
 
     auto sn2ns = Type{FunctionType{&stringAndNumberPack, &numberAndStringPack}};
-    auto ns2sn = Type{FunctionType(frontend.globals.globalScope->level, &numberAndStringPack, &stringAndNumberPack)};
+    auto ns2sn = Type{FunctionType(getFrontend().globals.globalScope->level, &numberAndStringPack, &stringAndNumberPack)};
 
     auto utv = Type{UnionType{{&ns2sn, &sn2ns}}};
     auto itv = Type{IntersectionType{{&ns2sn, &sn2ns}}};
@@ -339,34 +327,28 @@ TEST_CASE_FIXTURE(Fixture, "quit_stringifying_table_type_when_length_is_exceeded
 {
     TableType ttv{};
     for (char c : std::string("abcdefghijklmno"))
-        ttv.props[std::string(1, c)] = {builtinTypes->numberType};
+        ttv.props[std::string(1, c)] = {getBuiltins()->numberType};
 
     Type tv{ttv};
 
     ToStringOptions o;
     o.exhaustive = false;
     o.maxTableLength = 40;
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ(toString(&tv, o), "{| a: number, b: number, c: number, d: number, e: number, ... 10 more ... |}");
-    else
-        CHECK_EQ(toString(&tv, o), "{ a: number, b: number, c: number, d: number, e: number, ... 10 more ... }");
+    CHECK_EQ(toString(&tv, o), "{| a: number, b: number, c: number, d: number, e: number, ... 10 more ... |}");
 }
 
 TEST_CASE_FIXTURE(Fixture, "stringifying_table_type_is_still_capped_when_exhaustive")
 {
     TableType ttv{};
     for (char c : std::string("abcdefg"))
-        ttv.props[std::string(1, c)] = {builtinTypes->numberType};
+        ttv.props[std::string(1, c)] = {getBuiltins()->numberType};
 
     Type tv{ttv};
 
     ToStringOptions o;
     o.exhaustive = true;
     o.maxTableLength = 40;
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ(toString(&tv, o), "{| a: number, b: number, c: number, d: number, e: number, ... 2 more ... |}");
-    else
-        CHECK_EQ(toString(&tv, o), "{ a: number, b: number, c: number, d: number, e: number, ... 2 more ... }");
+    CHECK_EQ(toString(&tv, o), "{| a: number, b: number, c: number, d: number, e: number, ... 2 more ... |}");
 }
 
 TEST_CASE_FIXTURE(Fixture, "quit_stringifying_type_when_length_is_exceeded")
@@ -442,21 +424,18 @@ TEST_CASE_FIXTURE(Fixture, "stringifying_table_type_correctly_use_matching_table
 {
     TableType ttv{TableState::Sealed, TypeLevel{}};
     for (char c : std::string("abcdefghij"))
-        ttv.props[std::string(1, c)] = {builtinTypes->numberType};
+        ttv.props[std::string(1, c)] = {getBuiltins()->numberType};
 
     Type tv{ttv};
 
     ToStringOptions o;
     o.maxTableLength = 40;
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ(toString(&tv, o), "{ a: number, b: number, c: number, d: number, e: number, ... 5 more ... }");
-    else
-        CHECK_EQ(toString(&tv, o), "{| a: number, b: number, c: number, d: number, e: number, ... 5 more ... |}");
+    CHECK_EQ(toString(&tv, o), "{ a: number, b: number, c: number, d: number, e: number, ... 5 more ... }");
 }
 
 TEST_CASE_FIXTURE(Fixture, "stringifying_cyclic_union_type_bails_early")
 {
-    Type tv{UnionType{{builtinTypes->stringType, builtinTypes->numberType}}};
+    Type tv{UnionType{{getBuiltins()->stringType, getBuiltins()->numberType}}};
     UnionType* utv = getMutable<UnionType>(&tv);
     utv->options.push_back(&tv);
     utv->options.push_back(&tv);
@@ -477,28 +456,34 @@ TEST_CASE_FIXTURE(Fixture, "stringifying_cyclic_intersection_type_bails_early")
 TEST_CASE_FIXTURE(Fixture, "stringifying_array_uses_array_syntax")
 {
     TableType ttv{TableState::Sealed, TypeLevel{}};
-    ttv.indexer = TableIndexer{builtinTypes->numberType, builtinTypes->stringType};
+    ttv.indexer = TableIndexer{getBuiltins()->numberType, getBuiltins()->stringType};
 
     CHECK_EQ("{string}", toString(Type{ttv}));
 
-    ttv.props["A"] = {builtinTypes->numberType};
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ("{ [number]: string, A: number }", toString(Type{ttv}));
-    else
-        CHECK_EQ("{| [number]: string, A: number |}", toString(Type{ttv}));
+    ttv.props["A"] = {getBuiltins()->numberType};
+    CHECK_EQ("{ [number]: string, A: number }", toString(Type{ttv}));
 
     ttv.props.clear();
     ttv.state = TableState::Unsealed;
     CHECK_EQ("{string}", toString(Type{ttv}));
 }
 
+TEST_CASE_FIXTURE(Fixture, "the_empty_type_pack_should_be_parenthesized")
+{
+    TypePackVar emptyTypePack{TypePack{}};
+    CHECK_EQ(toString(&emptyTypePack), "()");
+
+    auto unitToUnit = Type{FunctionType{&emptyTypePack, &emptyTypePack}};
+    CHECK_EQ(toString(&unitToUnit), "() -> ()");
+}
+
 
 TEST_CASE_FIXTURE(Fixture, "generic_packs_are_stringified_differently_from_generic_types")
 {
     TypePackVar tpv{GenericTypePack{"a"}};
     CHECK_EQ(toString(&tpv), "a...");
 
-    Type tv{GenericType{"a"}};
+    Type tv{GenericType{"a", Polarity::Mixed}};
     CHECK_EQ(toString(&tv), "a");
 }
 
@@ -619,39 +604,31 @@ TEST_CASE_FIXTURE(Fixture, "toString_the_boundTo_table_type_contained_within_a_T
     Type tv1{TableType{}};
     TableType* ttv = getMutable<TableType>(&tv1);
     ttv->state = TableState::Sealed;
-    ttv->props["hello"] = {builtinTypes->numberType};
-    ttv->props["world"] = {builtinTypes->numberType};
+    ttv->props["hello"] = {getBuiltins()->numberType};
+    ttv->props["world"] = {getBuiltins()->numberType};
 
     TypePackVar tpv1{TypePack{{&tv1}}};
 
     Type tv2{TableType{}};
     TableType* bttv = getMutable<TableType>(&tv2);
     bttv->state = TableState::Free;
-    bttv->props["hello"] = {builtinTypes->numberType};
+    bttv->props["hello"] = {getBuiltins()->numberType};
     bttv->boundTo = &tv1;
 
     TypePackVar tpv2{TypePack{{&tv2}}};
 
 
-    if (FFlag::LuauSolverV2)
-    {
-        CHECK_EQ("{ hello: number, world: number }", toString(&tpv1));
-        CHECK_EQ("{ hello: number, world: number }", toString(&tpv2));
-    }
-    else
-    {
-        CHECK_EQ("{| hello: number, world: number |}", toString(&tpv1));
-        CHECK_EQ("{| hello: number, world: number |}", toString(&tpv2));
-    }
+    CHECK_EQ("{ hello: number, world: number }", toString(&tpv1));
+    CHECK_EQ("{ hello: number, world: number }", toString(&tpv2));
 }
 
 TEST_CASE_FIXTURE(Fixture, "no_parentheses_around_return_type_if_pack_has_an_empty_head_link")
 {
     TypeArena arena;
-    TypePackId realTail = arena.addTypePack({builtinTypes->stringType});
+    TypePackId realTail = arena.addTypePack({getBuiltins()->stringType});
     TypePackId emptyTail = arena.addTypePack({}, realTail);
 
-    TypePackId argList = arena.addTypePack({builtinTypes->stringType});
+    TypePackId argList = arena.addTypePack({getBuiltins()->stringType});
 
     TypeId functionType = arena.addType(FunctionType{argList, emptyTail});
 
@@ -680,10 +657,7 @@ TEST_CASE_FIXTURE(Fixture, "no_parentheses_around_cyclic_function_type_in_inters
 
     LUAU_REQUIRE_NO_ERRORS(result);
 
-    if (FFlag::LuauSolverV2)
-        CHECK("(() -> t1) & ((number) -> ()) where t1 = () -> t1" == toString(requireType("a")));
-    else
-        CHECK_EQ("((number) -> ()) & t1 where t1 = () -> t1", toString(requireType("a")));
+    CHECK_EQ("((number) -> ()) & t1 where t1 = () -> t1", toString(requireType("a")));
 }
 
 TEST_CASE_FIXTURE(Fixture, "self_recursive_instantiated_param")
@@ -867,53 +841,57 @@ TEST_CASE_FIXTURE(Fixture, "tostring_error_mismatch")
 {
     CheckResult result = check(R"(
         --!strict
-        function f1() : {a : number, b : string, c : { d : number}}
-            return { a = 1, b = "b", c = {d = "d"}}
+        function f1(t: {a : number, b: string, c: {d: string}}) : {a : number, b : string, c : { d : number}}
+            return t
         end
     )");
 
     std::string expected;
-    if (FFlag::LuauSolverV2 && FFlag::LuauImproveTypePathsInErrors)
-        expected =
-            "Type pack '{ a: number, b: string, c: { d: string } }' could not be converted into '{ a: number, b: string, c: { d: number } }'; \n"
-            "this is because in the 1st entry in the type pack, accessing `c.d` results in `string` in the former type and `number` in the latter "
-            "type, and `string` is not exactly `number`";
+    if (FFlag::LuauSolverV2 && FFlag::LuauBetterTypeMismatchErrors)
+        expected = "Expected this to be\n\t"
+                   "'{ a: number, b: string, c: { d: number } }'\n"
+                   "but got\n\t"
+                   "'{ a: number, b: string, c: { d: string } }'; \n"
+                   "accessing `c.d` results in `string` in the latter type and `number` in the former "
+                   "type, and `string` is not exactly `number`";
     else if (FFlag::LuauSolverV2)
-        expected =
-            R"(Type pack '{ a: number, b: string, c: { d: string } }' could not be converted into '{ a: number, b: string, c: { d: number } }'; at [0][read "c"][read "d"], string is not exactly number)";
-    else if (FFlag::LuauImproveTypePathsInErrors)
-        expected = R"(Type
-	'{ a: number, b: string, c: { d: string } }'
-could not be converted into
-	'{| a: number, b: string, c: {| d: number |} |}'
-caused by:
-  Property 'c' is not compatible.
-Type
-	'{ d: string }'
-could not be converted into
-	'{| d: number |}'
-caused by:
-  Property 'd' is not compatible.
-Type 'string' could not be converted into 'number' in an invariant context)";
+        expected = "Type\n\t"
+                   "'{ a: number, b: string, c: { d: string } }'\n"
+                   "could not be converted into\n\t"
+                   "'{ a: number, b: string, c: { d: number } }'; \n"
+                   "this is because accessing `c.d` results in `string` in the former type and `number` in the latter "
+                   "type, and `string` is not exactly `number`";
+    else if (FFlag::LuauBetterTypeMismatchErrors)
+        expected = "Expected this to be exactly\n\t"
+                   "'{ a: number, b: string, c: { d: number } }'\n"
+                   "but got\n\t"
+                   "'{ a: number, b: string, c: { d: string } }'\n"
+                   "caused by:\n  "
+                   "Property 'c' is not compatible.\n"
+                   "Expected this to be exactly\n\t"
+                   "'{ d: number }'\n"
+                   "but got\n\t"
+                   "'{ d: string }'\n"
+                   "caused by:\n  "
+                   "Property 'd' is not compatible.\n"
+                   "Expected this to be exactly 'number', but got 'string'";
     else
-        expected = R"(Type
-    '{ a: number, b: string, c: { d: string } }'
-could not be converted into
-    '{| a: number, b: string, c: {| d: number |} |}'
-caused by:
-  Property 'c' is not compatible.
-Type
-    '{ d: string }'
-could not be converted into
-    '{| d: number |}'
-caused by:
-  Property 'd' is not compatible.
-Type 'string' could not be converted into 'number' in an invariant context)";
+        expected = "Type\n\t"
+                   "'{ a: number, b: string, c: { d: string } }'\n"
+                   "could not be converted into\n\t"
+                   "'{ a: number, b: string, c: { d: number } }'\n"
+                   "caused by:\n  "
+                   "Property 'c' is not compatible.\n"
+                   "Type\n\t"
+                   "'{ d: string }'\n"
+                   "could not be converted into\n\t"
+                   "'{ d: number }'\n"
+                   "caused by:\n  "
+                   "Property 'd' is not compatible.\n"
+                   "Type 'string' could not be converted into 'number' in an invariant context";
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-
     std::string actual = toString(result.errors[0]);
-
     CHECK(expected == actual);
 }
 
@@ -956,12 +934,12 @@ TEST_CASE_FIXTURE(Fixture, "cycle_rooted_in_a_pack")
 {
     TypeArena arena;
 
-    TypePackId thePack = arena.addTypePack({builtinTypes->numberType, builtinTypes->numberType});
+    TypePackId thePack = arena.addTypePack({getBuiltins()->numberType, getBuiltins()->numberType});
     TypePack* packPtr = getMutable<TypePack>(thePack);
     REQUIRE(packPtr);
 
     const TableType::Props theProps = {
-        {"BaseField", Property::readonly(builtinTypes->unknownType)},
+        {"BaseField", Property::readonly(getBuiltins()->unknownType)},
         {"BaseMethod", Property::readonly(arena.addType(FunctionType{thePack, arena.addTypePack({})}))}
     };
 
@@ -969,10 +947,7 @@ TEST_CASE_FIXTURE(Fixture, "cycle_rooted_in_a_pack")
 
     packPtr->head[0] = theTable;
 
-    if (FFlag::LuauSolverV2)
-        CHECK("tp1 where tp1 = { read BaseField: unknown, read BaseMethod: (tp1) -> () }, number" == toString(thePack));
-    else
-        CHECK("tp1 where tp1 = {| BaseField: unknown, BaseMethod: (tp1) -> () |}, number" == toString(thePack));
+    CHECK("tp1 where tp1 = { read BaseField: unknown, read BaseMethod: (tp1) -> () }, number" == toString(thePack));
 }
 
 TEST_CASE_FIXTURE(Fixture, "correct_stringification_user_defined_type_functions")
@@ -980,7 +955,7 @@ TEST_CASE_FIXTURE(Fixture, "correct_stringification_user_defined_type_functions"
     TypeFunction user{"user", nullptr};
     TypeFunctionInstanceType tftt{
         NotNull{&user},
-        std::vector<TypeId>{builtinTypes->numberType}, // Type Function Arguments
+        std::vector<TypeId>{getBuiltins()->numberType}, // Type Function Arguments
         {},
         {AstName{"woohoo"}}, // Type Function Name
         {},
@@ -992,4 +967,127 @@ TEST_CASE_FIXTURE(Fixture, "correct_stringification_user_defined_type_functions"
         CHECK_EQ(toString(&tv, {}), "woohoo<number>");
 }
 
+TEST_CASE_FIXTURE(Fixture, "record_type_compositions_table")
+{
+    ScopedFastFlag _{FFlag::LuauToStringDecomposition, true};
+
+    CheckResult checkResult = check(R"(
+        type Table = {}
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(checkResult);
+
+    ToStringOptions opts;
+
+    TypeId ty = requireTypeAlias("Table");
+    ToStringResult result = toStringDetailed(ty, opts);
+
+    REQUIRE_EQ(result.typeSpans.size(), 1);
+
+    auto [startPos, endPos, recordedTy] = result.typeSpans[0];
+    CHECK_EQ(startPos, 0);
+    CHECK_EQ(endPos, 5);
+    CHECK_EQ(recordedTy, ty);
+}
+
+TEST_CASE_FIXTURE(Fixture, "record_type_compositions_union_intersection")
+{
+    ScopedFastFlag _{FFlag::LuauToStringDecomposition, true};
+
+    CheckResult checkResult = check(R"(
+        type TableA = {}
+        type TableB = {}
+
+        type Composite1 = TableA | TableB
+        type Composite2 = TableA & TableB
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(checkResult);
+
+    ToStringOptions opts;
+
+    for (const auto& aliasName : {"Composite1", "Composite2"})
+    {
+        TypeId ty = requireTypeAlias(aliasName);
+        ToStringResult result = toStringDetailed(ty, opts);
+
+        REQUIRE_EQ(result.typeSpans.size(), 2);
+
+        auto [startPosA, endPosA, recordedTyA] = result.typeSpans[0];
+        CHECK_EQ(startPosA, 0);
+        CHECK_EQ(endPosA, 6);
+        CHECK_EQ(recordedTyA, requireTypeAlias("TableA"));
+
+        auto [startPosB, endPosB, recordedTyB] = result.typeSpans[1];
+        CHECK_EQ(startPosB, 9);
+        CHECK_EQ(endPosB, 15);
+        CHECK_EQ(recordedTyB, requireTypeAlias("TableB"));
+    }
+}
+
+TEST_CASE_FIXTURE(Fixture, "record_type_compositions_union_handle_resorted_results")
+{
+    ScopedFastFlag _{FFlag::LuauToStringDecomposition, true};
+
+    CheckResult checkResult = check(R"(
+        type Zebra = {}
+        type Alpha = {}
+
+        type Composite = Zebra | Alpha
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(checkResult);
+
+    ToStringOptions opts;
+
+    TypeId ty = requireTypeAlias("Composite");
+    ToStringResult result = toStringDetailed(ty, opts);
+
+    CHECK_EQ(result.name, "Alpha | Zebra");
+
+    REQUIRE_EQ(result.typeSpans.size(), 2);
+
+    auto [startPosAlpha, endPosAlpha, recordedTyAlpha] = result.typeSpans[0];
+    CHECK_EQ(startPosAlpha, 0);
+    CHECK_EQ(endPosAlpha, 5);
+    CHECK_EQ(recordedTyAlpha, requireTypeAlias("Alpha"));
+
+    auto [startPosZebra, endPosZebra, recordedTyZebra] = result.typeSpans[1];
+    CHECK_EQ(startPosZebra, 8);
+    CHECK_EQ(endPosZebra, 13);
+    CHECK_EQ(recordedTyZebra, requireTypeAlias("Zebra"));
+}
+
+
+TEST_CASE_FIXTURE(Fixture, "record_type_compositions_generic")
+{
+    ScopedFastFlag _{FFlag::LuauToStringDecomposition, true};
+
+    CheckResult checkResult = check(R"(
+        type Object = {}
+        type Box<T> = { inner: T }
+
+        local x: Box<Object>
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(checkResult);
+
+    ToStringOptions opts;
+
+    TypeId ty = requireType("x");
+    ToStringResult result = toStringDetailed(ty, opts);
+
+    REQUIRE_EQ(result.typeSpans.size(), 2);
+
+    auto [startPosBox, endPosBox, recordedTyBox] = result.typeSpans[0];
+    CHECK_EQ(startPosBox, 0);
+    CHECK_EQ(endPosBox, 3);
+    CHECK_EQ(recordedTyBox, ty);
+
+    auto [startPosObject, endPosObject, recordedTyObject] = result.typeSpans[1];
+    CHECK_EQ(startPosObject, 4);
+    CHECK_EQ(endPosObject, 10);
+    CHECK_EQ(recordedTyObject, requireTypeAlias("Object"));
+}
+
 TEST_SUITE_END();
