@@ -10,11 +10,7 @@
 using namespace Luau;
 
 LUAU_FASTFLAG(LuauSolverV2)
-LUAU_FASTFLAG(LuauFixInfiniteRecursionInNormalization)
-LUAU_FASTFLAG(LuauImproveTypePathsInErrors)
-LUAU_FASTFLAG(LuauBidirectionalInferenceUpcast)
-LUAU_FASTFLAG(LuauBidirectionalInferenceCollectIndexerTypes)
-LUAU_FASTFLAG(LuauRetainDefinitionAliasLocations)
+LUAU_FASTFLAG(LuauBetterTypeMismatchErrors)
 
 TEST_SUITE_BEGIN("TypeAliases");
 
@@ -80,29 +76,27 @@ TEST_CASE_FIXTURE(Fixture, "cannot_steal_hoisted_type_alias")
     if (FFlag::LuauSolverV2)
     {
         CHECK(
-            result.errors[0] ==
-            TypeError{
-                Location{{1, 21}, {1, 26}},
-                getMainSourceModule()->name,
-                TypeMismatch{
-                    builtinTypes->numberType,
-                    builtinTypes->stringType,
-                },
-            }
+            result.errors[0] == TypeError{
+                                    Location{{1, 21}, {1, 26}},
+                                    getMainSourceModule()->name,
+                                    TypeMismatch{
+                                        getBuiltins()->numberType,
+                                        getBuiltins()->stringType,
+                                    },
+                                }
         );
     }
     else
     {
         CHECK(
-            result.errors[0] ==
-            TypeError{
-                Location{{1, 8}, {1, 26}},
-                getMainSourceModule()->name,
-                TypeMismatch{
-                    builtinTypes->numberType,
-                    builtinTypes->stringType,
-                },
-            }
+            result.errors[0] == TypeError{
+                                    Location{{1, 8}, {1, 26}},
+                                    getMainSourceModule()->name,
+                                    TypeMismatch{
+                                        getBuiltins()->numberType,
+                                        getBuiltins()->stringType,
+                                    },
+                                }
         );
     }
 }
@@ -189,7 +183,7 @@ TEST_CASE_FIXTURE(Fixture, "cyclic_types_of_named_table_fields_do_not_expand_whe
     TypeMismatch* tm = get<TypeMismatch>(result.errors[0]);
     REQUIRE_MESSAGE(tm, result.errors[0]);
     CHECK_EQ("Node?", toString(tm->wantedType));
-    CHECK_EQ(builtinTypes->numberType, tm->givenType);
+    CHECK_EQ(getBuiltins()->numberType, tm->givenType);
 }
 
 TEST_CASE_FIXTURE(Fixture, "mutually_recursive_aliases")
@@ -209,9 +203,8 @@ TEST_CASE_FIXTURE(Fixture, "mutually_recursive_aliases")
 
 TEST_CASE_FIXTURE(Fixture, "generic_aliases")
 {
-    ScopedFastFlag sff[] = {
-        {FFlag::LuauSolverV2, true},
-    };
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+
     CheckResult result = check(R"(
         type T<a> = { v: a }
         local x: T<number> = { v = 123 }
@@ -220,20 +213,16 @@ TEST_CASE_FIXTURE(Fixture, "generic_aliases")
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-    const std::string expected = (FFlag::LuauImproveTypePathsInErrors)
-                                     ? "Type '{ v: string }' could not be converted into 'T<number>'; \n"
-                                       "this is because accessing `v` results in `string` in the former type and `number` in the latter type, and "
-                                       "`string` is not exactly `number`"
-                                     : R"(Type '{ v: string }' could not be converted into 'T<number>'; at [read "v"], string is not exactly number)";
-    CHECK(result.errors[0].location == Location{{4, 31}, {4, 44}});
-    CHECK_EQ(expected, toString(result.errors[0]));
+    CHECK(result.errors[0].location == Location{{4, 37}, {4, 42}});
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ("Expected this to be 'number', but got 'string'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type 'string' could not be converted into 'number'", toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(Fixture, "dependent_generic_aliases")
 {
-    ScopedFastFlag sff[] = {
-        {FFlag::LuauSolverV2, true},
-    };
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
 
     CheckResult result = check(R"(
         type T<a> = { v: a }
@@ -243,24 +232,15 @@ TEST_CASE_FIXTURE(Fixture, "dependent_generic_aliases")
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-    const std::string expected =
-        (FFlag::LuauImproveTypePathsInErrors)
-            ? "Type '{ t: { v: string } }' could not be converted into 'U<number>'; \n"
-              "this is because accessing `t.v` results in `string` in the former type and `number` in the latter type, and `string` is not exactly "
-              "`number`"
-            : R"(Type '{ t: { v: string } }' could not be converted into 'U<number>'; at [read "t"][read "v"], string is not exactly number)";
-
-    CHECK(result.errors[0].location == Location{{4, 31}, {4, 52}});
-    CHECK_EQ(expected, toString(result.errors[0]));
+    CHECK(result.errors[0].location == Location{{4, 43}, {4, 48}});
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ("Expected this to be 'number', but got 'string'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type 'string' could not be converted into 'number'", toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(Fixture, "mutually_recursive_generic_aliases")
 {
-    ScopedFastFlag sffs[] = {
-        {FFlag::LuauBidirectionalInferenceUpcast, true},
-        {FFlag::LuauBidirectionalInferenceCollectIndexerTypes, true},
-    };
-
     CheckResult result = check(R"(
         --!strict
         type T<a> = { f: a, g: U<a> }
@@ -289,9 +269,9 @@ TEST_CASE_FIXTURE(Fixture, "mutually_recursive_types_errors")
     LUAU_REQUIRE_ERRORS(result);
 
     // We had a UAF in this example caused by not cloning type function arguments
-    ModulePtr module = frontend.moduleResolver.getModule("MainModule");
+    ModulePtr module = getFrontend().moduleResolver.getModule("MainModule");
     unfreeze(module->interfaceTypes);
-    copyErrors(module->errors, module->interfaceTypes, builtinTypes);
+    copyErrors(module->errors, module->interfaceTypes, getBuiltins());
     freeze(module->interfaceTypes);
     module->internalTypes.clear();
     module->astTypes.clear();
@@ -333,6 +313,7 @@ TEST_CASE_FIXTURE(Fixture, "dont_stop_typechecking_after_reporting_duplicate_typ
 
 TEST_CASE_FIXTURE(Fixture, "stringify_type_alias_of_recursive_template_table_type")
 {
+
     CheckResult result = check(R"(
         type Table<T> = { a: T }
         type Wrapped = Table<Wrapped>
@@ -344,11 +325,12 @@ TEST_CASE_FIXTURE(Fixture, "stringify_type_alias_of_recursive_template_table_typ
     TypeMismatch* tm = get<TypeMismatch>(result.errors[0]);
     REQUIRE(tm);
     CHECK_EQ("Wrapped", toString(tm->wantedType));
-    CHECK_EQ(builtinTypes->numberType, tm->givenType);
+    CHECK_EQ(getBuiltins()->numberType, tm->givenType);
 }
 
 TEST_CASE_FIXTURE(Fixture, "stringify_type_alias_of_recursive_template_table_type2")
 {
+
     CheckResult result = check(R"(
         type Table<T> = { a: T }
         type Wrapped = (Table<Wrapped>) -> string
@@ -359,18 +341,13 @@ TEST_CASE_FIXTURE(Fixture, "stringify_type_alias_of_recursive_template_table_typ
 
     TypeMismatch* tm = get<TypeMismatch>(result.errors[0]);
     REQUIRE(tm);
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ("t1 where t1 = ({ a: t1 }) -> string", toString(tm->wantedType));
-    else
-        CHECK_EQ("t1 where t1 = ({| a: t1 |}) -> string", toString(tm->wantedType));
-    CHECK_EQ(builtinTypes->numberType, tm->givenType);
+    CHECK_EQ("t1 where t1 = ({ a: t1 }) -> string", toString(tm->wantedType));
+    CHECK_EQ(getBuiltins()->numberType, tm->givenType);
 }
 
-// Check that recursive intersection type doesn't generate an OOM
 TEST_CASE_FIXTURE(Fixture, "cli_38393_recursive_intersection_oom")
 {
-    DOES_NOT_PASS_NEW_SOLVER_GUARD();
-
+    // Check that recursive intersection type doesn't generate an OOM
     CheckResult result = check(R"(
         function _(l0:(t0)&((t0)&(((t0)&((t0)->()))->(typeof(_),typeof(# _)))),l39,...):any
         end
@@ -525,7 +502,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "general_require_multi_assign")
         local b: Bar.myvec3
     )";
 
-    CheckResult result = frontend.check("workspace/C");
+    CheckResult result = getFrontend().check("workspace/C");
     LUAU_REQUIRE_NO_ERRORS(result);
 
     TypeId aTypeId = requireType("workspace/C", "a");
@@ -544,7 +521,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "type_alias_import_mutation")
     CheckResult result = check("type t10<x> = typeof(table)");
     LUAU_REQUIRE_NO_ERRORS(result);
 
-    TypeId ty = getGlobalBinding(frontend.globals, "table");
+    TypeId ty = getGlobalBinding(getFrontend().globals, "table");
 
     CHECK(toString(ty) == "typeof(table)");
 
@@ -614,12 +591,13 @@ type Cool = typeof(c)
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "type_alias_of_an_imported_recursive_type")
 {
+
     fileResolver.source["game/A"] = R"(
 export type X = { a: number, b: X? }
 return {}
     )";
 
-    CheckResult aResult = frontend.check("game/A");
+    CheckResult aResult = getFrontend().check("game/A");
     LUAU_REQUIRE_NO_ERRORS(aResult);
 
     CheckResult bResult = check(R"(
@@ -639,12 +617,13 @@ type X = Import.X
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "type_alias_of_an_imported_recursive_generic_type")
 {
+
     fileResolver.source["game/A"] = R"(
         export type X<T, U> = { a: T, b: U, C: X<T, U>? }
         return {}
     )";
 
-    CheckResult aResult = frontend.check("game/A");
+    CheckResult aResult = getFrontend().check("game/A");
     LUAU_REQUIRE_NO_ERRORS(aResult);
 
     CheckResult bResult = check(R"(
@@ -680,8 +659,8 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "type_alias_of_an_imported_recursive_generic_
     }
     else
     {
-        CHECK_EQ(toString(*ty1, {true}), "t1 where t1 = {| C: t1?, a: T, b: U |}");
-        CHECK_EQ(toString(*ty2, {true}), "{| C: t1, a: U, b: T |} where t1 = {| C: t1, a: U, b: T |}?");
+        CHECK_EQ(toString(*ty1, {true}), "t1 where t1 = { C: t1?, a: T, b: U }");
+        CHECK_EQ(toString(*ty2, {true}), "{ C: t1, a: U, b: T } where t1 = { C: t1, a: U, b: T }?");
     }
 }
 
@@ -859,6 +838,7 @@ TEST_CASE_FIXTURE(Fixture, "generic_typevars_are_not_considered_to_escape_their_
  */
 TEST_CASE_FIXTURE(Fixture, "forward_declared_alias_is_not_clobbered_by_prior_unification_with_any")
 {
+
     CheckResult result = check(R"(
         local function x()
             local y: FutureType = {}::any
@@ -868,10 +848,7 @@ TEST_CASE_FIXTURE(Fixture, "forward_declared_alias_is_not_clobbered_by_prior_uni
         local d: FutureType = { smth = true } -- missing error, 'd' is resolved to 'any'
     )");
 
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ("{ foo: number }", toString(requireType("d"), {true}));
-    else
-        CHECK_EQ("{| foo: number |}", toString(requireType("d"), {true}));
+    CHECK_EQ("{ foo: number }", toString(requireType("d"), {true}));
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
 }
@@ -984,9 +961,6 @@ TEST_CASE_FIXTURE(Fixture, "type_alias_locations")
  */
 TEST_CASE_FIXTURE(BuiltinsFixture, "dont_lose_track_of_PendingExpansionTypes_after_substitution")
 {
-    // CLI-114134 - We need egraphs to properly simplify these types.
-    DOES_NOT_PASS_NEW_SOLVER_GUARD();
-
     fileResolver.source["game/ReactCurrentDispatcher"] = R"(
         export type BasicStateAction<S> = ((S) -> S) | S
         export type Dispatch<A> = (A) -> ()
@@ -1015,7 +989,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "dont_lose_track_of_PendingExpansionTypes_aft
         end
     )";
 
-    CheckResult result = frontend.check("game/React/React/ReactHooks");
+    CheckResult result = getFrontend().check("game/React/React/ReactHooks");
 
     LUAU_REQUIRE_NO_ERRORS(result);
 }
@@ -1068,7 +1042,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "alias_expands_to_bare_reference_to_imported_
         end
     )";
 
-    CheckResult result = frontend.check("game/B");
+    CheckResult result = getFrontend().check("game/B");
     LUAU_REQUIRE_NO_ERRORS(result);
 }
 
@@ -1103,7 +1077,14 @@ TEST_CASE_FIXTURE(Fixture, "typeof_is_not_a_valid_alias_name")
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
 
-    CHECK("Type aliases cannot be named typeof" == toString(result.errors[0]));
+    if (FFlag::LuauSolverV2)
+    {
+        CHECK("typeof cannot be used as an identifier for a type function or alias" == toString(result.errors[0]));
+    }
+    else
+    {
+        CHECK("Type aliases cannot be named typeof" == toString(result.errors[0]));
+    }
 }
 
 TEST_CASE_FIXTURE(Fixture, "fuzzer_bug_doesnt_crash")
@@ -1128,14 +1109,16 @@ type Foo<T> = Foo<T>
 
 TEST_CASE_FIXTURE(Fixture, "recursive_type_alias_bad_pack_use_warns")
 {
-    if (!FFlag::LuauSolverV2)
-        return;
+    ScopedFastFlag sff{FFlag::LuauSolverV2, true};
 
     CheckResult result = check(R"(
 type Foo<T> = Foo<T...>
 )");
 
-    LUAU_REQUIRE_ERROR_COUNT(4, result);
+    LUAU_REQUIRE_ERROR_COUNT(5, result);
+    LUAU_CHECK_ERROR(result, GenericError);
+    CHECK_EQ(toString(result.errors[4]), "Generic type 'Foo<T>' expects 1 type argument, but none are specified");
+
     auto occursCheckFailed = get<OccursCheckFailed>(result.errors[1]);
     REQUIRE(occursCheckFailed);
 
@@ -1184,21 +1167,22 @@ TEST_CASE_FIXTURE(Fixture, "bound_type_in_alias_segfault")
 {
     ScopedFastFlag sff{FFlag::LuauSolverV2, true};
 
-    LUAU_CHECK_NO_ERRORS(check(R"(
+    CheckResult result = check(R"(
         --!nonstrict
-        type Map<T, V> = {[ K]: V}
+        type Map<T, V> = {[K]: V}
         function foo:bar(): Config<any, any> end
         type Config<TSource, TContext> = Map<TSource, TContext> & { fields: FieldConfigMap<any, any>}
-        export type FieldConfig<TSource, TContext, TArgs> = {[ string]: any}
+        export type FieldConfig<TSource, TContext, TArgs> = {[string]: any}
         export type FieldConfigMap<TSource, TContext> = Map<string, FieldConfig<TSource, TContext>>
-    )"));
+    )");
+
+    LUAU_CHECK_ERROR_COUNT(2, result);
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "gh1632_no_infinite_recursion_in_normalization")
 {
     ScopedFastFlag flags[] = {
         {FFlag::LuauSolverV2, true},
-        {FFlag::LuauFixInfiniteRecursionInNormalization, true},
     };
 
     CheckResult result = check(R"(
@@ -1225,8 +1209,6 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "gh1632_no_infinite_recursion_in_normalizatio
 
 TEST_CASE_FIXTURE(Fixture, "exported_alias_location_is_accessible_on_module")
 {
-    ScopedFastFlag sff{FFlag::LuauRetainDefinitionAliasLocations, true};
-
     CheckResult result = check(R"(
         export type Value = string
     )");
@@ -1243,7 +1225,6 @@ TEST_CASE_FIXTURE(Fixture, "exported_type_function_location_is_accessible_on_mod
 {
     ScopedFastFlag flags[] = {
         {FFlag::LuauSolverV2, true},
-        {FFlag::LuauRetainDefinitionAliasLocations, true},
     };
 
     CheckResult result = check(R"(
@@ -1259,4 +1240,111 @@ TEST_CASE_FIXTURE(Fixture, "exported_type_function_location_is_accessible_on_mod
     CHECK_EQ(tfun->second.definitionLocation, Location{{1, 8}, {2, 11}});
 }
 
+TEST_CASE_FIXTURE(Fixture, "fuzzer_cursed_type_aliases")
+{
+    // This used to crash under the new solver: we would like this to continue
+    // to not crash.
+    LUAU_REQUIRE_ERRORS(check(R"(
+        export type t1<t0...> = t4<t0...>
+        export type t4<t2, t0...> = t4<t0...>
+    )"));
+}
+
+TEST_CASE_FIXTURE(Fixture, "type_alias_dont_crash_on_bad_name")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+
+    CheckResult result = check(R"(
+        type typeof = typeof(nil :: any)
+    )");
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    CHECK(get<ReservedIdentifier>(result.errors[0]));
+}
+
+TEST_CASE_FIXTURE(Fixture, "type_alias_dont_crash_on_duplicate_with_typeof")
+{
+    // NOTE: This pattern looks quite silly, but it's pretty common in the old
+    // solver in:
+    //
+    //  type Foo = typeof(setmetatable({} :: SomeType, {} :: SomeMetatableType))
+    //
+    CheckResult result = check(R"(
+        type A = typeof(nil :: any)
+        type A = typeof(nil :: any)
+    )");
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    CHECK(get<DuplicateTypeDefinition>(result.errors[0]));
+}
+
+TEST_CASE_FIXTURE(Fixture, "fuzzer_more_cursed_aliases")
+{
+    LUAU_REQUIRE_ERRORS(check(R"(
+export type t138 = t0<t138>
+export type t0<t0,t10,t10,t109> = t0
+    )"));
+}
+
+TEST_CASE_FIXTURE(Fixture, "evaluating_generic_default_type_shouldnt_ice")
+{
+    auto result = check(R"(
+local A = {}
+type B<T = typeof(A)> = unknown
+)");
+
+    if (FFlag::LuauSolverV2)
+        LUAU_REQUIRE_NO_ERRORS(result);
+    else
+    {
+        LUAU_REQUIRE_ERROR_COUNT(1, result);
+        LUAU_CHECK_ERROR(result, UnknownSymbol);
+    }
+}
+
+TEST_CASE_FIXTURE(Fixture, "evaluating_generic_default_type_pack_shouldnt_ice")
+{
+    auto result = check(R"(
+local A = {}
+type B<T... = ...typeof(A)> = unknown
+)");
+
+    if (FFlag::LuauSolverV2)
+        LUAU_REQUIRE_NO_ERRORS(result);
+    else
+    {
+        LUAU_REQUIRE_ERROR_COUNT(1, result);
+        LUAU_CHECK_ERROR(result, UnknownSymbol);
+    }
+}
+
+TEST_CASE_FIXTURE(Fixture, "evaluating_generic_default_type_for_symbol_before_definition_is_an_error")
+{
+    ScopedFastFlag sff[] = {
+        {FFlag::LuauSolverV2, true},
+    };
+
+    auto result = check(R"(
+type B<T = typeof(A)> = unknown
+local A = {}
+)");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    LUAU_CHECK_ERROR(result, UnknownSymbol);
+}
+
+TEST_CASE_FIXTURE(Fixture, "evaluating_generic_default_type_pack_for_symbol_before_definition_is_an_error")
+{
+    ScopedFastFlag sff[] = {
+        {FFlag::LuauSolverV2, true},
+    };
+
+    auto result = check(R"(
+type B<T... = ...typeof(A)> = unknown
+local A = {}
+)");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    LUAU_CHECK_ERROR(result, UnknownSymbol);
+}
+
+
 TEST_SUITE_END();
