@@ -10,8 +10,12 @@
 using namespace Luau;
 
 LUAU_FASTFLAG(LuauSolverV2)
-LUAU_FASTFLAG(LuauTableCloneClonesType3)
-LUAU_FASTFLAG(LuauImproveTypePathsInErrors)
+LUAU_FASTFLAG(LuauTableCloneClonesType4)
+LUAU_FASTFLAG(LuauBetterTypeMismatchErrors)
+LUAU_FASTFLAG(LuauExplicitTypeInstantiationSyntax)
+LUAU_FASTFLAG(LuauExplicitTypeInstantiationSupport)
+LUAU_FASTFLAG(LuauCloneForIntersectionsUnions)
+LUAU_FASTFLAG(LuauTableFreezeCheckIsSubtype)
 
 TEST_SUITE_BEGIN("BuiltinTests");
 
@@ -108,7 +112,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "table_concat_returns_string")
     )");
 
     LUAU_REQUIRE_NO_ERRORS(result);
-    CHECK_EQ(*builtinTypes->stringType, *requireType("r"));
+    CHECK("string" == toString(requireType("r")));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "sort")
@@ -145,32 +149,32 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "sort_with_bad_predicate")
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-    const std::string expected = (FFlag::LuauImproveTypePathsInErrors) ? "Type\n\t"
-                                                                         "'(number, number) -> boolean'"
-                                                                         "\ncould not be converted into\n\t"
-                                                                         "'((string, string) -> boolean)?'"
-                                                                         "\ncaused by:\n"
-                                                                         "  None of the union options are compatible. For example:\n"
-                                                                         "Type\n\t"
-                                                                         "'(number, number) -> boolean'"
-                                                                         "\ncould not be converted into\n\t"
-                                                                         "'(string, string) -> boolean'"
-                                                                         "\ncaused by:\n"
-                                                                         "  Argument #1 type is not compatible.\n"
-                                                                         "Type 'string' could not be converted into 'number'"
-                                                                       : R"(Type
-    '(number, number) -> boolean'
-could not be converted into
-    '((string, string) -> boolean)?'
-caused by:
-  None of the union options are compatible. For example:
-Type
-    '(number, number) -> boolean'
-could not be converted into
-    '(string, string) -> boolean'
-caused by:
-  Argument #1 type is not compatible.
-Type 'string' could not be converted into 'number')";
+    const std::string expected = FFlag::LuauBetterTypeMismatchErrors ? "Expected this to be\n\t"
+                                                                       "'((string, string) -> boolean)?'"
+                                                                       "\nbut got\n\t"
+                                                                       "'(number, number) -> boolean'"
+                                                                       "\ncaused by:\n"
+                                                                       "  None of the union options are compatible. For example:\n"
+                                                                       "Expected this to be\n\t"
+                                                                       "'(string, string) -> boolean'"
+                                                                       "\nbut got\n\t"
+                                                                       "'(number, number) -> boolean'"
+                                                                       "\ncaused by:\n"
+                                                                       "  Argument #1 type is not compatible.\n"
+                                                                       "Expected this to be 'number', but got 'string'"
+                                                                     : "Type\n\t"
+                                                                       "'(number, number) -> boolean'"
+                                                                       "\ncould not be converted into\n\t"
+                                                                       "'((string, string) -> boolean)?'"
+                                                                       "\ncaused by:\n"
+                                                                       "  None of the union options are compatible. For example:\n"
+                                                                       "Type\n\t"
+                                                                       "'(number, number) -> boolean'"
+                                                                       "\ncould not be converted into\n\t"
+                                                                       "'(string, string) -> boolean'"
+                                                                       "\ncaused by:\n"
+                                                                       "  Argument #1 type is not compatible.\n"
+                                                                       "Type 'string' could not be converted into 'number'";
     CHECK_EQ(expected, toString(result.errors[0]));
 }
 
@@ -181,7 +185,7 @@ TEST_CASE_FIXTURE(Fixture, "strings_have_methods")
     )LUA");
 
     LUAU_REQUIRE_NO_ERRORS(result);
-    CHECK_EQ(*builtinTypes->stringType, *requireType("s"));
+    CHECK("string" == toString(requireType("s")));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "math_max_variatic")
@@ -191,7 +195,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "math_max_variatic")
     )");
 
     LUAU_REQUIRE_NO_ERRORS(result);
-    CHECK_EQ(*builtinTypes->numberType, *requireType("n"));
+    CHECK("number" == toString(requireType("n")));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "math_max_checks_for_numbers")
@@ -201,7 +205,10 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "math_max_checks_for_numbers")
     )");
 
     LUAU_REQUIRE_ERRORS(result);
-    CHECK_EQ("Type 'string' could not be converted into 'number'", toString(result.errors[0]));
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ("Expected this to be 'number', but got 'string'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type 'string' could not be converted into 'number'", toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "builtin_tables_sealed")
@@ -395,8 +402,10 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "setmetatable_on_union_of_tables")
     )");
 
     LUAU_REQUIRE_NO_ERRORS(result);
-
-    CHECK("{ @metatable {  }, A } | { @metatable {  }, B }" == toString(requireTypeAlias("X")));
+    if (FFlag::LuauSolverV2)
+        CHECK("{ @metatable {  }, A } | { @metatable {  }, B }" == toString(requireTypeAlias("X")));
+    else
+        CHECK("{ @metatable {|  |}, A } | { @metatable {|  |}, B }" == toString(requireTypeAlias("X")));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "table_insert_correctly_infers_type_of_array_2_args_overload")
@@ -408,7 +417,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "table_insert_correctly_infers_type_of_array_
     )");
 
     LUAU_REQUIRE_NO_ERRORS(result);
-    CHECK_EQ(builtinTypes->stringType, requireType("s"));
+    CHECK_EQ(getBuiltins()->stringType, requireType("s"));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "table_insert_correctly_infers_type_of_array_3_args_overload")
@@ -430,10 +439,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "table_pack")
     )");
 
     LUAU_REQUIRE_NO_ERRORS(result);
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ("{ [number]: boolean | number | string, n: number }", toString(requireType("t")));
-    else
-        CHECK_EQ("{| [number]: boolean | number | string, n: number |}", toString(requireType("t")));
+    CHECK_EQ("{ [number]: boolean | number | string, n: number }", toString(requireType("t")));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "table_pack_variadic")
@@ -448,10 +454,7 @@ local t = table.pack(f())
     )");
 
     LUAU_REQUIRE_NO_ERRORS(result);
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ("{ [number]: number | string, n: number }", toString(requireType("t")));
-    else
-        CHECK_EQ("{| [number]: number | string, n: number |}", toString(requireType("t")));
+    CHECK_EQ("{ [number]: number | string, n: number }", toString(requireType("t")));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "table_pack_reduce")
@@ -461,20 +464,14 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "table_pack_reduce")
     )");
 
     LUAU_REQUIRE_NO_ERRORS(result);
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ("{ [number]: boolean | number, n: number }", toString(requireType("t")));
-    else
-        CHECK_EQ("{| [number]: boolean | number, n: number |}", toString(requireType("t")));
+    CHECK_EQ("{ [number]: boolean | number, n: number }", toString(requireType("t")));
 
     result = check(R"(
         local t = table.pack("a", "b", "c")
     )");
 
     LUAU_REQUIRE_NO_ERRORS(result);
-    if (FFlag::LuauSolverV2)
-        CHECK_EQ("{ [number]: string, n: number }", toString(requireType("t")));
-    else
-        CHECK_EQ("{| [number]: string, n: number |}", toString(requireType("t")));
+    CHECK_EQ("{ [number]: string, n: number }", toString(requireType("t")));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "gcinfo")
@@ -484,7 +481,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "gcinfo")
     )");
 
     LUAU_REQUIRE_NO_ERRORS(result);
-    CHECK_EQ(*builtinTypes->numberType, *requireType("n"));
+    CHECK("number" == toString(requireType("n")));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "getfenv")
@@ -501,9 +498,9 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "os_time_takes_optional_date_table")
     )");
 
     LUAU_REQUIRE_NO_ERRORS(result);
-    CHECK_EQ(*builtinTypes->numberType, *requireType("n1"));
-    CHECK_EQ(*builtinTypes->numberType, *requireType("n2"));
-    CHECK_EQ(*builtinTypes->numberType, *requireType("n3"));
+    CHECK("number" == toString(requireType("n1")));
+    CHECK("number" == toString(requireType("n2")));
+    CHECK("number" == toString(requireType("n3")));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "thread_is_a_type")
@@ -626,8 +623,8 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "string_format_correctly_ordered_types")
     LUAU_REQUIRE_ERROR_COUNT(1, result);
     TypeMismatch* tm = get<TypeMismatch>(result.errors[0]);
     REQUIRE(tm);
-    CHECK_EQ(tm->wantedType, builtinTypes->stringType);
-    CHECK_EQ(tm->givenType, builtinTypes->numberType);
+    CHECK_EQ(tm->wantedType, getBuiltins()->stringType);
+    CHECK_EQ(tm->givenType, getBuiltins()->numberType);
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "string_format_tostring_specifier")
@@ -721,10 +718,24 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "bad_select_should_not_crash")
 
     if (FFlag::LuauSolverV2)
     {
-        // Counterintuitively, the parameter l0 is unconstrained and therefore it is valid to pass nil.
-        // The new solver therefore considers that parameter to be optional.
+        // Note, the function "_" places no constraints on its arguments.  They
+        // can therefore be nil.  They are therefore optional.  Only the
+        // select() call is invalid here.
         LUAU_REQUIRE_ERROR_COUNT(1, result);
-        CHECK("Argument count mismatch. Function expects at least 1 argument, but none are specified" == toString(result.errors[0]));
+
+        CHECK(Location{{6, 17}, {6, 23}} == result.errors.at(0).location);
+        const CountMismatch* err = get<CountMismatch>(result.errors.at(0));
+        REQUIRE(err);
+        CHECK(1 == err->expected);
+
+        // "_" returns 0 values.
+        CHECK(0 == err->actual);
+    }
+    else if (FFlag::LuauSolverV2)
+    {
+        LUAU_REQUIRE_ERROR_COUNT(2, result);
+        CHECK_EQ("Argument count mismatch. Function expects at least 1 argument, but none are specified", toString(result.errors[0]));
+        CHECK_EQ("Argument count mismatch. Function expects at least 1 argument, but none are specified", toString(result.errors[1]));
     }
     else
     {
@@ -814,8 +825,8 @@ TEST_CASE_FIXTURE(Fixture, "string_format_as_method")
 
     TypeMismatch* tm = get<TypeMismatch>(result.errors[0]);
     REQUIRE(tm);
-    CHECK_EQ(tm->wantedType, builtinTypes->stringType);
-    CHECK_EQ(tm->givenType, builtinTypes->numberType);
+    CHECK_EQ(tm->wantedType, getBuiltins()->stringType);
+    CHECK_EQ(tm->givenType, getBuiltins()->numberType);
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "string_format_trivial_arity")
@@ -848,8 +859,17 @@ TEST_CASE_FIXTURE(Fixture, "string_format_use_correct_argument2")
 
     LUAU_REQUIRE_ERROR_COUNT(2, result);
 
-    CHECK_EQ("Type 'string' could not be converted into 'number'", toString(result.errors[0]));
-    CHECK_EQ("Type 'number' could not be converted into 'string'", toString(result.errors[1]));
+
+    if (FFlag::LuauBetterTypeMismatchErrors)
+    {
+        CHECK_EQ("Expected this to be 'number', but got 'string'", toString(result.errors[0]));
+        CHECK_EQ("Expected this to be 'string', but got 'number'", toString(result.errors[1]));
+    }
+    else
+    {
+        CHECK_EQ("Type 'string' could not be converted into 'number'", toString(result.errors[0]));
+        CHECK_EQ("Type 'number' could not be converted into 'string'", toString(result.errors[1]));
+    }
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "string_format_use_correct_argument3")
@@ -906,7 +926,10 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "aliased_string_format")
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-    CHECK_EQ("Type 'string' could not be converted into 'number'", toString(result.errors[0]));
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ("Expected this to be 'number', but got 'string'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type 'string' could not be converted into 'number'", toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "string_lib_self_noself")
@@ -964,9 +987,9 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "string_format_report_all_type_errors_at_corr
         string.format("%s%d%s", 1, "hello", true)
     )");
 
-    TypeId stringType = builtinTypes->stringType;
-    TypeId numberType = builtinTypes->numberType;
-    TypeId booleanType = builtinTypes->booleanType;
+    TypeId stringType = getBuiltins()->stringType;
+    TypeId numberType = getBuiltins()->numberType;
+    TypeId booleanType = getBuiltins()->booleanType;
 
     LUAU_REQUIRE_ERROR_COUNT(6, result);
 
@@ -998,19 +1021,28 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "tonumber_returns_optional_number_type")
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
 
-    if (FFlag::LuauSolverV2 && FFlag::LuauImproveTypePathsInErrors)
-        CHECK_EQ(
-            "Type 'number?' could not be converted into 'number'; \n"
-            "this is because the 2nd component of the union is `nil`, which is not a subtype of `number`",
-            toString(result.errors[0])
-        );
-    else if (FFlag::LuauSolverV2)
-        CHECK_EQ(
-            "Type 'number?' could not be converted into 'number'; type number?[1] (nil) is not a subtype of number (number)",
-            toString(result.errors[0])
-        );
+    if (FFlag::LuauSolverV2)
+    {
+        if (FFlag::LuauBetterTypeMismatchErrors)
+            CHECK_EQ(
+                "Expected this to be 'number', but got 'number?'; \n"
+                "the 2nd component of the union is `nil`, which is not a subtype of `number`",
+                toString(result.errors[0])
+            );
+        else
+            CHECK_EQ(
+                "Type 'number?' could not be converted into 'number'; \n"
+                "this is because the 2nd component of the union is `nil`, which is not a subtype of `number`",
+                toString(result.errors[0])
+            );
+    }
     else
-        CHECK_EQ("Type 'number?' could not be converted into 'number'", toString(result.errors[0]));
+    {
+        if (FFlag::LuauBetterTypeMismatchErrors)
+            CHECK_EQ("Expected this to be 'number', but got 'number?'", toString(result.errors[0]));
+        else
+            CHECK_EQ("Type 'number?' could not be converted into 'number'", toString(result.errors[0]));
+    }
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "tonumber_returns_optional_number_type2")
@@ -1153,7 +1185,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "table_freeze_is_generic")
     if (FFlag::LuauSolverV2)
         CHECK("Key 'b' not found in table '{ read a: number }'" == toString(result.errors[0]));
     else
-        CHECK_EQ("Key 'b' not found in table '{| a: number |}'", toString(result.errors[0]));
+        CHECK_EQ("Key 'b' not found in table '{ a: number }'", toString(result.errors[0]));
     CHECK(Location({13, 18}, {13, 23}) == result.errors[0].location);
 
     if (FFlag::LuauSolverV2)
@@ -1291,6 +1323,26 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "table_clone_persistent_skip")
     LUAU_REQUIRE_NO_ERRORS(result);
 }
 
+TEST_CASE_FIXTURE(BuiltinsFixture, "table_clone_should_support_variadic_any_in_old_solver")
+{
+    ScopedFastFlag _{FFlag::LuauTableCloneClonesType4, true};
+
+    fileResolver.source["game/A"] = R"(
+        --!nonstrict
+        return function()
+            return {}
+        end
+    )";
+
+    fileResolver.source["game/B"] = R"(
+        local A = require(game.A)
+        local _ = table.clone(A())
+    )";
+
+    CheckResult result = getFrontend().check("game/B");
+    LUAU_REQUIRE_ERROR_COUNT(0, result);
+}
+
 TEST_CASE_FIXTURE(BuiltinsFixture, "set_metatable_needs_arguments")
 {
     // In the new solver, nil can certainly be used where a generic is required, so all generic parameters are optional.
@@ -1307,22 +1359,50 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "set_metatable_needs_arguments")
     CHECK_EQ(toString(result.errors[1]), "Argument count mismatch. Function 'a.b' expects 2 arguments, but only 1 is specified");
 }
 
+TEST_CASE_FIXTURE(BuiltinsFixture, "table_clone_intersection_of_tables")
+{
+    CheckResult result = check(R"(
+        type FIRST = {
+            some: string,
+        }
+
+        type SECOND = FIRST & {
+            thing: string,
+        }
+
+        local b: SECOND
+        -- c's type used to be FIRST, but should be the full type of SECOND
+        local c = table.clone(b)
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+
+    if (FFlag::LuauSolverV2 || FFlag::LuauCloneForIntersectionsUnions)
+    {
+        CHECK_EQ("{ some: string } & { thing: string }", toString(requireType("c"), {true}));
+        CHECK_EQ("FIRST & { thing: string }", toString(requireType("c")));
+    }
+}
+
 TEST_CASE_FIXTURE(Fixture, "typeof_unresolved_function")
 {
     CheckResult result = check(R"(
         local function f(a: typeof(f)) end
         )");
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-    CHECK_EQ("Unknown global 'f'", toString(result.errors[0]));
+    CHECK_EQ("Unknown global 'f'; consider assigning to it first", toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "no_persistent_typelevel_change")
 {
-    TypeId mathTy = requireType(frontend.globals.globalScope, "math");
+    TypeId mathTy = requireType(getFrontend().globals.globalScope, "math");
     REQUIRE(mathTy);
     TableType* ttv = getMutable<TableType>(mathTy);
     REQUIRE(ttv);
-    const FunctionType* ftv = get<FunctionType>(ttv->props["frexp"].type());
+
+    REQUIRE(ttv->props["frexp"].readTy);
+    const FunctionType* ftv = get<FunctionType>(*ttv->props["frexp"].readTy);
+
     REQUIRE(ftv);
     auto original = ftv->level;
 
@@ -1606,6 +1686,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "string_find_should_not_crash")
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "table_dot_clone_type_states")
 {
+    ScopedFastFlag sff{FFlag::LuauTableCloneClonesType4, true};
     CheckResult result = check(R"(
         local t1 = {}
         t1.x = 5
@@ -1616,15 +1697,15 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "table_dot_clone_type_states")
 
     LUAU_REQUIRE_NO_ERRORS(result);
 
-    if (FFlag::LuauTableCloneClonesType3)
+    if (FFlag::LuauSolverV2)
     {
         CHECK_EQ(toString(requireType("t1"), {true}), "{ x: number, z: number }");
         CHECK_EQ(toString(requireType("t2"), {true}), "{ x: number, y: number }");
     }
     else
     {
-        CHECK_EQ(toString(requireType("t1"), {true}), "{ x: number, y: number, z: number }");
-        CHECK_EQ(toString(requireType("t2"), {true}), "{ x: number, y: number, z: number }");
+        CHECK_EQ(toString(requireType("t1"), {true}), "{| x: number, z: number |}");
+        CHECK_EQ(toString(requireType("t2"), {true}), "{| x: number, y: number |}");
     }
 }
 
@@ -1674,4 +1755,340 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "string_format_should_support_any")
     LUAU_REQUIRE_NO_ERRORS(result);
 }
 
+TEST_CASE_FIXTURE(BuiltinsFixture, "string_format_should_support_any_2")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+
+    CheckResult result = check(R"(
+        local fmt = "Hello, %s!" :: any
+        local x = "world" :: any
+        print(string.format(fmt, x))
+        print(string.format(fmt, "hello"))
+        print(string.format(fmt, 5)) -- unchecked because the format string is `any`!
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "string_format_should_support_singleton_types")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+
+    CheckResult result = check(R"(
+        local fmt: "Hello, %s!" = "Hello, %s!"
+        print(string.format(fmt, "hello"))
+        print(string.format(fmt, 5)) -- should still produce an error since the expected type is `string`!
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    TypeMismatch* tm = get<TypeMismatch>(result.errors[0]);
+    REQUIRE(tm);
+    CHECK_EQ(tm->wantedType, getBuiltins()->stringType);
+    CHECK_EQ(tm->givenType, getBuiltins()->numberType);
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "better_string_format_error_when_format_string_is_dynamic")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+
+    CheckResult result = check(R"(
+        local fmt: string = "Hello, %s!"
+        print(string.format(fmt, "hello"))
+        print(string.format(fmt :: any, "hello")) -- no error
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    CHECK_EQ(
+        "We cannot statically check the type of `string.format` when called with a format string that is not statically known.\n"
+        "If you'd like to use an unchecked `string.format` call, you can cast the format string to `any` using `:: any`.",
+        toString(result.errors[0])
+    );
+}
+
+TEST_CASE_FIXTURE(Fixture, "write_only_table_assertion")
+{
+    ScopedFastFlag sff{FFlag::LuauSolverV2, true};
+
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local function accept(t: { write foo: number })
+        end
+
+        accept({ foo = "lol", foo = true })
+    )"));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "table_insert_into_any")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+table.insert(1::any, 2::any)
+    )"));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "table_insert_requires_all_fields")
+{
+    CheckResult result = check(R"(
+        local function huh(): { { x: number, y: string } }
+            local ret = {}
+            while true do
+                table.insert(ret, { x = 42 })
+            end
+            return ret
+        end
+    )");
+
+    LUAU_REQUIRE_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "read_refinements_on_persistent_tables_known_property_identity")
+{
+    // This will not result in a real refinement, as we refine `bnot`, a function, to be truthy
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        if bit32.bnot then
+        end
+    )"));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "read_refinements_on_persistent_tables_unknown_property")
+{
+    CheckResult results = check(R"(
+        if bit32.scrambleEggs then
+        end
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, results);
+    auto err = get<UnknownProperty>(results.errors[0]);
+    CHECK_EQ(err->key, "scrambleEggs");
+    CHECK_EQ(toString(err->table), "typeof(bit32)");
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "read_refinements_on_persistent_tables_known_property_narrow")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local myutf8 = utf8
+        if myutf8.charpattern == "lol" then
+            local foobar = myutf8.charpattern
+            local _ = foobar
+        end
+    )"));
+    CHECK_EQ("\"lol\"", toString(requireTypeAtPosition(Position{4, 23})));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "next_with_refined_any")
+{
+    ScopedFastFlag lsv2{FFlag::LuauSolverV2, true};
+
+    CheckResult result = check(R"(
+        --!strict
+        local t: any = {"hello", "world"}
+        if type(t) == "table" and next(t) then
+            local foo, bar = next(t)
+            local _ = foo
+            local _ = bar
+        end
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+
+    CHECK_EQ(toString(requireTypeAtPosition(Position{5, 23})), "unknown?");
+    CHECK_EQ(toString(requireTypeAtPosition(Position{6, 23})), "unknown");
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "pairs_with_refined_any")
+{
+    ScopedFastFlag lsv2{FFlag::LuauSolverV2, true};
+
+    CheckResult result = check(R"(
+        --!strict
+        local t: any = {"hello", "world"}
+        if type(t) == "table" and pairs(t) then
+	        local foo, bar, lorem = pairs(t)
+            local _ = foo
+            local _ = bar
+            local _ = lorem
+        end
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+
+    CHECK_EQ(toString(requireTypeAtPosition(Position{5, 23})), "({+ [unknown]: unknown +}, unknown?) -> (unknown?, unknown)");
+    CHECK_EQ(toString(requireTypeAtPosition(Position{6, 23})), "{+ [unknown]: unknown +}");
+    CHECK_EQ(toString(requireTypeAtPosition(Position{7, 23})), "nil");
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "vector_lerp_should_not_crash")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local function half(x: number, y: number, z: number): vector
+            return vector.lerp(vector.zero, vector.create(x, y, z), 0.5)
+        end
+    )"));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "instantiation_works_on_builtins")
+{
+    ScopedFastFlag sffs[] = {
+        {FFlag::LuauExplicitTypeInstantiationSyntax, true},
+        {FFlag::LuauExplicitTypeInstantiationSupport, true},
+    };
+
+    CheckResult result = check(R"(
+        local foo = table.create<<string>>(4)
+        local bar = table.unpack<<string>>({})
+        local baz = table.find<<string>>({}, 1) -- should error
+        assert<<string>>("asd", "lol")
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ("Expected this to be 'string', but got 'number'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type 'number' could not be converted into 'string'", toString(result.errors[0]));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "table_freeze_on_any_should_not_error")
+{
+    ScopedFastFlag sffs[] = {{FFlag::LuauSolverV2, true}, {FFlag::LuauTableFreezeCheckIsSubtype, true}};
+
+    CheckResult result = check(R"(
+        local function foo(): any
+            return true
+        end
+
+        table.freeze(foo())
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "table_freeze_with_type_check_should_not_error")
+{
+    ScopedFastFlag sffs[] = {{FFlag::LuauSolverV2, true}, {FFlag::LuauTableFreezeCheckIsSubtype, true}};
+
+    CheckResult result = check(R"(
+        local function maybeFreeze(t: any)
+            if type(t) == "table" then
+                table.freeze(t)
+            end
+        end
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "table_freeze_no_args")
+{
+    ScopedFastFlag sffs[] = {{FFlag::LuauSolverV2, true}, {FFlag::LuauTableFreezeCheckIsSubtype, true}};
+
+    CheckResult result = check(R"(
+        table.freeze()
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    CHECK_EQ("Argument count mismatch. Function 'table.freeze' expects 1 argument, but none are specified", toString(result.errors[0]));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "table_freeze_with_type_pack_should_error")
+{
+    ScopedFastFlag sffs[] = {{FFlag::LuauSolverV2, true}, {FFlag::LuauTableFreezeCheckIsSubtype, true}};
+
+    CheckResult result = check(R"(
+        table.freeze({x = 5}, {y = "hello"})
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    CHECK_EQ("Argument count mismatch. Function 'table.freeze' expects 1 argument, but 2 are specified", toString(result.errors[0]));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "table_freeze_with_variadic_any_should_not_error")
+{
+    ScopedFastFlag sffs[] = {{FFlag::LuauSolverV2, true}, {FFlag::LuauTableFreezeCheckIsSubtype, true}};
+
+    CheckResult result = check(R"(
+        local function bar(): ...any
+            return true
+        end
+
+        table.freeze(bar())
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "table_freeze_with_variadic_non_error_suppressing_should_error")
+{
+    ScopedFastFlag sffs[] = {{FFlag::LuauSolverV2, true}, {FFlag::LuauTableFreezeCheckIsSubtype, true}};
+
+    CheckResult result = check(R"(
+        local function bar(): ...string
+            return
+        end
+
+        table.freeze(bar())
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    // TODO (CLI-185019): We probably want a count mismatch error here instead.
+    if (FFlag::LuauBetterTypeMismatchErrors)
+    {
+        CHECK_EQ("Expected this to be 'table', but got 'string'", toString(result.errors[0]));
+    }
+    else
+    {
+        CHECK_EQ("Type 'string' could not be converted into 'table'", toString(result.errors[0]));
+    }
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "variadic_return_to_single_parameter_function")
+{
+    CheckResult result = check(R"(
+        local function bar(): ...string
+            return
+        end
+
+        local function foo(x: string)
+            print(x) -- nil
+        end
+
+        foo(bar())
+    )");
+
+    // TODO (CLI-185019): We want to error here in some way, maybe a count mismatch?
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "table_freeze_generic_pack")
+{
+    ScopedFastFlag sffs[] = {{FFlag::LuauSolverV2, true}, {FFlag::LuauTableFreezeCheckIsSubtype, true}};
+
+    CheckResult result = check(R"(
+        local function foo<T...>(...: T...)
+            table.freeze(...)
+        end
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+
+    TypePackMismatch* tpm = get<TypePackMismatch>(result.errors[0]);
+    REQUIRE(tpm);
+    CHECK_EQ(toString(tpm->wantedTp), "table");
+    CHECK_EQ(toString(tpm->givenTp), "T...");
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "table_freeze_function")
+{
+    ScopedFastFlag sffs[] = {{FFlag::LuauSolverV2, true}, {FFlag::LuauTableFreezeCheckIsSubtype, true}};
+
+    CheckResult result = check(R"(
+        local function foo(f: () -> ())
+            table.freeze(f())
+        end
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    CHECK_EQ("Argument count mismatch. Function 'table.freeze' expects 1 argument, but none are specified", toString(result.errors[0]));
+}
+
 TEST_SUITE_END();
