@@ -4,6 +4,8 @@
 
 using namespace Luau;
 
+LUAU_FASTFLAG(LuauBetterTypeMismatchErrors)
+
 TEST_SUITE_BEGIN("ControlFlowAnalysis");
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "if_not_x_return")
@@ -811,7 +813,10 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "prototyping_and_visiting_alias_has_the_same_
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
 
-    CHECK_EQ("Type 'nil' could not be converted into 'number'", toString(result.errors[0]));
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ("Expected this to be 'number', but got 'nil'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type 'nil' could not be converted into 'number'", toString(result.errors[0]));
 
     CHECK_EQ("nil", toString(requireTypeAtPosition({8, 29})));
 }
@@ -834,7 +839,10 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "prototyping_and_visiting_alias_has_the_same_
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
 
-    CHECK_EQ("Type 'nil' could not be converted into 'number'", toString(result.errors[0]));
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ("Expected this to be 'number', but got 'nil'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type 'nil' could not be converted into 'number'", toString(result.errors[0]));
 
     CHECK_EQ("nil", toString(requireTypeAtPosition({9, 43})));
 }
@@ -857,7 +865,10 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "prototyping_and_visiting_alias_has_the_same_
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
 
-    CHECK_EQ("Type 'nil' could not be converted into 'number'", toString(result.errors[0]));
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ("Expected this to be 'number', but got 'nil'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type 'nil' could not be converted into 'number'", toString(result.errors[0]));
 
     CHECK_EQ("nil", toString(requireTypeAtPosition({9, 43})));
 }
