@@ -9,8 +9,8 @@
 
 using namespace Luau;
 
+
 LUAU_FASTINT(LuauTypeInferRecursionLimit)
-LUAU_FASTFLAG(LuauDontForgetToReduceUnionFunc)
 
 TEST_SUITE_BEGIN("DefinitionTests");
 
@@ -22,13 +22,13 @@ TEST_CASE_FIXTURE(Fixture, "definition_file_simple")
         declare foo2: typeof(foo)
     )");
 
-    TypeId globalFooTy = getGlobalBinding(frontend.globals, "foo");
+    TypeId globalFooTy = getGlobalBinding(getFrontend().globals, "foo");
     CHECK_EQ(toString(globalFooTy), "number");
 
-    TypeId globalBarTy = getGlobalBinding(frontend.globals, "bar");
+    TypeId globalBarTy = getGlobalBinding(getFrontend().globals, "bar");
     CHECK_EQ(toString(globalBarTy), "(number) -> string");
 
-    TypeId globalFoo2Ty = getGlobalBinding(frontend.globals, "foo2");
+    TypeId globalFoo2Ty = getGlobalBinding(getFrontend().globals, "foo2");
     CHECK_EQ(toString(globalFoo2Ty), "number");
 
     CheckResult result = check(R"(
@@ -51,20 +51,20 @@ TEST_CASE_FIXTURE(Fixture, "definition_file_loading")
         declare function var(...: any): string
     )");
 
-    TypeId globalFooTy = getGlobalBinding(frontend.globals, "foo");
+    TypeId globalFooTy = getGlobalBinding(getFrontend().globals, "foo");
     CHECK_EQ(toString(globalFooTy), "number");
 
-    std::optional<TypeFun> globalAsdfTy = frontend.globals.globalScope->lookupType("Asdf");
+    std::optional<TypeFun> globalAsdfTy = getFrontend().globals.globalScope->lookupType("Asdf");
     REQUIRE(bool(globalAsdfTy));
     CHECK_EQ(toString(globalAsdfTy->type), "number | string");
 
-    TypeId globalBarTy = getGlobalBinding(frontend.globals, "bar");
+    TypeId globalBarTy = getGlobalBinding(getFrontend().globals, "bar");
     CHECK_EQ(toString(globalBarTy), "(number) -> string");
 
-    TypeId globalFoo2Ty = getGlobalBinding(frontend.globals, "foo2");
+    TypeId globalFoo2Ty = getGlobalBinding(getFrontend().globals, "foo2");
     CHECK_EQ(toString(globalFoo2Ty), "number");
 
-    TypeId globalVarTy = getGlobalBinding(frontend.globals, "var");
+    TypeId globalVarTy = getGlobalBinding(getFrontend().globals, "var");
 
     CHECK_EQ(toString(globalVarTy), "(...any) -> string");
 
@@ -80,25 +80,25 @@ TEST_CASE_FIXTURE(Fixture, "definition_file_loading")
 
 TEST_CASE_FIXTURE(Fixture, "load_definition_file_errors_do_not_pollute_global_scope")
 {
-    unfreeze(frontend.globals.globalTypes);
-    LoadDefinitionFileResult parseFailResult = frontend.loadDefinitionFile(
-        frontend.globals,
-        frontend.globals.globalScope,
+    unfreeze(getFrontend().globals.globalTypes);
+    LoadDefinitionFileResult parseFailResult = getFrontend().loadDefinitionFile(
+        getFrontend().globals,
+        getFrontend().globals.globalScope,
         R"(
         declare foo
     )",
         "@test",
         /* captureComments */ false
     );
-    freeze(frontend.globals.globalTypes);
+    freeze(getFrontend().globals.globalTypes);
 
     REQUIRE(!parseFailResult.success);
-    std::optional<Binding> fooTy = tryGetGlobalBinding(frontend.globals, "foo");
+    std::optional<Binding> fooTy = tryGetGlobalBinding(getFrontend().globals, "foo");
     CHECK(!fooTy.has_value());
 
-    LoadDefinitionFileResult checkFailResult = frontend.loadDefinitionFile(
-        frontend.globals,
-        frontend.globals.globalScope,
+    LoadDefinitionFileResult checkFailResult = getFrontend().loadDefinitionFile(
+        getFrontend().globals,
+        getFrontend().globals.globalScope,
         R"(
         local foo: string = 123
         declare bar: typeof(foo)
@@ -108,11 +108,11 @@ TEST_CASE_FIXTURE(Fixture, "load_definition_file_errors_do_not_pollute_global_sc
     );
 
     REQUIRE(!checkFailResult.success);
-    std::optional<Binding> barTy = tryGetGlobalBinding(frontend.globals, "bar");
+    std::optional<Binding> barTy = tryGetGlobalBinding(getFrontend().globals, "bar");
     CHECK(!barTy.has_value());
 }
 
-TEST_CASE_FIXTURE(Fixture, "definition_file_classes")
+TEST_CASE_FIXTURE(Fixture, "definition_file_extern_types")
 {
     loadDefinition(R"(
         declare class Foo
@@ -152,10 +152,10 @@ TEST_CASE_FIXTURE(Fixture, "definition_file_classes")
 
 TEST_CASE_FIXTURE(Fixture, "class_definitions_cannot_overload_non_function")
 {
-    unfreeze(frontend.globals.globalTypes);
-    LoadDefinitionFileResult result = frontend.loadDefinitionFile(
-        frontend.globals,
-        frontend.globals.globalScope,
+    unfreeze(getFrontend().globals.globalTypes);
+    LoadDefinitionFileResult result = getFrontend().loadDefinitionFile(
+        getFrontend().globals,
+        getFrontend().globals.globalScope,
         R"(
         declare class A
             X: number
@@ -165,23 +165,37 @@ TEST_CASE_FIXTURE(Fixture, "class_definitions_cannot_overload_non_function")
         "@test",
         /* captureComments */ false
     );
-    freeze(frontend.globals.globalTypes);
+    freeze(getFrontend().globals.globalTypes);
 
     REQUIRE(!result.success);
     CHECK_EQ(result.parseResult.errors.size(), 0);
     REQUIRE(bool(result.module));
-    REQUIRE_EQ(result.module->errors.size(), 1);
+    if (FFlag::LuauSolverV2)
+        REQUIRE_EQ(result.module->errors.size(), 2);
+    else
+        REQUIRE_EQ(result.module->errors.size(), 1);
+
     GenericError* ge = get<GenericError>(result.module->errors[0]);
     REQUIRE(ge);
-    CHECK_EQ("Cannot overload non-function class member 'X'", ge->message);
+    if (FFlag::LuauSolverV2)
+        CHECK_EQ("Cannot overload read type of non-function class member 'X'", ge->message);
+    else
+        CHECK_EQ("Cannot overload non-function class member 'X'", ge->message);
+
+    if (FFlag::LuauSolverV2)
+    {
+        GenericError* ge2 = get<GenericError>(result.module->errors[1]);
+        REQUIRE(ge2);
+        CHECK_EQ("Cannot overload write type of non-function class member 'X'", ge2->message);
+    }
 }
 
 TEST_CASE_FIXTURE(Fixture, "class_definitions_cannot_extend_non_class")
 {
-    unfreeze(frontend.globals.globalTypes);
-    LoadDefinitionFileResult result = frontend.loadDefinitionFile(
-        frontend.globals,
-        frontend.globals.globalScope,
+    unfreeze(getFrontend().globals.globalTypes);
+    LoadDefinitionFileResult result = getFrontend().loadDefinitionFile(
+        getFrontend().globals,
+        getFrontend().globals.globalScope,
         R"(
         type NotAClass = {}
 
@@ -191,7 +205,7 @@ TEST_CASE_FIXTURE(Fixture, "class_definitions_cannot_extend_non_class")
         "@test",
         /* captureComments */ false
     );
-    freeze(frontend.globals.globalTypes);
+    freeze(getFrontend().globals.globalTypes);
 
     REQUIRE(!result.success);
     CHECK_EQ(result.parseResult.errors.size(), 0);
@@ -202,12 +216,12 @@ TEST_CASE_FIXTURE(Fixture, "class_definitions_cannot_extend_non_class")
     CHECK_EQ("Cannot use non-class type 'NotAClass' as a superclass of class 'Foo'", ge->message);
 }
 
-TEST_CASE_FIXTURE(Fixture, "no_cyclic_defined_classes")
+TEST_CASE_FIXTURE(Fixture, "no_cyclic_defined_extern_types")
 {
-    unfreeze(frontend.globals.globalTypes);
-    LoadDefinitionFileResult result = frontend.loadDefinitionFile(
-        frontend.globals,
-        frontend.globals.globalScope,
+    unfreeze(getFrontend().globals.globalTypes);
+    LoadDefinitionFileResult result = getFrontend().loadDefinitionFile(
+        getFrontend().globals,
+        getFrontend().globals.globalScope,
         R"(
         declare class Foo extends Bar
         end
@@ -218,7 +232,7 @@ TEST_CASE_FIXTURE(Fixture, "no_cyclic_defined_classes")
         "@test",
         /* captureComments */ false
     );
-    freeze(frontend.globals.globalTypes);
+    freeze(getFrontend().globals.globalTypes);
 
     REQUIRE(!result.success);
 }
@@ -317,25 +331,25 @@ TEST_CASE_FIXTURE(Fixture, "definitions_documentation_symbols")
         }
     )");
 
-    std::optional<Binding> xBinding = frontend.globals.globalScope->linearSearchForBinding("x");
+    std::optional<Binding> xBinding = getFrontend().globals.globalScope->linearSearchForBinding("x");
     REQUIRE(bool(xBinding));
     // note: loadDefinition uses the @test package name.
     CHECK_EQ(xBinding->documentationSymbol, "@test/global/x");
 
-    std::optional<TypeFun> fooTy = frontend.globals.globalScope->lookupType("Foo");
+    std::optional<TypeFun> fooTy = getFrontend().globals.globalScope->lookupType("Foo");
     REQUIRE(bool(fooTy));
     CHECK_EQ(fooTy->type->documentationSymbol, "@test/globaltype/Foo");
 
-    std::optional<TypeFun> barTy = frontend.globals.globalScope->lookupType("Bar");
+    std::optional<TypeFun> barTy = getFrontend().globals.globalScope->lookupType("Bar");
     REQUIRE(bool(barTy));
     CHECK_EQ(barTy->type->documentationSymbol, "@test/globaltype/Bar");
 
-    ClassType* barClass = getMutable<ClassType>(barTy->type);
+    ExternType* barClass = getMutable<ExternType>(barTy->type);
     REQUIRE(bool(barClass));
     REQUIRE_EQ(barClass->props.count("prop"), 1);
     CHECK_EQ(barClass->props["prop"].documentationSymbol, "@test/globaltype/Bar.prop");
 
-    std::optional<Binding> yBinding = frontend.globals.globalScope->linearSearchForBinding("y");
+    std::optional<Binding> yBinding = getFrontend().globals.globalScope->linearSearchForBinding("y");
     REQUIRE(bool(yBinding));
     CHECK_EQ(yBinding->documentationSymbol, "@test/global/y");
 
@@ -355,18 +369,19 @@ TEST_CASE_FIXTURE(Fixture, "definitions_symbols_are_generated_for_recursively_re
         declare function myFunc(): MyClass
     )");
 
-    std::optional<TypeFun> myClassTy = frontend.globals.globalScope->lookupType("MyClass");
+    std::optional<TypeFun> myClassTy = getFrontend().globals.globalScope->lookupType("MyClass");
     REQUIRE(bool(myClassTy));
     CHECK_EQ(myClassTy->type->documentationSymbol, "@test/globaltype/MyClass");
 
-    ClassType* cls = getMutable<ClassType>(myClassTy->type);
+    ExternType* cls = getMutable<ExternType>(myClassTy->type);
     REQUIRE(bool(cls));
     REQUIRE_EQ(cls->props.count("myMethod"), 1);
 
     const auto& method = cls->props["myMethod"];
     CHECK_EQ(method.documentationSymbol, "@test/globaltype/MyClass.myMethod");
 
-    FunctionType* function = getMutable<FunctionType>(method.type());
+    REQUIRE(method.readTy);
+    FunctionType* function = getMutable<FunctionType>(*method.readTy);
     REQUIRE(function);
 
     REQUIRE(function->definition.has_value());
@@ -382,7 +397,7 @@ TEST_CASE_FIXTURE(Fixture, "documentation_symbols_dont_attach_to_persistent_type
         export type Evil = string
     )");
 
-    std::optional<TypeFun> ty = frontend.globals.globalScope->lookupType("Evil");
+    std::optional<TypeFun> ty = getFrontend().globals.globalScope->lookupType("Evil");
     REQUIRE(bool(ty));
     CHECK_EQ(ty->type->documentationSymbol, std::nullopt);
 }
@@ -448,10 +463,10 @@ TEST_CASE_FIXTURE(Fixture, "class_definition_string_props")
 
 TEST_CASE_FIXTURE(Fixture, "class_definition_malformed_string")
 {
-    unfreeze(frontend.globals.globalTypes);
-    LoadDefinitionFileResult result = frontend.loadDefinitionFile(
-        frontend.globals,
-        frontend.globals.globalScope,
+    unfreeze(getFrontend().globals.globalTypes);
+    LoadDefinitionFileResult result = getFrontend().loadDefinitionFile(
+        getFrontend().globals,
+        getFrontend().globals.globalScope,
         R"(
         declare class Foo
             ["a\0property"]: string
@@ -460,7 +475,7 @@ TEST_CASE_FIXTURE(Fixture, "class_definition_malformed_string")
         "@test",
         /* captureComments */ false
     );
-    freeze(frontend.globals.globalTypes);
+    freeze(getFrontend().globals.globalTypes);
 
     REQUIRE(!result.success);
     REQUIRE_EQ(result.parseResult.errors.size(), 1);
@@ -482,18 +497,18 @@ TEST_CASE_FIXTURE(Fixture, "class_definition_indexer")
 
     LUAU_REQUIRE_NO_ERRORS(result);
 
-    const ClassType* ctv = get<ClassType>(requireType("x"));
-    REQUIRE(ctv != nullptr);
+    const ExternType* etv = get<ExternType>(requireType("x"));
+    REQUIRE(etv != nullptr);
 
-    REQUIRE(bool(ctv->indexer));
+    REQUIRE(bool(etv->indexer));
 
-    CHECK_EQ(*ctv->indexer->indexType, *builtinTypes->numberType);
-    CHECK_EQ(*ctv->indexer->indexResultType, *builtinTypes->stringType);
+    CHECK("number" == toString(etv->indexer->indexType));
+    CHECK("string" == toString(etv->indexer->indexResultType));
 
     CHECK_EQ(toString(requireType("y")), "string");
 }
 
-TEST_CASE_FIXTURE(Fixture, "class_definitions_reference_other_classes")
+TEST_CASE_FIXTURE(Fixture, "class_definitions_reference_other_extern_types")
 {
     loadDefinition(R"(
         declare class Channel
@@ -532,13 +547,13 @@ TEST_CASE_FIXTURE(Fixture, "definition_file_has_source_module_name_set")
     CHECK_EQ(result.sourceModule.name, "@test");
     CHECK_EQ(result.sourceModule.humanReadableName, "@test");
 
-    std::optional<TypeFun> fooTy = frontend.globals.globalScope->lookupType("Foo");
+    std::optional<TypeFun> fooTy = getFrontend().globals.globalScope->lookupType("Foo");
     REQUIRE(fooTy);
 
-    const ClassType* ctv = get<ClassType>(fooTy->type);
+    const ExternType* etv = get<ExternType>(fooTy->type);
 
-    REQUIRE(ctv);
-    CHECK_EQ(ctv->definitionModuleName, "@test");
+    REQUIRE(etv);
+    CHECK_EQ(etv->definitionModuleName, "@test");
 }
 
 TEST_CASE_FIXTURE(Fixture, "recursive_redefinition_reduces_rightfully")
@@ -556,7 +571,9 @@ TEST_CASE_FIXTURE(Fixture, "recursive_redefinition_reduces_rightfully")
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "cli_142285_reduce_minted_union_func")
 {
-    ScopedFastFlag sffs[] = {{FFlag::LuauSolverV2, true}, {FFlag::LuauDontForgetToReduceUnionFunc, true}};
+    ScopedFastFlag sff[] = {
+        {FFlag::LuauSolverV2, true},
+    };
 
     CheckResult result = check(R"(
         local function middle(a: number, b: number): number
@@ -576,11 +593,11 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "cli_142285_reduce_minted_union_func")
         return nil
         end
     )");
-    LUAU_REQUIRE_ERROR_COUNT(3, result);
-    // There are three errors in the above snippet, but they should all be where
-    // clause needed errors.
-    for (const auto& e : result.errors)
-        CHECK(get<WhereClauseNeeded>(e));
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    auto err1 = get<CannotInferBinaryOperation>(result.errors[0]);
+    REQUIRE(err1);
+    CHECK_EQ(err1->suggestedToAnnotate, "item");
+    CHECK_EQ(err1->op, AstExprBinary::Op::CompareLe);
 }
 
 TEST_CASE_FIXTURE(Fixture, "vector3_overflow")
