@@ -2,6 +2,7 @@
 
 #include "Luau/AstQuery.h"
 #include "Luau/BuiltinDefinitions.h"
+#include "Luau/Common.h"
 #include "Luau/Frontend.h"
 #include "Luau/Scope.h"
 #include "Luau/TypeInfer.h"
@@ -15,7 +16,10 @@
 using namespace Luau;
 
 LUAU_FASTFLAG(LuauSolverV2)
-LUAU_FASTFLAG(LuauStatForInFix)
+LUAU_FASTFLAG(LuauCheckForInWithSubtyping3)
+LUAU_FASTFLAG(LuauInstantiationUsesGenericPolarity2)
+LUAU_FASTFLAG(LuauBetterTypeMismatchErrors)
+LUAU_FASTFLAG(LuauPropagateTypeAnnotationsInForInLoops)
 
 TEST_SUITE_BEGIN("TypeInferLoops");
 
@@ -37,7 +41,7 @@ TEST_CASE_FIXTURE(Fixture, "for_loop")
         CHECK("number?" == toString(requireType("q")));
     }
     else
-        CHECK_EQ(*builtinTypes->numberType, *requireType("q"));
+        CHECK("number" == toString(requireType("q")));
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "iteration_no_table_passed")
@@ -133,6 +137,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_loop")
         for i, v in pairs({ "foo" }) do
             n = i
             s = v
+            print(i, v)
         end
     )");
 
@@ -142,32 +147,69 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_loop")
     {
         CHECK("number?" == toString(requireType("n")));
         CHECK("string?" == toString(requireType("s")));
+        CHECK_EQ("number", toString(requireTypeAtPosition({6, 18})));
+        CHECK_EQ("string", toString(requireTypeAtPosition({6, 21})));
     }
     else
     {
-        CHECK_EQ(*builtinTypes->numberType, *requireType("n"));
-        CHECK_EQ(*builtinTypes->stringType, *requireType("s"));
+        CHECK("number" == toString(requireType("n")));
+        CHECK("string" == toString(requireType("s")));
     }
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_loop_with_next")
 {
-    // CLI-116494 The generics K and V are leaking out of the next() function somehow.
-    DOES_NOT_PASS_NEW_SOLVER_GUARD();
+    CheckResult result = check(R"(
+        local n
+        local s
+        for i, v in next, { "foo" } do
+            n = i
+            s = v
+            print(i, v)
+        end
+    )");
 
+    LUAU_REQUIRE_NO_ERRORS(result);
+
+    if (FFlag::LuauSolverV2)
+    {
+        CHECK("number?" == toString(requireType("n")));
+        CHECK("string?" == toString(requireType("s")));
+        CHECK_EQ("number", toString(requireTypeAtPosition({6, 18})));
+        CHECK_EQ("string", toString(requireTypeAtPosition({6, 21})));
+    }
+    else
+    {
+        CHECK("number" == toString(requireType("n")));
+        CHECK("string" == toString(requireType("s")));
+    }
+}
+TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_loop_with_next_and_multiple_elements")
+{
     CheckResult result = check(R"(
         local n
         local s
         for i, v in next, { "foo", "bar" } do
             n = i
             s = v
+            print(i, v)
         end
     )");
 
     LUAU_REQUIRE_NO_ERRORS(result);
 
-    CHECK_EQ(*builtinTypes->numberType, *requireType("n"));
-    CHECK_EQ(*builtinTypes->stringType, *requireType("s"));
+    if (FFlag::LuauSolverV2)
+    {
+        CHECK("number?" == toString(requireType("n")));
+        CHECK("string?" == toString(requireType("s")));
+        CHECK_EQ("number", toString(requireTypeAtPosition({6, 18})));
+        CHECK_EQ("string", toString(requireTypeAtPosition({6, 21})));
+    }
+    else
+    {
+        CHECK("number" == toString(requireType("n")));
+        CHECK("string" == toString(requireType("s")));
+    }
 }
 
 TEST_CASE_FIXTURE(Fixture, "for_in_with_an_iterator_of_type_any")
@@ -224,11 +266,16 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_loop_with_zero_iterators_dcr")
         for key in no_iter() do end
     )");
 
-    LUAU_REQUIRE_ERROR_COUNT(2, result);
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    auto err = get<GenericError>(result.errors[0]);
+    REQUIRE(err);
+    CHECK_EQ("for..in loops require at least one value to iterate over.  Got zero", err->message);
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_with_a_custom_iterator_should_type_check")
 {
+    ScopedFastFlag _{FFlag::LuauPropagateTypeAnnotationsInForInLoops, true};
+
     CheckResult result = check(R"(
         local function range(l, h): () -> number
             return function()
@@ -241,7 +288,10 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_with_a_custom_iterator_should_type_ch
         end
     )");
 
-    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    if (FFlag::LuauPropagateTypeAnnotationsInForInLoops)
+        LUAU_REQUIRE_ERROR_COUNT(1, result);
+    else
+        LUAU_REQUIRE_NO_ERRORS(result);
 }
 
 TEST_CASE_FIXTURE(Fixture, "for_in_loop_on_error")
@@ -270,9 +320,6 @@ TEST_CASE_FIXTURE(Fixture, "for_in_loop_on_error")
 
 TEST_CASE_FIXTURE(Fixture, "for_in_loop_on_non_function")
 {
-    // We report a spuriouus duplicate error here.
-    DOES_NOT_PASS_NEW_SOLVER_GUARD();
-
     CheckResult result = check(R"(
         local bad_iter = 5
 
@@ -333,8 +380,8 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_loop_error_on_factory_not_returning_t
 
     TypeMismatch* tm = get<TypeMismatch>(result.errors[1]);
     REQUIRE(tm);
-    CHECK_EQ(builtinTypes->numberType, tm->wantedType);
-    CHECK_EQ(builtinTypes->stringType, tm->givenType);
+    CHECK_EQ(getBuiltins()->numberType, tm->wantedType);
+    CHECK_EQ(getBuiltins()->stringType, tm->givenType);
 }
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_loop_error_on_iterator_requiring_args_but_none_given")
@@ -361,6 +408,11 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_loop_error_on_iterator_requiring_args
 
 TEST_CASE_FIXTURE(Fixture, "for_in_loop_with_incompatible_args_to_iterator")
 {
+    ScopedFastFlag sffs[] = {
+        {FFlag::LuauSolverV2, true},
+        {FFlag::LuauCheckForInWithSubtyping3, true},
+    };
+
     CheckResult result = check(R"(
         function my_iter(state: string, index: number)
             return state, index
@@ -374,13 +426,23 @@ TEST_CASE_FIXTURE(Fixture, "for_in_loop_with_incompatible_args_to_iterator")
         end
     )");
 
-    LUAU_REQUIRE_ERROR_COUNT(2, result);
+    // TODO, CLI-177651: The rough bidirectional rule with for-in loops ought to be:
+    //
+    //  for a, b in c, d, e
+    //  end
+    //
+    //  c => (A, B) -> (A, B)
+    //  d <= A
+    //  e <= B
+    //  ---
+    //  a => A
+    //  b => B
+    //
+    // That would give us the nice errors here of `my_state </: string` and `first_index </: number`
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
 
-    CHECK(get<TypeMismatch>(result.errors[1]));
-    CHECK(Location{{9, 29}, {9, 37}} == result.errors[0].location);
-
-    CHECK(get<TypeMismatch>(result.errors[1]));
-    CHECK(Location{{9, 39}, {9, 50}} == result.errors[1].location);
+    CHECK(get<TypeMismatch>(result.errors[0]));
+    CHECK(Location{{9, 20}, {9, 27}} == result.errors[0].location);
 }
 
 TEST_CASE_FIXTURE(Fixture, "for_in_loop_with_custom_iterator")
@@ -399,8 +461,8 @@ TEST_CASE_FIXTURE(Fixture, "for_in_loop_with_custom_iterator")
 
     TypeMismatch* tm = get<TypeMismatch>(result.errors[0]);
     REQUIRE(tm);
-    CHECK_EQ(builtinTypes->numberType, tm->wantedType);
-    CHECK_EQ(builtinTypes->stringType, tm->givenType);
+    CHECK_EQ(getBuiltins()->numberType, tm->wantedType);
+    CHECK_EQ(getBuiltins()->stringType, tm->givenType);
 }
 
 TEST_CASE_FIXTURE(Fixture, "while_loop")
@@ -417,7 +479,7 @@ TEST_CASE_FIXTURE(Fixture, "while_loop")
     if (FFlag::LuauSolverV2)
         CHECK("number?" == toString(requireType("i")));
     else
-        CHECK_EQ(*builtinTypes->numberType, *requireType("i"));
+        CHECK("number" == toString(requireType("i")));
 }
 
 TEST_CASE_FIXTURE(Fixture, "repeat_loop")
@@ -434,7 +496,7 @@ TEST_CASE_FIXTURE(Fixture, "repeat_loop")
     if (FFlag::LuauSolverV2)
         CHECK("string?" == toString(requireType("i")));
     else
-        CHECK_EQ(*builtinTypes->stringType, *requireType("i"));
+        CHECK("string" == toString(requireType("i")));
 }
 
 TEST_CASE_FIXTURE(Fixture, "repeat_loop_condition_binds_to_its_block")
@@ -752,7 +814,7 @@ TEST_CASE_FIXTURE(Fixture, "loop_iter_basic")
         CHECK("number?" == toString(keyTy));
     }
     else
-        CHECK_EQ(*builtinTypes->numberType, *requireType("key"));
+        CHECK("number" == toString(requireType("key")));
 }
 
 TEST_CASE_FIXTURE(Fixture, "loop_iter_trailing_nil")
@@ -769,7 +831,7 @@ TEST_CASE_FIXTURE(Fixture, "loop_iter_trailing_nil")
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(0, result);
-    CHECK_EQ(*builtinTypes->nilType, *requireType("extra"));
+    CHECK("nil" == toString(requireType("extra")));
 }
 
 TEST_CASE_FIXTURE(Fixture, "loop_iter_no_indexer_strict")
@@ -827,11 +889,10 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "loop_iter_metamethod_not_enough_returns")
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
     CHECK(
-        result.errors[0] ==
-        TypeError{
-            Location{{2, 36}, {2, 37}},
-            GenericError{"__iter must return at least one value"},
-        }
+        result.errors[0] == TypeError{
+                                Location{{2, 36}, {2, 37}},
+                                GenericError{"__iter must return at least one value"},
+                            }
     );
 }
 
@@ -891,7 +952,10 @@ TEST_CASE_FIXTURE(Fixture, "for_loop_lower_bound_is_string_2")
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-    CHECK_EQ("Type 'number' could not be converted into 'never'", toString(result.errors[0]));
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ("Expected this to be unreachable, but got 'number'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type 'number' could not be converted into 'never'", toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(Fixture, "for_loop_lower_bound_is_string_3")
@@ -922,7 +986,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "cli_68448_iterators_need_not_accept_nil")
     LUAU_REQUIRE_NO_ERRORS(result);
     // HACK (CLI-68453): We name this inner table `enum`. For now, use the
     // exhaustive switch to see past it.
-    CHECK(toString(requireType("makeEnum"), {true}) == "<a>({a}) -> {| [a]: a |}");
+    CHECK(toString(requireType("makeEnum"), {true}) == "<a>({a}) -> { [a]: a }");
 }
 
 TEST_CASE_FIXTURE(Fixture, "iterate_over_free_table")
@@ -1073,7 +1137,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "dcr_iteration_on_never_gives_never")
     LUAU_REQUIRE_NO_ERRORS(result);
 
     if (FFlag::LuauSolverV2)
-        CHECK("never?" == toString(requireType("ans"))); // CLI-114134 egraph simplification.  Should just be nil.
+        CHECK("nil" == toString(requireType("ans")));
     else
         CHECK(toString(requireType("ans")) == "never");
 }
@@ -1232,6 +1296,8 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "forin_metatable_iter_mm")
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "iteration_preserves_error_suppression")
 {
+    ScopedFastFlag v1{FFlag::LuauSolverV2, true};
+
     CheckResult result = check(R"(
         function first(x: any)
             for k, v in pairs(x) do
@@ -1242,7 +1308,7 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "iteration_preserves_error_suppression")
 
     LUAU_REQUIRE_NO_ERRORS(result);
 
-    CHECK("any" == toString(requireTypeAtPosition({3, 22})));
+    CHECK("*error-type* | ~nil" == toString(requireTypeAtPosition({3, 22})));
     CHECK("any" == toString(requireTypeAtPosition({3, 25})));
 }
 
@@ -1258,8 +1324,6 @@ end
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_surprising_iterator")
 {
-    ScopedFastFlag luauStatForInFix{FFlag::LuauStatForInFix, true};
-
     CheckResult result = check(R"(
 function broken(): (...() -> ())
     return function() end, function() end
@@ -1271,4 +1335,325 @@ for p in broken() do print(p) end
     LUAU_REQUIRE_ERRORS(result);
 }
 
+TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_require")
+{
+    ScopedFastFlag sff{FFlag::LuauSolverV2, true};
+
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        for _ in require do
+        end
+    )"));
+}
+
+TEST_CASE_FIXTURE(Fixture, "oss_1480")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        type Part = { Parent: Part? }
+        type Instance = Part
+
+        local part = {} :: Part
+
+        local currentParent: Instance? = part.Parent
+        while currentParent ~= nil do
+            currentParent = currentParent.Parent
+        end
+    )"));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "oss_1413")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local function KahanSum(values: {number}): number
+            local sum: number = 0
+            local compensator: number = 0
+            for _, value in values do
+                local y = value - compensator
+                local t = sum + y
+                compensator = (t - sum) - y
+                sum = t
+            end
+            return sum
+        end
+    )"));
+
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local function HistogramString(values: {number})
+            local histogram = {}
+            values = table.clone(values)
+            table.sort(values)
+
+            local count = #values
+            local range = (count - 1)
+
+            local digitIndex = range // 2 + 1
+            while digitIndex < count and values[digitIndex] == 0 do
+                digitIndex = count - ((count - digitIndex) // 2)
+            end
+        end
+    )"));
+
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local function fun1()
+            local foo = 1
+            local bar = foo - foo + foo
+            while false do
+                foo = bar
+            end
+        end
+        local function fun2()
+            local foo = 1
+            while false do
+                local bar = foo - foo + foo
+                foo = bar
+            end
+        end
+    )"));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "while_loop_error_in_body")
+{
+    if (!FFlag::LuauSolverV2)
+        return;
+
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local function foo()
+            local x = ""
+            while math.random () > 0.5 do
+                x = nil
+                error("why did you make x nil tho")
+            end
+            return x
+        end
+    )"));
+
+    CHECK_EQ("() -> string", toString(requireType("foo")));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "while_loop_assign_different_type")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local function takesString(_: string) end
+        local function takesNil(_: nil) end
+        local function foo()
+            local x = ""
+            takesString(x)
+            while math.random () > 0.5 do
+                x = nil
+                takesNil(x)
+            end
+            return x
+        end
+    )"));
+
+    CHECK_EQ("() -> string?", toString(requireType("foo")));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "repeat_loop_assignment")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local x = nil
+        repeat
+            x = 42
+        until math.random() > 0.5
+        local y = x
+    )"));
+
+    CHECK_EQ("number", toString(requireType("y")));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "repeat_loop_assignment_with_break")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local x = nil
+        repeat
+            x = 42
+        until math.random() > 0.5
+        local y = x
+    )"));
+
+    CHECK_EQ("number", toString(requireType("y")));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "repeat_unconditionally_fires_error")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local x = nil
+        repeat
+            x = 42
+        until true
+        -- `x` should unconditionally be `number` here as the assignment
+        -- above will _always_ run.
+        local y = x
+    )"));
+
+    CHECK_EQ("number", toString(requireType("y")));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "repeat_is_linearish")
+{
+    if (!FFlag::LuauSolverV2)
+        return;
+
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local x = nil
+        if math.random () > 0.5 then
+            x = ""
+            repeat
+                error("spooky scary error")
+            until true
+        end
+        -- The repeat in the above branch unconditionally fires the error, so
+        -- this should _always_ be `nil`
+        local y = x
+    )"));
+
+    CHECK_EQ("nil", toString(requireType("y")));
+}
+
+TEST_CASE_FIXTURE(Fixture, "ensure_local_in_loop_does_not_escape")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local x = 42
+        repeat
+            local x = ""
+        until true
+        -- The local inside the loop should have no effect on the local
+        -- outside the loop.
+        local y = x
+    )"));
+
+    CHECK_EQ("number", toString(requireType("y")));
+}
+
+TEST_CASE_FIXTURE(Fixture, "oss_1851_union_of_many_strings")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+--!strict
+type union = "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8" | "9"
+
+local example: { [union]: number } = {}
+
+for key in example do
+end
+    )"));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "any_type_in_for_loop_should_propagate")
+{
+    ScopedFastFlag _{FFlag::LuauPropagateTypeAnnotationsInForInLoops, true};
+
+    CheckResult result = check(R"(
+        --!strict
+        function my_iter(): any
+            return {}
+        end
+
+        for index: number, value: string in my_iter() do
+            print(index, value)
+        end
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+
+    CHECK("number" == toString(requireTypeAtPosition(Position{7, 18})));
+    CHECK("string" == toString(requireTypeAtPosition(Position{7, 25})));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "explicit_types_in_for_loop_should_propagate")
+{
+    ScopedFastFlag _{FFlag::LuauPropagateTypeAnnotationsInForInLoops, true};
+
+    CheckResult result = check(R"(
+        --!strict
+        function my_iter(): {[number]: string}
+            return {}
+        end
+
+        for index: number, value: string in my_iter() do
+            print(index, value)
+        end
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+
+    CHECK("number" == toString(requireTypeAtPosition(Position{7, 18})));
+    CHECK("string" == toString(requireTypeAtPosition(Position{7, 25})));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "incorrect_type_annotation_types_in_loop_should_propagate_with_errors")
+{
+    ScopedFastFlag _{FFlag::LuauPropagateTypeAnnotationsInForInLoops, true};
+
+    CheckResult result = check(R"(
+        --!strict
+        function my_iter(): any
+            return {}
+        end
+        for index: number, value: string in my_iter() do
+            index = ""
+            print(index)
+        end
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+
+    auto err = get<TypeMismatch>(result.errors[0]);
+    REQUIRE(err);
+    CHECK_EQ("number", toString(err->wantedType));
+    CHECK_EQ("string", toString(err->givenType));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_loop_annotations_apply_to_function_expressions")
+{
+    ScopedFastFlag _{FFlag::LuauPropagateTypeAnnotationsInForInLoops, true};
+
+    CheckResult result = check(R"(
+        --!strict
+        function my_iter(): any
+            return {}
+        end
+
+        local function takesString(s: string) end
+
+        for index: number in my_iter() do
+            takesString(index)
+        end
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+
+    auto err = get<TypeMismatch>(result.errors[0]);
+    REQUIRE(err);
+    CHECK_EQ("string", toString(err->wantedType));
+    CHECK_EQ("number", toString(err->givenType));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "for_in_loop_annotations_apply_inside_lambdas")
+{
+    ScopedFastFlag _{FFlag::LuauPropagateTypeAnnotationsInForInLoops, true};
+
+    CheckResult result = check(R"(
+        --!strict
+        function my_iter(): any
+            return {}
+        end
+
+        for index: number in my_iter() do
+            local fn = function()
+                index = ""
+            end
+            fn()
+        end
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+
+    auto err = get<TypeMismatch>(result.errors[0]);
+    REQUIRE(err);
+    CHECK_EQ("number", toString(err->wantedType));
+    CHECK_EQ("string", toString(err->givenType));
+}
+
 TEST_SUITE_END();
