@@ -2,13 +2,15 @@
 
 #include "Fixture.h"
 
+#include "ScopedFlags.h"
 #include "doctest.h"
 
 using namespace Luau;
 
-LUAU_FASTFLAG(LuauSolverV2)
-LUAU_FASTFLAG(LuauPropagateExpectedTypesForCalls)
-LUAU_FASTFLAG(LuauImproveTypePathsInErrors)
+LUAU_FASTFLAG(LuauBetterTypeMismatchErrors)
+LUAU_FASTFLAG(LuauMorePreciseErrorSuppression)
+LUAU_FASTFLAG(LuauBetterTypeMismatchErrors)
+LUAU_FASTFLAG(LuauPushTypeUnifyConstantHandling)
 
 TEST_SUITE_BEGIN("TypeSingletons");
 
@@ -67,7 +69,10 @@ TEST_CASE_FIXTURE(Fixture, "bool_singletons_mismatch")
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-    CHECK_EQ("Type 'false' could not be converted into 'true'", toString(result.errors[0]));
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ("Expected this to be 'true', but got 'false'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type 'false' could not be converted into 'true'", toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(Fixture, "string_singletons_mismatch")
@@ -77,7 +82,10 @@ TEST_CASE_FIXTURE(Fixture, "string_singletons_mismatch")
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-    CHECK_EQ("Type '\"bar\"' could not be converted into '\"foo\"'", toString(result.errors[0]));
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ("Expected this to be '\"foo\"', but got '\"bar\"'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type '\"bar\"' could not be converted into '\"foo\"'", toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(Fixture, "string_singletons_escape_chars")
@@ -87,7 +95,10 @@ TEST_CASE_FIXTURE(Fixture, "string_singletons_escape_chars")
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-    CHECK_EQ(R"(Type '"\000\r"' could not be converted into '"\n"')", toString(result.errors[0]));
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ(R"(Expected this to be '"\n"', but got '"\000\r"')", toString(result.errors[0]));
+    else
+        CHECK_EQ(R"(Type '"\000\r"' could not be converted into '"\n"')", toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(Fixture, "bool_singleton_subtype")
@@ -138,7 +149,10 @@ TEST_CASE_FIXTURE(Fixture, "function_call_with_singletons_mismatch")
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-    CHECK_EQ("Type '\"bar\"' could not be converted into '\"foo\"'", toString(result.errors[0]));
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ("Expected this to be '\"foo\"', but got '\"bar\"'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type '\"bar\"' could not be converted into '\"foo\"'", toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(Fixture, "overloaded_function_call_with_singletons")
@@ -155,8 +169,6 @@ TEST_CASE_FIXTURE(Fixture, "overloaded_function_call_with_singletons")
 
 TEST_CASE_FIXTURE(Fixture, "overloaded_function_resolution_singleton_parameters")
 {
-    ScopedFastFlag sff{FFlag::LuauPropagateExpectedTypesForCalls, true};
-
     CheckResult result = check(R"(
         type A = ("A") -> string
         type B = ("B") -> number
@@ -191,7 +203,10 @@ TEST_CASE_FIXTURE(Fixture, "overloaded_function_call_with_singletons_mismatch")
     }
     else
     {
-        CHECK_EQ("Type 'number' could not be converted into 'string'", toString(result.errors[0]));
+        if (FFlag::LuauBetterTypeMismatchErrors)
+            CHECK_EQ("Expected this to be 'string', but got 'number'", toString(result.errors[0]));
+        else
+            CHECK_EQ("Type 'number' could not be converted into 'string'", toString(result.errors[0]));
         CHECK_EQ("Other overloads are also not viable: (false, number) -> ()", toString(result.errors[1]));
     }
 }
@@ -217,8 +232,32 @@ TEST_CASE_FIXTURE(Fixture, "enums_using_singletons_mismatch")
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
 
-    if (FFlag::LuauSolverV2)
-        CHECK("Type '\"bang\"' could not be converted into '\"bar\" | \"baz\" | \"foo\"'" == toString(result.errors[0]));
+    if (FFlag::LuauSolverV2 && FFlag::LuauMorePreciseErrorSuppression)
+    {
+        // clang-format off
+        const std::string expected =
+            "Expected this to be '\"bar\" | \"baz\" | \"foo\"', but got '\"bang\"'; \n"
+            "this is because \n"
+            "	 * the 1st component of the union is `\"foo\"`, and `\"bang\"` is not a subtype of `\"foo\"`\n"
+            "	 * the 2nd component of the union is `\"bar\"`, and `\"bang\"` is not a subtype of `\"bar\"`\n"
+            "	 * the 3rd component of the union is `\"baz\"`, and `\"bang\"` is not a subtype of `\"baz\"`"
+        ;
+        // clang-format on
+
+        CHECK_LONG_STRINGS_EQ(expected, toString(result.errors[0]));
+    }
+    else if (FFlag::LuauSolverV2)
+    {
+        if (FFlag::LuauBetterTypeMismatchErrors)
+            CHECK("Expected this to be '\"bar\" | \"baz\" | \"foo\"', but got '\"bang\"'" == toString(result.errors[0]));
+        else
+            CHECK("Type '\"bang\"' could not be converted into '\"bar\" | \"baz\" | \"foo\"'" == toString(result.errors[0]));
+    }
+    else if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ(
+            "Expected this to be '\"bar\" | \"baz\" | \"foo\"', but got '\"bang\"'; none of the union options are compatible",
+            toString(result.errors[0])
+        );
     else
         CHECK_EQ(
             "Type '\"bang\"' could not be converted into '\"bar\" | \"baz\" | \"foo\"'; none of the union options are compatible",
@@ -283,7 +322,7 @@ TEST_CASE_FIXTURE(Fixture, "tagged_unions_immutable_tag")
         CannotAssignToNever* tm = get<CannotAssignToNever>(result.errors[0]);
         REQUIRE(tm);
 
-        CHECK(builtinTypes->stringType == tm->rhsType);
+        CHECK(getBuiltins()->stringType == tm->rhsType);
         CHECK(CannotAssignToNever::Reason::PropertyNarrowed == tm->reason);
         REQUIRE(tm->cause.size() == 2);
         CHECK("\"Dog\"" == toString(tm->cause[0]));
@@ -297,7 +336,10 @@ TEST_CASE_FIXTURE(Fixture, "table_has_a_boolean")
         local t={a=1,b=false}
     )");
 
-    CHECK("{ a: number, b: boolean }" == toString(requireType("t"), {true}));
+    if (FFlag::LuauSolverV2)
+        CHECK("{ a: number, b: boolean }" == toString(requireType("t"), {true}));
+    else
+        CHECK("{| a: number, b: boolean |}" == toString(requireType("t"), {true}));
 }
 
 TEST_CASE_FIXTURE(Fixture, "table_properties_singleton_strings")
@@ -338,7 +380,10 @@ TEST_CASE_FIXTURE(Fixture, "table_properties_singleton_strings_mismatch")
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-    CHECK_EQ("Type 'number' could not be converted into 'string'", toString(result.errors[0]));
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK_EQ("Expected this to be 'string', but got 'number'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type 'number' could not be converted into 'string'", toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(Fixture, "table_properties_alias_or_parens_is_indexer")
@@ -379,6 +424,8 @@ TEST_CASE_FIXTURE(Fixture, "indexer_can_be_union_of_singletons")
 
 TEST_CASE_FIXTURE(Fixture, "table_properties_type_error_escapes")
 {
+    ScopedFastFlag sff{FFlag::LuauSolverV2, true};
+
     CheckResult result = check(R"(
         --!strict
         local x: { ["<>"] : number }
@@ -387,26 +434,9 @@ TEST_CASE_FIXTURE(Fixture, "table_properties_type_error_escapes")
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
 
-    if (FFlag::LuauSolverV2 && FFlag::LuauImproveTypePathsInErrors)
-    {
-        const std::string expected = "Type\n\t"
-                                     "'{ [\"\\n\"]: number }'"
-                                     "\ncould not be converted into\n\t"
-                                     "'{ [\"<>\"]: number }'";
-        CHECK(expected == toString(result.errors[0]));
-    }
-    else if (FFlag::LuauSolverV2)
-        CHECK(
-            "Type\n"
-            "    '{ [\"\\n\"]: number }'\n"
-            "could not be converted into\n"
-            "    '{ [\"<>\"]: number }'" == toString(result.errors[0])
-        );
-    else
-        CHECK_EQ(
-            R"(Table type '{ ["\n"]: number }' not compatible with type '{| ["<>"]: number |}' because the former is missing field '<>')",
-            toString(result.errors[0])
-        );
+    const std::string expected =
+        R"(Table type '{ ["\n"]: number }' not compatible with type '{ ["<>"]: number }' because the former is missing field '<>')";
+    CHECK(expected == toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(Fixture, "error_detailed_tagged_union_mismatch_string")
@@ -421,7 +451,18 @@ local a: Animal = { tag = 'cat', cafood = 'something' }
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
     if (FFlag::LuauSolverV2)
-        CHECK("Type '{ cafood: string, tag: \"cat\" }' could not be converted into 'Cat | Dog'" == toString(result.errors[0]));
+        CHECK(
+            R"(Table type '{ cafood: string, tag: "cat" }' not compatible with type 'Cat' because the former is missing field 'catfood')" ==
+            toString(result.errors[0])
+        );
+    else if (FFlag::LuauBetterTypeMismatchErrors)
+    {
+        const std::string expected = R"(Expected this to be 'Cat | Dog', but got 'a'
+caused by:
+  None of the union options are compatible. For example:
+Table type 'a' not compatible with type 'Cat' because the former is missing field 'catfood')";
+        CHECK_EQ(expected, toString(result.errors[0]));
+    }
     else
     {
         const std::string expected = R"(Type 'a' could not be converted into 'Cat | Dog'
@@ -434,6 +475,10 @@ Table type 'a' not compatible with type 'Cat' because the former is missing fiel
 
 TEST_CASE_FIXTURE(Fixture, "error_detailed_tagged_union_mismatch_bool")
 {
+    ScopedFastFlag sffs[] = {
+        {FFlag::LuauBetterTypeMismatchErrors, true},
+        {FFlag::LuauPushTypeUnifyConstantHandling, true},
+    };
     CheckResult result = check(R"(
 type Good = { success: true, result: string }
 type Bad = { success: false, error: string }
@@ -444,10 +489,15 @@ local a: Result = { success = false, result = 'something' }
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
     if (FFlag::LuauSolverV2)
-        CHECK("Type '{ result: string, success: boolean }' could not be converted into 'Bad | Good'" == toString(result.errors[0]));
+    {
+        CHECK_EQ(
+            "Table type '{ result: string, success: false }' not compatible with type 'Bad' because the former is missing field 'error'",
+            toString(result.errors[0])
+        );
+    }
     else
     {
-        const std::string expected = R"(Type 'a' could not be converted into 'Bad | Good'
+        const std::string expected = R"(Expected this to be 'Bad | Good', but got 'a'
 caused by:
   None of the union options are compatible. For example:
 Table type 'a' not compatible with type 'Bad' because the former is missing field 'error')";
@@ -459,6 +509,8 @@ TEST_CASE_FIXTURE(Fixture, "parametric_tagged_union_alias")
 {
     ScopedFastFlag sff[] = {
         {FFlag::LuauSolverV2, true},
+        {FFlag::LuauBetterTypeMismatchErrors, true},
+        {FFlag::LuauPushTypeUnifyConstantHandling, true},
     };
     CheckResult result = check(R"(
         type Ok<T> = {success: true, result: T}
@@ -471,23 +523,10 @@ TEST_CASE_FIXTURE(Fixture, "parametric_tagged_union_alias")
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
 
-    if (FFlag::LuauImproveTypePathsInErrors)
-    {
-        const std::string expectedError = "Type\n\t"
-                                          "'{ result: string, success: boolean }'"
-                                          "\ncould not be converted into\n\t"
-                                          "'Err<number> | Ok<string>'";
-        CHECK(toString(result.errors[0]) == expectedError);
-    }
-    else
-    {
-        const std::string expectedError = R"(Type
-    '{ result: string, success: boolean }'
-could not be converted into
-    'Err<number> | Ok<string>')";
-
-        CHECK(toString(result.errors[0]) == expectedError);
-    }
+    CHECK_EQ(
+        "Table type '{ result: string, success: false }' not compatible with type 'Err<number>' because the former is missing field 'error'",
+        toString(result.errors[0])
+    );
 }
 
 TEST_CASE_FIXTURE(Fixture, "if_then_else_expression_singleton_options")
@@ -682,4 +721,193 @@ TEST_CASE_FIXTURE(BuiltinsFixture, "singletons_stick_around_under_assignment")
         LUAU_REQUIRE_ERROR_COUNT(1, result);
 }
 
+TEST_CASE_FIXTURE(Fixture, "tagged_union_in_ternary")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        type Result = { type: "ok", value: unknown } | { type: "error" }
+
+        local function coinflip(): boolean return true end
+
+        local function readFromDB(): Result
+            return if coinflip() then { type = "ok", value = 42 } else { type = "error" }
+        end
+    )"));
+}
+
+TEST_CASE_FIXTURE(Fixture, "table_literal_with_singleton_union_values")
+{
+    CheckResult result = check(R"(
+        local t1: {[string]: "a" | "b"} = { a = "a", b = "b" }
+        local t2: {[string]: "a" | true} = { a = "a", b = true }
+        local t3: {[string]: "a" | nil} = { a = "a" }
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(Fixture, "singleton_type_mismatch_via_variable")
+{
+    CheckResult result = check(R"(
+        local c = "c"
+        local x: "a" = c
+        local y: "a" | "b" = c
+        local z: "a"? = c
+        local w: "a" | "b" = "c"
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(4, result);
+
+    REQUIRE(get<TypeMismatch>(result.errors[0]));
+    REQUIRE(get<TypeMismatch>(result.errors[1]));
+    REQUIRE(get<TypeMismatch>(result.errors[2]));
+    REQUIRE(get<TypeMismatch>(result.errors[3]));
+}
+
+TEST_CASE_FIXTURE(Fixture, "cli_163481_any_indexer_pushes_type")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        --!strict
+
+        type test = "A"
+        type test2 = "A"|"B"|"C"
+
+        local t: { [any]: test } = { A = "A" }
+
+        local t2: { [any]: test2 } = {
+            A = "A",
+            B = "B",
+            C = "C"
+        }
+    )"));
+}
+
+TEST_CASE_FIXTURE(Fixture, "oss_2010")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local function foo<T>(my_enum: "" | T): T
+            return my_enum :: T
+        end
+
+        local var = foo("meow")
+    )"));
+
+    CHECK_EQ("\"meow\"", toString(requireType("var")));
+}
+
+TEST_CASE_FIXTURE(Fixture, "oss_1773")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        --!strict
+
+        export type T = "foo" | "bar" | "toto"
+
+        local object: T = "foo"
+
+        local getOpposite: {[T]: T} = {
+            ["foo"] = "bar",
+            ["bar"] = "toto",
+            ["toto"] = "foo"
+        }
+
+        local function hello()
+            local x = getOpposite[object]
+
+            if x then
+                object = x
+            end
+        end
+    )"));
+}
+
+TEST_CASE_FIXTURE(Fixture, "bidirectionally_infer_indexers_errored")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+
+    CheckResult result = check(R"(
+        --!strict
+
+        export type T = "foo" | "bar" | "toto"
+
+        local getOpposite: { [number]: T } = {
+            ["foo"] = "bar",
+            ["bar"] = "toto",
+            ["toto"] = "foo"
+        }
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(3, result);
+
+    for (const auto& e : result.errors)
+        CHECK(get<TypeMismatch>(e));
+}
+
+TEST_CASE_FIXTURE(Fixture, "oss_2018")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local rule: { rule: "AppendTextComment" } | { rule: "Other" } = { rule = "AppendTextComment" }
+    )"));
+}
+
+TEST_CASE_FIXTURE(Fixture, "oss_2010_but_with_booleans")
+{
+    ScopedFastFlag sffs[] = {
+        {FFlag::LuauSolverV2, true},
+        {FFlag::LuauPushTypeUnifyConstantHandling, true},
+    };
+
+    CheckResult results = check(R"(
+        local function foo<T>(my_enum: true | T): T
+            return my_enum :: T
+        end
+
+        local function bar<T>(my_enum: true & T): T
+            return my_enum :: T
+        end
+
+        local var1 = foo(true)
+        local var2 = foo(false)
+
+        local var3 = bar(true)
+        local var4 = bar(false)
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, results);
+    auto err = get<TypeMismatch>(results.errors[0]);
+    REQUIRE(err);
+    CHECK_EQ("true", toString(err->wantedType));
+    CHECK_EQ("false", toString(err->givenType));
+
+    // FIXME: That one seems arguably correct.
+    CHECK_EQ("unknown", toString(requireType("var1")));
+    CHECK_EQ("false", toString(requireType("var2")));
+    CHECK_EQ("true", toString(requireType("var3")));
+    CHECK_EQ("false", toString(requireType("var4")));
+}
+
+TEST_CASE_FIXTURE(Fixture, "cli_184125")
+{
+    ScopedFastFlag sffs[] = {
+        {FFlag::LuauSolverV2, true},
+        {FFlag::LuauPushTypeUnifyConstantHandling, true},
+    };
+
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        type MyTypeA = {Value: true}
+        type MyTypeB = {Value: false}
+
+        local function Func(input: number) : (MyTypeA | MyTypeB)
+            if input == 1 then
+                return {Value = true}
+            else
+                return {Value = false}
+            end
+        end
+    )"));
+}
+
+
 TEST_SUITE_END();
