@@ -4,6 +4,8 @@
 #include "doctest.h"
 
 LUAU_FASTFLAG(LuauSolverV2)
+LUAU_FASTFLAG(LuauBetterTypeMismatchErrors)
+LUAU_FASTFLAG(LuauUnionOfTablesPreservesReadWrite)
 
 using namespace Luau;
 
@@ -82,11 +84,10 @@ TEST_CASE_FIXTURE(TypeStateFixture, "parameter_x_was_constrained_by_two_types")
         // as a type for `x`, but it's a limitation we can accept for now.
         LUAU_REQUIRE_ERRORS(result);
 
-        TypePackMismatch* tpm = get<TypePackMismatch>(result.errors[0]);
-        REQUIRE_MESSAGE(tpm, "Expected TypePackMismatch but got " << result.errors[0]);
-        CHECK("string?" == toString(tpm->wantedTp));
-        CHECK("number | string" == toString(tpm->givenTp));
-
+        TypeMismatch* tm = get<TypeMismatch>(result.errors[0]);
+        REQUIRE_MESSAGE(tm, "Expected TypeMismatch but got " << result.errors[0]);
+        CHECK("string?" == toString(tm->wantedType));
+        CHECK("number | string" == toString(tm->givenType));
         CHECK("(number | string) -> string?" == toString(requireType("f")));
     }
     else
@@ -121,6 +122,7 @@ TEST_CASE_FIXTURE(TypeStateFixture, "refine_a_local_and_then_assign_it")
 
     LUAU_REQUIRE_NO_ERRORS(result);
 }
+#endif
 
 TEST_CASE_FIXTURE(TypeStateFixture, "assign_a_local_and_then_refine_it")
 {
@@ -135,9 +137,11 @@ TEST_CASE_FIXTURE(TypeStateFixture, "assign_a_local_and_then_refine_it")
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-    CHECK("Type 'string' could not be converted into 'never'" == toString(result.errors[0]));
+    if (FFlag::LuauBetterTypeMismatchErrors)
+        CHECK("Expected this to be unreachable, but got 'string'" == toString(result.errors[0]));
+    else
+        CHECK("Type 'string' could not be converted into 'never'" == toString(result.errors[0]));
 }
-#endif
 
 TEST_CASE_FIXTURE(TypeStateFixture, "recursive_local_function")
 {
@@ -274,7 +278,7 @@ TEST_CASE_FIXTURE(TypeStateFixture, "then_branch_assigns_and_else_branch_also_as
     )");
 
     LUAU_REQUIRE_NO_ERRORS(result);
-    CHECK("number?" == toString(requireType("y")));
+    CHECK("number" == toString(requireType("y")));
 }
 
 TEST_CASE_FIXTURE(TypeStateFixture, "then_branch_assigns_but_is_met_with_return_and_else_branch_assigns")
@@ -291,7 +295,7 @@ TEST_CASE_FIXTURE(TypeStateFixture, "then_branch_assigns_but_is_met_with_return_
     )");
 
     LUAU_REQUIRE_NO_ERRORS(result);
-    CHECK("string?" == toString(requireType("y")));
+    CHECK("string" == toString(requireType("y")));
 }
 
 TEST_CASE_FIXTURE(TypeStateFixture, "invalidate_type_refinements_upon_assignments")
@@ -337,7 +341,7 @@ TEST_CASE_FIXTURE(TypeStateFixture, "local_t_is_assigned_a_fresh_table_with_x_as
 }
 #endif
 
-TEST_CASE_FIXTURE(TypeStateFixture, "captured_locals_are_unions_of_all_assignments")
+TEST_CASE_FIXTURE(TypeStateFixture, "captured_locals_do_not_mutate_upvalue_type")
 {
     CheckResult result = check(R"(
         local x = nil
@@ -351,12 +355,17 @@ TEST_CASE_FIXTURE(TypeStateFixture, "captured_locals_are_unions_of_all_assignmen
         f()
     )");
 
-    LUAU_REQUIRE_NO_ERRORS(result);
-    CHECK("(number | string)?" == toString(requireTypeAtPosition({4, 18})));
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    auto err = get<TypeMismatch>(result.errors[0]);
+    CHECK_EQ("number?", toString(err->wantedType));
+    CHECK_EQ("string", toString(err->givenType));
+    CHECK("number?" == toString(requireTypeAtPosition({4, 18})));
 }
 
-TEST_CASE_FIXTURE(TypeStateFixture, "captured_locals_are_unions_of_all_assignments_2")
+TEST_CASE_FIXTURE(TypeStateFixture, "captured_locals_do_not_mutate_upvalue_type_2")
 {
+    ScopedFastFlag sffs[] = {{FFlag::LuauSolverV2, true}, {FFlag::LuauUnionOfTablesPreservesReadWrite, true}};
+
     CheckResult result = check(R"(
         local t = {x = nil}
 
@@ -369,9 +378,12 @@ TEST_CASE_FIXTURE(TypeStateFixture, "captured_locals_are_unions_of_all_assignmen
         f()
     )");
 
-    LUAU_REQUIRE_NO_ERRORS(result);
-    CHECK("{ x: nil } | { x: number } | { x: string }" == toString(requireTypeAtPosition({4, 18}), {true}));
-    CHECK("(number | string)?" == toString(requireTypeAtPosition({4, 20})));
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    auto err = get<TypeMismatch>(result.errors[0]);
+    CHECK_EQ("{ x: nil } | { x: number }", toString(err->wantedType, {/* exhaustive */ true}));
+    CHECK_EQ("{ x: string }", toString(err->givenType));
+    CHECK("{ x: nil } | { x: number }" == toString(requireTypeAtPosition({4, 18}), {true}));
+    CHECK("number?" == toString(requireTypeAtPosition({4, 20})));
 }
 
 TEST_CASE_FIXTURE(TypeStateFixture, "prototyped_recursive_functions")
@@ -391,9 +403,9 @@ TEST_CASE_FIXTURE(TypeStateFixture, "prototyped_recursive_functions")
 
 TEST_CASE_FIXTURE(BuiltinsFixture, "prototyped_recursive_functions_but_has_future_assignments")
 {
-    // early return if the flag isn't set since this is blocking gated commits
-    if (!FFlag::LuauSolverV2)
-        return;
+    ScopedFastFlag sffs[] = {
+        {FFlag::LuauSolverV2, true},
+    };
 
     CheckResult result = check(R"(
         local f
@@ -455,25 +467,6 @@ TEST_CASE_FIXTURE(TypeStateFixture, "typestates_preserve_error_suppression")
     CHECK("*error-type* | string" == toString(requireTypeAtPosition({3, 14}), {true}));
 }
 
-
-TEST_CASE_FIXTURE(BuiltinsFixture, "typestates_preserve_error_suppression_properties")
-{
-    // early return if the flag isn't set since this is blocking gated commits
-    // unconditional return
-    // CLI-117098 Type states with error suppressing properties doesn't infer the correct type for properties.
-    if (!FFlag::LuauSolverV2 || FFlag::LuauSolverV2)
-        return;
-
-    CheckResult result = check(R"(
-        local a: {x: any} = {x = 51}
-        a.x = "pickles" -- We'll have a new DefId for this iteration of `a.x`.  Its type must also be error-suppressing
-        print(a.x)
-    )");
-
-    LUAU_REQUIRE_NO_ERRORS(result);
-    CHECK("*error-type* | string" == toString(requireTypeAtPosition({3, 16}), {true}));
-}
-
 TEST_CASE_FIXTURE(BuiltinsFixture, "typestates_do_not_apply_to_the_initial_local_definition")
 {
     // early return if the flag isn't set since this is blocking gated commits
@@ -523,4 +516,330 @@ TEST_CASE_FIXTURE(Fixture, "typestate_unknown_global")
     CHECK(get<UnknownSymbol>(result.errors[0]));
 }
 
+TEST_CASE_FIXTURE(BuiltinsFixture, "fuzzer_normalized_type_variables_are_bad" * doctest::timeout(0.5))
+{
+    // We do not care about the errors here, only that this finishes typing
+    // in a sensible amount of time.
+    LUAU_REQUIRE_ERRORS(check(R"(
+        local _
+        while _[""] do
+            _, _ = nil
+            while _.n0 do
+                _, _ = nil
+            end
+            _, _ = nil
+        end
+        while _[""] do
+            while if _ then if _ then _ else "" else "" do
+                _, _ = nil
+                do
+                end
+                _, _, _ = nil
+            end
+            _, _ = nil
+            _, _, _ = nil
+            while _.readi16 do
+                _, _ = nil
+            end
+            _, _ = nil
+        end
+    )"));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "oss_1547_simple")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local rand = 0
+
+        function a()
+            rand = (rand % 4) + 1;
+        end
+    )"));
+
+    auto randTy = getType("rand");
+    REQUIRE(randTy);
+    CHECK_EQ("number", toString(*randTy));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "oss_1547")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local rand = 0
+
+        function a()
+            rand = (rand % 4) + 1;
+        end
+
+        function b()
+            rand = math.max(rand - 1, 0);
+        end
+    )"));
+
+    auto randTy = getType("rand");
+    REQUIRE(randTy);
+    CHECK_EQ("number", toString(*randTy));
+}
+
+TEST_CASE_FIXTURE(Fixture, "modify_captured_table_field")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local state = { x = 0 }
+        function incr()
+            state.x = state.x + 1
+        end
+    )"));
+
+    auto randTy = getType("state");
+    REQUIRE(randTy);
+    if (FFlag::LuauSolverV2)
+        CHECK_EQ("{ x: number }", toString(*randTy, {true}));
+    else
+        CHECK_EQ("{| x: number |}", toString(*randTy, {true}));
+}
+
+TEST_CASE_FIXTURE(Fixture, "oss_1561")
+{
+    loadDefinition(R"(
+        declare class Vector3
+            X: number
+            Y: number
+            Z: number
+        end
+
+        declare Vector3: {
+            new: (number?, number?, number?) -> Vector3
+        }
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local targetVelocity: Vector3 = Vector3.new()
+        function set2D(X: number, Y: number)
+            targetVelocity = Vector3.new(X, Y, targetVelocity.Z)
+        end
+    )"));
+
+    CHECK_EQ("(number, number) -> ()", toString(requireType("set2D")));
+}
+
+TEST_CASE_FIXTURE(Fixture, "oss_1575")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        local flag = true
+        local function Flip()
+            flag = not flag
+        end
+    )"));
+}
+
+TEST_CASE_FIXTURE(Fixture, "capture_upvalue_in_returned_function")
+{
+    LUAU_REQUIRE_NO_ERRORS(check(R"(
+        function def()
+            local i : number = 0
+            local function Counter()
+                i = i + 1
+                return i
+            end
+            return Counter
+        end
+    )"));
+    CHECK_EQ("() -> () -> number", toString(requireType("def")));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "throw_in_else_branch")
+{
+    CheckResult result = check(R"(
+        --!strict
+        local x
+        local coinflip : () -> boolean = (nil :: any)
+
+        if coinflip () then
+            x = "I win."
+        else
+            error("You lose.")
+        end
+
+        print(x)
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+
+    CHECK_EQ("string", toString(requireTypeAtPosition({11, 14})));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "throw_in_if_branch")
+{
+    CheckResult result = check(R"(
+        --!strict
+        local x
+        local coinflip : () -> boolean = (nil :: any)
+
+        if coinflip () then
+            error("You lose.")
+        else
+            x = "I win."
+        end
+
+        print(x)
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+
+    CHECK_EQ("string", toString(requireTypeAtPosition({11, 14})));
+}
+
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "refinement_through_erroring")
+{
+    CheckResult result = check(R"(
+        --!strict
+        type Payload = { payload: number }
+
+        local function decode(s: string): Payload?
+            return (nil :: any)
+        end
+
+        local function decodeEx(s: string): Payload
+            local p = decode(s)
+            if not p then
+                error("failed to decode payload!!!")
+            end
+            return p
+        end
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "refinement_through_erroring_in_loop")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+
+    CheckResult result = check(R"(
+        --!strict
+
+        local x = nil
+
+        while math.random() > 0.5 do
+            x = 42
+            return
+        end
+
+        print(x)
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+    CHECK_EQ("nil", toString(requireTypeAtPosition({10, 14})));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "type_refinement_in_loop")
+{
+    CheckResult result = check(R"(
+        --!strict
+        local function onEachString(t: { string | number })
+            for _, v in t do
+                if type(v) ~= "string" then
+                    continue
+                end
+                print(v)
+            end
+        end
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+
+    CHECK_EQ("number | string", toString(requireTypeAtPosition({4, 24})));
+    CHECK_EQ("string", toString(requireTypeAtPosition({7, 22})));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "throw_in_if_branch_and_do_nothing_in_else")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+
+    CheckResult result = check(R"(
+        --!strict
+        local x
+        local coinflip : () -> boolean = (nil :: any)
+
+        if coinflip () then
+            error("You lose.")
+        else
+        end
+
+        print(x)
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+
+    CHECK_EQ("nil", toString(requireTypeAtPosition({10, 14})));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "assign_in_an_if_branch_without_else")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+
+    CheckResult result = check(R"(
+        --!strict
+        local x
+        local coinflip : () -> boolean = (nil :: any)
+
+        if coinflip () then
+            x = "I win."
+        end
+
+        print(x)
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+
+    CHECK_EQ("string?", toString(requireTypeAtPosition({9, 14})));
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "fuzzer_table_freeze_in_binary_expr")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+    // CLI-154237: This currently throws an exception due to a mismatch between
+    // the scopes created in the data flow graph versus the constraint generator.
+    CHECK_THROWS_AS(
+        check(R"(
+            local _
+            if _ or table.freeze(_,_) or table.freeze(_,_) then
+            end
+        )"),
+        Luau::InternalCompilerError
+    );
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "table_freeze_in_conditional")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+    // NOTE: This _probably_ should be disallowed, but it is representing that
+    // type stating functions in short circuiting binary expressions do not
+    // reflect their type states.
+    CheckResult result = check(R"(
+        local t = { x = 42 }
+        if math.random() > 0.5 and table.freeze(t) then
+        end
+        t.y = 13
+    )");
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(BuiltinsFixture, "fuzzer_table_freeze_in_conditional_expr")
+{
+    ScopedFastFlag _{FFlag::LuauSolverV2, true};
+    // CLI-154237: This currently throws an exception due to a mismatch between
+    // the scopes created in the data flow graph versus the constraint generator.
+    CHECK_THROWS_AS(
+        check(R"(
+            local _
+            if
+                if table.freeze(_,_) then _ else _
+            then
+            end
+        )"),
+        Luau::InternalCompilerError
+    );
+}
+
 TEST_SUITE_END();
