@@ -17,16 +17,19 @@ using namespace Luau::TypePath;
 
 LUAU_FASTFLAG(LuauSolverV2);
 LUAU_DYNAMIC_FASTINT(LuauTypePathMaximumTraverseSteps);
-LUAU_FASTFLAG(LuauFreeTypesMustHaveBounds);
 
 struct TypePathFixture : Fixture
 {
     ScopedFastFlag sff1{FFlag::LuauSolverV2, true};
+    TypeArena arena;
+    const DenseHashMap<TypePackId, TypePackId> emptyMap_DEPRECATED{nullptr};
 };
 
 struct TypePathBuiltinsFixture : BuiltinsFixture
 {
     ScopedFastFlag sff1{FFlag::LuauSolverV2, true};
+    TypeArena arena;
+    const DenseHashMap<TypePackId, TypePackId> emptyMap_DEPRECATED{nullptr};
 };
 
 TEST_SUITE_BEGIN("TypePathManipulation");
@@ -109,7 +112,7 @@ TEST_SUITE_BEGIN("TypePathTraversal");
 
 TEST_CASE_FIXTURE(TypePathFixture, "empty_traversal")
 {
-    CHECK(traverseForType(builtinTypes->numberType, kEmpty, builtinTypes) == builtinTypes->numberType);
+    CHECK(traverseForType(getBuiltins()->numberType, kEmpty, getBuiltins(), NotNull{&arena}) == getBuiltins()->numberType);
 }
 
 TEST_CASE_FIXTURE(TypePathFixture, "table_property")
@@ -118,12 +121,15 @@ TEST_CASE_FIXTURE(TypePathFixture, "table_property")
         local x = { y = 123 }
     )");
 
-    CHECK(traverseForType(requireType("x"), Path(TypePath::Property{"y", true}), builtinTypes) == builtinTypes->numberType);
+    CHECK(traverseForType(requireType("x"), Path(TypePath::Property{"y", true}), getBuiltins(), NotNull{&arena}) == getBuiltins()->numberType);
 }
 
-TEST_CASE_FIXTURE(ClassFixture, "class_property")
+TEST_CASE_FIXTURE(ExternTypeFixture, "class_property")
 {
-    CHECK(traverseForType(vector2InstanceType, Path(TypePath::Property{"X", true}), builtinTypes) == builtinTypes->numberType);
+    // Force this here because vector2InstanceType won't get initialized until the frontend has been forced
+    getFrontend();
+    TypeArena arena;
+    CHECK(traverseForType(vector2InstanceType, Path(TypePath::Property{"X", true}), getBuiltins(), NotNull{&arena}) == getBuiltins()->numberType);
 }
 
 TEST_CASE_FIXTURE(TypePathBuiltinsFixture, "metatable_property")
@@ -150,7 +156,7 @@ TEST_CASE_FIXTURE(TypePathBuiltinsFixture, "metatable_property")
         )");
     }
 
-    CHECK(traverseForType(requireType("x"), Path(TypePath::Property::read("x")), builtinTypes) == builtinTypes->numberType);
+    CHECK(traverseForType(requireType("x"), Path(TypePath::Property::read("x")), getBuiltins(), NotNull{&arena}) == getBuiltins()->numberType);
 }
 
 TEST_CASE_FIXTURE(TypePathFixture, "index")
@@ -163,12 +169,12 @@ TEST_CASE_FIXTURE(TypePathFixture, "index")
 
         SUBCASE("in_bounds")
         {
-            CHECK(traverseForType(requireTypeAlias("T"), Path(TypePath::Index{1}), builtinTypes) == builtinTypes->stringType);
+            CHECK(traverseForType(requireTypeAlias("T"), Path(TypePath::Index{1}), getBuiltins(), NotNull{&arena}) == getBuiltins()->stringType);
         }
 
         SUBCASE("out_of_bounds")
         {
-            CHECK(traverseForType(requireTypeAlias("T"), Path(TypePath::Index{97}), builtinTypes) == std::nullopt);
+            CHECK(traverseForType(requireTypeAlias("T"), Path(TypePath::Index{97}), getBuiltins(), NotNull{&arena}) == std::nullopt);
         }
     }
 
@@ -181,7 +187,7 @@ TEST_CASE_FIXTURE(TypePathFixture, "index")
 
         SUBCASE("in_bounds")
         {
-            auto result = traverseForType(requireTypeAlias("T"), Path(TypePath::Index{1}), builtinTypes);
+            auto result = traverseForType(requireTypeAlias("T"), Path(TypePath::Index{1}), getBuiltins(), NotNull{&arena});
             CHECK(result);
 
             if (result)
@@ -190,7 +196,7 @@ TEST_CASE_FIXTURE(TypePathFixture, "index")
 
         SUBCASE("out_of_bounds")
         {
-            CHECK(traverseForType(requireTypeAlias("T"), Path(TypePath::Index{97}), builtinTypes) == std::nullopt);
+            CHECK(traverseForType(requireTypeAlias("T"), Path(TypePath::Index{97}), getBuiltins(), NotNull{&arena}) == std::nullopt);
         }
     }
 
@@ -204,25 +210,28 @@ TEST_CASE_FIXTURE(TypePathFixture, "index")
         SUBCASE("in_bounds")
         {
             Path path = Path({TypePath::PackField::Arguments, TypePath::Index{1}});
-            auto result = traverseForType(requireTypeAlias("T"), path, builtinTypes);
-            CHECK(result == builtinTypes->stringType);
+            auto result = traverseForType(requireTypeAlias("T"), path, getBuiltins(), NotNull{&arena});
+            CHECK(result == getBuiltins()->stringType);
         }
 
         SUBCASE("out_of_bounds")
         {
             Path path = Path({TypePath::PackField::Arguments, TypePath::Index{72}});
-            auto result = traverseForType(requireTypeAlias("T"), path, builtinTypes);
+            auto result = traverseForType(requireTypeAlias("T"), path, getBuiltins(), NotNull{&arena});
             CHECK(result == std::nullopt);
         }
     }
 }
 
-TEST_CASE_FIXTURE(ClassFixture, "metatables")
+TEST_CASE_FIXTURE(ExternTypeFixture, "metatables")
 {
+    getFrontend();
+    TypeArena arena;
+
     SUBCASE("string")
     {
-        auto result = traverseForType(builtinTypes->stringType, Path(TypeField::Metatable), builtinTypes);
-        CHECK(result == getMetatable(builtinTypes->stringType, builtinTypes));
+        auto result = traverseForType(getBuiltins()->stringType, Path(TypeField::Metatable), getBuiltins(), NotNull{&arena});
+        CHECK(result == getMetatable(getBuiltins()->stringType, getBuiltins()));
     }
 
     SUBCASE("string_singleton")
@@ -231,8 +240,8 @@ TEST_CASE_FIXTURE(ClassFixture, "metatables")
             type T = "foo"
         )");
 
-        auto result = traverseForType(requireTypeAlias("T"), Path(TypeField::Metatable), builtinTypes);
-        CHECK(result == getMetatable(builtinTypes->stringType, builtinTypes));
+        auto result = traverseForType(requireTypeAlias("T"), Path(TypeField::Metatable), getBuiltins(), NotNull{&arena});
+        CHECK(result == getMetatable(getBuiltins()->stringType, getBuiltins()));
     }
 
     SUBCASE("table")
@@ -246,7 +255,7 @@ TEST_CASE_FIXTURE(ClassFixture, "metatables")
         )");
 
         // Tricky test setup because 'setmetatable' mutates the argument 'tbl' type
-        auto result = traverseForType(requireType("res"), Path(TypeField::Table), builtinTypes);
+        auto result = traverseForType(requireType("res"), Path(TypeField::Table), getBuiltins(), NotNull{&arena});
         auto expected = lookupType("Table");
         REQUIRE(expected);
         CHECK(result == follow(*expected));
@@ -259,14 +268,14 @@ TEST_CASE_FIXTURE(ClassFixture, "metatables")
             local tbl = setmetatable({}, mt)
         )");
 
-        auto result = traverseForType(requireType("tbl"), Path(TypeField::Metatable), builtinTypes);
+        auto result = traverseForType(requireType("tbl"), Path(TypeField::Metatable), getBuiltins(), NotNull{&arena});
         CHECK(result == requireType("mt"));
     }
 
     SUBCASE("class")
     {
-        auto result = traverseForType(vector2InstanceType, Path(TypeField::Metatable), builtinTypes);
-        // ClassFixture's Vector2 metatable is just an empty table, but it's there.
+        auto result = traverseForType(vector2InstanceType, Path(TypeField::Metatable), getBuiltins(), NotNull{&arena});
+        // ExternTypeFixture's Vector2 metatable is just an empty table, but it's there.
         CHECK(result);
     }
 }
@@ -275,31 +284,31 @@ TEST_CASE_FIXTURE(TypePathFixture, "bounds")
 {
     SUBCASE("free_type")
     {
-        TypeArena& arena = frontend.globals.globalTypes;
+        TypeArena& arena = getFrontend().globals.globalTypes;
         unfreeze(arena);
 
-        TypeId ty = arena.freshType(frontend.builtinTypes, frontend.globals.globalScope.get());
+        TypeId ty = arena.freshType(getBuiltins(), getFrontend().globals.globalScope.get());
         FreeType* ft = getMutable<FreeType>(ty);
 
         SUBCASE("upper")
         {
-            ft->upperBound = builtinTypes->numberType;
-            auto result = traverseForType(ty, Path(TypeField::UpperBound), builtinTypes);
-            CHECK(result == builtinTypes->numberType);
+            ft->upperBound = getBuiltins()->numberType;
+            auto result = traverseForType(ty, Path(TypeField::UpperBound), getBuiltins(), NotNull{&arena});
+            CHECK(result == getBuiltins()->numberType);
         }
 
         SUBCASE("lower")
         {
-            ft->lowerBound = builtinTypes->booleanType;
-            auto result = traverseForType(ty, Path(TypeField::LowerBound), builtinTypes);
-            CHECK(result == builtinTypes->booleanType);
+            ft->lowerBound = getBuiltins()->booleanType;
+            auto result = traverseForType(ty, Path(TypeField::LowerBound), getBuiltins(), NotNull{&arena});
+            CHECK(result == getBuiltins()->booleanType);
         }
     }
 
     SUBCASE("unbounded_type")
     {
-        CHECK(traverseForType(builtinTypes->numberType, Path(TypeField::UpperBound), builtinTypes) == std::nullopt);
-        CHECK(traverseForType(builtinTypes->numberType, Path(TypeField::LowerBound), builtinTypes) == std::nullopt);
+        CHECK(traverseForType(getBuiltins()->numberType, Path(TypeField::UpperBound), getBuiltins(), NotNull{&arena}) == std::nullopt);
+        CHECK(traverseForType(getBuiltins()->numberType, Path(TypeField::LowerBound), getBuiltins(), NotNull{&arena}) == std::nullopt);
     }
 }
 
@@ -313,11 +322,11 @@ TEST_CASE_FIXTURE(TypePathFixture, "indexers")
                 type T = { [string]: boolean }
             )");
 
-            auto lookupResult = traverseForType(requireTypeAlias("T"), Path(TypeField::IndexLookup), builtinTypes);
-            auto resultResult = traverseForType(requireTypeAlias("T"), Path(TypeField::IndexResult), builtinTypes);
+            auto lookupResult = traverseForType(requireTypeAlias("T"), Path(TypeField::IndexLookup), getBuiltins(), NotNull{&arena});
+            auto resultResult = traverseForType(requireTypeAlias("T"), Path(TypeField::IndexResult), getBuiltins(), NotNull{&arena});
 
-            CHECK(lookupResult == builtinTypes->stringType);
-            CHECK(resultResult == builtinTypes->booleanType);
+            CHECK(lookupResult == getBuiltins()->stringType);
+            CHECK(resultResult == getBuiltins()->booleanType);
         }
 
         SUBCASE("no_indexer")
@@ -326,32 +335,32 @@ TEST_CASE_FIXTURE(TypePathFixture, "indexers")
                 type T = { y: number }
             )");
 
-            auto lookupResult = traverseForType(requireTypeAlias("T"), Path(TypeField::IndexLookup), builtinTypes);
-            auto resultResult = traverseForType(requireTypeAlias("T"), Path(TypeField::IndexResult), builtinTypes);
+            auto lookupResult = traverseForType(requireTypeAlias("T"), Path(TypeField::IndexLookup), getBuiltins(), NotNull{&arena});
+            auto resultResult = traverseForType(requireTypeAlias("T"), Path(TypeField::IndexResult), getBuiltins(), NotNull{&arena});
 
             CHECK(lookupResult == std::nullopt);
             CHECK(resultResult == std::nullopt);
         }
     }
 
-    // TODO: Class types
+    // TODO: Extern types
 }
 
 TEST_CASE_FIXTURE(TypePathFixture, "negated")
 {
     SUBCASE("valid")
     {
-        TypeArena& arena = frontend.globals.globalTypes;
+        TypeArena& arena = getFrontend().globals.globalTypes;
         unfreeze(arena);
 
-        TypeId ty = arena.addType(NegationType{builtinTypes->numberType});
-        auto result = traverseForType(ty, Path(TypeField::Negated), builtinTypes);
-        CHECK(result == builtinTypes->numberType);
+        TypeId ty = arena.addType(NegationType{getBuiltins()->numberType});
+        auto result = traverseForType(ty, Path(TypeField::Negated), getBuiltins(), NotNull{&arena});
+        CHECK(result == getBuiltins()->numberType);
     }
 
     SUBCASE("not_negation")
     {
-        auto result = traverseForType(builtinTypes->numberType, Path(TypeField::Negated), builtinTypes);
+        auto result = traverseForType(getBuiltins()->numberType, Path(TypeField::Negated), getBuiltins(), NotNull{&arena});
         CHECK(result == std::nullopt);
     }
 }
@@ -360,17 +369,17 @@ TEST_CASE_FIXTURE(TypePathFixture, "variadic")
 {
     SUBCASE("valid")
     {
-        TypeArena& arena = frontend.globals.globalTypes;
+        TypeArena& arena = getFrontend().globals.globalTypes;
         unfreeze(arena);
 
-        TypePackId tp = arena.addTypePack(VariadicTypePack{builtinTypes->numberType});
-        auto result = traverseForType(tp, Path(TypeField::Variadic), builtinTypes);
-        CHECK(result == builtinTypes->numberType);
+        TypePackId tp = arena.addTypePack(VariadicTypePack{getBuiltins()->numberType});
+        auto result = traverseForType(tp, Path(TypeField::Variadic), getBuiltins(), NotNull{&arena});
+        CHECK(result == getBuiltins()->numberType);
     }
 
     SUBCASE("not_variadic")
     {
-        auto result = traverseForType(builtinTypes->numberType, Path(TypeField::Variadic), builtinTypes);
+        auto result = traverseForType(getBuiltins()->numberType, Path(TypeField::Variadic), getBuiltins(), NotNull{&arena});
         CHECK(result == std::nullopt);
     }
 }
@@ -384,7 +393,7 @@ TEST_CASE_FIXTURE(TypePathFixture, "arguments")
             end
         )");
 
-        auto result = traverseForPack(requireType("f"), Path(PackField::Arguments), builtinTypes);
+        auto result = traverseForPack(requireType("f"), Path(PackField::Arguments), getBuiltins(), NotNull{&arena});
         CHECK(result);
         if (result)
             CHECK(toString(*result) == "number, string");
@@ -392,7 +401,7 @@ TEST_CASE_FIXTURE(TypePathFixture, "arguments")
 
     SUBCASE("not_function")
     {
-        auto result = traverseForPack(builtinTypes->booleanType, Path(PackField::Arguments), builtinTypes);
+        auto result = traverseForPack(getBuiltins()->booleanType, Path(PackField::Arguments), getBuiltins(), NotNull{&arena});
         CHECK(result == std::nullopt);
     }
 }
@@ -407,7 +416,7 @@ TEST_CASE_FIXTURE(TypePathFixture, "returns")
             end
         )");
 
-        auto result = traverseForPack(requireType("f"), Path(PackField::Returns), builtinTypes);
+        auto result = traverseForPack(requireType("f"), Path(PackField::Returns), getBuiltins(), NotNull{&arena});
         CHECK(result);
         if (result)
             CHECK(toString(*result) == "number, string");
@@ -415,7 +424,7 @@ TEST_CASE_FIXTURE(TypePathFixture, "returns")
 
     SUBCASE("not_function")
     {
-        auto result = traverseForPack(builtinTypes->booleanType, Path(PackField::Returns), builtinTypes);
+        auto result = traverseForPack(getBuiltins()->booleanType, Path(PackField::Returns), getBuiltins(), NotNull{&arena});
         CHECK(result == std::nullopt);
     }
 }
@@ -428,7 +437,7 @@ TEST_CASE_FIXTURE(TypePathFixture, "tail")
             type T = (number, string, ...boolean) -> ()
         )");
 
-        auto result = traverseForPack(requireTypeAlias("T"), Path({PackField::Arguments, PackField::Tail}), builtinTypes);
+        auto result = traverseForPack(requireTypeAlias("T"), Path({PackField::Arguments, PackField::Tail}), getBuiltins(), NotNull{&arena});
         CHECK(result);
         if (result)
             CHECK(toString(*result) == "...boolean");
@@ -440,42 +449,81 @@ TEST_CASE_FIXTURE(TypePathFixture, "tail")
             type T = (number, string) -> ()
         )");
 
-        auto result = traverseForPack(requireTypeAlias("T"), Path({PackField::Arguments, PackField::Tail}), builtinTypes);
+        auto result = traverseForPack(requireTypeAlias("T"), Path({PackField::Arguments, PackField::Tail}), getBuiltins(), NotNull{&arena});
         CHECK(result == std::nullopt);
     }
 
     SUBCASE("type")
     {
-        auto result = traverseForPack(builtinTypes->stringType, Path({PackField::Arguments, PackField::Tail}), builtinTypes);
+        auto result = traverseForPack(getBuiltins()->stringType, Path({PackField::Arguments, PackField::Tail}), getBuiltins(), NotNull{&arena});
         CHECK(result == std::nullopt);
     }
 }
 
+TEST_CASE_FIXTURE(TypePathFixture, "pack_slice_has_tail")
+{
+    TypeArena& arena = getFrontend().globals.globalTypes;
+    unfreeze(arena);
+
+    TYPESOLVE_CODE(R"(
+        type T = (number, string, ...boolean) -> ()
+    )");
+
+    auto result = traverseForPack(requireTypeAlias("T"), Path({PackField::Arguments, PackSlice{1}}), getBuiltins(), NotNull{&arena});
+    CHECK(result);
+    if (result)
+        CHECK(toString(*result) == "string, ...boolean");
+}
+
+TEST_CASE_FIXTURE(TypePathFixture, "pack_slice_finite_pack")
+{
+    TypeArena& arena = getFrontend().globals.globalTypes;
+    unfreeze(arena);
+
+    TYPESOLVE_CODE(R"(
+        type T = (number, string) -> ()
+    )");
+
+    auto result = traverseForPack(requireTypeAlias("T"), Path({PackField::Arguments, PackSlice{1}}), getBuiltins(), NotNull{&arena});
+    CHECK(result);
+    if (result)
+        CHECK(toString(*result) == "string");
+}
+
+TEST_CASE_FIXTURE(TypePathFixture, "pack_slice_type")
+{
+    TypeArena& arena = getFrontend().globals.globalTypes;
+    unfreeze(arena);
+
+    auto result = traverseForPack(builtinTypes->stringType, Path({PackField::Arguments, PackSlice{1}}), getBuiltins(), NotNull{&arena});
+    CHECK(result == std::nullopt);
+}
+
 TEST_CASE_FIXTURE(TypePathFixture, "cycles" * doctest::timeout(0.5))
 {
     // This will fail an occurs check, but it's a quick example of a cyclic type
     // where there _is_ no traversal.
     SUBCASE("bound_cycle")
     {
-        TypeArena& arena = frontend.globals.globalTypes;
+        TypeArena& arena = getFrontend().globals.globalTypes;
         unfreeze(arena);
 
         TypeId a = arena.addType(BlockedType{});
         TypeId b = arena.addType(BoundType{a});
         asMutable(a)->ty.emplace<BoundType>(b);
 
-        CHECK_THROWS(traverseForType(a, Path(TypeField::IndexResult), builtinTypes));
+        CHECK_THROWS(traverseForType(a, Path(TypeField::IndexResult), getBuiltins(), NotNull{&arena}));
     }
 
     SUBCASE("table_contains_itself")
     {
-        TypeArena& arena = frontend.globals.globalTypes;
+        TypeArena& arena = getFrontend().globals.globalTypes;
         unfreeze(arena);
 
         TypeId tbl = arena.addType(TableType{});
         getMutable<TableType>(tbl)->props["a"] = Luau::Property(tbl);
 
-        auto result = traverseForType(tbl, Path(TypePath::Property{"a", true}), builtinTypes);
+        auto result = traverseForType(tbl, Path(TypePath::Property{"a", true}), getBuiltins(), NotNull{&arena});
         CHECK(result == tbl);
     }
 }
@@ -496,7 +544,7 @@ TEST_CASE_FIXTURE(TypePathFixture, "step_limit")
 
     TypeId root = requireTypeAlias("T");
     Path path = PathBuilder().readProp("x").readProp("y").readProp("z").build();
-    auto result = traverseForType(root, path, builtinTypes);
+    auto result = traverseForType(root, path, getBuiltins(), NotNull{&arena});
     CHECK(!result);
 }
 
@@ -508,14 +556,14 @@ TEST_CASE_FIXTURE(TypePathBuiltinsFixture, "complex_chains")
             type Meta = {
                 __add: (Tab, Tab) -> number,
             }
-            
+
             type Tab = typeof(setmetatable({}, {} :: Meta))
         )");
 
         TypeId root = requireTypeAlias("Tab");
         Path path = PathBuilder().mt().readProp("__add").rets().index(0).build();
-        auto result = traverseForType(root, path, builtinTypes);
-        CHECK(result == builtinTypes->numberType);
+        auto result = traverseForType(root, path, getBuiltins(), NotNull{&arena});
+        CHECK(result == getBuiltins()->numberType);
     }
 
     SUBCASE("overloaded_fn_overload_one_argument_two")
@@ -528,8 +576,8 @@ TEST_CASE_FIXTURE(TypePathBuiltinsFixture, "complex_chains")
 
         TypeId root = requireTypeAlias("Obj");
         Path path = PathBuilder().readProp("method").index(0).args().index(1).build();
-        auto result = traverseForType(root, path, builtinTypes);
-        CHECK(*result == builtinTypes->falseType);
+        auto result = traverseForType(root, path, getBuiltins(), NotNull{&arena});
+        CHECK(*result == getBuiltins()->falseType);
     }
 }
 
@@ -562,6 +610,14 @@ TEST_CASE("human_property_then_metatable_portion")
     CHECK(toStringHuman(PathBuilder().writeProp("a").mt().build()) == "writing to `a` has the metatable portion as ");
 }
 
+TEST_CASE("pack_slice")
+{
+    ScopedFastFlag sff{FFlag::LuauSolverV2, true};
+
+    CHECK(toString(PathBuilder().packSlice(1).build()) == "[1:]");
+    CHECK(toStringHuman(PathBuilder().packSlice(1).build()) == "the portion of the type pack starting at index 1 to the end");
+}
+
 TEST_SUITE_END(); // TypePathToString
 
 TEST_SUITE_BEGIN("TypePathBuilder");
@@ -621,5 +677,9 @@ TEST_CASE("chained")
         Path({Index{0}, TypePath::Property::read("foo"), TypeField::Metatable, TypePath::Property::read("bar"), PackField::Arguments, Index{1}})
     );
 }
+TEST_CASE("pack_slice")
+{
+    CHECK(PathBuilder().packSlice(3).build() == Path(PackSlice{3}));
+}
 
 TEST_SUITE_END(); // TypePathBuilder
